/* @lastChanged: "1998-06-26  9:57"
 *
 * @filename:    vcn32.cpp
 * @purpose:     "backup history management"
 * @release:     7.1.0.0
 * @see:         "-.-"
 *
 * @copyright:   "Copyright by SAP AG, 1998"
 *
 * ==================================================================
 *
 * responsible:   Martin Reinecke (D024853 - Martin.Reinecke@SAP-AG.de)
 *
 * created:       1998-05-25 by Martin Reinecke
 *
 * purpose:       basic module for backup history
 *                               
 * ==================================================================
 */

/* ------------------------------------------------------------------
 * list of functions:
 *
 * cn32_InitHistoryEBID
 * cn32NewHistoryEBID
 * cn32SetHistoryEBID
 * cn32_SetHistoryEBIDLine
 *
 * cn32_CheckHistoryFile
 * cn32_OpenHistoryFile
 * cn32_ReadHistoryLine
 * cn32_OpenHistoryMediaFile 
 * cn32_ReadHistoryMediaLine
 * cn32_OpenFile
 * cn32_ReadHistoryEBIDLine
 * cn32_CloseFile
 * cn32_InitHistoryEntry
 * cn32_NewHistoryEntry
 * cn32_InitHistoryMedium
 * cn32_NewHistoryMedium
 * cn32_InitHistory
 * cn32_InitHistoryMedia
 * cn32InitHistoryEBIDs
 * cn32_FreeHistory
 * cn32_FreeHistoryMedia
 * cn32_FreeHistoryEBID
 * cn32FreeHistoryEBIDs
 * cn32AddToEndOfHistoryEBIDs
 * cn32_ReadHistoryFile
 * cn32_ReadHistoryMediaFile
 * cn32_ReadHistoryEBIDFile
 * cn32_CheckAvailabilityOfExternalBackups
 * cn32_AllocHistorySession
 * cn32_HistorySession
 * cn32_FreeHistorySession
 * cn32_PrintEntry
 * cn32_GetEntryLine
 * cn32_GetMediumLine
 * cn32_GetEBIDAsLine
 * cn32_ListMedia
 * cn32_ListEBIDsForHistoryEntry
 * cn32_ListHistory
 * cn32_FindFirstRestoreItem
 * cn32_RemoveDoublequotes
 * 
 * cn32BackupHistoryOpen
 * cn32BackupHistoryList
 * cn32BackupHistoryListNext
 * cn32BackupHistoryClose
 * cn32BackupHistoryDate
 * cn32ListExternalBackupIDs
 * ------------------------------------------------------------------
 */

/* ------------------------------------------------------------------
 * includes
 * ------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>

#include "heo02.h"
#include "heo06.h"
#include "hcn32.h"
#include "hcn42.h"
#include "hcn90.h"
#include "hcn35.h"
#include "hcn31.h"

/* PRIVATE FUNCTION cn32_InitHistoryEBID
 * ------------------------------------------------------------------
 * purpose: initialize the history EBID struct
 *
 */
static void cn32_InitHistoryEBID(tcn003_BackupHistoryEBID * pEBID)
{
    if(0!=pEBID)
    {
        pEBID->Status=EBIDStatusUnknown;
        pEBID->Line.EmptyString='\0';

        pEBID->Line.szKey=&(pEBID->Line.EmptyString);       // init as empty string
        pEBID->Line.szLabel=&(pEBID->Line.EmptyString);     // init as empty string
        pEBID->Line.szEBID=&(pEBID->Line.EmptyString);      // init as empty string
        pEBID->Line.szType=&(pEBID->Line.EmptyString);      // init as empty string
        pEBID->Line.szDateTime=&(pEBID->Line.EmptyString);  // init as empty string

        pEBID->Line.CompleteMemory=0; //no memory is used at the moment

        pEBID->pNext = NULL;
        pEBID->pPrev = NULL;
    }
} /* end cn32_InitHistoryEBID */

/* FUNCTION cn32NewHistoryEBID
 * ------------------------------------------------------------------
 * purpose: alloc memory for new history EBID
 *
 */
tcn00_Error cn32NewHistoryEBID(char                      * replyData, 
                               int                       * replyLen,
                               tcn003_BackupHistoryEBID ** pEBID )
{
    tcn00_Error nFuncStatus=OK_CN00;
    bool        bMallocOK=true;
    //bool        bOK=false;

    cn00MyMalloc(sizeof (tcn003_BackupHistoryEBID), pEBID, &bMallocOK);

    if((NULL==(*pEBID)) || (!bMallocOK))
    {
        nFuncStatus=ERR_MEM_CN00;
        cn90AnswerIError(replyData, replyLen, nFuncStatus);
    }
    else
        cn32_InitHistoryEBID(*pEBID);

    return nFuncStatus;
} /* end cn32NewHistoryEBID */

/* PRIVATE FUNCTION cn32_StrNCpyAndMoveBehindEnd
 * ------------------------------------------------------------------
 * purpose: make an strncpy and set pointer Dest to Src+Length+1
 *
 */

static void cn32_StrNCpyAndMoveBehindEnd(char *& Dest, const char * Src, size_t Length)
{
    strncpy(Dest, Src, Length);
    Dest+=Length;
    *(Dest++)='\0';
}

/* PRIVATE FUNCTION cn32_SetHistoryEBIDLine
 * ------------------------------------------------------------------
 * purpose: set information of a *pLine
 *
 */
static tcn00_Error cn32_SetHistoryEBIDLine(char       * replyData, 
                                           int        * replyLen,
                                           const char * Key,
                                           size_t       KeyLength,
                                           const char * Label,
                                           size_t       LabelLength,
                                           const char * EBID,
                                           size_t       EBIDLength,
                                           const char * Type,
                                           size_t       TypeLength,
                                           const char * DateTime,
                                           size_t       DateTimeLength,
                                           tcn003_BackupHistoryEBIDLine * pLine)
{
    tcn00_Error rc=OK_CN00;
    bool bMallocOK=true;

    if(0!=pLine->CompleteMemory)
    {
        cn00MyFree(pLine->CompleteMemory);
        pLine->CompleteMemory=0;
    }

    cn00MyMalloc(KeyLength+1+LabelLength+1+EBIDLength+1+TypeLength+1+DateTimeLength+1, &(pLine->CompleteMemory), &bMallocOK);

    if((0==pLine->CompleteMemory) || (!bMallocOK))
    {
        rc=ERR_MEM_CN00;
        cn90AnswerIError(replyData, replyLen, rc);
    }
    else
    {
        char * CurrentPosition=pLine->CompleteMemory;

        pLine->szKey=CurrentPosition;
        cn32_StrNCpyAndMoveBehindEnd(CurrentPosition, Key, KeyLength);
        
        pLine->szLabel=CurrentPosition;
        cn32_StrNCpyAndMoveBehindEnd(CurrentPosition, Label, LabelLength);

        pLine->szEBID=CurrentPosition;
        cn32_StrNCpyAndMoveBehindEnd(CurrentPosition, EBID, EBIDLength);

        pLine->szType=CurrentPosition;
        cn32_StrNCpyAndMoveBehindEnd(CurrentPosition, Type, TypeLength);

        pLine->szDateTime=CurrentPosition;
        cn32_StrNCpyAndMoveBehindEnd(CurrentPosition, DateTime, DateTimeLength);
    }

    return rc;
} /* end cn32_SetHistoryEBIDLine */

/* FUNCTION cn32SetHistoryEBID
 * ------------------------------------------------------------------
 * purpose: set information of of *pEBID
 *
 */
tcn00_Error cn32SetHistoryEBID(char       * replyData, 
                               int        * replyLen,
                               tcn003_BackupHistoryEBIDStatus Status,
                               const char * Key,
                               size_t       KeyLength,
                               const char * Label,
                               size_t       LabelLength,
                               const char * EBID,
                               size_t       EBIDLength,
                               const char * Type,
                               //size_t       TypeLength,
                               const char * DateTime,
                               //size_t       DateTimeLength,
                               tcn003_BackupHistoryEBID * pEBID)
{
    tcn00_Error rc;

    pEBID->Status=Status;
    
    rc=cn32_SetHistoryEBIDLine(replyData,
                               replyLen,
                               Key,
                               KeyLength,
                               Label,
                               LabelLength,
                               EBID,
                               EBIDLength,
                               Type,
                               strlen(Type),
                               DateTime,
                               strlen(DateTime),
                               &(pEBID->Line));

    return rc;
} /* end cn32SetHistoryEBID */

/* ------------------------------------------------------------------
 * PRIVATE FUNCTION cn30_CheckHistoryFile
 * ------------------------------------------------------------------
 * purpose: exists the history file for specified database
 *
 */
static tcn00_Error cn32_CheckHistoryFile (
      VControlDataT  * vcontrol,
      char           * replyData,
      int            * replyLen,
      bool           * bExists,
      char           * szDate)
{
  tcn00_Error          nFuncStatus  = OK_CN00;
  tsp5_rte_fileerror   rteFileError;

  char                 szFile[255];

  tsp5_rte_fileinfo    rteFileInfo;

  char                 * szPos = NULL;

  /**/

  *bExists = false;

  /* exists history file */
  cn42GetFileName(vcontrol->dbname, FGET_BACKHIST_CN42, *(tsp00_C264c *) szFile);
  sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  if (rteFileError.sp5fe_result == vf_ok && !rteFileInfo.sp5fi_exists) {
    cn42RestoreFile(vcontrol->dbname, FGET_BACKHIST_CN42);
    sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  } // end if

  if (rteFileError.sp5fe_result == vf_ok) {
    if (rteFileInfo.sp5fi_exists) {

      *bExists = true;

      szPos = szDate;        
      memcpy(szPos, rteFileInfo.sp5fi_date_modified, sizeof(rteFileInfo.sp5fi_date_modified));
      szPos += sizeof(rteFileInfo.sp5fi_date_modified);
      memcpy(szPos, rteFileInfo.sp5fi_time_modified, sizeof(rteFileInfo.sp5fi_time_modified));
      szPos += sizeof(rteFileInfo.sp5fi_time_modified);
      *szPos = '\0';

    } else {
      nFuncStatus = ERR_NOHISTORY_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    }
  } else {
    nFuncStatus = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
  }

  return nFuncStatus;
} /* end cn30_CheckHistoryFile */


/* PRIVATE FUNCTION cn32_OpenHistoryFile
 * ------------------------------------------------------------------
 * purpose: open the backup history file for specified database
 *
 */
static tcn00_Error cn32_OpenHistoryFile ( 
      VControlDataT  * vcontrol, 
      char           * replyData,
      int            * replyLen,
      tsp_int4       * hHistoryFile )
{
  tcn00_Error          nFuncStatus  = OK_CN00;
  tsp5_rte_fileerror   rteFileError;

  char                 szFile[255];

  tsp5_rte_fileinfo    rteFileInfo;

  /**/

  /* exists history file */
  cn42GetFileName(vcontrol->dbname, FGET_BACKHIST_CN42, *(tsp00_C264c *) szFile);
  sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  if (rteFileError.sp5fe_result == vf_ok && !rteFileInfo.sp5fi_exists) {
    cn42RestoreFile(vcontrol->dbname, FGET_BACKHIST_CN42);
    sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  } // end if

  sqlfinfoc (szFile, &rteFileInfo, &rteFileError);

  if (rteFileError.sp5fe_result == vf_ok) {
    if (rteFileInfo.sp5fi_exists == false) {
      nFuncStatus = ERR_NOHISTORY_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    }
  } else {
    nFuncStatus = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
  }
    
  /* open history file */
  if (nFuncStatus == OK_CN00) {
    sqlfopenc (szFile, sp5vf_text, sp5vf_read, sp5bk_buffered, hHistoryFile, &rteFileError);

    if (rteFileError.sp5fe_result != vf_ok) {
      nFuncStatus = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
    }
  }

  return nFuncStatus;
} /* end cn32_OpenHistoryFile */

/* PRIVATE FUNCTION cn32_ReadHistoryLine
 * ------------------------------------------------------------------
 * purpose: read one history line from file
 *
 */
static tcn00_Error cn32_ReadHistoryLine (
      tsp_int4                    hFile,
      tcn003_BackupHistoryLine  * pLine,
      tsp5_rte_fileerror        * pFileError)
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  char                szLine[BA_HISTORY_LINE_LENGTH_MXCN003 + 1];
  long                nLen = 0;
  tsp_longint         nRead;

  /**/
 
  sqlfreadc(hFile, szLine, BA_HISTORY_LINE_LENGTH_MXCN003, &nRead, pFileError);
  
  if (pFileError->sp5fe_result == vf_ok) {

//    cn90Strip(szLine);
    cn90StripTrailingChar(szLine, '\t');
    cn90StripTrailingChar(szLine, ' ');

    nLen = (long)strlen(szLine);

    if (nLen > 0) {
      memcpy(pLine, szLine, nLen);
      pLine->szKEY[BA_HISTORY_LEN_KEY_CN003] = '\0';
      pLine->szLABEL[BA_HISTORY_LEN_LBL_CN003] = '\0';
      pLine->szACTION[BA_HISTORY_LEN_ACT_CN003] = '\0';
      pLine->szDBSTAMP1[BA_HISTORY_LEN_ST1_CN003] = '\0';
      pLine->szDBSTAMP2[BA_HISTORY_LEN_ST2_CN003] = '\0';
      pLine->szSTARTDATE[BA_HISTORY_LEN_STA_CN003] = '\0';
      pLine->szSTOPDATE[BA_HISTORY_LEN_STO_CN003] = '\0';
      pLine->szSTARTLOGPAGE[BA_HISTORY_LEN_FLO_CN003] = '\0';
      pLine->szSTOPLOGPAGE[BA_HISTORY_LEN_LLO_CN003] = '\0';
      pLine->szLOGREQUIRED[BA_HISTORY_LEN_LOG_CN003] = '\0';
      pLine->szMEDIANAME[BA_HISTORY_LEN_MED_CN003] = '\0';
      pLine->szPAGESTRANSFERED[BA_HISTORY_LEN_PAG_CN003] = '\0';
      pLine->szVOLUMES[BA_HISTORY_LEN_VOL_CN003] = '\0';
      pLine->szRETURNCODE[BA_HISTORY_LEN_RCO_CN003] = '\0';
      pLine->szRETURNTEXT[BA_HISTORY_LEN_RTX_CN003] = '\0';
    }
  } else {
    nFuncStatus = ERR_RTE_CN00;
  }
  /**/

  return nFuncStatus;
} /* end cn32_ReadHistoryLine */

/* PRIVATE FUNCTION cn32_OpenHistoryMediaFile
 * ------------------------------------------------------------------
 * purpose: open the backup history media file for specified database
 *
 */
static tcn00_Error cn32_OpenHistoryMediaFile ( 
      VControlDataT  * vcontrol, 
      char           * replyData,
      int            * replyLen,
      tsp_int4       * hHistoryMediaFile )
{
  tcn00_Error          nFuncStatus  = OK_CN00;
  tsp5_rte_fileerror   rteFileError;

  char                 szFile[255];

  tsp5_rte_fileinfo    rteFileInfo;

  /**/

  /* exists media file */
  cn42GetFileName(vcontrol->dbname, FGET_BACKMDF_CN42, *(tsp00_C264c *) szFile);
  sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  if (rteFileError.sp5fe_result == vf_ok && !rteFileInfo.sp5fi_exists) {
    cn42RestoreFile(vcontrol->dbname, FGET_BACKMDF_CN42);
    sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
  } // end if

  sqlfinfoc (szFile, &rteFileInfo, &rteFileError);

  if (rteFileError.sp5fe_result == vf_ok) {
    if (rteFileInfo.sp5fi_exists == false) {
      nFuncStatus = ERR_NOHISTORY_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    }
  } else {
    nFuncStatus = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
  }
    
  /* open history media file */
  if (nFuncStatus == OK_CN00) {
    sqlfopenc (szFile, sp5vf_text, sp5vf_read, sp5bk_buffered, hHistoryMediaFile, &rteFileError);

    if (rteFileError.sp5fe_result != vf_ok) {
      nFuncStatus = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
    }
  }

  return nFuncStatus;
} /* end cn32_OpenHistoryMediaFile */

/* PRIVATE FUNCTION cn32_ReadHistoryMediaLine
 * ------------------------------------------------------------------
 * purpose: read one history media line from file
 *
 */
static tcn00_Error cn32_ReadHistoryMediaLine (
      tsp_int4                        hFile,
      tcn003_BackupHistoryMediumLine  * pLine,
      tsp5_rte_fileerror              * pFileError)
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  char                szLine[BA_HISTORY_LINE_LENGTH_MXCN003 + 1];
  long                nLen = 0;
  tsp_longint         nRead;

  /**/
 
  sqlfreadc(hFile, szLine, BA_HISTORY_LINE_LENGTH_MXCN003, &nRead, pFileError);
  
  if (pFileError->sp5fe_result == vf_ok) {
/*
    cn90Strip(szLine);
 */
   nLen = (long)strlen(szLine);

    if (nLen > 0) {
      memcpy(pLine, szLine, nLen);
      pLine->szKEY[BA_HISTMED_LEN_KEY_CN003] = '\0';
      pLine->szLABEL[BA_HISTMED_LEN_LBL_CN003] = '\0';
      pLine->szMEDIANAME[BA_HISTMED_LEN_MED_CN003] = '\0';
      pLine->szTYPE[BA_HISTMED_LEN_TYP_CN003] = '\0';
      pLine->szOVERWRITE[BA_HISTMED_LEN_OVR_CN003] = '\0';
      pLine->szSIZE[BA_HISTMED_LEN_SIZ_CN003] = '\0';
      pLine->szBLOCKSIZE[BA_HISTMED_LEN_BSZ_CN003] = '\0';
      pLine->szKIND[BA_HISTMED_LEN_KND_CN003] = '\0';
      pLine->szLOCATION[BA_HISTMED_LEN_LOC_CN003] = '\0';
    }
  } else {
    nFuncStatus = ERR_RTE_CN00;
  }
  /**/

  return nFuncStatus;
} /* end cn32_ReadHistoryMediaLine */

/* PRIVATE FUNCTION cn32_OpenFile
 * ------------------------------------------------------------------
 * purpose: open the backup history media file for specified database
 *
 */
static tcn00_Error cn32_OpenFile(const tsp00_Namec   FileType,
                                 VControlDataT     * vcontrol, 
                                 char              * replyData,
                                 int               * replyLen,
                                 tsp_int4          * hFile)
{
    tcn00_Error          nFuncStatus=OK_CN00;
    tsp5_rte_fileerror   rteFileError;
    char                 szFile[255];
    tsp5_rte_fileinfo    rteFileInfo;

    // Does the file exist? If not, can it be restored ?
    cn42GetFileName(vcontrol->dbname, FileType, *(tsp00_C264c *) szFile);

    sqlfinfoc(szFile, &rteFileInfo, &rteFileError);

    if(vf_ok==rteFileError.sp5fe_result &&
       !rteFileInfo.sp5fi_exists)
    {
        cn42RestoreFile(vcontrol->dbname, FileType);
        sqlfinfoc (szFile, &rteFileInfo, &rteFileError);
    }

    if(vf_ok==rteFileError.sp5fe_result)
    {
        if(!rteFileInfo.sp5fi_exists)
        {
            nFuncStatus=ERR_NOHISTORY_CN00;
            cn90AnswerIError(replyData, replyLen, nFuncStatus);
        }
    }
    else
    {
        nFuncStatus=ERR_RTE_CN00;
        cn90AnswerRTEError(replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
    }

    // open file
    if(OK_CN00==nFuncStatus)
    {
        sqlfopenc(szFile, sp5vf_text, sp5vf_read, sp5bk_buffered, hFile, &rteFileError);

        if(vf_ok!=rteFileError.sp5fe_result)
        {
            nFuncStatus=ERR_RTE_CN00;
            cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
        }
    }

    return nFuncStatus;
} /* end cn32_OpenFile */

/* PRIVATE FUNCTION cn32_ReadHistoryEBIDLine
 * ------------------------------------------------------------------
 * purpose: read one history media line from file
 *          writes non RTE Errors directly to replyData
 */
static tcn00_Error cn32_ReadHistoryEBIDLine(
      char                          * replyData,
      int                           * replyLen,
      tsp_int4                        hFile,
      tcn003_BackupHistoryEBIDLine  * pLine,
      tsp5_rte_fileerror            * pFileError)
{
    tcn00_Error	        nFuncStatus=OK_CN00;

    char        szLine[BA_HISTORY_LINE_LENGTH_MXCN003 + 1];
    long        nLen=0;
    tsp_longint nRead;

    sqlfreadc(hFile, szLine, BA_HISTORY_LINE_LENGTH_MXCN003, &nRead, pFileError);
  
    if(vf_ok==pFileError->sp5fe_result)
    {
        nLen = (long)strlen(szLine);

        if(nLen>0) 
        {
            char * pDelimiterBetweenKeyAndLabel=0;
            char * pDelimiterBetweenLabelAndEBID=0;
            char * pDelimiterEBID=0;
            
            pDelimiterBetweenKeyAndLabel=strstr(szLine, BA_HISTORY_COLSEP_CN003C);

            if(0!=pDelimiterBetweenKeyAndLabel)
                pDelimiterBetweenLabelAndEBID=strstr(pDelimiterBetweenKeyAndLabel+1, BA_HISTORY_COLSEP_CN003C);

            if(0!=pDelimiterBetweenLabelAndEBID)
                pDelimiterEBID=strstr(pDelimiterBetweenLabelAndEBID+1, BA_HISTORY_COLSEP_CN003C);

            if(0==pDelimiterEBID)
            {
                nFuncStatus=ERR_EBFDIRTY_CN00;
                cn90AnswerIError(replyData, replyLen, nFuncStatus);
            }
            else
                nFuncStatus=cn32_SetHistoryEBIDLine(replyData,
                                                    replyLen,
                                                    szLine,
                                                    pDelimiterBetweenKeyAndLabel-szLine,
                                                    pDelimiterBetweenKeyAndLabel+1,
                                                    pDelimiterBetweenLabelAndEBID-(pDelimiterBetweenKeyAndLabel+1),
                                                    pDelimiterBetweenLabelAndEBID+1,
                                                    pDelimiterEBID-(pDelimiterBetweenLabelAndEBID+1),
                                                    "",                                                             // no extra save type information
                                                    0,
                                                    "",                                                             // no extra date and time information
                                                    0,
                                                    pLine);
        }
    }
    else
        nFuncStatus = ERR_RTE_CN00;

  return nFuncStatus;
} /* end cn32_ReadHistoryEBIDLine */

/* PRIVATE FUNCTION cn32_CloseFile
 * ------------------------------------------------------------------
 * purpose: close file
 *
 */
static tcn00_Error cn32_CloseFile (
      char           * replyData,
      int            * replyLen,
      tsp_int4       hFile )
{
  tcn00_Error          nFuncStatus  = OK_CN00;
  tsp5_rte_fileerror   rteFileError;

  /**/

  sqlfclosec (hFile, sp5vf_close_normal, &rteFileError);
  if (rteFileError.sp5fe_result != vf_ok) {
    nFuncStatus = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
  }

  return nFuncStatus;
} /* end cn32_CloseFile */

/* PRIVATE FUNCTION cn32_InitHistoryEntry
 * ------------------------------------------------------------------
 * purpose: initialize the history entry struct
 *
 */
static void cn32_InitHistoryEntry(
      tcn003_BackupHistoryEntry   * pEntry)
{
  pEntry->tLine.szKEY[0] = '\0';
  pEntry->tLine.szLABEL[0] = '\0';
  pEntry->tLine.szACTION[0] = '\0';
  pEntry->tLine.szDBSTAMP1[0] = '\0';
  pEntry->tLine.szDBSTAMP2[0] = '\0';
  pEntry->tLine.szSTARTDATE[0] = '\0';
  pEntry->tLine.szSTOPDATE[0] = '\0';
  pEntry->tLine.szSTARTLOGPAGE[0] = '\0';
  pEntry->tLine.szSTOPLOGPAGE[0] = '\0';
  pEntry->tLine.szLOGREQUIRED[0] = '\0';
  pEntry->tLine.szMEDIANAME[0] = '\0';
  pEntry->tLine.szPAGESTRANSFERED[0] = '\0';
  pEntry->tLine.szVOLUMES[0] = '\0';
  pEntry->tLine.szRETURNCODE[0] = '\0';
  pEntry->tLine.szRETURNTEXT[0] = '\0';
  
  pEntry->pNext = NULL;
  pEntry->pPrev = NULL;
} /* end cn32_InitHistoryEntry */

/* PRIVATE FUNCTION cn32_NewHistoryEntry
 * ------------------------------------------------------------------
 * purpose: alloc memory for new history entry
 *
 */
static tcn00_Error cn32_NewHistoryEntry (
      char                        * replyData, 
      int                         * replyLen,
      tcn003_BackupHistoryEntry  ** pEntry )
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  bool                bMallocOK = true;

  bool                bOK = false;

  /**/

  cn00MyMalloc (sizeof (tcn003_BackupHistoryEntry), pEntry, &bMallocOK);
  if ((*pEntry == NULL) || (!bMallocOK)) {

    nFuncStatus = ERR_MEM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncStatus);

  } else {

    cn32_InitHistoryEntry(*pEntry);

  }

  return nFuncStatus;
} /* end cn32_NewHistoryEntry */

/* PRIVATE FUNCTION cn32_InitHistoryMedium
 * ------------------------------------------------------------------
 * purpose: initialize the history medium struct
 *
 */
static void cn32_InitHistoryMedium(
        tcn003_BackupHistoryMedium   * pMedium)
{
  pMedium->tLine.szKEY[0] = '\0';
  pMedium->tLine.szLABEL[0] = '\0';
  pMedium->tLine.szMEDIANAME[0] = '\0';
  pMedium->tLine.szTYPE[0] = '\0';
  pMedium->tLine.szOVERWRITE[0] = '\0';
  pMedium->tLine.szSIZE[0] = '\0';
  pMedium->tLine.szBLOCKSIZE[0] = '\0';
  pMedium->tLine.szKIND[0] = '\0';
  pMedium->tLine.szLOCATION[0] = '\0';
  
  pMedium->pNext = NULL;
  pMedium->pPrev = NULL;
} /* end cn32_InitHistoryMedium */

/* PRIVATE FUNCTION cn32_NewHistoryMedium
 * ------------------------------------------------------------------
 * purpose: alloc memory for new history medium
 *
 */
static tcn00_Error cn32_NewHistoryMedium (
      char                        * replyData, 
      int                         * replyLen,
      tcn003_BackupHistoryMedium  ** pMedium )
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  bool                bMallocOK = true;

  bool                bOK = false;

  /**/

  cn00MyMalloc (sizeof (tcn003_BackupHistoryMedium), pMedium, &bMallocOK);
  if ((*pMedium == NULL) || (!bMallocOK)) {

    nFuncStatus = ERR_MEM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncStatus);

  } else {

    cn32_InitHistoryMedium(*pMedium);

  }

  return nFuncStatus;
} /* end cn32_NewHistoryMedium */

/* PRIVATE FUNCTION cn32_InitHistory
 * ------------------------------------------------------------------
 * purpose: initialize the history struct
 *
 */
static void cn32_InitHistory(
        tcn003_BackupHistory   * pHistory)
{
  pHistory->pFirst   = NULL;
  pHistory->pLast    = NULL;
  pHistory->pCurrent = NULL;

  strcpy(pHistory->szCols,    BA_HISTORY_COL_ALL_CN003C);
  strcpy(pHistory->szKey,     BA_HISTORY_KEY_ALL_CN003C);
  strcpy(pHistory->szLabel,   BA_HISTORY_LBL_ALL_CN003C);
  strcpy(pHistory->szAction,  BA_HISTORY_ACT_ALL_CN003C);
  strcpy(pHistory->szRestore, BA_HISTORY_RES_ALL_CN003C);
  strcpy(pHistory->szUntil,   BA_HISTORY_UNT_ALL_CN003C);

  pHistory->bForRestore   = false;
  pHistory->bRcOKOnly     = false;
  pHistory->bWithMedia    = false;
  pHistory->bWithEBIDs    = false;
} /* end cn32_InitHistory */

/* PRIVATE FUNCTION cn32_InitHistoryMedia
 * ------------------------------------------------------------------
 * purpose: initialize the history media struct
 *
 */
static void cn32_InitHistoryMedia(
        tcn003_BackupHistoryMedia   * pMedia)
{
  pMedia->pFirst = NULL;
  pMedia->pLast = NULL;
} /* end cn32_InitHistoryMedia */

/* PRIVATE FUNCTION cn32InitHistoryEBIDs
 * ------------------------------------------------------------------
 * purpose: initialize the history media struct
 *
 */
void cn32InitHistoryEBIDs(tcn003_BackupHistoryEBIDList * pEBIDs)
{
    pEBIDs->pFirst=0;
    pEBIDs->pLast=0;

    pEBIDs->pCurrent=0;
} /* end cn32InitHistoryEBIDs */

/* PRIVATE FUNCTION cn32_FreeHistory
 * ------------------------------------------------------------------
 * purpose: frees the history structure
 *
 */
static void cn32_FreeHistory (
      tcn003_BackupHistory   * pHistory)
{
  tcn003_BackupHistoryEntry * pEntry = NULL;
  tcn003_BackupHistoryEntry * pNextEntry = NULL;

  if (pHistory != NULL) {
    pEntry = pHistory->pFirst;

    while (pEntry != NULL) {
      pNextEntry = pEntry->pNext;
      cn00MyFree(pEntry);
      pEntry = pNextEntry;
    } 
  }
} /* end cn32_FreeHistory */

/* PUBLIC FUNCTION cn32FreeHistory
 * ------------------------------------------------------------------
 * purpose: frees the history structure
 *
 */
void cn32FreeHistory (
      VControlDataT * vcontrol)
{
  cn32_FreeHistory ((tcn003_BackupHistory *) vcontrol->pHistorySession);
} /* end cn32FreeHistory */

/* PRIVATE FUNCTION cn32_FreeHistoryMedia
 * ------------------------------------------------------------------
 * purpose: frees the history media structure
 *
 */
static void cn32_FreeHistoryMedia (
      tcn003_BackupHistoryMedia   * pMedia)
{
  tcn003_BackupHistoryMedium * pMedium = NULL;
  tcn003_BackupHistoryMedium * pNextMedium = NULL;

  if (pMedia != NULL) {
    pMedium = pMedia->pFirst;

    while (pMedium != NULL) {
      pNextMedium = pMedium->pNext;
      cn00MyFree(pMedium);
      pMedium = pNextMedium;
    } 
  }
} /* end cn32_FreeHistoryMedia */

/* FUNCTION cn32_FreeHistoryEBID
 * ------------------------------------------------------------------
 * purpose: frees the history EBID structure (counter part of cn32NewHistoryEBID)
 *
 */
static void cn32_FreeHistoryEBID(tcn003_BackupHistoryEBID * &pEBID)
{
    if(0!=pEBID) // if a EBID is given
    {
        if(0!=pEBID->Line.CompleteMemory)           // free any memory used to
            cn00MyFree(pEBID->Line.CompleteMemory); // hold EBID information

        cn00MyFree(pEBID); // free the list element itself

        pEBID=0;
    }
} /* end cn32_FreeHistoryEBID */

/* FUNCTION cn32FreeHistoryEBIDs
 * ------------------------------------------------------------------
 * purpose: frees the history EBID structure (a list of EBIDs)
 *
 */
void cn32FreeHistoryEBIDs(tcn003_BackupHistoryEBIDList * pEBIDs)
{
    if(0!=pEBIDs) // if a list of EBIDs is given
    {
        tcn003_BackupHistoryEBID * pEBID=pEBIDs->pFirst;
        tcn003_BackupHistoryEBID * pNextEBID=0;

        while(0!=pEBID) // while something in the list
        {
            pNextEBID=pEBID->pNext;

            cn32_FreeHistoryEBID(pEBID);

            pEBID=pNextEBID; // go on to the next list element 
        }
    }
} /* end cn32FreeHistoryEBIDs */

/* FUNCTION cn32AddToEndOfHistoryEBIDs
 * ------------------------------------------------------------------
 * purpose: inserts *pEBID at the end of *pEBIDs
 *
 */
void cn32AddToEndOfHistoryEBIDs(tcn003_BackupHistoryEBID * pEBID,
                                tcn003_BackupHistoryEBIDList * pEBIDs)
{
    if(NULL==pEBIDs->pFirst)
        pEBIDs->pFirst=pEBID;

    if(NULL!=pEBIDs->pLast)
        pEBIDs->pLast->pNext=pEBID;

    pEBID->pNext=NULL;
    pEBID->pPrev=pEBIDs->pLast;

    pEBIDs->pLast=pEBID;
} /* end cn32AddToEndOfHistoryEBIDs */

/* PRIVATE FUNCTION cn32_ReadHistoryFile
 * ------------------------------------------------------------------
 * purpose: reading contents of history
 *
 */
static tcn00_Error cn32_ReadHistoryFile (
      VControlDataT         * vcontrol,
      char                  * replyData,
      int                   * replyLen,
      tcn003_BackupHistory  * pHistory)
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  tsp_int4            hHistoryFile;

  tsp5_rte_fileerror  rteFileError;

  tcn003_BackupHistoryEntry * pEntry = NULL;

  /**/
 
  cn32_InitHistory(pHistory);
  
  nFuncStatus = cn32_OpenHistoryFile(vcontrol, replyData, replyLen, &hHistoryFile);
  if (nFuncStatus == OK_CN00) {

    do {

      nFuncStatus = cn32_NewHistoryEntry(replyData, replyLen, &pEntry);
      if (nFuncStatus == OK_CN00) {
        nFuncStatus = cn32_ReadHistoryLine(hHistoryFile, &(pEntry->tLine), &rteFileError);
      }

      // copy timestamp without format characters
      if (nFuncStatus == OK_CN00) {
        int nDest   = 0;
        int nSource = 0;

        if (strncmp(BA_HISTORY_ACTHIS_CN003C, pEntry->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0) {
          strcpy(pEntry->tLine.szDBSTAMP1, pEntry->tLine.szSTARTDATE);
        } // end if

        while (pEntry->tLine.szDBSTAMP1[nSource] != '\0') {
          if (isdigit(pEntry->tLine.szDBSTAMP1[nSource])) {
            pEntry->szStamp[nDest] = pEntry->tLine.szDBSTAMP1[nSource];
            ++nDest;
          } // end if
          ++nSource;
        } // while
        pEntry->szStamp[nDest] = pEntry->tLine.szDBSTAMP1[nSource];
      } // end if

      if (nFuncStatus == OK_CN00) {
        if (pHistory->pFirst == NULL) {
          pHistory->pFirst = pEntry;
        } 
        if (pHistory->pLast != NULL) {
          pHistory->pLast->pNext = pEntry;
        } 
        pEntry->pNext = NULL;
        pEntry->pPrev = pHistory->pLast;
        pHistory->pLast = pEntry;
      }

    } while ( ( rteFileError.sp5fe_result  == vf_ok             ) &&
              ( rteFileError.sp5fe_warning == sp5vfw_no_warning ) &&
              ( nFuncStatus                == OK_CN00           )    );

    if (rteFileError.sp5fe_result == vf_eof) {
      nFuncStatus = cn32_CloseFile(replyData, replyLen, hHistoryFile);
    } else {
      nFuncStatus = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
    }
  }

  /**/

  return nFuncStatus;
} /* end cn32_ReadHistoryFile */

/* PRIVATE FUNCTION cn32_ReadHistoryMediaFile
 * ------------------------------------------------------------------
 * purpose: reading contents of history media
 *
 */
static tcn00_Error cn32_ReadHistoryMediaFile (
      VControlDataT              * vcontrol,
      char                       * replyData,
      int                        * replyLen,
      tcn003_BackupHistoryMedia  * pHistoryMedia)
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  tsp_int4            hHistoryMediaFile;

  tsp5_rte_fileerror  rteFileError;

  tcn003_BackupHistoryMedium * pMedium = NULL;

  /**/
 
  cn32_InitHistoryMedia(pHistoryMedia);
  
  nFuncStatus = cn32_OpenHistoryMediaFile(vcontrol, replyData, replyLen, &hHistoryMediaFile);
  if (nFuncStatus == OK_CN00) {

    do {

      nFuncStatus = cn32_NewHistoryMedium(replyData, replyLen, &pMedium);
      if (nFuncStatus == OK_CN00) {
        nFuncStatus = cn32_ReadHistoryMediaLine(hHistoryMediaFile, &(pMedium->tLine), &rteFileError);
      }

      if (nFuncStatus == OK_CN00) {
        if (pHistoryMedia->pFirst == NULL) {
          pHistoryMedia->pFirst = pMedium;
        } 
        if (pHistoryMedia->pLast != NULL) {
          pHistoryMedia->pLast->pNext = pMedium;
        } 
        pMedium->pNext = NULL;
        pMedium->pPrev = pHistoryMedia->pLast;
        pHistoryMedia->pLast = pMedium;
      }

    } while ( ( rteFileError.sp5fe_result  == vf_ok             ) &&
              ( rteFileError.sp5fe_warning == sp5vfw_no_warning ) &&
              ( nFuncStatus                == OK_CN00           )    );

    if (rteFileError.sp5fe_result == vf_eof) {
      nFuncStatus = cn32_CloseFile(replyData, replyLen, hHistoryMediaFile);
    } else {
      nFuncStatus = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);
    }
  }

  /**/

  return nFuncStatus;
} /* end cn32_ReadHistoryMediaFile */

/* PRIVATE FUNCTION cn32_ReadHistoryEBIDFile
 * ------------------------------------------------------------------
 * purpose: reading contents of history EBID (external backup
 * history, dbm.ebf)
 *
 */
static tcn00_Error cn32_ReadHistoryEBIDFile (
      VControlDataT                * vcontrol,
      char                         * replyData,
      int                          * replyLen,
      tcn003_BackupHistoryEBIDList * pHistoryEBIDs)
{
    tcn00_Error	               nFuncStatus=OK_CN00;
    tsp_int4                   hHistoryEBIDsFile;
    tsp5_rte_fileerror         rteFileError;
    tcn003_BackupHistoryEBID * pEBID=NULL;

    cn32InitHistoryEBIDs(pHistoryEBIDs);

    nFuncStatus=cn32_OpenFile(FGET_BACKEBF_CN42, vcontrol, replyData, replyLen, &hHistoryEBIDsFile);

    if(OK_CN00==nFuncStatus)
    {
        do
        {
            nFuncStatus=cn32NewHistoryEBID(replyData, replyLen, &pEBID);

            if(OK_CN00==nFuncStatus)
            {
                nFuncStatus=cn32_ReadHistoryEBIDLine(replyData, replyLen, hHistoryEBIDsFile, &(pEBID->Line), &rteFileError);

                if(OK_CN00==nFuncStatus) // if all ok, insert (*pEBID) into (*pHistoryEBIDs)
                {
                    pEBID->Status=EBIDStatusUnavailable; //(*pEBID) was found in the history -> at least an unavailable external save
                    cn32AddToEndOfHistoryEBIDs(pEBID, pHistoryEBIDs); //now insert (*pEBID) into (*pHistoryEBIDs)
                }
                else
                    cn32_FreeHistoryEBID(pEBID); //free the memory before it's lost (if we get here pEBID is not added to the list)
            }
        }
        while((rteFileError.sp5fe_result  == vf_ok             ) &&
              (rteFileError.sp5fe_warning == sp5vfw_no_warning ) &&
              (nFuncStatus                == OK_CN00           )    );

        if(vf_eof==rteFileError.sp5fe_result)
            nFuncStatus=cn32_CloseFile(replyData, replyLen, hHistoryEBIDsFile);
        else
        {
            nFuncStatus = ERR_RTE_CN00;
            cn90AnswerRTEError (replyData, replyLen, nFuncStatus, rteFileError.sp5fe_text, rteFileError.sp5fe_text.length(), rteFileError.sp5fe_result);

            cn32_CloseFile(replyData, replyLen, hHistoryEBIDsFile);
        }
    }
    else
        nFuncStatus=OK_CN00; // if external backup history and its copy are
                             // present -> it's no error, all are EBIDs are simply unknown

    return nFuncStatus;
} /* end cn32_ReadHistoryEBIDFile */

/* PRIVATE FUNCTION cn32_CheckAvailabilityOfExternalBackups
 * ------------------------------------------------------------------
 * purpose: check the availability of external backups and set
 *          status of EBIDs from the list pHistoryEBIDs respective
 */
static tcn00_Error cn32_CheckAvailabilityOfExternalBackups(                                                   
                        VControlDataT                   * vcontrol,
                        char                            * replyData,
                        int                             * replyLen,
                        const tcn003_BackupHistoryMedia * pHistoryMedia,
                        tcn003_BackupHistoryEBIDList    * pHistoryEBIDs)
{
    tcn00_Error nFuncStatus=OK_CN00;

    if(0!=pHistoryMedia &&           // we must have a list of media
       0!=pHistoryMedia->pFirst &&   // with at least one media and
       0!=pHistoryEBIDs &&           // we must have a list of known EBIDs
       0!=pHistoryEBIDs->pFirst)     // with at least one EBID
    {
        bool      * AllreadyChecked;
        const int   NumberOfBackupTools=tcn35_BackupToolConnectorEnum::NumberOfBackupToolConnectors();
        bool        bMallocOK;

        cn00MyMalloc(sizeof(bool)*(NumberOfBackupTools+1), //+1 to have also space for tcn35_BackupToolConnectorEnum::None
                     &AllreadyChecked,
                     &bMallocOK);

        if(0==AllreadyChecked || (!bMallocOK))
        {
            nFuncStatus=ERR_MEM_CN00;
            cn90AnswerIError(replyData, replyLen, nFuncStatus);
        }
        else
        {
            int   i;
            const tcn003_BackupHistoryMedium * pMedium=pHistoryMedia->pFirst;
            const tcn35_SaveInfo Info(replyData,
                                      replyLen,
                                      vcontrol->dbname,
                                      0,                // no support of migrations with history at the moment
                                      0);               // no support of migrations with history at the moment

            //init AllreadyChecked
            for(i=0; i<NumberOfBackupTools; ++i)
                AllreadyChecked[i]=false;

            AllreadyChecked[i]=true;

            //determine the available backups for every used external backup tool (mentioned in pHistoryMedia)
            //and check against all backups known through pHistoryEBIDs
            if(Info.IsGood)
            {
                while(0!=pMedium && OK_CN00==nFuncStatus) //for all used media 
                {
                    tcn35_BackupToolConnectorEnum BackupToolType(&(pMedium->tLine));    //get the Type of backup tool for the medium (can be also "None")

                    if(false==AllreadyChecked[int(BackupToolType)]) //already checked that tool?
                    {
                        tcn35_BackupToolConnector * CurrentTool=0;

                        AllreadyChecked[int(BackupToolType)]=true; //do not check this tool again

                        if(!cn35_ConstructBackupToolConnector(tcn35_BackupToolConnector::GiveBackupInfo,
                                                              BackupToolType,
                                                              Info,
                                                              CurrentTool))
                        {
                            nFuncStatus=TheDBMSrvErrorHandler_cn36.GiveLastError();
                        }
                        else
                        {
                            tcn003_BackupHistoryEBIDList AvailableEBIDs;

                            cn32InitHistoryEBIDs(&AvailableEBIDs);

                            if(!CurrentTool->ListEBIDsOfAvailableBackups(AvailableEBIDs))   // do not use CurrentTool->GiveLastError() if Tool is not installed anymore
                                nFuncStatus=CurrentTool->GiveLastError();                   // do not ignore "real" errors

                            if(OK_CN00==nFuncStatus &&      //all ok
                               0!=AvailableEBIDs.pFirst)    //have got at least one available EBID from the tool
                            {
                                tcn003_BackupHistoryEBID *pEBIDFromHistory=pHistoryEBIDs->pFirst;   // pointer for looping through pHistoryEBIDs
                                
                                while(0!=pEBIDFromHistory)  // test every EBID from history
                                {
                                    const tcn003_BackupHistoryMedium * pMediumH=pHistoryMedia->pFirst;  // pointer for looping through pHistoryMedia
                                    int   EBIDCreatedByCurrentTool=0;

                                    // check that EBID belongs to the current tool (specified by *pMedium and stored in BackupToolType)
                                    while(!EBIDCreatedByCurrentTool && 0!=pMediumH)
                                    {
                                        if( !strcmp(pEBIDFromHistory->Line.szKey, pMediumH->tLine.szKEY) &&     // if key and
                                            !strcmp(pEBIDFromHistory->Line.szLabel, pMediumH->tLine.szLABEL) && // label are equal, pMediumH points to a medium used for the EBID from history
                                            BackupToolType==tcn35_BackupToolConnectorEnum(&(pMediumH->tLine)))  // the tool specified through that medium must be equal to the CurrentTool
                                        {
                                            EBIDCreatedByCurrentTool=1;
                                        }
                                    
                                        pMediumH=pMediumH->pNext;
                                    }

                                    if(EBIDCreatedByCurrentTool)    // if EBID belongs to the current tool, we can determine the status with AvailableEBIDs
                                    {
                                        tcn003_BackupHistoryEBID *pEBIDFromTool=AvailableEBIDs.pFirst;  // pointer for looping through AvailableEBIDs

                                        while(pEBIDFromHistory->Status!=EBIDStatusAvailable && 0!=pEBIDFromTool)        // test until EBID is marked as available or end of list is reached
                                        {
                                            if(0==strcmp(pEBIDFromHistory->Line.szEBID, pEBIDFromTool->Line.szEBID))    // if the EBIDs are equal
                                                pEBIDFromHistory->Status=EBIDStatusAvailable;                           // (*pEBIDFromHistory) is available

                                            pEBIDFromTool=pEBIDFromTool->pNext;
                                        }
                                    }
                    
                                    pEBIDFromHistory=pEBIDFromHistory->pNext;
                                }
                            }
            
                            cn32FreeHistoryEBIDs(&AvailableEBIDs);
                        }

                        if(0!=CurrentTool)
                            delete CurrentTool;
                    }

                    pMedium=pMedium->pNext;
                }
            }
            else
                nFuncStatus=TheDBMSrvErrorHandler_cn36.GiveLastError();

            // free the allocated memory
            cn00MyFree(AllreadyChecked);
        }
    }

    return nFuncStatus;
}/* end cn32_CheckAvailabilityOfExternalBackups */

/* PRIVATE FUNCTION cn32_AllocHistorySession
 * ------------------------------------------------------------------
 * purpose: allocate new memory for backup session
 *
 */
static tcn00_Error cn32_AllocHistorySession (
      VControlDataT                 * vcontrol,
      char                          * replyData,
      int                           * replyLen,
      tcn003_BackupHistorySession   ** pHistorySession )
{
  tcn00_Error   nFuncStatus  = OK_CN00;

  bool          bMallocOK = true;

  /**/

  if (nFuncStatus == OK_CN00) {

    if (*pHistorySession == NULL) {
      cn00MyMalloc (sizeof (tcn003_BackupHistorySession), pHistorySession, &bMallocOK);
    }

    if ((*pHistorySession == NULL) || (!bMallocOK)) {
      nFuncStatus = ERR_MEM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    } else {
      vcontrol->pHistorySession = (void *) *pHistorySession;

      cn32_InitHistory(&((*pHistorySession)->tHistory));
      cn32_InitHistoryMedia(&((*pHistorySession)->tMedia));
      cn32InitHistoryEBIDs(&((*pHistorySession)->tEBIDs));
    } 
  }

  return nFuncStatus;
} /* end cn32_AllocHistorySession */

/* PRIVATE FUNCTION cn32_HistorySession
 * ------------------------------------------------------------------
 * purpose: get the history session from global struct and check
 *
 */
static tcn00_Error cn32_HistorySession (
        VControlDataT                 * vcontrol,
        char                          * replyData,
        int                           * replyLen,
        tcn003_BackupHistorySession   ** pHistorySession )
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  /**/

  *pHistorySession = (tcn003_BackupHistorySession *) vcontrol->pHistorySession;

  /* is there an active history session */
  if (nFuncStatus == OK_CN00) {
    if (*pHistorySession == NULL) {
      nFuncStatus = ERR_NOHISTORYSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    } 
  }
 
  /**/

  return nFuncStatus;
} /* end cn32_HistorySession */

/* PRIVATE FUNCTION cn32_FreeHistorySession
 * ------------------------------------------------------------------
 * purpose: frees the history session from the VControlDataT structure
 *
 */
static void cn32_FreeHistorySession(
      VControlDataT     * vcontrol)
{
  tcn003_BackupHistorySession  * pHistorySession=NULL;

  /**/

  if (vcontrol != NULL) {
    pHistorySession =  (tcn003_BackupHistorySession *) vcontrol->pHistorySession;
  }

  if (pHistorySession != NULL) {

    cn32_FreeHistory(&(pHistorySession->tHistory));
    cn32_FreeHistoryMedia(&(pHistorySession->tMedia));
    cn32FreeHistoryEBIDs(&(pHistorySession->tEBIDs));

    cn00MyFree(pHistorySession);
    vcontrol->pHistorySession = NULL;
  }
} /* end cn32_FreeHistorySession */


/* PRIVATE FUNCTION cn32_AllocEBIDList
 * ------------------------------------------------------------------
 * purpose: allocate new memory for a EBID list
 *
 */
static tcn00_Error cn32_AllocEBIDList(VControlDataT                 * vcontrol,
                                      char                          * replyData,
                                      int                           * replyLen,
                                      tcn003_BackupHistoryEBIDList ** pEBIDList)
{
    tcn00_Error rc=OK_CN00;
    bool        bMallocOK=true;

    if(0!=vcontrol)
    {
        if(0==(*pEBIDList))
            cn00MyMalloc(sizeof(tcn003_BackupHistoryEBIDList), pEBIDList, &bMallocOK);

        if(0==(*pEBIDList) || (!bMallocOK))
        {
            rc=ERR_MEM_CN00;
            cn90AnswerIError(replyData, replyLen, rc);
        }
        else
        {
          vcontrol->pEBIDList=(void *) *pEBIDList;
          cn32InitHistoryEBIDs(*pEBIDList);
        }
    }

    return rc;
} /* end cn32_AllocEBIDList */

/* PRIVATE FUNCTION cn32_EBIDList
 * ------------------------------------------------------------------
 * purpose: get the EBID list from global struct and check
 *
 */
static tcn00_Error cn32_EBIDList(VControlDataT                 * vcontrol,
                                 char                          * replyData,
                                 int                           * replyLen,
                                 tcn003_BackupHistoryEBIDList ** pEBIDList)
{
    tcn00_Error	rc=OK_CN00;

    *pEBIDList=(tcn003_BackupHistoryEBIDList *) vcontrol->pEBIDList;

    /* is there an active history session */
    if(0==(*pEBIDList))
    {
        rc=ERR_NOEBIDLIST_CN00;
        cn90AnswerIError(replyData, replyLen, rc);
    } 

    return rc;
} /* end cn32_EBIDList */

/* PRIVATE FUNCTION cn32_FreeEBIDList
 * ------------------------------------------------------------------
 * purpose: frees the EBID List from the VControlDataT structure
 *
 */
static void cn32_FreeEBIDList(VControlDataT * vcontrol)
{
    tcn003_BackupHistoryEBIDList * pEBIDList=0;

    if(0!=vcontrol)
        pEBIDList=(tcn003_BackupHistoryEBIDList *)vcontrol->pEBIDList;

    if(0!=pEBIDList)
    {
        cn32FreeHistoryEBIDs(pEBIDList);

        cn00MyFree(pEBIDList);
        vcontrol->pEBIDList=0;
    }
} /* end cn32_FreeEBIDList */
  
/* PUBLIC FUNCTION cn32FreeEBIDList
 * ------------------------------------------------------------------
 * purpose: frees the EBID list
 *
 */
void cn32FreeEBIDList(VControlDataT * vcontrol)
{
    cn32_FreeEBIDList(vcontrol);
} /* end cn32FreeEBIDList */

/* PRIVATE FUNCTION cn32_GetEntryLine
 * ------------------------------------------------------------------
 * purpose: build a line from one history entry
 *
 */
static void cn32_GetEntryLine(
      tcn003_BackupHistoryEntry  * pEntry,
      char                       * szCols, 
      char                       * szLine)

{
  char * szSep = NULL;;
  char * szNextCol = NULL;;
  char   szCol[32];

  /**/

  szLine[0] = '\0';

  if (pEntry != NULL) {
    if (strcmp(szCols, BA_HISTORY_COL_ALL_CN003C) == 0) {
      strcpy(szLine, pEntry->tLine.szKEY);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szLABEL);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szACTION);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szDBSTAMP1);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szDBSTAMP2);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szSTARTDATE);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szSTOPDATE);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szSTARTLOGPAGE);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szSTOPLOGPAGE);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szLOGREQUIRED);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szMEDIANAME);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szPAGESTRANSFERED);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szVOLUMES);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szRETURNCODE);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
      strcat(szLine, pEntry->tLine.szRETURNTEXT);      
      strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    } else {
      
      szNextCol = szCols;

      do {
        szSep = strchr(szNextCol, ',');
        if (szSep) {
          strncpy(szCol, szNextCol, szSep - szNextCol);
          szCol[szSep - szNextCol] = '\0';
          szNextCol = szSep + 1;
        } else {
          strcpy(szCol, szNextCol);
        }
      
        if (strcmp(szCol, BA_HISTORY_COL_KEY_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szKEY);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_LBL_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szLABEL);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_ACT_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szACTION);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_ST1_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szDBSTAMP1);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_ST2_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szDBSTAMP2);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_STA_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szSTARTDATE);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_STO_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szSTOPDATE);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_FLO_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szSTARTLOGPAGE);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_LLO_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szSTOPLOGPAGE);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_LOG_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szLOGREQUIRED);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_MED_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szMEDIANAME);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_PAG_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szPAGESTRANSFERED);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_VOL_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szVOLUMES);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_RCO_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szRETURNCODE);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
        if (strcmp(szCol, BA_HISTORY_COL_RTX_CN003C) == 0) {
          strcat(szLine, pEntry->tLine.szRETURNTEXT);      
          strcat(szLine, BA_HISTORY_COLSEP_CN003C);
        }
      } while (szSep);

    }

  }
} /* end cn32_GetEntryLine */

/* PRIVATE FUNCTION cn32_GetMediumLine
 * ------------------------------------------------------------------
 * purpose: build a line from one history medium
 *
 */
static void cn32_GetMediumLine(
      tcn003_BackupHistoryMedium * pMedium,
      char                       * szLine)

{
  char * szSep = NULL;;
  char * szNextCol = NULL;;

  /**/

  szLine[0] = '\0';

  if (pMedium != NULL) {
    strcpy(szLine, pMedium->tLine.szMEDIANAME);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szTYPE);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szOVERWRITE);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szSIZE);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szBLOCKSIZE);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szKIND);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
    strcat(szLine, pMedium->tLine.szLOCATION);      
    strcat(szLine, BA_HISTORY_COLSEP_CN003C);      
  } 
} /* end cn32_GetMediumLine */

/* PRIVATE FUNCTION cn32_GetEBIDAsLine
 * ------------------------------------------------------------------
 * purpose: build a line from one history EBID
 *
 */
static void cn32_GetEBIDAsLine(const tcn003_BackupHistoryEBID * pEBID,
                        char                           * Line,
                        size_t                           MaxLengthOfLine,
                        int                              WithExtendedInformation)
{
    Line[0]='\0';

    if(0!=pEBID)
    {
        switch(pEBID->Status)   // start line with status
        {
            default:
            case EBIDStatusUnknown:     cn36_StrNCpy(Line, BA_HISTORY_EBIDSTATUS_UNKNOWN_CN003C, MaxLengthOfLine);
                break;
            case EBIDStatusNone:        cn36_StrNCpy(Line, BA_HISTORY_EBIDSTATUS_NONE_CN003C, MaxLengthOfLine);
                break;
            case EBIDStatusAvailable:   cn36_StrNCpy(Line, BA_HISTORY_EBIDSTATUS_AVAILABLE_CN003C, MaxLengthOfLine);
                break;
            case EBIDStatusUnavailable: cn36_StrNCpy(Line, BA_HISTORY_EBIDSTATUS_UNAVAILABLE_CN003C, MaxLengthOfLine);
                break;
        }

        cn36_StrNCat(Line, BA_HISTORY_COLSEP_CN003C, MaxLengthOfLine);  // separate fields
        cn36_StrNCat(Line, pEBID->Line.szEBID, MaxLengthOfLine);        // continue with the EBID itself

        if(WithExtendedInformation)
        {
            cn36_StrNCat(Line, BA_HISTORY_COLSEP_CN003C, MaxLengthOfLine);  // separate fields
            cn36_StrNCat(Line, pEBID->Line.szType, MaxLengthOfLine);        // continue with type of save ("DATA RECOVERY", ...)
            cn36_StrNCat(Line, BA_HISTORY_COLSEP_CN003C, MaxLengthOfLine);  // separate fields
            cn36_StrNCat(Line, pEBID->Line.szDateTime, MaxLengthOfLine);    // continue with date of save ("yyyy-mm-dd hh:mm:ss")
        }

        cn36_StrNCat(Line, BA_HISTORY_COLSEP_CN003C, MaxLengthOfLine);  // mark end with separator
    } 
} /* end cn32_GetEBIDAsLine */

/* PRIVATE FUNCTION cn32_ListMedia
 * ------------------------------------------------------------------
 * purpose: list media for one history entry
 *
 */
static bool cn32_ListMedia(
      tcn003_BackupHistoryEntry  *  pEntry,
      tcn003_BackupHistoryMedia  *  pMedia,
      char                       *& pCurrent,
      size_t                      & nSizeAvailable )
{
  char                         szLine[BA_HISTORY_LINE_LENGTH_MXCN003];
  size_t                       nLen     = 0;
  bool                         bGetLine = false;
  bool                         bFull    = false;
  tcn003_BackupHistoryMedium * pMedium  = NULL;

  /**/

  if ( (pMedia != NULL) && (pEntry != NULL) ) {
    if (pMedia->pFirst != NULL) {
      pMedium = pMedia->pFirst;

      do {
  
        szLine[0] = '\0';
        bGetLine = false;

        if ( (strcmp(pEntry->tLine.szKEY, pMedium->tLine.szKEY) == 0)     && 
             (strcmp(pEntry->tLine.szLABEL, pMedium->tLine.szLABEL) == 0)    ) {
          bGetLine = true;
        } // end if

        if (bGetLine) {
          cn32_GetMediumLine(pMedium, szLine);
          nLen=strlen("M:")+strlen(szLine)+strlen(LINE_SEPSTRING_CN00);

          if (nLen > 0) {
            if (nLen < nSizeAvailable) {
              sprintf(pCurrent, "M:%s%s", szLine, LINE_SEPSTRING_CN00);
              nSizeAvailable = nSizeAvailable - nLen;
              pCurrent       = pCurrent       + nLen;
            } else {
              bFull = true;
            } // end if
          } // end if
        } // end if
        
        pMedium = pMedium->pNext;

      } while ( (pMedium != NULL) &&
                (bFull == false)     );
    } // end if
  } // end if

  /**/

  return bFull;
} /* end cn32_ListMedia */

/* PRIVATE FUNCTION cn32_AppendEBID
 * ------------------------------------------------------------------
 * purpose: append line with EBID at replyData
 *
 */ 
static bool cn32_AppendEBID(const tcn003_BackupHistoryEBID *  pEBID,
                           char                            *& pCurrent,
                           size_t                           & nSizeAvailable )
{
    bool    bFull = false;
    size_t  nLen;
    char    Line[BA_HISTORY_LINE_LENGTH_MXCN003];

    Line[0]='\0';
    cn32_GetEBIDAsLine(pEBID, Line, BA_HISTORY_LINE_LENGTH_MXCN003-1, 0);   //without extra information (date, time and savetype)

    nLen=strlen("E:")+strlen(Line)+strlen(LINE_SEPSTRING_CN00);

    if(nLen < nSizeAvailable)
    {
        sprintf(pCurrent, "E:%s%s", Line, LINE_SEPSTRING_CN00);
        nSizeAvailable = nSizeAvailable - nLen;
        pCurrent       = pCurrent       + nLen;
    }
    else
    {
        bFull = true;
    }

    return bFull;
} /* end cn32_AppendEBID */

/* PRIVATE FUNCTION cn32_ListEBIDsForHistoryEntry
 * ------------------------------------------------------------------
 * purpose: list EBID or EBIDs for one history entry
 *
 */
static bool cn32_ListEBIDsForHistoryEntry(
      tcn003_BackupHistoryEntry    *  pEntry,
      tcn003_BackupHistoryMedia    *  pMedia,
      tcn003_BackupHistoryEBIDList *  pEBIDs,
      char                         *& pCurrent,
      size_t                        & nSizeAvailable )
{
    tcn00_Error   nFuncStatus=OK_CN00;
    bool          FoundEBIDForEntry=false;
    bool          bFull=false;

    if(0!=pEntry)  // an history entry is given
    {
        //try to find the EBID corresponding to pEntry in the list of EBIDs
        if(0!=pEBIDs)   // list of EBIDs is given
        {
            tcn003_BackupHistoryEBID * pEBID=pEBIDs->pFirst;

            while(0!=pEBID && !(bFull)) //while not at end of list and replyData not full
            {
                if(0==strcmp(pEntry->tLine.szKEY, pEBID->Line.szKey) &&    // if key and
                   0==strcmp(pEntry->tLine.szLABEL, pEBID->Line.szLabel))  // label of pEntry and pEBID are equal 
                {
                    FoundEBIDForEntry=true; //we have found an EBID for (*pEntry)

                    bFull = cn32_AppendEBID(pEBID, pCurrent, nSizeAvailable);
                }
        
                pEBID=pEBID->pNext;
            }  // end while
        }

        //if no corresponding EBID from the list of EBIDs could be found, decide between "Unkown" and "None" as status
        if(!FoundEBIDForEntry && !(bFull) && 0!=pMedia)    //all ok and pEntry not found in pEBIDs -> look for pEBID in pMedia
        {
            tcn003_BackupHistoryMedium *pMedium=pMedia->pFirst;

            while(0!=pMedium && !(bFull))
            {
                if(0==strcmp(pEntry->tLine.szKEY, pMedium->tLine.szKEY) &&      // if key and
                   0==strcmp(pEntry->tLine.szLABEL, pMedium->tLine.szLABEL))    // label of pEntry and pMedium are equal
                {
                    tcn003_BackupHistoryEBID TemporaryEBID;

                    FoundEBIDForEntry=true; //we have found an EBID ("NONE") for (*pEntry)
                
                    cn32_InitHistoryEBID(&TemporaryEBID);

                    if(cn35_IsExternalBackupTool(&(pMedium->tLine)))
                        TemporaryEBID.Status=EBIDStatusUnknown; // if it was a tool ->EBID is unknown
                    else
                        TemporaryEBID.Status=EBIDStatusNone; // if it was no tool ->no EBID at all

                    bFull = cn32_AppendEBID(&TemporaryEBID, pCurrent, nSizeAvailable);
                }

                pMedium=pMedium->pNext;
            } // end while
        }
    }

    return bFull;
} /* end cn32_ListEBIDsForHistoryEntry */
                    
/* PRIVATE FUNCTION cn32_PrintEntry
 * ------------------------------------------------------------------
 * purpose: build a line from one history entry
 *
 */
static bool cn32_PrintEntry(
      tcn003_BackupHistoryEntry    * pEntry,
      tcn003_BackupHistory         * pHistory,
      tcn003_BackupHistoryMedia    * pMedia,
      tcn003_BackupHistoryEBIDList * pEBIDs,
      char                         * pCurrent,
      size_t                         nSizeAvailable )
{
  bool   bFull = false;
  char   szLine[BA_HISTORY_LINE_LENGTH_MXCN003];
  size_t nLen  = 0;

  cn32_GetEntryLine(pEntry, pHistory->szCols, szLine);

  nLen = strlen(szLine) + strlen(LINE_SEPSTRING_CN00);

  if ( nLen > 0 ) {
    if ( nLen < nSizeAvailable ) {

      sprintf(pCurrent, "%s%s", szLine, LINE_SEPSTRING_CN00);
      nSizeAvailable = nSizeAvailable - nLen;
      pCurrent       = pCurrent       + nLen;

      if (pHistory->bWithMedia) {
        bFull = cn32_ListMedia(pEntry, pMedia, pCurrent, nSizeAvailable);
      } // end if

      if(pHistory->bWithEBIDs && !bFull) {
        bFull = cn32_ListEBIDsForHistoryEntry(pEntry, pMedia, pEBIDs, pCurrent, nSizeAvailable);
      } // end if

    } else {
      bFull = true;
    } // end if

  } // end if

  return bFull;
} // end cn32_PrintEntry

/* PRIVATE FUNCTION cn32_ListHistory
 * ------------------------------------------------------------------
 * purpose: list contents of history structure
 *
 */
static tcn00_Error cn32_ListHistory(
      char                         * replyData,
      int                          * replyLen,
      int                            replyLenMax,
      tcn003_BackupHistory         * pHistory,
      tcn003_BackupHistoryMedia    * pMedia,
      tcn003_BackupHistoryEBIDList * pEBIDs,
      tcn003_BackupHistoryEntry   ** pEntry)
{
  tcn00_Error   nFuncStatus  = OK_CN00;

  char          * pCurrPos;
  char          * pFlagPos;

  char          szLine[BA_HISTORY_LINE_LENGTH_MXCN003];
  long          nLen = 0;
  bool          bGetLine     = false;

  bool          bFull                 = false;
  bool          bFirstDataSavePassed  = false;
  bool          bSecondDataSavePassed = false;

  tcn003_BackupHistoryEntry * pLog    = NULL;
  tcn003_BackupHistoryEntry * pTemp   = NULL;
  tcn003_BackupHistoryEntry * pData   = NULL;
  
  /**/

  sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  pCurrPos = replyData + strlen(replyData);

  sprintf(pCurrPos, "%s", BA_HISTORY_CONT_FLAG_CONTINUE_CN003C);
  pFlagPos = pCurrPos;
  pCurrPos  = pCurrPos + BA_HISTORY_CONT_LEN_CN003C;

  if (*pEntry != NULL) {

    do {

      szLine[0] = '\0';
      bGetLine = true;

      // check the key (full match)
      if (( strcmp(pHistory->szKey, BA_HISTORY_KEY_ALL_CN003C) != 0) &&
          ( strcmp(pHistory->szKey, (*pEntry)->tLine.szKEY   ) != 0)    ) {
        // key doesn't match
        bGetLine = false;
      } // end if

      // check the label (pattern match <label>*) (execpt is is a history line)
      if (( strcmp(pHistory->szLabel, BA_HISTORY_LBL_ALL_CN003C)                                           != 0) &&
          ( strncmp(pHistory->szLabel, (*pEntry)->tLine.szLABEL, strlen(pHistory->szLabel))                != 0) && 
          ( strncmp(BA_HISTORY_ACTHIS_CN003C, (*pEntry)->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) != 0)    ) {
        // label doesn't match
        bGetLine = false;
      } // end if

      // check the action (pattern match <action>*) (execpt is is a history line)
      if (( strcmp(pHistory->szAction, BA_HISTORY_ACT_ALL_CN003C)                                          != 0) &&
          ( strncmp(pHistory->szAction, (*pEntry)->tLine.szACTION, strlen(pHistory->szAction))             != 0) &&
          ( strncmp(BA_HISTORY_ACTHIS_CN003C, (*pEntry)->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) != 0)    ) {
        // action doesn't match
        bGetLine = false;
      } // end if

      // check the until timestamp (pattern match <timestamp>*)
      if (( strcmp(pHistory->szUntil, BA_HISTORY_UNT_ALL_CN003C) != 0) &&
            strncmp(pHistory->szUntil, (*pEntry)->szStamp, strlen(pHistory->szUntil)) < 0) {
        // timestamp doesn't match
        bGetLine = false;
      } // end if

      // check existing log against log save when restore manager is active
      if ( pHistory->bForRestore                                                                                 && 
           (strncmp(BA_HISTORY_LBLLOG_CN003C, (*pEntry)->tLine.szLABEL,  strlen(BA_HISTORY_LBLLOG_CN003C)) == 0) && 
           (pHistory->nFirstLogPage != BA_HISTORY_INVLOG_CN003C                                                ) &&
           (pHistory->nUsedLogPage  >= pHistory->nFirstLogPage                                                 ) &&
           (pHistory->nFirstLogPage <= atoi((*pEntry)->tLine.szSTARTLOGPAGE)                                   )    ) {
        bGetLine = false;
      } // end if

      // check rc 
      if ( pHistory->bRcOKOnly                         && 
          ( atoi((*pEntry)->tLine.szRETURNCODE) != 0 )    ) {
        bGetLine = false;
      } // end if

      // skip datasaves during restore manager
      if ( pHistory->bForRestore &&  
           (pData == NULL)       &&
           (strncmp(BA_HISTORY_LBLDAT_CN003C, (*pEntry)->tLine.szLABEL,  strlen(BA_HISTORY_LBLDAT_CN003C)) == 0)  ) {
        if (bFirstDataSavePassed) {
          bGetLine = false;
          bSecondDataSavePassed = true;
        } else {
          bFirstDataSavePassed = true;
          
          pData = *pEntry;

          // look for logs before
          pTemp = (*pEntry)->pPrev;
          while (pTemp != NULL) {

            if (strncmp(BA_HISTORY_LBLLOG_CN003C, pTemp->tLine.szLABEL,  strlen(BA_HISTORY_LBLLOG_CN003C)) == 0) {
              if ( atoi(pData->tLine.szSTARTLOGPAGE) < atoi(pTemp->tLine.szSTOPLOGPAGE)) {
                pLog = pTemp;
              } else {
                pTemp = NULL;
              } // end if
            } else if ( strncmp(BA_HISTORY_ACTHIS_CN003C, pTemp->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0 ) {
              // break at HISTLOST
              pTemp = NULL;
            } // end if

            pTemp = (pTemp != NULL) ? pTemp->pPrev : NULL;
          } // end while

          pData = (pLog == NULL) ? NULL : pData;
        } // end if
      } // end if

      // reaching  the data save while using logs before data save in restore mode
      if ( pHistory->bForRestore &&  
           (pData != NULL)       &&
           (pLog  == NULL)       &&
           (pData == (*pEntry))      ) {
        bGetLine = false;
        pData = NULL;
      } // end if

      // while using logs before data save in restore mode ignore all saves except logs
      if ( pHistory->bForRestore &&  
           (pData != NULL)       &&
           (pLog  == NULL)       &&
           (strncmp(BA_HISTORY_LBLLOG_CN003C, (*pEntry)->tLine.szLABEL,  strlen(BA_HISTORY_LBLLOG_CN003C)) != 0) ) {
        bGetLine = false;
      } // end if

      // skip pagesaves during restore manager not according to the first datasave
      if ( pHistory->bForRestore && 
           bSecondDataSavePassed &&
           (strncmp(BA_HISTORY_LBLPAG_CN003C, (*pEntry)->tLine.szLABEL,  strlen(BA_HISTORY_LBLPAG_CN003C)) == 0)  ) {
        bGetLine = false;
      } // end if

      if (bGetLine) {
        bFull = cn32_PrintEntry(*pEntry, pHistory, pMedia, pEBIDs, pCurrPos, replyLenMax - strlen(replyData));
      } // end if

      if (!bFull) {
        pCurrPos = replyData + strlen(replyData);

        // in restore mode use logs before data save
        if (pHistory->bForRestore && (pLog != NULL)) {
          *pEntry = pLog;
          pLog = NULL;
        } else {
          *pEntry = (*pEntry)->pNext;
        } // end if

        // restore manager aborts at next HIST LOST
        if ((*pEntry != NULL) && pHistory->bForRestore) {
          if ( strncmp(BA_HISTORY_ACTHIS_CN003C, (*pEntry)->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0 ) {
            *pEntry = NULL;
          } // end if
        } // end if

      } // end if

    } while ( (*pEntry != NULL) &&
              ( bFull == false)     );

  } // end if

  if (*pEntry == NULL) {
    strncpy(pFlagPos, BA_HISTORY_CONT_FLAG_END_CN003C, BA_HISTORY_CONT_LEN_CN003C);
  } else {
    strncpy(pFlagPos, BA_HISTORY_CONT_FLAG_CONTINUE_CN003C, BA_HISTORY_CONT_LEN_CN003C);
  }

  /**/

  return nFuncStatus;
} /* end cn32_ListHistory */

/* PRIVATE FUNCTION cn32_FindFirstRestoreItem
 * ------------------------------------------------------------------
 * purpose: find the first item in histroy for restore
 *
 */
static tcn00_Error cn32_FindFirstRestoreItem(
      tcn003_BackupHistory         * pHistory,
      tcn003_BackupHistoryEntry   ** pEntry)
{
  tcn00_Error                 nFuncStatus  = OK_CN00;
  bool                        bFound       = false;
  tcn003_BackupHistoryEntry * pMatch       = NULL;
  
  /**/

  if (strcmp(pHistory->szRestore, BA_HISTORY_RES_ALL_CN003C) == 0) {

    // list all possible datasaves

    // oldest possible datasave (depends from restart info)
    //   - if there is a valid restart info use only data saves after the
    //     last histlost
    //   - if there no data save after the last hist lost iuse all data save
    //   - if there no valid restart info use all data saves

    if ((pHistory->nFirstLogPage != BA_HISTORY_INVLOG_CN003C) &&
        (pHistory->nUsedLogPage  >= pHistory->nFirstLogPage) ) {

      // search the last histlost (backward)
      pMatch = pHistory->pLast;
      while (pMatch != NULL) {
        if ( strncmp(BA_HISTORY_ACTHIS_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0 ) {
          break;
        } // end if
        pMatch = pMatch->pPrev;
      } // end while

      // search the first data save after the last histlost (forward)
      while (pMatch != NULL) {
        if ( ( strncmp(BA_HISTORY_LBLDAT_CN003C, pMatch->tLine.szLABEL,  strlen(BA_HISTORY_LBLDAT_CN003C)) == 0) &&
             ( strncmp(BA_HISTORY_ACTSAV_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTSAV_CN003C)) == 0)    ) {
          break;
        } // end if
        pMatch = pMatch->pNext;
      } // end while
    } // end if
    
    // no datasave after the last histlost? shit! use all data saves
    pMatch = (pMatch == NULL) ? pHistory->pFirst : pMatch;

    // disable restoremanager (this action is only a datasave listing with a specific
    // start position
    pHistory->bForRestore = false;
    strcpy(pHistory->szLabel, BA_HISTORY_LBLDAT_CN003C);
    strcpy(pHistory->szAction, BA_HISTORY_ACTSAV_CN003C);

  } else if (strcmp(pHistory->szRestore, BA_HISTORY_RES_LST_CN003C) == 0) {

    // search the last data save (backward)
    pMatch = pHistory->pLast;
    while (pMatch != NULL) {

      // HISTLOST passed -> maked restartinfo invalid
      if ( strncmp(BA_HISTORY_ACTHIS_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0 ) {
        pHistory->nFirstLogPage = BA_HISTORY_INVLOG_CN003C;
      } // end if

      int n1 = strncmp(BA_HISTORY_LBLDAT_CN003C, pMatch->tLine.szLABEL, strlen(BA_HISTORY_LBLDAT_CN003C));
      int n2 = atoi(pMatch->tLine.szRETURNCODE);
      int n3 = strcmp(pHistory->szUntil, BA_HISTORY_UNT_ALL_CN003C);
      int n4 = strncmp(pHistory->szUntil, pMatch->szStamp, strlen(pHistory->szUntil));

      if ( ( strncmp(BA_HISTORY_LBLDAT_CN003C, pMatch->tLine.szLABEL,  strlen(BA_HISTORY_LBLDAT_CN003C)) == 0 ) &&
           ( strncmp(BA_HISTORY_ACTSAV_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTSAV_CN003C)) == 0 ) &&
           ( atoi(pMatch->tLine.szRETURNCODE)                                                            == 0 ) &&
           ((strcmp(pHistory->szUntil, BA_HISTORY_UNT_ALL_CN003C)                      == 0) || 
            (strncmp(pHistory->szUntil, pMatch->szStamp, strlen(pHistory->szUntil))    >  0)                  )     ) {
        break;
      } // end if
      pMatch = pMatch->pPrev;
    } // end while

    // view only save actions
    strcpy(pHistory->szAction, BA_HISTORY_ACTSAV_CN003C);

  } else {

    // search the data save with this key (backward)
    pMatch = pHistory->pLast;
    while (pMatch != NULL) {

      // HISTLOST passed -> maked restartinfo invalid
      if ( strncmp(BA_HISTORY_ACTHIS_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTHIS_CN003C)) == 0 ) {
        pHistory->nFirstLogPage = BA_HISTORY_INVLOG_CN003C;
      } // end if

      // check key
      if ( ( strcmp(pHistory->szRestore, pMatch->tLine.szKEY) == 0 ) &&
           ( atoi(pMatch->tLine.szRETURNCODE)                 == 0 )    ) {
        break;
      } // end if
      pMatch = pMatch->pPrev;
    } // end while
    if (pMatch != NULL) {
      if ( ( strncmp(BA_HISTORY_LBLDAT_CN003C, pMatch->tLine.szLABEL,  strlen(BA_HISTORY_LBLDAT_CN003C)) != 0 ) || 
           ( strncmp(BA_HISTORY_ACTSAV_CN003C, pMatch->tLine.szACTION, strlen(BA_HISTORY_ACTSAV_CN003C)) != 0 )    ) {
        nFuncStatus = ERR_PARAM_CN00;
      } // end if
    } // end if

    // view only save actions
    strcpy(pHistory->szAction, BA_HISTORY_ACTSAV_CN003C);

  } // end if

  *pEntry = pMatch;
  
  /**/
  
  return nFuncStatus;
} /* end cn32_FindFirstRestoreItem */

/* SPECIFICATION PUBLIC FUNCTIONS
 * ------------------------------------------------------------------
 */

/* PUBLIC FUNCTION cn32BackupHistoryOpen
 * ------------------------------------------------------------------
 * purpose: open backup history session and read contents
 *
 */
tcn00_Error cn32BackupHistoryOpen (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	                  nFuncStatus = OK_CN00;
  char                          szParam[255];
  int                           nParam = 1; 
  bool                          bWithEBIDs = false;
  tcn003_BackupHistorySession * pHistorySession = NULL;

  /**/

  nFuncStatus = cn32_HistorySession(vcontrol, replyData, replyLen, &pHistorySession);
  if (nFuncStatus == OK_CN00) {
    cn32_FreeHistorySession(vcontrol);
    pHistorySession=0; // (*pHistorySession) was freed one line above, so don't reference to it anymore
  } else {
    nFuncStatus = OK_CN00;
  }
  
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn32_AllocHistorySession(vcontrol, replyData, replyLen, &pHistorySession);
  }
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn32_ReadHistoryFile(vcontrol, replyData, replyLen, &(pHistorySession->tHistory));
  }
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn32_ReadHistoryMediaFile(vcontrol, replyData, replyLen, &(pHistorySession->tMedia));
  }

  // look for switch -e
  if (nFuncStatus == OK_CN00) {
    szParam[0] = '\0';
    nParam = 1;
    
    while ( (cn90GetToken(command->args, szParam, nParam)) && 
            (nFuncStatus == OK_CN00)                          ) {
      /* -e */
      if(0 == cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_EBIDS_CN003C))  {
        bWithEBIDs = true;
      } // end if
      nParam++;
    } // end while
  } // end if

  // open "external" history
  if (bWithEBIDs) {
    if (nFuncStatus == OK_CN00) {
      nFuncStatus=cn32_ReadHistoryEBIDFile(vcontrol, replyData, replyLen, &(pHistorySession->tEBIDs));
    }
    if (nFuncStatus == OK_CN00) {
      nFuncStatus=cn32_CheckAvailabilityOfExternalBackups(vcontrol, replyData, replyLen, &(pHistorySession->tMedia), &(pHistorySession->tEBIDs));
    }
  } // end if

  if (nFuncStatus == OK_CN00) {
    sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  } else {
    cn32_FreeHistorySession(vcontrol);
    pHistorySession = NULL; 
  }

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncStatus;
} /* end cn32BackupHistoryOpen */

/* PUBLIC FUNCTION cn32BackupHistoryList
 * ------------------------------------------------------------------
 * purpose: list contents of history
 *
 */
tcn00_Error cn32BackupHistoryList (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax)
{
  tcn00_Error	                 nFuncStatus = OK_CN00;

  tcn003_BackupHistorySession  * pHistorySession = NULL;
  tcn003_BackupHistoryEntry    * pEntry = NULL;

  char                         szParam[255];
  int                          nParam = 1; 
  bool                         bHistoryOpened = false;

  /**/

  // get or open the history session
  nFuncStatus = cn32_HistorySession(vcontrol, replyData, replyLen, &pHistorySession);
  if (nFuncStatus == ERR_NOHISTORYSESSION_CN00) {
    nFuncStatus = cn32BackupHistoryOpen(vcontrol, command, replyData, replyLen, replyLenMax);
    if (nFuncStatus == OK_CN00) {
      bHistoryOpened = true;
      nFuncStatus = cn32_HistorySession(vcontrol, replyData, replyLen, &pHistorySession);
    } // end if
  } // end if

  // read the list parameters
  if (nFuncStatus == OK_CN00) {

    strcpy(pHistorySession->tHistory.szCols,    BA_HISTORY_COL_ALL_CN003C);
    strcpy(pHistorySession->tHistory.szKey,     BA_HISTORY_KEY_ALL_CN003C);
    strcpy(pHistorySession->tHistory.szLabel,   BA_HISTORY_LBL_ALL_CN003C);
    strcpy(pHistorySession->tHistory.szAction,  BA_HISTORY_ACT_ALL_CN003C);
    strcpy(pHistorySession->tHistory.szRestore, BA_HISTORY_RES_ALL_CN003C);
    strcpy(pHistorySession->tHistory.szUntil,   BA_HISTORY_UNT_ALL_CN003C);
    pHistorySession->tHistory.bForRestore   = false;
    pHistorySession->tHistory.bRcOKOnly     = false;
    pHistorySession->tHistory.bWithMedia    = false;
    pHistorySession->tHistory.bWithEBIDs    = false;

    szParam[0] = '\0';
    nParam = 1;
    
    while ( (cn90GetToken(command->args, szParam, nParam)) && 
            (nFuncStatus == OK_CN00)                          ) {
  
      nParam++;

      if (szParam[0] != '\0') {
        /* -c <col1,col2,...> */
        if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_COLS_CN003C) == 0) {
          if (cn90GetToken(command->args, pHistorySession->tHistory.szCols, nParam)) {
            nParam++;
            cn90StrUpperCopy(pHistorySession->tHistory.szCols, pHistorySession->tHistory.szCols, false);
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        /* -k <key> */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_KEY_CN003C) == 0) {
          if (cn90GetToken(command->args, pHistorySession->tHistory.szKey, nParam)) {
            nParam++;
            cn90StrUpperCopy(pHistorySession->tHistory.szKey, pHistorySession->tHistory.szKey, false);
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        /* -l <label> */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_LABEL_CN003C) == 0) {
          if (cn90GetToken(command->args, pHistorySession->tHistory.szLabel, nParam)) {
            nParam++;
            cn90StrUpperCopy(pHistorySession->tHistory.szLabel, pHistorySession->tHistory.szLabel, false);
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        /* -a <action> */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_ACTION_CN003C) == 0) {
          if (cn90GetToken(command->args, pHistorySession->tHistory.szAction, nParam)) {
            nParam++;
            cn90StrUpperCopy(pHistorySession->tHistory.szAction, pHistorySession->tHistory.szAction, false);
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        /* -r <what> */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_RESTORE_CN003C) == 0) {
          pHistorySession->tHistory.bForRestore = true;
          pHistorySession->tHistory.bRcOKOnly   = true;
          if (cn90GetToken(command->args, pHistorySession->tHistory.szRestore, nParam)) {
            nParam++;
            cn90StrUpperCopy(pHistorySession->tHistory.szRestore, pHistorySession->tHistory.szRestore, false);
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        /* -m */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_MEDIA_CN003C) == 0) {
          pHistorySession->tHistory.bWithMedia = true;
        /* -e */
        } else if(0==cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_EBIDS_CN003C))  {
          pHistorySession->tHistory.bWithEBIDs=true;
        /* -u <when> */
        } else if (cn90Stricmp(szParam, BA_HISTORY_LISTPARAM_UNTIL_CN003C) == 0) {
          if (cn90GetToken(command->args, pHistorySession->tHistory.szUntil, nParam)) {
            nParam++;
          } else {
            nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
          }
        } else {
          nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
        } // end if
      }
    } // end while
  } // end if

  // checking parameters (option mix)
  if (nFuncStatus == OK_CN00) {
    if ( ( pHistorySession->tHistory.bForRestore   &&
           ( (strcmp(pHistorySession->tHistory.szKey,     BA_HISTORY_KEY_ALL_CN003C) != 0) ||
             (strcmp(pHistorySession->tHistory.szLabel,   BA_HISTORY_LBL_ALL_CN003C) != 0) || 
             (strcmp(pHistorySession->tHistory.szAction,  BA_HISTORY_ACT_ALL_CN003C) != 0)    ) ) || 
         ( !pHistorySession->tHistory.bForRestore   &&
           (strcmp(pHistorySession->tHistory.szUntil,   BA_HISTORY_UNT_ALL_CN003C) != 0)        )    ) {
      nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end 

  // checking db state for restore manager
/*
  if (nFuncStatus == OK_CN00) {
    if (pHistorySession->tHistory.bForRestore) {
      if (cn90DBState(vcontrol->dbname) != STATE_COLD_CN00) {
        nFuncStatus = cn90AnswerIError(replyData, replyLen, ERR_NOTCOLD_CN00);
      } // end if
    } // end if
  } // end 
*/

  // read the restart info
  if (nFuncStatus == OK_CN00) {
    if (pHistorySession->tHistory.bForRestore) {
      tcn003_BackupRestartInfo tRestartInfo;
      if (cn90DBState(vcontrol->dbname) != STATE_COLD_CN00) {
        nFuncStatus = ERR_NOTCOLD_CN00;
      } else {      
        nFuncStatus = cn31BackupRestartInfo(vcontrol, replyData, replyLen, &tRestartInfo);
      } // end if
      if (nFuncStatus == OK_CN00) {
        pHistorySession->tHistory.nUsedLogPage   = atoi(tRestartInfo.szUsedLogPage);
        pHistorySession->tHistory.nFirstLogPage  = (tRestartInfo.szFirstLogPage[0] == '*') ? BA_HISTORY_INVLOG_CN003C : atoi(tRestartInfo.szFirstLogPage);
        pHistorySession->tHistory.bRestartable   = (tRestartInfo.szRestartable[0]  == '1'); 
      } else {
        nFuncStatus = OK_CN00;
        pHistorySession->tHistory.nUsedLogPage   = BA_HISTORY_INVLOG_CN003C;
        pHistorySession->tHistory.nFirstLogPage  = BA_HISTORY_INVLOG_CN003C;
        pHistorySession->tHistory.bRestartable   = false;
      } // end if
    } // end if
  } // end 

  // list the history
  if (nFuncStatus == OK_CN00) {
    pEntry = pHistorySession->tHistory.pFirst;

    if (pHistorySession->tHistory.bForRestore) {
      nFuncStatus = cn32_FindFirstRestoreItem(&(pHistorySession->tHistory),
                                              &pEntry);
      if (nFuncStatus != OK_CN00) {
        cn90AnswerIError(replyData, replyLen, nFuncStatus);
      } // end if
    } // end if

    if (nFuncStatus == OK_CN00) {
      nFuncStatus = cn32_ListHistory(replyData, replyLen, replyLenMax, 
                                     &(pHistorySession->tHistory),
                                     &(pHistorySession->tMedia),
                                     &(pHistorySession->tEBIDs),
                                     &pEntry);
    }

    if (nFuncStatus == OK_CN00) {
      pHistorySession->tHistory.pCurrent = pEntry;
    }
  }

  // close the session if opened
  if (bHistoryOpened) {
    cn32_FreeHistorySession(vcontrol);
    pHistorySession = NULL; 
  } // end if

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncStatus;
} /* end cn32BackupHistoryList */

/* PUBLIC FUNCTION cn32BackupHistoryListNext
 * ------------------------------------------------------------------
 * purpose: continue listing contents of history
 *
 */
tcn00_Error cn32BackupHistoryListNext (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax)
{
  tcn00_Error	                 nFuncStatus = OK_CN00;

  tcn003_BackupHistorySession  * pHistorySession = NULL;
  tcn003_BackupHistoryEntry    * pEntry = NULL;

  /**/

  nFuncStatus = cn32_HistorySession(vcontrol, replyData, replyLen, &pHistorySession);
  if (nFuncStatus == OK_CN00) {
    if (pHistorySession->tHistory.pCurrent != NULL) {
      pEntry = pHistorySession->tHistory.pCurrent;
    } else {
      pEntry = pHistorySession->tHistory.pFirst;
    }

    nFuncStatus = cn32_ListHistory(replyData, replyLen, replyLenMax, 
                                   &(pHistorySession->tHistory),
                                   &(pHistorySession->tMedia),
                                   &(pHistorySession->tEBIDs),
                                   &pEntry);

    if (nFuncStatus == OK_CN00) {
      pHistorySession->tHistory.pCurrent = pEntry;
    }
  }

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncStatus;
} /* end cn32BackupHistoryListNext */

/* PUBLIC FUNCTION cn32BackupHistoryClose
 * ------------------------------------------------------------------
 * purpose: close backup history session
 *
 */
tcn00_Error cn32BackupHistoryClose (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	                 nFuncStatus = OK_CN00;

  tcn003_BackupHistorySession  * pHistorySession = NULL;

  /**/

  nFuncStatus = cn32_HistorySession(vcontrol, replyData, replyLen, &pHistorySession);
  if (nFuncStatus == OK_CN00) {
    cn32_FreeHistorySession(vcontrol);
  }

  if (nFuncStatus == OK_CN00) {
    sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  }

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncStatus;
} /* end cn32BackupHistoryClose */

/* PUBLIC FUNCTION cn32BackupHistoryDate
 * ------------------------------------------------------------------
 * purpose: get date of history file
 *
 */
tcn00_Error cn32BackupHistoryDate (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	        nFuncStatus = OK_CN00;

  char                * pCurrPos;

  bool                bExists;
  char                szDate[64];

  /**/

  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn32_CheckHistoryFile(vcontrol, replyData, replyLen, &bExists, szDate);
  }

  if (nFuncStatus == OK_CN00) {
    if (!bExists) {
      nFuncStatus = ERR_NOHISTORY_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncStatus);
    } else {
      sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
      pCurrPos = replyData + strlen(replyData);
       
      sprintf(pCurrPos, "%s%s", szDate, LINE_SEPSTRING_CN00);
      pCurrPos = replyData + strlen (replyData);
    }
  }

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncStatus;
} /* end cn32BackupHistoryDate */

/* PRIVATE FUNCTION cn32_RemoveDoublequotes
 * ------------------------------------------------------------------
 * purpose: removes enclosing double qoutes, if they exist
 *
 */
static void cn32_RemoveDoublequotes(char * String)
{
    size_t Length=strlen(String);

    if(Length>1 &&              //if the string has more than one character and
       '"'==String[0] &&        //the first is a '"' and
       '"'==String[Length-1])   //the last is a '"'
    {
        size_t i;

        // do a secure strncpy(String, String+1, Length-2);
        for(i=0; i<Length-2; i++) 
            String[i]=String[i+1];  //take all chars but not the first and the last
    }
} /* end cn32_RemoveDoublequotes */

/* PUBLIC FUNCTION cn32GetEBIDs
 * ------------------------------------------------------------------
 * purpose: queries an external backup tool for a list of external
 *          backup IDs (EBIDs). The tool is determined through a
 *          backup medium and the list consists of EBIDs for backups
 *          which are currently available.
 *
 */
tcn00_Error cn32GetEBIDs(VControlDataT * vcontrol,
                         CommandT      * command,
                         char          * replyData,
                         int           * replyLen,
                         int             replyLenMax)
{
    tcn00_Error	rc=OK_CN00;

    char MediumName[255];
    char DBName[255];
    char HostName[255];

    MediumName[0]='\0';
    DBName[0]='\0';
    HostName[0]='\0';

    tcn003_BackupHistoryEBIDList * pEBIDList=0;

    rc=cn32_EBIDList(vcontrol, replyData, replyLen, &pEBIDList);

    if(rc==OK_CN00)
    { 
        cn32_FreeEBIDList(vcontrol);
        pEBIDList=0;                    // (*pEBIDList) was freed one line above, so don't reference to it anymore
    }

    rc=cn32_AllocEBIDList(vcontrol, replyData, replyLen, &pEBIDList);

    /* get media name */
    if(OK_CN00==rc)
    {
        if(cn90GetToken(command->args, MediumName, 1))
            cn32_RemoveDoublequotes(MediumName);
        else
        {
            rc=ERR_PARAM_CN00;                                  //we need a medium, if it's not there we have a error
            cn90AnswerIError(replyData, replyLen, rc);
        }
    }

    if(OK_CN00==rc)
    {
        cn90GetToken(command->args, DBName, 2);     // if the parameter is not supplied, DBName[0]=='\0'
        cn90GetToken(command->args, HostName, 3);   // if the parameter is not supplied, HostName[0]=='\0'

        const tcn35_SaveInfo Info(replyData,
                                  replyLen,
                                  vcontrol->dbname,
                                  '\0'==DBName[0] ? 0 :DBName,
                                  '\0'==HostName[0] ? 0 : HostName);

        if(Info.IsGood)
        {
            tcn35_BackupToolConnectorEnum BackupToolType(MediumName);

            if(tcn35_BackupToolConnectorEnum::None!=BackupToolType) //if an external backup tool is specified be the medium
            {
                tcn35_BackupToolConnector * Tool=0;

                if(!cn35_ConstructBackupToolConnector(tcn35_BackupToolConnector::GiveBackupInfo,
                                                      BackupToolType,
                                                      Info,
                                                      Tool))
                {
                    rc=TheDBMSrvErrorHandler_cn36.GiveLastError();
                }
                else
                {
                    Tool->ListEBIDsOfAvailableBackups(*pEBIDList);
                    rc=Tool->GiveLastError();                       // report all errors (also if tool is just not installed)
                }

                if(0!=Tool)
                    delete Tool;
            }
        }
        else
            rc=TheDBMSrvErrorHandler_cn36.GiveLastError();
    }

    if(OK_CN00==rc)
        sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);

    *replyLen = (int)strlen(replyData);

    return rc;
} /* end cn32GetEBIDs */

/* PRIVATE FUNCTION cn32_ListNextEBIDs
 * ------------------------------------------------------------------
 * purpose: does the real printing work for cn32ListNextEBIDs
 *
 */
static tcn00_Error cn32_ListNextEBIDs(char                               * replyData,
                                      int                                * replyLen,
                                      int                                  replyLenMax,
                                      const tcn003_BackupHistoryEBIDList & EBIDs,
                                      tcn003_BackupHistoryEBID          *& pEntry)
{
    tcn00_Error rc=OK_CN00;

    char * pCurrPos;
    char * pFlagPos;

    char   Line[BA_HISTORY_LINE_LENGTH_MXCN003];
    size_t Length=0;

    bool bFull=false;

    sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    pCurrPos=replyData+strlen(replyData);

    sprintf(pCurrPos, "%s", BA_HISTORY_CONT_FLAG_CONTINUE_CN003C);
    pFlagPos=pCurrPos;
    pCurrPos=pCurrPos+BA_HISTORY_CONT_LEN_CN003C;

    while(0!=pEntry && false==bFull)     // not at end of list and reply buffer not full
    {
        Line[0]='\0';

        cn32_GetEBIDAsLine(pEntry, Line, BA_HISTORY_LINE_LENGTH_MXCN003-1, 1);  // get also extended information (type and date/time)

        Length=strlen(Line)+strlen(LINE_SEPSTRING_CN00);

        if(((int)(strlen(replyData)+Length))<replyLenMax)   //if(int(strlen(replyData)+Length)<replyLenMax) is not accepted by the sun compiler
        {
            sprintf(pCurrPos, "%s%s", Line, LINE_SEPSTRING_CN00);
            pCurrPos=replyData+strlen(replyData);

            pEntry=pEntry->pNext;   // go to the next list element only, if all was ok (enough space in reply buffer)
        }
        else
            bFull=true;
    }

    if(0==pEntry)
        strncpy(pFlagPos, BA_HISTORY_CONT_FLAG_END_CN003C, BA_HISTORY_CONT_LEN_CN003C);
    else
        strncpy(pFlagPos, BA_HISTORY_CONT_FLAG_CONTINUE_CN003C, BA_HISTORY_CONT_LEN_CN003C);

    return rc;
} /* end tcn00_Error cn32_ListNextEBIDs */

/* PUBLIC FUNCTION cn32ListNextEBIDs
 * ------------------------------------------------------------------
 * purpose: prints a list of EBIDs to the reply buffer
 *
 */
tcn00_Error cn32ListNextEBIDs(VControlDataT * vcontrol,
                              CommandT      * command,
                              char          * replyData,
                              int           * replyLen,
                              int             replyLenMax)
{
    tcn00_Error	rc;

    tcn003_BackupHistoryEBIDList * pEBIDList=0;

    rc=cn32_EBIDList(vcontrol, replyData, replyLen, &pEBIDList);

    if(OK_CN00==rc)
    {
        tcn003_BackupHistoryEBID * pEntry=0;

        if(0!=pEBIDList->pCurrent)
            pEntry=pEBIDList->pCurrent;
        else
            pEntry=pEBIDList->pFirst;

        rc=cn32_ListNextEBIDs(replyData,replyLen,replyLenMax, *pEBIDList, pEntry);
            
        if(OK_CN00==rc)
            pEBIDList->pCurrent=pEntry;
    }

    *replyLen = (int)strlen(replyData);

    return rc;
} /* end cn32ListNextEBIDs */

/* PUBLIC FUNCTION cn32ForgetEBIDs
 * ------------------------------------------------------------------
 * purpose: frees EBID list
 *
 */
tcn00_Error cn32ForgetEBIDs(VControlDataT * vcontrol,
                            CommandT      * command,
                            char          * replyData,
                            int           * replyLen,
                            int             replyLenMax )
{
    tcn00_Error rc=OK_CN00;

    tcn003_BackupHistoryEBIDList * pEBIDList=0;

    rc=cn32_EBIDList(vcontrol, replyData, replyLen, &pEBIDList);

    if(OK_CN00==rc)
    {
        cn32_FreeEBIDList(vcontrol);
        sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    }

    *replyLen = (int)strlen(replyData);

    return rc;
} /* end cn32ForgetEBIDs */
