/*!*****************************************************************************
  module:       vcn35.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing connection between dbmsrv and abstract external
                archiving tool, declared in hcn35.h 
  last change:  2000-11-24 14:37
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hcn20.h"
#include "hcn32.h"
#include "hcn35.h"
#include "hcn351.h"
#include "hcn352.h"
#include "hcn353.h"
#include "hcn42.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>


// -----------------------------------------------------------------------------
// constants and types
// -----------------------------------------------------------------------------
static const char * Action_String_DataMigration_cn35="SAVEDATA";
static const char * Action_String_DataRecovery_cn35="SAVEDTNCHK";
static const char * Action_String_PagesMigration_cn35="SAVEPAGES";
static const char * Action_String_PagesRecovery_cn35="SAVEPGNCHK";
static const char * Action_String_Log_cn35="SAVELOG";

struct SaveTypeMapEntry
{
    const char * DBMSrvSaveType;
    const char * ToolSaveTypeNoCheckpoint;
    const char * ToolSaveTypeCheckpoint;
    const char * ListSaveTypeNoCheckpoint;
    const char * ListSaveTypeCheckpoint;
};

const SaveTypeMapEntry TheSaveTypeMap[]={{ BA_KEY_DATA_CN003,
                                           Action_String_DataRecovery_cn35,
                                           Action_String_DataMigration_cn35,
                                           BA_EBIDS_TYPE_DAT_REC_CN003C,
                                           BA_EBIDS_TYPE_DAT_MIG_CN003C},

                                         { BA_KEY_PAGES_CN003,
                                           Action_String_PagesRecovery_cn35,
                                           Action_String_PagesMigration_cn35,
                                           BA_EBIDS_TYPE_PAG_REC_CN003C,
                                           BA_EBIDS_TYPE_PAG_MIG_CN003C},

                                         { BA_KEY_LOG_CN003,
                                           Action_String_Log_cn35,
                                           Action_String_Log_cn35,
                                           BA_EBIDS_TYPE_LOG_REC_CN003C,
                                           BA_EBIDS_TYPE_LOG_REC_CN003C},

                                         { "",
                                           "",
                                           "",
                                           "",
                                           ""}};

static const char * cn35_ActionStringForDBMSaveType(const char * DBMSaveType, int WithoutCheckpoint)
{
    int i=0;
    const char *rc=0;

    while(TheSaveTypeMap[i].DBMSrvSaveType[0]!='\0' &&                 //while not at end of the list
          0!=strcmp(TheSaveTypeMap[i].DBMSrvSaveType, DBMSaveType))    //and type keywords are not equal
    {
        i++;
    }

    if('\0'==TheSaveTypeMap[i].DBMSrvSaveType[0]) // we should find always the correct type (before reaching end of map)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    else
    {
        if(WithoutCheckpoint) // do not forget the diffrence between checkpoint and savepoint
            rc=TheSaveTypeMap[i].ToolSaveTypeNoCheckpoint;
        else
            rc=TheSaveTypeMap[i].ToolSaveTypeCheckpoint;
    }

    return rc;
}

static const char * cn35_ListSaveTypeStringForActionString(const char * ActionString)
{
    const char *rc=0;

    if(0!=ActionString)
    {
        int i=0;

        // check actions with checkpoint
        while(TheSaveTypeMap[i].DBMSrvSaveType[0]!='\0' &&                          //while not at end of the list
              0!=strcmp(TheSaveTypeMap[i].ToolSaveTypeCheckpoint, ActionString))    //and type keywords are not equal
        {
            i++;
        }

        if('\0'!=TheSaveTypeMap[i].DBMSrvSaveType[0])       // action string found?
            rc=TheSaveTypeMap[i].ListSaveTypeCheckpoint;    
        else
        {                                                   // check actions without checkpoint
            i=0;

            while(TheSaveTypeMap[i].DBMSrvSaveType[0]!='\0' &&                          //while not at end of the list
                  0!=strcmp(TheSaveTypeMap[i].ToolSaveTypeNoCheckpoint, ActionString))  //and type keywords are not equal
            {
                i++;
            }
        
            if('\0'!=TheSaveTypeMap[i].DBMSrvSaveType[0])  // we should find always the correct type (before reaching end of map)
                rc=TheSaveTypeMap[i].ListSaveTypeNoCheckpoint;
            //else 0==rc holds already
        }
    }

    return rc;
}

// -----------------------------------------------------------------------------
// member functions of class tcn35_SaveInfo
// -----------------------------------------------------------------------------

const char * tcn35_SaveInfo::Action_String_DataMigration    =Action_String_DataMigration_cn35;
const char * tcn35_SaveInfo::Action_String_DataRecovery     =Action_String_DataRecovery_cn35;
const char * tcn35_SaveInfo::Action_String_PagesMigration   =Action_String_PagesMigration_cn35;
const char * tcn35_SaveInfo::Action_String_PagesRecovery    =Action_String_PagesRecovery_cn35;
const char * tcn35_SaveInfo::Action_String_Log              =Action_String_Log_cn35;

tcn35_SaveInfo::tcn35_SaveInfo(char                    * pcReplyData,
                               int                     * piReplyLen,
                               tin01_sql_session       * ptSQLSession,
                               const char              * pcszKernelCommand,
                               const tcn003_Medium     * pMedium,
                               const tcn003_MediaGroup * pMediaGroup,
                               const char              * pcszBackupType,
                               const int                 WithoutCheckpoint,
                               const tsp00_DbNamec       tDBName,
                               const char              * StartOfExternalBackupID,
                               const char              * EndOfExternalBackupID)
{
    long PagesPerBlock=0;

    SetAllMembersToZero();

    IsGood=InitErrorHandlerAndProtocol(pcReplyData, piReplyLen, tDBName);

    //get the name of the computer the program is running on
    if(IsGood)
    {
        pszMachName=cn36_GetHostName();

        if(0==pszMachName)
            IsGood=0;

        if(!cn36_StrAllocCpy(SourceDBHost, pszMachName))
            IsGood=0;
    }

    //make a copy of the database name
    if(IsGood)
        IsGood=cn36_StrAllocCpy(pszDBName, (const char *)tDBName);

    if(IsGood)
    {
        tsp00_C64c  szValue;
        tcn00_Error XParamRC=cn20XParamGetValue(tDBName, XP_ID_RUNDIRECTORY_CN00, szValue);

        if(OK_CN00==XParamRC)
            IsGood=cn36_StrAllocCpy(DBRunDir, (const char *)szValue);
        else
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CN00Error.SetI(int(XParamRC)));

            cn36_Verbose("Could not determine the run directory of the database.\n");

            IsGood=0;
        }
    }

    // determine the type of save as known from database version 6.1 and up (needed for ADSM) 
    if(IsGood)
    {
        int i=0;

        char *BackupTypeInUpperLetters=0;

        IsGood=cn36_StrAllocCpy(BackupTypeInUpperLetters, pcszBackupType);

        if(IsGood)
        {
            cn36_StrUpr(BackupTypeInUpperLetters);
            
            pszActionString=cn35_ActionStringForDBMSaveType(BackupTypeInUpperLetters, WithoutCheckpoint);

            if(0==pszActionString)
                IsGood=0;   //error message already given
        }

        cn36_StrDealloc(BackupTypeInUpperLetters); // last chance to free the memory
    }

    //allocate (*ppszFiles) and initialize it,
    //determine also the PagesPerBlock of the medium or the first medium of a group
    if(IsGood)
    {
        const tcn003_Medium * pSingleMedium;

        if(0!=pMedium) //only one medium
        {
            pSingleMedium=pMedium;
            pszMediumName=pMedium->szName;
            NumFiles=1;
        }
        else
            if(0!=pMediaGroup && 0!=pMediaGroup->pFirst) // we have a group of media with at least one element
            {
                pSingleMedium=pMediaGroup->pFirst;
                pszMediumName=pMediaGroup->szName;

                //determine the number of media in the group
                NumFiles=0;

                while(0!=pSingleMedium)
                {
                    NumFiles++;
                    pSingleMedium=pSingleMedium->pNext;
                }
                
                pSingleMedium=pMediaGroup->pFirst; //go back to the first medium of the group
            }
            else
            {
                TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_MENOTFOUND_CN00);
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::NoMediumFound.Set());
                IsGood=0;
            }

        if(IsGood) //now get the pages per block of the medium or of the first medium of a group
        {
            PagesPerBlock=atoi(pSingleMedium->szBlocksize);

            if(0==PagesPerBlock)
            {
                IsGood=0;
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotGetBlockingFactor.Set(pSingleMedium->szBlocksize));
            }
        }

        if(IsGood) // now allocate (*ppszFiles) and initialize it
        {
            ppszFiles=new const char *[NumFiles]; //place for an array of NumFiles pointer to const char

            if(0==ppszFiles)
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                IsGood=0;
            }
            else
            {
                int i;

                for(i=0; i<NumFiles; ++i, pSingleMedium=pSingleMedium->pNext)
                    ppszFiles[i]=pSingleMedium->szLocation;
            }
        }
    }

    //determine size of a block 
    if(IsGood)
    {
        tsp00_C64c szPageSize;
        tcn00_Error RCOfGetValue=cn20XParamGetValue(tDBName, XP_ID_PAGE_SIZE_CN00, szPageSize);

        if(OK_CN00!=RCOfGetValue)
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CN00Error.SetI(int(RCOfGetValue)));
            IsGood=0;
        }
        else
        {
            PageSize=atoi((char *)szPageSize);

            if(0==PageSize)
            {
                IsGood=0;
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotGetPageSize.Set(szPageSize));
            }
            else
                BlockSize=PageSize*PagesPerBlock;
        }
    }

    if(IsGood &&                                            //all ok so far
       0!=StartOfExternalBackupID &&                        //ExternalBackupID has a start
       0!=EndOfExternalBackupID &&                          //ExternalBackupID has an end
       0<EndOfExternalBackupID-StartOfExternalBackupID)     //ExternalBackupID is not empty
    {
        if('"'==(*StartOfExternalBackupID) && '"'==(*(EndOfExternalBackupID-1)) &&  //the whole thing is enclosed in double qoutes
           EndOfExternalBackupID!=StartOfExternalBackupID-1)                        //and it contains more than only one '"'
        {
            StartOfExternalBackupID++;
            EndOfExternalBackupID--;
        }

        IsGood=cn36_StrNAllocCpy(ExternalSaveID, StartOfExternalBackupID, EndOfExternalBackupID-StartOfExternalBackupID);
    }

    //initialize all other members
    if(IsGood)
    {
        ReplyData=pcReplyData;
        ReplyLen=piReplyLen;
        SQLSession=ptSQLSession;
        KernelCommand=pcszKernelCommand;
        SourceDBName=0;
    }
}

void tcn35_SaveInfo::SetAllMembersToZero()
{
    pszMachName=0;
    pszDBName=0;
    DBRunDir=0;
    pszActionString=0;
    BlockSize=0;
    pszMediumName=0;
    ppszFiles=0;
    NumFiles=0;
    
    ReplyData=0;
    ReplyLen=0;
    SQLSession=0;
    KernelCommand=0;

    ExternalSaveID=0;
    SourceDBName=0;
    SourceDBHost=0;
    
    IsGood=0;
}

void tcn35_SaveInfo::FreeAllDynamicMembers()
{
    if(0!=ppszFiles)
    {
        delete [] ppszFiles;
        ppszFiles=0;
    }

    cn36_StrDealloc(pszDBName);
    cn36_StrDealloc(DBRunDir);
    cn36_StrDealloc(ExternalSaveID);
    cn36_StrDealloc(SourceDBName);
    cn36_StrDealloc(SourceDBHost);
}

tcn35_SaveInfo::tcn35_SaveInfo(char * pcReplyData,
                               int  * piReplyLen,
                               tsp00_DbNamec tDBName,
                               const char * TheSourceDBName,
                               const char * TheSourceDBHost)
{
    SetAllMembersToZero();

    IsGood=InitErrorHandlerAndProtocol(pcReplyData, piReplyLen, tDBName);

    if(IsGood)
    {
        tsp00_C64c  szValue;
        tcn00_Error XParamRC=cn20XParamGetValue(tDBName, XP_ID_RUNDIRECTORY_CN00, szValue);

        if(OK_CN00==XParamRC)
            IsGood=cn36_StrAllocCpy(DBRunDir, (const char *)szValue);
        else
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CN00Error.SetI(int(XParamRC)));

            cn36_Verbose("Could not determine the run directory of the database.\n");

            IsGood=0;
        }
    }

    if(IsGood)
    {
        int StringOK;

        if(0==TheSourceDBName)
            StringOK=cn36_StrAllocCpy(SourceDBName, (char *)tDBName);   // use name of current db
        else
            StringOK=cn36_StrAllocCpy(SourceDBName, TheSourceDBName);   // use specified name of source db

        if(StringOK)
        {
            if(0==TheSourceDBHost)
            {                                                               // use name of current computer
                const char * HostName=cn36_GetHostName();               

                if(0==HostName)
                    StringOK=0;
                else
                    StringOK=cn36_StrAllocCpy(SourceDBHost, HostName);
            }
            else
                StringOK=cn36_StrAllocCpy(SourceDBHost, TheSourceDBHost);   // use specified name of source computer
        }

        if(StringOK)
        {
            if(!cn36_StrAllocCpy(pszDBName, (char *)tDBName))
                IsGood=0;
            else
            {
                ReplyData=pcReplyData;
                ReplyLen=piReplyLen;
            }
        }
        else
            IsGood=0;
    }
}

tcn35_SaveInfo::tcn35_SaveInfo(const tcn35_SaveInfo & Original)
{
    *this=Original; // duplicate object bit-wise

    if(0!=Original.pszDBName)
        pszDBName=new char [strlen(Original.pszDBName)+1];

    if(0!=Original.DBRunDir)
        DBRunDir=new char [strlen(Original.DBRunDir)+1];

    if(0!=Original.ppszFiles)
        ppszFiles=new const char *[Original.NumFiles];

    if(0!=Original.ExternalSaveID)
        cn36_StrAllocCpy(ExternalSaveID=0, Original.ExternalSaveID);    //ExternalSaveID=0 to avoid deleting content of Original.ExternalSaveID

    if(0!=Original.SourceDBName)
        SourceDBName=new char [strlen(Original.SourceDBName)+1];
    
    if(0!=Original.SourceDBHost)
        SourceDBHost=new char [strlen(Original.SourceDBHost)+1];

    if(0!=Original.ExternalSaveID && 0==ExternalSaveID)
        IsGood=0;

    if((0!=Original.pszDBName && 0==pszDBName) ||              //if one of the necessary allocations failed
       (0!=Original.DBRunDir && 0==DBRunDir) ||
       (0!=Original.ppszFiles && 0==ppszFiles) ||
       (0!=Original.SourceDBName && 0==SourceDBName) ||
       (0!=Original.SourceDBHost && 0==SourceDBHost))
    {
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());

        IsGood=0;
    }

    if(!IsGood)
        FreeAllDynamicMembers();
    else
    {
        if(0!=Original.pszDBName)
            strcpy(pszDBName, Original.pszDBName);

        if(0!=Original.DBRunDir)
            strcpy(DBRunDir, Original.DBRunDir);

        if(0!=Original.ppszFiles)
            memcpy(ppszFiles, Original.ppszFiles, NumFiles*sizeof(const char *));

        if(0!=Original.SourceDBName)
            strcpy(SourceDBName, Original.SourceDBName);

        if(0!=Original.SourceDBHost)
            strcpy(SourceDBHost, Original.SourceDBHost);
    }
}

tcn35_SaveInfo::~tcn35_SaveInfo()
{
    FreeAllDynamicMembers();
}

int tcn35_SaveInfo::InitErrorHandlerAndProtocol(char * pcReplyData,
                                                int * piReplyLen,
                                                tsp00_DbNamec tDBName)
{
    int rc=1;

    TheDBMSrvErrorHandler_cn36.SetReplyArrayTo(pcReplyData, piReplyLen);
    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_CN00);

    if(PathOfProtFile_cn36[0]=='\0') //init name of dbm.ebp here (if needed) because it's the first function using cn36_Verbose() 
    {
        tsp00_C264c tAbsoluteNameOfProtocolFile;
        tcn00_Error Err=cn42GetFileName(tDBName, FGET_BACKEBP_CN42, tAbsoluteNameOfProtocolFile);

        if(OK_CN00!=Err || strlen((char *)tAbsoluteNameOfProtocolFile)>MaxLengthOfProtFileName_cn36)
        {
            rc=IsGood=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotGetNameOfToolProtFile.Set());
        }
        else
            strcpy(PathOfProtFile_cn36, (char *)tAbsoluteNameOfProtocolFile);
    }

    return rc;
}


// -----------------------------------------------------------------------------
// member functions of class tcn35_BackupToolConnector
// -----------------------------------------------------------------------------

tcn35_BackupToolConnector::tcn35_BackupToolConnector(tcn35_ActionType Action, const tcn35_SaveInfo & rInfo)
:Info(rInfo)
{
    int i;
    
    StatusOfRequest=NoRequestSend;
    SQLCodeOfRequest=0;
    FirstLogPageNo=0;
    LastLogPageNo=0;
    StatusOfTool=NotStarted;
    
    NumberOfCommandLines=0;
    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        CommandLine[i]=0;

    NumberOfRunningToolProcesses=0;
    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
    {
        ToolProcess[i]=0;
        ToolProcessWasRunning[i]=0;
    }

    ToolProcessOutput=0;
    ToolProcessError=0;
    
    NumberOfExternalSaveIDsFound=0;
    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        ExternalSaveID[i]=0;

    TempDateTimeString=0;
    ToolProcessOutputAsText=0;

    ActionType=Action;

    NumberOfExistingPipes=0;
    for(i=0; i<MAX_NUMBER_OF_DATAPIPES_CN35; i++)
        TheDataPipe[i]=0;

    CurrentWaitingPhase=NoWaiting;
    StartOfCurrentWaitingPhase=-1;
    WaitForeverOnDB=0;

    PathOfExternalBackupHistory=0;
    PathOfBackupHistory=0;

    if(ARestore==ActionType && (0==Info.ExternalSaveID || 0>=strlen(Info.ExternalSaveID)))
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::NeedExternalSaveIDForRestore.Set());
}

tcn35_BackupToolConnector::~tcn35_BackupToolConnector()
{
    int i;

    cn36_StrDealloc(FirstLogPageNo);
    cn36_StrDealloc(LastLogPageNo);

    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        cn36_StrDealloc(CommandLine[i]);

    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        if(0!=ToolProcess[i])
            delete ToolProcess[i];

    if(0!=ToolProcessOutput)
        delete ToolProcessOutput;

    if(0!=ToolProcessError)
        delete ToolProcessError;
        
    for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        cn36_StrDealloc(ExternalSaveID[i]);

    cn36_StrDealloc(TempDateTimeString);
    if(0!=ToolProcessOutputAsText)
        delete ToolProcessOutputAsText;

    for(i=0;i<MAX_NUMBER_OF_DATAPIPES_CN35;i++)
        if(0!=TheDataPipe[i])
            delete TheDataPipe[i];

    cn36_StrDealloc(PathOfExternalBackupHistory);
    cn36_StrDealloc(PathOfBackupHistory);
}

void tcn35_BackupToolConnector::BeginBackupOperation()
{
    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_TOOLCHK_CN00);

    int ToolIsInstalled=0;

    if(CheckToolInstallation(ToolIsInstalled) && ToolIsInstalled)
    {
        TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_MEDIUMCHK_CN00);
        
        if(CheckMedium())
        {
            int AllOk;

            TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_PREPARE_CN00);
            if(ActionType==ABackup)
            {
                cn36_Verbose("Preparing backup.\n", AddALevelToPrefix);
                AllOk=Prepare();
                cn36_Verbose(AllOk?"Prepare passed successfully.\n\n":"Prepare failed.\n\n", SubALevelFromPrefix);
            }
            else
            {
                cn36_Verbose("Preparing restore.\n", AddALevelToPrefix);
                AllOk=Prepare();
                cn36_Verbose(AllOk? "Prepare passed successfully.\n\n": "Prepare failed.\n\n", SubALevelFromPrefix);
            }
            
            if(AllOk)
            {
                AllOk=0;

                if(CreatePipes())
                {
                    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_DBREQ_CN00);
                    
                    if(StartDBAction())  // start the tool only, when Request was send
                        if(DBIsReady())
                        {
                            TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_TOOLREQ_CN00);

                            if(StartTool())
                            {
                                StatusOfTool=Started;
                                CheckTool();

                                AllOk=1;    // only if we get here, we have started the db request and the backup tool correctly
                            }
                        }
                }
            }

            if(OK_CN00!=GiveLastError() || !AllOk)  // if something went wrong, stop everything and clean up the allocated resources (pipes, files, ...)
                EndBackupOperation();
        }
    }
}

int tcn35_BackupToolConnector::BackupOperationIsRunning()
{
    //function returns 0 if backup operation is not running or has already finished
    //                 1 if database or backup tool is working
    //         uses RequestSend==StatusOfRequest to indicate running database request
    //              Started==StatusOfTool to indicate running backup tool       

    int rc=1;
    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_OPCHK_CN00);
    
    //check that db and tool were started correctly at first entrance!
    if(NoWaiting==CurrentWaitingPhase)
    {
        WaitForeverOnDB=0;

        if(NotStarted==StatusOfTool &&                  // the tool was not started
           NoRequestSend==StatusOfRequest &&            // no request was send to the db
           !IsBad())                                    // no error was reported
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::NoBackupIsRunning.Set()); // no error was reported and nothing is running -> no backup/restore on the way
        }

        if(RequestSend==StatusOfRequest && Started==StatusOfTool)
            CurrentWaitingPhase=WaitForDBAndBackupTool; //database and tool are working, start waiting for the end of one or the other
        else
            CurrentWaitingPhase=WaitForDB;              //database or tool or both are not working, same as one has already ended
    }

    if(WaitForDBAndBackupTool==CurrentWaitingPhase) // (RequestSend==StatusOfRequest && Started==StatusOfTool), db and backup tool were working, when checked last time.
    {                                               // then
        if(!ToolIsRunning())                        // check backup tool again,
            StatusOfTool=EndedPartially;

        if(ReplyAvailable(*Info.SQLSession))        // check db again
            ReceiveBackupRestoreRequestReplyFromDBKernel();

        if(RequestReplied==StatusOfRequest || EndedPartially==StatusOfTool) // if something has changed, set maximal reaction time and get the current time
        {
            CurrentWaitingPhase=WaitForDB;  //go to next phase

            if(KernelMayBeUsingLogDevspaceForLogRestore())
                WaitForeverOnDB=1;

            StartOfCurrentWaitingPhase=cn36_GetCurrentTime();
        }
    }

    if(WaitForDB==CurrentWaitingPhase)  // wait maximal MaxWaitingTimeOfPhase[CurrentWaitingPhase] seconds for reaction of the db kernel, if it is still running
    {
        int ChangePhase=1;

        if(RequestSend==StatusOfRequest)
        {
            if(WaitForeverOnDB ||
               MaxWaitingTimeOfCurrentWaitingPhase()>=cn36_SecondsBetween(StartOfCurrentWaitingPhase, cn36_GetCurrentTime()))
            {
                if(ReplyAvailable(*Info.SQLSession)) // check db again
                    ReceiveBackupRestoreRequestReplyFromDBKernel();
                else
                    ChangePhase=0;
            }
        }

        if(ChangePhase)
        {
            CurrentWaitingPhase=WaitForBackupTool;  //go to next phase

            StartOfCurrentWaitingPhase=cn36_GetCurrentTime();

            if(RequestSend==StatusOfRequest)
                SignalEndToDB();    //start cancel DB action -> StatusOfRequest changes to CancelSendByDBMSrv -> should start cancel db only once

            if(ToolIsRunning())
                SignalEndToTool();
        }
    }

    if(WaitForBackupTool==CurrentWaitingPhase)
    {
        int ChangePhase=1;

        if(Started==StatusOfTool || EndedPartially==StatusOfTool)  //was the tool running or partially running?
        {
            if(MaxWaitingTimeOfCurrentWaitingPhase()>=cn36_SecondsBetween(StartOfCurrentWaitingPhase, cn36_GetCurrentTime()))
            {
                if(!ExistsToolProcesses())
                    StatusOfTool=Ended;
                else
                    ChangePhase=0;
            }
            else
            {
                if(ExistsToolProcesses())
                {
                    sprintf(MsgBuf_cn36, "Canceling tool after the timeout of %d seconds elapsed.\n", int(MaxWaitingTimeOfCurrentWaitingPhase()));
                    cn36_Verbose(MsgBuf_cn36);

                    CancelTool();
                }
                else
                    StatusOfTool=Ended;
            }
        }
        
        if(ChangePhase)
        {
            CurrentWaitingPhase=WaitForKilled;  //go to next phase

            StartOfCurrentWaitingPhase=cn36_GetCurrentTime();
        }
    }

    if(WaitForKilled==CurrentWaitingPhase)  // wait maximal MaxWaitingTimeOfPhase[CurrentWaitingPhase] seconds for reaction of the db kernel, if it is still running or the tool if it is still running
    {
        if(RequestSend==StatusOfRequest || CancelSendByDBMSrv==StatusOfRequest) // was the DB working on the request?
        {
            if(MaxWaitingTimeOfCurrentWaitingPhase()>=cn36_SecondsBetween(StartOfCurrentWaitingPhase,cn36_GetCurrentTime()))
            {
                if(ReplyAvailable(*Info.SQLSession)) // check db again
                    ReceiveBackupRestoreRequestReplyFromDBKernel();
            }
            else 
                TerminateDBAction();
        }

        if(Started==StatusOfTool || EndedPartially==StatusOfTool)  //was the tool running or partialy running?
        {
            if(MaxWaitingTimeOfCurrentWaitingPhase()>=cn36_SecondsBetween(StartOfCurrentWaitingPhase, cn36_GetCurrentTime()))
            {
                if(!ExistsToolProcesses())
                    StatusOfTool=Ended;
            }
            else
                if(KilledByDBMSrv!=StatusOfTool)
                {
                    if(ExistsToolProcesses())
                    {
                        sprintf(MsgBuf_cn36, "Terminating tool after the timeout of %d seconds elapsed.\n", int(MaxWaitingTimeOfCurrentWaitingPhase()));
                        cn36_Verbose(MsgBuf_cn36);

                        TerminateTool();
                        StatusOfTool=KilledByDBMSrv;    // calling TerminateTool() only once
                    }
                    else
                        StatusOfTool=Ended;
                }              
        }

        if((RequestSend!=StatusOfRequest && CancelSendByDBMSrv!=StatusOfRequest) &&      // if db request &&
           (Started!=StatusOfTool && EndedPartially!=StatusOfTool))                      // tool are not working
            rc=0;                                                                        // set return-code appropriate
    }

    return rc;
}

void tcn35_BackupToolConnector::EndBackupOperation()
{
    cn36_Verbose(GiveActionType()==ABackup?
                    "Waiting for end of the backup operation.\n":
                    "Waiting for end of the restore operation.\n",
                 AddALevelToPrefix);

    // must not check that Backup is/was running. BackupOperationIsRunning() reports error if necessary
    int iCheckTime=5, LastWaitingPhase=CurrentWaitingPhase;

    while(BackupOperationIsRunning())
    {
        if(LastWaitingPhase!=CurrentWaitingPhase)  // the waiting phase has changed,
        {
            iCheckTime=5;                          // reset check-time 
            LastWaitingPhase=CurrentWaitingPhase;
        }

        cn36_Sleep(iCheckTime++);
    }
    
    cn36_Verbose(GiveActionType()==ABackup?
                    "The backup operation has ended.\n\n":
                    "The restore operation has ended.\n\n",
                 SubALevelFromPrefix);

    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_POSTOP_CN00);

    if(ABackup==ActionType && DBActionWasSuccessful() && ToolWasSuccessful())
    {
        WriteLogFile();
        
        if(OK_CN00==GiveLastError() &&                                          //no error occurred (also not in WriteLogFile())
           0==strcmp(Info.pszActionString, tcn35_SaveInfo::Action_String_Log))  //we make a log backup
        {
            FreeLogAfterLogBackup();
        }
    }

    if(OK_CN00==GiveLastError())  // no error reported so far -> fill the reply buffer
    {
        TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_BACKUPOP_CN00);   //reports of real backup/restore errors are constructed here -> set the correct error type
        FillReplyBuffer();
    }

    if(OK_CN00==GiveLastError() && DBActionWasSuccessful())    // no error reported so far -> do all necessary post-operations
    {
        TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_POSTOP_CN00); // set possible errorcode back to ERR_POSTOP_CN00
        PostOperationsAfterSuccess();
    }
    
    CleanUp();
}

int tcn35_BackupToolConnector::ListEBIDsOfAvailableBackups(tcn003_BackupHistoryEBIDList & AvailableEBIDs)
{
    int rc=1,
        ToolIsInstalled=0;

    cn32FreeHistoryEBIDs(&AvailableEBIDs);

    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_TOOLCHK_CN00);
    if(!CheckToolInstallation(ToolIsInstalled))
        rc=0;

    if(rc && ToolIsInstalled)
    {
        TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_PREPARE_CN00);
        cn36_Verbose("Preparing query for available backups.\n", AddALevelToPrefix);
        rc=Prepare();
        cn36_Verbose(rc? "Prepare passed successfully.\n\n": "Prepare failed.\n\n", SubALevelFromPrefix);
 
        if(rc)
        {
            int WaitingIntervall=1;

            TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_TOOLREQ_CN00);
            if(!StartTool())
                rc=0;

            if(rc)
            {
                TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_OPCHK_CN00);
                cn36_Verbose("Waiting for end of the query.\n", AddALevelToPrefix);

                while(ToolIsRunning())
                    cn36_Sleep(WaitingIntervall++);

                TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_BACKUPOP_CN00);
                if(!ToolWasSuccessful())
                {
                    rc=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseBackupToolFailed, SumOfToolExitCodes()));
                }
                
                sprintf(MsgBuf_cn36, "%s has finished the query %s.\n\n", GiveToolName(), 0==rc?"unsuccessfully":"successfully");
                cn36_Verbose(MsgBuf_cn36, SubALevelFromPrefix);

                TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_POSTOP_CN00);

                if(rc)
                    rc=PostOperationsAfterSuccess();
                    
                if(rc)  // generate list for output
                {
                    rc=PrepareToolOutputOfAvailableBackups();

                    if(rc)
                    {
                        tcn003_BackupHistoryEBIDStatus EBIDStatus;
                        const char * EBID=0;
                        size_t       EBIDLength=0;
                        const char * TypeString=0;
                        const char * DateTime=0;
                        int          LineContainsEBID=0;
                        int          ReachedEnd=0;

                        do
                        {
                            rc=GiveNextAvailableBackup(EBIDStatus, EBID, EBIDLength, TypeString, DateTime, ReachedEnd);

                            if(rc && !ReachedEnd)
                                TypeString=cn35_ListSaveTypeStringForActionString(TypeString);

                            if(rc && !ReachedEnd && 0!=TypeString)
                            {
                                if(0==EBID || 0==EBIDLength || 0==DateTime)
                                {
                                    rc=0;
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                                }
                                else
                                {
                                    tcn003_BackupHistoryEBID * pEBID;

                                    if(OK_CN00==cn32NewHistoryEBID(Info.ReplyData, Info.ReplyLen, &pEBID))
                                    {
                                        if(OK_CN00==cn32SetHistoryEBID(Info.ReplyData, 
                                                                       Info.ReplyLen,
                                                                       EBIDStatus,
                                                                       "",                          //we dont have a key in this situation
                                                                       0,
                                                                       "",                          //and no label too
                                                                       0,
                                                                       EBID,
                                                                       EBIDLength,
                                                                       TypeString,
                                                                       DateTime,
                                                                       pEBID))
                                        {
                                            cn32AddToEndOfHistoryEBIDs(pEBID, &AvailableEBIDs);
                                        }
                                        else
                                            rc=0;
                                    }
                                    else
                                        rc=0;

                                    if(0==rc)
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                                }
                            }
                        }
                        while(rc && !ReachedEnd);
                    }

                    CloseToolOutputOfAvailableBackups();
                }
            }
        }

        // cleanup used resources
        TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_CLEANUP_CN00);
        CleanUp();
    }

    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_CN00);
    cn36_Verbose("\n");

    return rc;
}

int tcn35_BackupToolConnector::PrepareToolOutputOfAvailableBackups()
{
    int rc=0;

    if(0!=ToolProcessOutputAsText)
        delete ToolProcessOutputAsText;
    
    ToolProcessOutputAsText=new tni34_ATextFile(ToolProcessOutput->GiveName(), tni34_AFile::FromDisk, 0, 0); // do not create, do not open now

    if(0==ToolProcessOutputAsText)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
    else
    {
        if(ToolProcessOutputAsText->IsGood() && ToolProcessOutputAsText->Open())
            rc=1;
    }

    return rc;
}

int tcn35_BackupToolConnector::GiveNextAvailableBackup(tcn003_BackupHistoryEBIDStatus & Status,
                                                       const char * & EBID,
                                                       size_t       & EBIDLength,
                                                       const char * & TypeString,
                                                       const char * & DateTimeString,
                                                       int          & ReachedEnd)
{
    int rc=1;
    const char *Line=0;

    ReachedEnd=0;

    if(!ToolProcessOutputAsText->IsGood())
        rc=0;
    else
    {
        int LineContainsEBID=0;

        while(!LineContainsEBID && rc && !ReachedEnd)
        {
            rc=ToolProcessOutputAsText->ReadLine(Line);
                
            if(rc)
            {
                if(0!=Line)
                {
                    rc=GiveInformationOfLine(Line,
                                             EBID,
                                             EBIDLength,
                                             TypeString,
                                             DateTimeString,
                                             LineContainsEBID);
                }
                else
                    ReachedEnd=1;
            }
        }

        if(LineContainsEBID)
            Status=EBIDStatusAvailable;
    }

    return rc;
}

int tcn35_BackupToolConnector::CloseToolOutputOfAvailableBackups()
{
    int rc=1;
    
    if(0!=ToolProcessOutputAsText)
    {
        rc=ToolProcessOutputAsText->Close();

        delete ToolProcessOutputAsText;
        ToolProcessOutputAsText=0;
    }

    return rc;
}

tcn00_Error tcn35_BackupToolConnector::GiveLastError()
{
    return TheDBMSrvErrorHandler_cn36.GiveLastError();
}

int tcn35_BackupToolConnector::GiveSQLResult()
{
    int rc;

    switch(StatusOfRequest)
    {
        case NoRequestSend:         // return SQLCodeOfRequest to comply with part for normal media (utility session can only be used for receiv the reply of the db)
        case RequestSend:           // return SQLCodeOfRequest to comply with part for normal media (utility session can only be used for receiv the reply of the db)
        case RequestReplied:        // return SQLCodeOfRequest to comply with part for normal media (utility session can be used for new backup commands)
        case CanceledByDBMSrv:      // have send a cancel to the utility task, utility task has ended and reply was read (SQLCodeOfRequest should contain an error)
            rc=SQLCodeOfRequest;
        break;
        
        case CancelSendByDBMSrv:    // have send a cancel to the utility task, no answer yet
        case TerminatedByDBMSrv:    // have no further idea to stop db-kernel
            rc=ERR_CN00;            // we can not give any reply of the db kernel -> we reply with an error at least
        break;

        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=ERR_CN00;       
    }

    return rc;
}

int tcn35_BackupToolConnector::IsBad()
{
    return (GiveLastError()!=OK_CN00);
}

const tcn35_BackupToolConnector::tcn35_ActionType tcn35_BackupToolConnector::GiveActionType() const
{
    return ActionType;
}

int tcn35_BackupToolConnector::CreatePipes()
{
    int rc=1;
    
    #if defined (WIN32)
    //database and backup tool have to create the pipes under NT
    #elif defined (UNIX)
    int i;
    cn36_Verbose("Creating pipes for data transfer.\n", AddALevelToPrefix);

    for(i=0; i<Info.NumFiles && NumberOfExistingPipes==i; i++)
    {
        TheDataPipe[i]=new tni34_APipe(Info.ppszFiles[i], tni34_APipe::In, 0, 0);
        if(0==TheDataPipe[i])
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
        else
        {
            sprintf(MsgBuf_cn36,"Creating pipe '%s' ... ", Info.ppszFiles[i]);
            cn36_Verbose(MsgBuf_cn36);

            if(!TheDataPipe[i]->Create())
            {
                cn36_Verbose("Failed.\n", NoPrefix);
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotCreateDataPipe.Set(TheDataPipe[i]->GiveName()));
                delete TheDataPipe[i];
                TheDataPipe[i]=0;
            }
            else
            {
                cn36_Verbose("Done.\n", NoPrefix);
                NumberOfExistingPipes++;
            }
        }
    }

    if(NumberOfExistingPipes==Info.NumFiles)
        cn36_Verbose("All data transfer pipes have been created.\n\n", SubALevelFromPrefix);
    else
    {
        rc=0;
        cn36_Verbose(0==NumberOfExistingPipes?"No data transfer pipes have been created.\n\n":"Not all data transfer pipes have been created.\n\n", SubALevelFromPrefix);
    }
    #else
        #error tcn35_BackupToolConnector::CreatePipes() was only coded for WIN32 and UNIX (define one of them).
    #endif

    return rc;
}

int tcn35_BackupToolConnector::RemovePipes()
{
    int i, rc=0;

    for(i=NumberOfExistingPipes-1; 0<=i; i--)
    {
        sprintf(MsgBuf_cn36, "Removing data transfer pipe %s ... ", TheDataPipe[i]->GiveName());
        cn36_Verbose(MsgBuf_cn36);

        if(TheDataPipe[i]->Remove())
        {
            NumberOfExistingPipes--;
            cn36_Verbose("Done.\n", NoPrefix);
        }
    }

    if(0==NumberOfExistingPipes)
        rc=1;
    else
        NumberOfExistingPipes=0;  // declare all pipes as removed, no way for handling unremoveable pipes

    return rc;
}

int tcn35_BackupToolConnector::CheckMedium()
{
    int rc=1;
    
    cn36_Verbose("Checking medium.\n", AddALevelToPrefix);

    // TODO?: check NumFiles against MAX_NUMBER_OF_PIPES_CN35 ?
    // TODO?: check type "P" of each single medium

    cn36_Verbose(rc?"Check passed successfully.\n\n":"Check failed.\n\n", SubALevelFromPrefix);

    return rc;
}

const char * tcn35_BackupToolConnector::GivePathOfExternalBackupHistory(void)   //returns 0 on any error
{
    if(0==PathOfExternalBackupHistory)
    {
        tsp00_DbNamec tDBName;
        tDBName.rawAssign(Info.pszDBName);  //==strcpy((char *)tDBName, Info.pszDBName);
        tsp00_C264c tAbsolutNameOfToolLogFile;

        tcn00_Error Err=cn42GetFileName(tDBName, FGET_BACKEBF_CN42, tAbsolutNameOfToolLogFile);

        if(OK_CN00==Err)
            cn36_StrAllocCpy(PathOfExternalBackupHistory, (char *)tAbsolutNameOfToolLogFile);   //  0==PathOfExternalBackupHistory, if cn36_StrAllocCpy fails
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotGetNameOfToolLogFile.SetI(int(Err)));
    }
    // else use the PathOfExternalBackupHistory, which was determined in earlier calls of this function

    return PathOfExternalBackupHistory;
}

const char * tcn35_BackupToolConnector::GivePathOfBackupHistory(void)   //returns 0 on any error
{
    if(0==PathOfBackupHistory)
    {
        tsp00_DbNamec tDBName;
        tDBName.rawAssign(Info.pszDBName);  //==strcpy((char *)tDBName, Info.pszDBName);
        tsp00_C264c tAbsoluteNameOfDBMKnl;

        tcn00_Error Err=cn42GetFileName(tDBName, FGET_BACKHIST_CN42, tAbsoluteNameOfDBMKnl);

        if(OK_CN00==Err)
            cn36_StrAllocCpy(PathOfBackupHistory, (char *)tAbsoluteNameOfDBMKnl);   // 0==PathOfBackupHistory, if cn36_StrAllocCpy fails
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotGetNameOfBackupHistory.SetI(int(Err)));
    }
    // else use the PathOfBackupHistory, which was determined in earlier calls of this function

    return PathOfBackupHistory;
}

const char * tcn35_BackupToolConnector::ActionStringForDBMSaveType(const char * DBMSaveType, int WithoutCheckpoint)
{
    return cn35_ActionStringForDBMSaveType(DBMSaveType, WithoutCheckpoint);
}

const char * tcn35_BackupToolConnector::ActionStringForActionStringN(const char *ActionString, size_t LengthActionString)
{ 
    const char *rc=0;

    if(0!=ActionString)
    {
        int i=0;

        // check actions with checkpoint
        while(TheSaveTypeMap[i].DBMSrvSaveType[0]!='\0' &&                 //while not at end of the list
              0!=cn36_StrStrNCmp(TheSaveTypeMap[i].ToolSaveTypeCheckpoint, ActionString, LengthActionString)) //and type keywords are not equal
        {
            i++;
        }

        if('\0'!=TheSaveTypeMap[i].DBMSrvSaveType[0])       // action string found?
            rc=TheSaveTypeMap[i].ToolSaveTypeCheckpoint;    
        else
        {                                                   // check actions without checkpoint
            i=0;

            while(TheSaveTypeMap[i].DBMSrvSaveType[0]!='\0' &&                 //while not at end of the list
                  0!=cn36_StrStrNCmp(TheSaveTypeMap[i].ToolSaveTypeNoCheckpoint, ActionString, LengthActionString)) //and type keywords are not equal
            {
                i++;
            }
        
            if('\0'!=TheSaveTypeMap[i].DBMSrvSaveType[0])  // we should find always the correct type (before reaching end of map)
                rc=TheSaveTypeMap[i].ToolSaveTypeNoCheckpoint;
            //else 0==rc holds already
        }
    }

    return rc;
}

int tcn35_BackupToolConnector::StartDBAction()
{
    int rc=1;

    cn36_Verbose("Starting database action for", AddALevelToPrefix);
    cn36_Verbose(GiveActionType()==ABackup?" the backup.\n":" the restore.\n", NoPrefix);

    rc=SendBackupRestoreRequestToKernel(*(Info.SQLSession), Info.KernelCommand, SQLCodeOfRequest);

    if(!rc)
    {
        cn36_Verbose("Could not start database action.\n", SubALevelFromPrefix);
        
        StatusOfRequest=NoRequestSend;
    }
    else
    {
        cn36_Verbose("The database is working on the request.\n\n", SubALevelFromPrefix);
        
        StatusOfRequest=RequestSend;
    }

    return rc;
}

void tcn35_BackupToolConnector::ReceiveReplyFromDBKernel(tin01_sql_session   & Session,
                                                         int                 & SQLCode,
                                                         tcn003_BackupResult & tResult,
                                                         boolean             & IsNoSQLError)
{
    const int RealParamCount=25;    //tFetchVars and sizeArray has at least 25 parts
                                    //Do not use BA_RESULT_PARAM_COUNT_CN003, because sizeArray maybe initialised with less parts.
                                    //This leads to a core.

    int             i;
    tin01_c_sqlca   tFetchCA;
    tin01_c_hostvar tFetchVars[BA_RESULT_PARAM_COUNT_CN003];
    tsp00_Int4      sizeArray[BA_RESULT_PARAM_COUNT_CN003]={sizeof (tResult.szDate),
                                                            sizeof (tResult.szTime),
                                                            sizeof (tResult.szDatabase),
                                                            sizeof (tResult.szServer),
                                                            sizeof (tResult.szKernelVersion),
                                                            sizeof (tResult.szPagesTransfered),
                                                            sizeof (tResult.szPagesLeft),
                                                            sizeof (tResult.szVolumes),
                                                            sizeof (tResult.szMediaName),
                                                            sizeof (tResult.szLocation),
                                                            sizeof (tResult.szErrortext),
                                                            sizeof (tResult.szLabel),
                                                            sizeof (tResult.szIsConsistent),
                                                            sizeof (tResult.szFirstLogPageNo),
                                                            sizeof (tResult.szLastLogPageNo),
                                                            sizeof (tResult.szDBStamp1Date),
                                                            sizeof (tResult.szDBStamp1Time),
                                                            sizeof (tResult.szDBStamp2Date),
                                                            sizeof (tResult.szDBStamp2Time),
                                                            sizeof (tResult.szBDPageCount),
                                                            sizeof (tResult.szDevicesUsed),
                                                            sizeof (tResult.szDatabaseID),
                                                            sizeof (tResult.szMaxUsedDataPage),
                                                            sizeof (tResult.szRedoTransRead),
                                                            sizeof (tResult.szRedoTransDone)};
                                                
    // init hostvar settings
    for(i=0;i<BA_RESULT_PARAM_COUNT_CN003;i++)
    {
        tFetchVars[i].len   =sizeArray[i];
        tFetchVars[i].c_type.becomes(cin01_c_charp);
    }

    tFetchCA.inUse     =false;
    //tFetchCA.varCount  =BA_RESULT_PARAM_COUNT_CN003;
    //tFetchCA.paramCount=BA_RESULT_PARAM_COUNT_CN003;
    tFetchCA.varCount  =RealParamCount;
    tFetchCA.paramCount=RealParamCount;
    tFetchCA.hostVar   =&tFetchVars[0];

    tResult.szDate[0] = '\0';
    tResult.szTime[0] = '\0';
    tResult.szDatabase[0] = '\0';
    tResult.szServer[0] = '\0';
    tResult.szKernelVersion[0] = '\0';
    tResult.szPagesTransfered[0] = '\0';
    tResult.szPagesLeft[0] = '\0';
    tResult.szVolumes[0] = '\0';
    tResult.szMediaName[0] = '\0';
    tResult.szLocation[0] = '\0';
    tResult.szErrortext[0] = '\0';
    tResult.szLabel[0] = '\0';
    tResult.szIsConsistent[0] = '\0';
    tResult.szFirstLogPageNo[0] = '\0';
    tResult.szLastLogPageNo[0] = '\0';
    tResult.szDBStamp1Date[0] = '\0';
    tResult.szDBStamp1Time[0] = '\0';
    tResult.szDBStamp2Date[0] = '\0';
    tResult.szDBStamp2Time[0] = '\0';
    tResult.szBDPageCount[0] = '\0';
    tResult.szDevicesUsed[0] = '\0';
    tResult.szDatabaseID[0] = '\0';
    tResult.szMaxUsedDataPage[0]='\0';
    tResult.szRedoTransRead[0]='\0';
    tResult.szRedoTransDone[0]='\0';

    SQLCode=i29utilbackupreceive(&Session, &tFetchCA, /*&bBackupRC*/&IsNoSQLError,
                            tResult.szDate,
                            tResult.szTime,
                            tResult.szDatabase,
                            tResult.szServer,
                            tResult.szKernelVersion,
                            tResult.szPagesTransfered,
                            tResult.szPagesLeft,
                            tResult.szVolumes,
                            tResult.szMediaName,
                            tResult.szLocation,
                            tResult.szErrortext,
                            tResult.szLabel,
                            tResult.szIsConsistent,
                            tResult.szFirstLogPageNo,
                            tResult.szLastLogPageNo,
                            tResult.szDBStamp1Date,
                            tResult.szDBStamp1Time,
                            tResult.szDBStamp2Date,
                            tResult.szDBStamp2Time,
                            tResult.szBDPageCount,
                            tResult.szDevicesUsed,
                            tResult.szDatabaseID,
                            tResult.szMaxUsedDataPage,
                            tResult.szRedoTransRead,
                            tResult.szRedoTransDone);


    cn36_Verbose("Got the following reply from db-kernel:\n", AddALevelToPrefix);

    sprintf(MsgBuf_cn36, "SQL-Code              :%d\n", int(SQLCode));
    cn36_Verbose(MsgBuf_cn36);
    
    if('\0'!=tResult.szDate[0])
    {
        sprintf(MsgBuf_cn36, "Date                  :%s\n", tResult.szDate);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szTime[0])
    {
        sprintf(MsgBuf_cn36, "Time                  :%s\n", tResult.szTime);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDatabase[0])
    {
        sprintf(MsgBuf_cn36, "Database              :%s\n", tResult.szDatabase);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szServer[0])
    {
        sprintf(MsgBuf_cn36, "Server                :%s\n", tResult.szServer);
        cn36_Verbose(MsgBuf_cn36);
    }
    
    if('\0'!=tResult.szKernelVersion[0])
    {
        sprintf(MsgBuf_cn36, "KernelVersion         :%s\n", tResult.szKernelVersion);
        cn36_Verbose(MsgBuf_cn36);
    }
    
    if('\0'!=tResult.szPagesTransfered[0])
    {
        sprintf(MsgBuf_cn36, "PagesTransfered       :%s\n", tResult.szPagesTransfered);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szPagesLeft[0])
    {
        sprintf(MsgBuf_cn36, "PagesLeft             :%s\n", tResult.szPagesLeft);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szVolumes[0])
    {
        sprintf(MsgBuf_cn36, "Volumes               :%s\n", tResult.szVolumes);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szMediaName[0])
    {
        sprintf(MsgBuf_cn36, "MediaName             :%s\n", tResult.szMediaName);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szLocation[0])
    {
        sprintf(MsgBuf_cn36, "Location              :%s\n", tResult.szLocation);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szErrortext[0])
    {
        sprintf(MsgBuf_cn36, "Errortext             :%s\n", tResult.szErrortext);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szLabel[0])
    {
        sprintf(MsgBuf_cn36, "Label                 :%s\n", tResult.szLabel);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szIsConsistent[0])
    {
        sprintf(MsgBuf_cn36, "IsConsistent          :%s\n", tResult.szIsConsistent);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szFirstLogPageNo[0])
    {
        sprintf(MsgBuf_cn36, "FirstLogPageNo        :%s\n", tResult.szFirstLogPageNo);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szLastLogPageNo[0])
    {
        sprintf(MsgBuf_cn36, "LastLogPageNo         :%s\n", tResult.szLastLogPageNo);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDBStamp1Date[0])
    {
        sprintf(MsgBuf_cn36, "DBStamp1Date          :%s\n", tResult.szDBStamp1Date);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDBStamp1Time[0])
    {
        sprintf(MsgBuf_cn36, "DBStamp1Time          :%s\n", tResult.szDBStamp1Time);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDBStamp2Date[0])
    {
        sprintf(MsgBuf_cn36, "DBStamp2Date          :%s\n", tResult.szDBStamp2Date);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDBStamp2Time[0])
    {
        sprintf(MsgBuf_cn36, "DBStamp2Date          :%s\n", tResult.szDBStamp2Time);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szBDPageCount[0])
    {
        sprintf(MsgBuf_cn36, "BDPageCount           :%s\n", tResult.szBDPageCount);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDevicesUsed[0])
    {
        sprintf(MsgBuf_cn36, "DevicesUsed           :%s\n", tResult.szDevicesUsed);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szDatabaseID[0])
    {
        sprintf(MsgBuf_cn36, "DatabaseID            :%s\n", tResult.szDatabaseID);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szMaxUsedDataPage[0])
    {
        sprintf(MsgBuf_cn36, "Max Used Data Page    :%s\n", tResult.szMaxUsedDataPage);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szRedoTransRead[0])
    {
        sprintf(MsgBuf_cn36, "Redo Transactions Read:%s\n", tResult.szRedoTransRead);
        cn36_Verbose(MsgBuf_cn36);
    }

    if('\0'!=tResult.szRedoTransDone[0])
    {
        sprintf(MsgBuf_cn36, "Redo Transactions Done:%s\n", tResult.szRedoTransDone);
        cn36_Verbose(MsgBuf_cn36);
    }

    cn36_Verbose(SubALevelFromPrefix);
}

void tcn35_BackupToolConnector::ReceiveBackupRestoreRequestReplyFromDBKernel(int ForLogDelete)
{
    boolean IsNoSQLError=true;
    tcn003_BackupResult TheDeleteLogResult;

    cn36_Verbose("Receiving Reply.\n");

    if(!ForLogDelete)
    {
        ReceiveReplyFromDBKernel(*Info.SQLSession, SQLCodeOfRequest, tTheDBResult, IsNoSQLError);

        if(0==SQLCodeOfRequest)
        {
            cn36_StrAllocCpy(FirstLogPageNo, tTheDBResult.szFirstLogPageNo);    // we check for error (0==FirstLogPageNo) before usage
            cn36_StrAllocCpy(LastLogPageNo, tTheDBResult.szLastLogPageNo);      // we check for error (0==LastLogPageNo) before usage
        }
        else
        {
            cn36_StrDealloc(FirstLogPageNo);
            cn36_StrDealloc(LastLogPageNo);
        }
    }
    else
        ReceiveReplyFromDBKernel(*Info.SQLSession, SQLCodeOfRequest, TheDeleteLogResult, IsNoSQLError); //using TheDeleteLogResult, to avoid overwriting of tTheDBResult if all is done correctly

    SQLErrForRequest=!IsNoSQLError;

    if(CancelSendByDBMSrv==StatusOfRequest)
        StatusOfRequest=CanceledByDBMSrv;
    else
        StatusOfRequest=RequestReplied;  // we got something (an error or a reply)!

    if(0!=SQLCodeOfRequest && SQLErrForRequest)
        i29errmsg(Info.SQLSession, SQLErrText);

    if(0!=SQLCodeOfRequest)
        if(ForLogDelete)
            tTheDBResult=TheDeleteLogResult;    //if we get an error, overwrite the tTheDBResult of the successful backup with the TheDeleteLogResult of the failed "DELETE LOG ..."
}

int tcn35_BackupToolConnector::GetStatusOfUtilityTask(tcn003_BackupResult &StatusInformation)
{
    tin01_sql_session sqlsession;
    int               rc;
    int               SQLCode;
    int               GotReplyFromKernel=0;

    // connect temporary
    i29initsession(&sqlsession, NULL);
    i29pascalstring((char *) sqlsession.xuser.xu_serverdb, 
                    sizeof (sqlsession.xuser.xu_serverdb), 
                    Info.pszDBName,
                    -1);
    i29pascalstring((char *) sqlsession.xuser.xu_user,     
                    sizeof (sqlsession.xuser.xu_user), 
                    (const char *) " ",
                    -1);
    i29pascalstring((char *) sqlsession.xuser.xu_password, 
                    sizeof (sqlsession.xuser.xu_password), 
                    (const char *) " ",
                    -1);
    
    rc=i29specialconnect(&sqlsession);  //returns 0 if successful!
    
    // ask for state
    if(0==rc)
    {
        cn36_Verbose("Asking for state of database.\n");

        const char * Command=0;

        switch(ActionType)
        {
            case ABackup:
                Command="STATE SAVE";
            break;
            case ARestore:
                Command="STATE RESTORE";
            break;
            default:
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set()); //0==Command already
        }

        if(0!=Command)
            rc=SendBackupRestoreRequestToKernel(sqlsession, Command, SQLCode, 1, 0);    //quiet and not for backup
    }

    // print reply
    if(0==SQLCode)
    {
        boolean IsNoSQLError;
        ReceiveReplyFromDBKernel(sqlsession, SQLCode, StatusInformation, IsNoSQLError);
        
        if(0==SQLCode)                // ignore IsNoSQLError 
            GotReplyFromKernel=1;
        else
        {    
            tsp00_C256c errtext;
            
            i29errmsg(&sqlsession, errtext);
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::SQLError.SetSQL(SQLCode, (char *)errtext));
        }
    }

    if(sqlsession.is_connected)
        i29specialrelease(&sqlsession);

    return GotReplyFromKernel;
}

int tcn35_BackupToolConnector::SendBackupRestoreRequestToKernel(tin01_sql_session & Session,
                                                                const char        * KernelCommand,
                                                                int               & SQLCode,
                                                                int                 BeQuiet,
                                                                int                 ForBackup)
{
    int rc=1;

    if(!BeQuiet)
    {
        sprintf(MsgBuf_cn36, "Requesting '%s' from db-kernel.\n", KernelCommand);
        cn36_Verbose(MsgBuf_cn36);
    }

    if(ForBackup)
        SQLCode=i29utilbackuprequest(&Session, false, KernelCommand);
    else
        SQLCode=i29utilbackuprequest(&Session, true, KernelCommand);

    if(0!=SQLCode)
    {
        tsp00_C256c errtext;
        i29errmsg(&Session, errtext);

        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::SQLError.SetSQL(SQLCode, (char *)errtext));

        rc=0;
    }

    return rc;
}

int tcn35_BackupToolConnector::ReplyAvailable(tin01_sql_session & Session)
{
  int rc=1;
  
  if(0==i29replyavailable(&Session))
      cn36_Verbose("Reply from database available.\n");
  else
  {
      cn36_Verbose("No reply from database available.\n");
      rc=0;
  }

  return rc;
}

int tcn35_BackupToolConnector::DBIsReady()
{
    int IsReady=0;

    if(ABackup==ActionType)
    {
        cn36_Verbose("Waiting until database has prepared the backup.\n", AddALevelToPrefix);

        tcn003_BackupResult StatusInformation;
        int ReplyFromDBKernel=ReplyAvailable(*Info.SQLSession),
            SleepTime=1;

        while(!ReplyFromDBKernel && !IsReady) // db-kernel should send reply in case of error or set szPagesLeft to a value
        {
            cn36_Sleep(SleepTime++);

            if(GetStatusOfUtilityTask(StatusInformation))
                if(StatusInformation.szPagesLeft[0]!='\0')
                    IsReady=1;

            cn36_Verbose("Checking for reply of backup request.\n");
            ReplyFromDBKernel=ReplyAvailable(*Info.SQLSession);
        }

        if(ReplyFromDBKernel) // Reply available means, that db-kernel has finished the action
        {
            IsReady=0;
            ReceiveBackupRestoreRequestReplyFromDBKernel();
        }

        cn36_Verbose(ReplyFromDBKernel?"Database has already finished the request.\n\n":"The database has prepared the backup successfully.\n\n", SubALevelFromPrefix);
    }
    else
        IsReady=1;

    return IsReady;
}

int tcn35_BackupToolConnector::StartTool()
{
    int i,rc=1;

    sprintf(MsgBuf_cn36,"Starting %s.\n", GiveToolName());
    cn36_Verbose(MsgBuf_cn36, AddALevelToPrefix);
    
    if(0==NumberOfRunningToolProcesses)
    {
        for(i=0; i<NumberOfCommandLines && rc; i++)
        {
            if(0==ToolProcessOutput || 0==ToolProcessError)
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                rc=0;
            }
            else
            {
                sprintf(MsgBuf_cn36,
                        "Starting %s process '%s >>%s 2>>%s'.\n",
                        GiveToolName(),
                        CommandLine[i],
                        ToolProcessOutput->GiveName(),
                        ToolProcessError->GiveName());

                cn36_Verbose(MsgBuf_cn36);

                ToolProcess[i]=new tni33_ABackgroundProcess(CommandLine[i], (tni34_AFile *)(0), ToolProcessOutput, 1, ToolProcessError, 1);
            }

            if(0==ToolProcess[i])
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                rc=0;
            }
            else
                if(!ToolProcess[i]->WasStarted())
                    rc=0;
                else
                {
                    NumberOfRunningToolProcesses++;
                    ToolProcessWasRunning[i]=1;
                }

            if(rc)
                cn36_Verbose("Process was started successfully.\n");
            else
                cn36_Verbose("Start of process failed.\n");
        }
    }
    else
        rc=0;

    if(rc)
        sprintf(MsgBuf_cn36, "%s has been started successful.\n\n", GiveToolName());
    else
        sprintf(MsgBuf_cn36, "Could not start %s correctly.\n\n", GiveToolName());

    cn36_Verbose(MsgBuf_cn36, SubALevelFromPrefix);

    return rc;
}

int tcn35_BackupToolConnector::MaxWaitingTimeOfCurrentWaitingPhase()
{
    int rc=60;

    if(WaitForBackupTool==CurrentWaitingPhase || WaitForKilled==CurrentWaitingPhase)
        if(DBActionWasSuccessful())
            rc=5*60;    //increase MaxWaitingTime, if db was successful

    return rc;
}

int tcn35_BackupToolConnector::CheckTool()
{
    return 1;
};

int tcn35_BackupToolConnector::ToolIsRunning()
{
    int i,
        rc=1,
        SomeThingRunning=0;

    cn36_Verbose("Checking backup tool.\n", AddALevelToPrefix);

    for(i=0;i<NumberOfRunningToolProcesses;i++)
    {
        if(!ToolProcessWasRunning[i])
            rc=0;
        else
        {
            if(!ToolProcess[i]->IsRunning())
            {
                rc=0;

                sprintf(MsgBuf_cn36,"%s archiving tool process has finished work with return code %d.\n", NumberOfRunningToolProcesses==1 ? "The" : "An", int(ToolProcess[i]->GiveReturnCode()));
                cn36_Verbose(MsgBuf_cn36);

                ToolProcessWasRunning[i]=0;
            }
            else
                SomeThingRunning=1;
        }
    }

    if(!SomeThingRunning)    // just to be sure in the case of 0>=NumberOfRunningToolProcesses
        rc=0;

    if(rc)
        cn36_Verbose("The backup tool is running.\n", SubALevelFromPrefix);
    else
        cn36_Verbose(SomeThingRunning?"The backup tool is not running completely.\n":"The backup tool is not running.\n", SubALevelFromPrefix);

    return rc;
}

int tcn35_BackupToolConnector::ExistsToolProcesses()
{
    int i,rc=0;

    for(i=0;i<MAX_NUMBER_OF_TOOLPROCESSES_CN35 && 0==rc;i++)
        if(0!=ToolProcess[i] && ToolProcess[i]->IsRunning())
            rc=1;

    return rc;
}

int tcn35_BackupToolConnector::SignalEndToTool()
{
    return 1;
}

int tcn35_BackupToolConnector::DBActionIsRunning()
{
    return (StatusOfRequest==RequestSend);
}

int  tcn35_BackupToolConnector::KernelMayBeUsingLogDevspaceForLogRestore()
{
    int rc=0;

    if(ARestore==GiveActionType() &&                                            //a restore...
       0==strcmp(Info.pszActionString, tcn35_SaveInfo::Action_String_Log) &&    //... of log...
       RequestSend==StatusOfRequest)                                            //... is still running
    {
        tcn003_BackupResult StatusInformation;

        if(GetStatusOfUtilityTask(StatusInformation) &&                                             //could get make a "STATE RESTORE"
           0!=StatusInformation.szPagesTransfered && '\0'!=StatusInformation.szPagesTransfered[0])  //szPagesTransfered is not empty
        {
            teo00_Uint8 PagesTransfered;

            if(cn36_StrToUInt8(PagesTransfered, StatusInformation.szPagesTransfered) && //StatusInformation.szPagesTransfered could be converted to a teo00_Uint8
               0<PagesTransfered)                                                       //the database kernel has read already read (at least) one page of the pipe
            {
                rc=1;
            }
        }
    }

    return rc;
}

void tcn35_BackupToolConnector::CancelTool()
{
    int PipeDirectionKnown=1,
        i;

    tni34_APipe::ADirection PipeDirection; 

    switch(GiveActionType())
    {
        case ABackup:
            PipeDirection=tni34_APipe::Out;
        break;

        case ARestore:
            PipeDirection=tni34_APipe::In;
        break;

        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            PipeDirectionKnown=0;
    }

    if(PipeDirectionKnown)
        for(i=0; i<Info.NumFiles; i++)
        {
            tni34_APipe Pipe(Info.ppszFiles[i], PipeDirection, 0, 0);    //do not create now, do not open now

            sprintf(MsgBuf_cn36, "Trying to break pipe '%s'.\n", Pipe.GiveName());
            cn36_Verbose(MsgBuf_cn36);

            if(Pipe.BreakAndRemovePipeIfOpened())
                cn36_Verbose("Pipe has been broken successfully.\n");
            else
                cn36_Verbose("Pipe couldn't be broken successfully.\n");
        }
}

void tcn35_BackupToolConnector::TerminateTool()
{
    int i;

    for(i=0;i<MAX_NUMBER_OF_TOOLPROCESSES_CN35;i++)
        if(0!=ToolProcess[i] && ToolProcess[i]->IsRunning())
            ToolProcess[i]->Stop();
}

int tcn35_BackupToolConnector::PostOperationsAfterSuccess()
{
    return 1;   // normaly we have no real post operations, we just do clean up (see CleanUp())
}

void tcn35_BackupToolConnector::CleanUp()
{
    int AllOk=1;

    TheDBMSrvErrorHandler_cn36.SetPhaseTo(ERR_CLEANUP_CN00);
    
    cn36_Verbose("Cleaning up.\n", AddALevelToPrefix);
    
    if(!RemovePipes())                  // remove all created "data-transfer" pipes (must not be all pipes of medium) 
        AllOk=0;

    if(!MoveToolOutputToProtocolFile()) // move stdout and stderr of started tool processes to the protocol (dbm.ebp)
        AllOk=0;

    if(!CleanUpTool())                  // do tool specific cleanup
        AllOk=0;
    
    cn36_Verbose(AllOk?"Have finished clean up successfully.\n\n":"Could not clean up all correctly.\n\n", SubALevelFromPrefix);
}

int tcn35_BackupToolConnector::MoveFileToProtocolFile(const char *FileDesignation, tni34_AFile & TheFileToMove)
{
    int rc=1;
    
    TheFileToMove.Close();

    if(TheFileToMove.IsGood())
    {
        if('\0'==PathOfProtFile_cn36[0] ||
           '\0'==(*TheFileToMove.GiveName()) ||
           '\0'==FileDesignation[0])
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
        }
        else
        {
            sprintf(MsgBuf_cn36, "Moving %s of %s to this file.\n", FileDesignation, GiveToolName());
            cn36_Verbose(MsgBuf_cn36);
            sprintf(MsgBuf_cn36, "---------- Begin of %s of %s (%s)----------\n", FileDesignation, GiveToolName(), TheFileToMove.GiveName());
            cn36_Verbose(MsgBuf_cn36);

            rc=cn36_AppendFileToFile(TheFileToMove.GiveName(), PathOfProtFile_cn36);

            sprintf(MsgBuf_cn36, "---------- End of %s of %s (%s)----------\n", FileDesignation, GiveToolName(), TheFileToMove.GiveName());
            cn36_Verbose(MsgBuf_cn36);

            if(!rc)
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotMoveToolOutputToToolProtFile.Set(TheFileToMove.GiveName())); // now report the write-Problem

                sprintf(MsgBuf_cn36,
                        "Not removing %s's temporary %s file (%s).\n",
                        GiveToolName(),
                        FileDesignation,
                        TheFileToMove.GiveName());

                cn36_Verbose(MsgBuf_cn36);
            }
        }

        if(rc && TheFileToMove.Exists() && TheFileToMove.IsAFile())
        {
            if(!TheFileToMove.Remove())
            {
                rc=0;

                sprintf(MsgBuf_cn36,
                        "Could not remove %s's temporary %s file '%s'.\n",
                        GiveToolName(),
                        FileDesignation,
                        TheFileToMove.GiveName());

                cn36_Verbose(MsgBuf_cn36);
            }
            else
            {
                sprintf(MsgBuf_cn36,
                        "Removed %s's temporary %s file '%s'.\n",
                        GiveToolName(),
                        FileDesignation,
                        TheFileToMove.GiveName());

                cn36_Verbose(MsgBuf_cn36);
            }
        }
        else
            rc=0;
    }
    else
        rc=0;

    return rc;
}

int tcn35_BackupToolConnector::MoveToolOutputToProtocolFile()
{
    int rc=1;

    if(0!=ToolProcessOutput)
    {
        if(!MoveFileToProtocolFile("output", *ToolProcessOutput))
            rc=0;

        delete ToolProcessOutput;
        ToolProcessOutput=0;
    }
    
    if(0!=ToolProcessError)
    {
        if(!MoveFileToProtocolFile("error output", *ToolProcessError))
            rc=0;

        delete ToolProcessError;
        ToolProcessError=0;
    }

    return rc;
}

int tcn35_BackupToolConnector::SumOfToolExitCodes()
{
    int i,sum;

    for(sum=i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35; i++)
        if(0!=ToolProcess[i])
            sum+=ToolProcess[i]->GiveReturnCode();

    return sum;
}

void tcn35_BackupToolConnector::FreeLogAfterLogBackup()
{
    if(DBActionWasSuccessful())
    //if(RequestReplied==StatusOfRequest &&
    //   0==SQLCodeOfRequest)
    {
        if(0!=FirstLogPageNo && 0<strlen(FirstLogPageNo) &&     //FirstLogPageNo is not empty
           0!=LastLogPageNo && 0<strlen(LastLogPageNo))         //LastLogPageNo is not empty
        {
            int AllOk=0;

            char * Command=0;
            const char * FreeCommandPart1="DELETE LOG FROM ";
            const char * FreeCommandPart2=" TO ";

            sprintf(MsgBuf_cn36, "Freeing log pages from %s to %s for overwriting.\n", FirstLogPageNo, LastLogPageNo);
            cn36_Verbose(MsgBuf_cn36, AddALevelToPrefix);

            if(cn36_StrAlloc(Command,
                             strlen(FreeCommandPart1)
                             +strlen(FirstLogPageNo)
                             +strlen(FreeCommandPart2)
                             +strlen(LastLogPageNo)))
            {
                sprintf(Command, "%s%s%s%s", FreeCommandPart1, FirstLogPageNo, FreeCommandPart2, LastLogPageNo);

                if(SendBackupRestoreRequestToKernel(*(Info.SQLSession), Command, SQLCodeOfRequest))
                {
                    cn36_Verbose("The database is working on the request.\n");
        
                    StatusOfRequest=RequestSend;

                    ReceiveBackupRestoreRequestReplyFromDBKernel(1);

                    if(0==SQLCodeOfRequest)
                        AllOk=1;
                }
                else
                    cn36_Verbose("Could not request freeing of log for overwriting.\n"/*, SubALevelFromPrefix*/);

                cn36_StrDealloc(Command);
            }

            sprintf(MsgBuf_cn36, AllOk?"Log pages freed successfully.\n\n":"Could not free logpages.\n\n");
            cn36_Verbose(MsgBuf_cn36, SubALevelFromPrefix);
        }
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseNoLogPageInformation));
    }
    else
        cn36_Verbose("Not freeing log pages for overwriting.\n\n");
}

void tcn35_BackupToolConnector::SignalEndToDB()
{
    cn36_Verbose("Canceling Utility-task ... ");
    i29cancel(Info.SQLSession);
    cn36_Verbose("OK.\n", NoPrefix);

    StatusOfRequest=CancelSendByDBMSrv;
}

void tcn35_BackupToolConnector::TerminateDBAction()
{
    if(TerminatedByDBMSrv!=StatusOfRequest)
    {
        if(ReplyAvailable(*Info.SQLSession))   // check db for a last time again
            ReceiveBackupRestoreRequestReplyFromDBKernel();
        else
        {
            //TODO?: something else to do here? kill utility-task ?
            StatusOfRequest=TerminatedByDBMSrv;
        }
    }
}

int tcn35_BackupToolConnector::ToolWasSuccessful()
{
    int rc=1;

    if(ExistsToolProcesses())
        rc=0;
    else
    {
        int i;
        for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35 && rc!=0; i++)
            if(0!=ToolProcess[i] && 0!=(ToolProcess[i]->GiveReturnCode()))
                rc=0;
    }
    
    return rc;
}

int tcn35_BackupToolConnector::DBActionWasSuccessful()
{
    int rc;

    if(ARestore==GiveActionType() && 0==strcmp(Info.pszActionString, tcn35_SaveInfo::Action_String_Log))
        rc=(StatusOfRequest==RequestReplied &&
            (0==SQLCodeOfRequest || -8020==SQLCodeOfRequest));
    else
        rc=(StatusOfRequest==RequestReplied && 0==SQLCodeOfRequest);

    return rc;
}

int tcn35_BackupToolConnector::ToolMustNotBeSuccessful()
{
    const int NumberOfErrorsAllowedForRestore=6;
    int rc=0,
        i,
        //ToolErrorAllowedForBackupSQLCode[]={-3004},
        ToolErrorAllowedForRestoreSQLCode[NumberOfErrorsAllowedForRestore]={0, -3004, -7075, -7076, -8003, -8020};

    switch(GiveActionType())
    {
        case ABackup:   
        {
            //The tool must always be successful on backup
            //for(i=0; i<1; i++)
            //    if(SQLCodeOfRequest==ToolErrorAllowedForBackupSQLCode[i])
            //        rc=1;
        }break;

        case ARestore:
        {
            if(0==strcmp(Info.pszActionString, tcn35_SaveInfo::Action_String_Log)) //while restoring log ignore the errors from ToolErrorAllowedForRestoreSQLCode
            {
                for(i=0; i<NumberOfErrorsAllowedForRestore; i++)
                    if(SQLCodeOfRequest==ToolErrorAllowedForRestoreSQLCode[i])
                        rc=1;
            }
        }break;

        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    }

    return rc;
}

const char * FindSectionInLine(int NoOfSection, const char * Line)
{
    int i=0;

    while(0!=Line && i<NoOfSection)
    {
        Line=strchr(Line, '|');
        if(0!=Line)
        {
            i++;
            Line++;
        }
    }

    return Line;    // returns 0 if not NoOfSections '|' can be found
}

const char * FindNextSectionInLine(const char * Line)
{
    if(0!=Line)
    {
        Line=strchr(Line, '|');
        if(0!=Line)
            Line++;
    }

    return Line; // returns 0 if the next Section can not be found
}

const char * tcn35_BackupToolConnector::GetSaveIDFromDBMKnl(const char * AbsoluteNameOfDBMKnl, const char * Label, const char * szDBStamp1Date, const char * szDBStamp1Time)
{
    int Found=0;
    static char SaveID[48]; // TODO: why 48????

    cn36_Verbose(AddALevelToPrefix);

    tni34_ATextFile DBMKnl(AbsoluteNameOfDBMKnl, tni34_AFile::FromDisk, 0);

    if(DBMKnl.IsGood())
    {
        char DateTimeInDBMKnlFormat[20]="yyyy-mm-dd hh:mm:ss";

        //cn36_StrNCpy(DateTimeInDBMKnlFormat, szDBStamp1Date, 4);        //get year
        //DateTimeInDBMKnlFormat[4]='-';
        //cn36_StrNCpy(DateTimeInDBMKnlFormat+5, szDBStamp1Date+4, 2);    //get month
        //DateTimeInDBMKnlFormat[7]='-';
        //cn36_StrNCpy(DateTimeInDBMKnlFormat+8, szDBStamp1Date+6, 2);    //get day
        //DateTimeInDBMKnlFormat[10]=' ';

        //cn36_StrNCpy(DateTimeInDBMKnlFormat+11, szDBStamp1Time+2, 2); //get hour, ignore first two 00
        //DateTimeInDBMKnlFormat[13]=':';
        //cn36_StrNCpy(DateTimeInDBMKnlFormat+14, szDBStamp1Time+4, 2); //get minute
        //DateTimeInDBMKnlFormat[16]=':';
        //cn36_StrNCpy(DateTimeInDBMKnlFormat+17, szDBStamp1Time+6, 2); //get second 

        char *CurrPos=DateTimeInDBMKnlFormat;

        cn36_StrNCpy(CurrPos, szDBStamp1Date, 4);       //copy year
        CurrPos+=strlen(CurrPos);

        *(CurrPos++)='-';

        cn36_StrNCpy(CurrPos, szDBStamp1Date+4, 2);     //copy month
        CurrPos+=strlen(CurrPos);

        *(CurrPos++)='-';

        cn36_StrNCpy(CurrPos, szDBStamp1Date+6, 2);     //copy day
        CurrPos+=strlen(CurrPos);

        *(CurrPos++)=' ';

        cn36_StrNCpy(CurrPos, szDBStamp1Time+2, 2);     //copy hour, without first two 00
        CurrPos+=strlen(CurrPos);
        
        *(CurrPos++)=':';

        cn36_StrNCpy(CurrPos, szDBStamp1Time+4, 2);     //copy minute
        CurrPos+=strlen(CurrPos);

        *(CurrPos++)=':';

        cn36_StrNCpy(CurrPos, szDBStamp1Time+6, 2);     //copy second 

        SaveID[0]='\0';

        sprintf(MsgBuf_cn36, "Looking for %s with DB Stamp1 %s.\n", Label, DateTimeInDBMKnlFormat);
        cn36_Verbose(MsgBuf_cn36);

        int AllOk=1;

        const char * Line;

        while(DBMKnl.ReadLine(Line) && Line!=0)
        {
            if(Line[0]!='\0')
            {
                const char * LabelInLine=FindSectionInLine(1, Line);
                const char * TypeInLine=FindNextSectionInLine(LabelInLine);
                const char * DateTimeInLine=FindNextSectionInLine(TypeInLine);

                if(0!=LabelInLine && 0!=TypeInLine && 0!=DateTimeInLine)
                {
                    int LabelOk   = (!strncmp(LabelInLine,Label,strlen(Label)) && '|'== *(LabelInLine+strlen(Label))); // are labels equal? 
                    int TypeOk    = ((!strncmp(TypeInLine, "SAVE WARM", strlen("SAVE WARM")) && '|'== *(TypeInLine+strlen("SAVE WARM")))
                                     || (!strncmp(TypeInLine, "SAVE COLD", strlen("SAVE COLD")) && '|'== *(TypeInLine+strlen("SAVE COLD")))
                                    ); //save type must be "SAVE COLD" or "SAVE WARM"
                    int DateTimeOk= (!strncmp(DateTimeInLine,DateTimeInDBMKnlFormat,strlen(DateTimeInDBMKnlFormat)) && '|'== *(DateTimeInLine+strlen(DateTimeInDBMKnlFormat))); //are date,time equal?

                    if(LabelOk && DateTimeOk && TypeOk)
                    {
                        Found++;

                        size_t IDLength=strchr(strchr(Line, '|')+1, '|')-Line; // there exist at least 3 '|' in Line
                        if(IDLength>47)
                        {
                            AllOk=0;
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile.Set(AbsoluteNameOfDBMKnl));
                        }

                        cn36_StrNCpy(SaveID, Line, IDLength);
                    }
                }
                else
                {
                    AllOk=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile.Set(AbsoluteNameOfDBMKnl));
                }
            }
        }

        sprintf(MsgBuf_cn36, "Found %d SaveID%s %s.\n", Found, Found>1?"'s. Last was":". It was", SaveID);
        cn36_Verbose(MsgBuf_cn36);
    }
    
    cn36_Verbose(SubALevelFromPrefix);

    if(1==Found)
        return SaveID;
    else
        return 0;
}


static void cn35_PrintBackupAnswer(char *replyData, int * replyLen, int SQLCode, boolean IsSQLErr, tsp00_C256c & SQLErrText, const tcn003_BackupResult & tResult)
{
    if(0!=SQLCode && IsSQLErr)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::SQLError.SetSQL(SQLCode, (char *)SQLErrText));
    else
    {
        char * pCurrPos = 0;

        sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%d%s",BA_RESULT_RC_CN003, VALUE_SEPSTRING_CN00, SQLCode, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DATE_CN003, VALUE_SEPSTRING_CN00, tResult.szDate, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_TIME_CN003, VALUE_SEPSTRING_CN00, tResult.szTime, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_SERVER_CN003, VALUE_SEPSTRING_CN00, tResult.szServer, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DATABASE_CN003, VALUE_SEPSTRING_CN00, tResult.szDatabase, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_KNLVER_CN003, VALUE_SEPSTRING_CN00, tResult.szKernelVersion, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_PGSTRAN_CN003, VALUE_SEPSTRING_CN00, tResult.szPagesTransfered, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_PGSLEFT_CN003, VALUE_SEPSTRING_CN00, tResult.szPagesLeft, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_VOLUMES_CN003, VALUE_SEPSTRING_CN00, tResult.szVolumes, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_MEDNAME_CN003, VALUE_SEPSTRING_CN00, tResult.szMediaName, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_LOCATION_CN003, VALUE_SEPSTRING_CN00, tResult.szLocation, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_ERRTEXT_CN003, VALUE_SEPSTRING_CN00, tResult.szErrortext, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_LABEL_CN003, VALUE_SEPSTRING_CN00, tResult.szLabel, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_ISCONS_CN003, VALUE_SEPSTRING_CN00, tResult.szIsConsistent, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_FLOGPG_CN003, VALUE_SEPSTRING_CN00, tResult.szFirstLogPageNo, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_LLOGPG_CN003, VALUE_SEPSTRING_CN00, tResult.szLastLogPageNo, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DBSTMD1_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp1Date, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DBSTMT1_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp1Time, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DBSTMD2_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp2Date, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DBSTMT2_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp2Time, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_PGCNT_CN003, VALUE_SEPSTRING_CN00, tResult.szBDPageCount, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DEVUSE_CN003, VALUE_SEPSTRING_CN00, tResult.szDevicesUsed, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_DBID_CN003, VALUE_SEPSTRING_CN00, tResult.szDatabaseID, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        sprintf(pCurrPos, "%-20s%s%s%s",BA_RESULT_MUSEDATAPG_CN003, VALUE_SEPSTRING_CN00, tResult.szMaxUsedDataPage, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen(replyData);

        //szRedoTransRead and szRedoTransDone should not be displayed for backup/restore commands

        cn36_Verbose("Constructed the following reply:\n", AddALevelToPrefix);

        sprintf(MsgBuf_cn36, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%d%s",BA_RESULT_RC_CN003, VALUE_SEPSTRING_CN00, SQLCode, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DATE_CN003, VALUE_SEPSTRING_CN00, tResult.szDate, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_TIME_CN003, VALUE_SEPSTRING_CN00, tResult.szTime, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_SERVER_CN003, VALUE_SEPSTRING_CN00, tResult.szServer, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DATABASE_CN003, VALUE_SEPSTRING_CN00, tResult.szDatabase, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_KNLVER_CN003, VALUE_SEPSTRING_CN00, tResult.szKernelVersion, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_PGSTRAN_CN003, VALUE_SEPSTRING_CN00, tResult.szPagesTransfered, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_PGSLEFT_CN003, VALUE_SEPSTRING_CN00, tResult.szPagesLeft, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_VOLUMES_CN003, VALUE_SEPSTRING_CN00, tResult.szVolumes, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_MEDNAME_CN003, VALUE_SEPSTRING_CN00, tResult.szMediaName, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_LOCATION_CN003, VALUE_SEPSTRING_CN00, tResult.szLocation, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_ERRTEXT_CN003, VALUE_SEPSTRING_CN00, tResult.szErrortext, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_LABEL_CN003, VALUE_SEPSTRING_CN00, tResult.szLabel, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_ISCONS_CN003, VALUE_SEPSTRING_CN00, tResult.szIsConsistent, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_FLOGPG_CN003, VALUE_SEPSTRING_CN00, tResult.szFirstLogPageNo, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_LLOGPG_CN003, VALUE_SEPSTRING_CN00, tResult.szLastLogPageNo, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DBSTMD1_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp1Date, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DBSTMT1_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp1Time, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DBSTMD2_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp2Date, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DBSTMT2_CN003, VALUE_SEPSTRING_CN00, tResult.szDBStamp2Time, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_PGCNT_CN003, VALUE_SEPSTRING_CN00, tResult.szBDPageCount, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DEVUSE_CN003, VALUE_SEPSTRING_CN00, tResult.szDevicesUsed, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_DBID_CN003, VALUE_SEPSTRING_CN00, tResult.szDatabaseID, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        sprintf(MsgBuf_cn36, "%-20s%s%s%s",BA_RESULT_MUSEDATAPG_CN003, VALUE_SEPSTRING_CN00, tResult.szMaxUsedDataPage, LINE_SEPSTRING_CN00);
        cn36_Verbose(MsgBuf_cn36);

        cn36_Verbose(SubALevelFromPrefix);
    }
}

void tcn35_BackupToolConnector::FillReplyBuffer()
{
    cn36_Verbose("Filling reply buffer.\n", AddALevelToPrefix);

    switch(StatusOfRequest)
    {
        case NoRequestSend:
        case RequestSend:
        case CancelSendByDBMSrv:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set()); // must have reported already an error!  (wheter tool was started or not)
        break;

        case RequestReplied:
        {
            switch(StatusOfTool)
            {
                case NotStarted:
                {
                    if(!IsBad()) // tool was not started, because db had finished already (no error reported, tool not started -> nobody tried to start the tool 
                        cn35_PrintBackupAnswer(Info.ReplyData, Info.ReplyLen, SQLCodeOfRequest, SQLErrForRequest, SQLErrText, tTheDBResult); //just tell what the db-kernel sayed
                    else  //the error was already reported (written in the reply buffer), FillReplyBuffer() should not have been called in this case!
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                }
                break;
                case Ended:
                {
                    if(ToolWasSuccessful() || ToolMustNotBeSuccessful())
                        cn35_PrintBackupAnswer(Info.ReplyData, Info.ReplyLen, SQLCodeOfRequest, SQLErrForRequest, SQLErrText,tTheDBResult);
                    else
                    {
                        if(DBActionWasSuccessful())
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolFailed.SetI(SumOfToolExitCodes()));
                        else
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolAndDBFailed.SetSQLI(SQLCodeOfRequest, SumOfToolExitCodes()));
                    }
                }
                break;
                case KilledByDBMSrv:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolKilled.SetSQLI(SQLCodeOfRequest, SumOfToolExitCodes()));
                }
                break;
                default: TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            }
        }
        break;

        case CanceledByDBMSrv:
        {
            switch(StatusOfTool)
            {
                case NotStarted:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    // if no error was reported so far(!IsBad()), tool must have been started 
                    // if error was reported already(IsBad()), this function must not have been called
                }
                break;
                case Ended:
                {
                    if(ToolWasSuccessful())  // do not override a tool error
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::DBKernelRequestCanceled.SetSQLI(SQLCodeOfRequest, SumOfToolExitCodes()));
                    else
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolFailedAndDBKernelRequestCanceled.SetSQLI(SQLCodeOfRequest, SumOfToolExitCodes()));
                }
                break;
                case KilledByDBMSrv:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolKilledAndDBKernelRequestCanceled.SetSQLI(SQLCodeOfRequest, SumOfToolExitCodes()));
                }
                break;
                default: TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            }
        }
        break;
        
        case TerminatedByDBMSrv:
        {
            switch(StatusOfTool)
            {
                case NotStarted:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    // if no error was reported so far(!IsBad()), tool must have been started 
                    // if error was reported already(IsBad()), this function must not have been called
                }
                break;
                case Ended:
                {
                    if(ToolWasSuccessful())  // do not override a tool error
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CancelDBKernelRequestFailed.SetI(SumOfToolExitCodes()));
                    else
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolFailedAndCancelDBKernelRequestFailed.SetI(SumOfToolExitCodes()));
                }
                break;
                case KilledByDBMSrv:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolKilledAndCancelDBKernelRequestFailed.Set());
                }
                break;
                default: TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            }
        }
        break;

        default: TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    }

    cn36_Verbose("Reply buffer filled.\n\n", SubALevelFromPrefix);
}

void tcn35_BackupToolConnector::WriteLogFile()
{
    if(OK_CN00==GiveLastError() && (RequestReplied==StatusOfRequest))
    {
        int rc=0;

        cn36_Verbose("Updating external backup file.\n",AddALevelToPrefix);

        if(0!=GivePathOfBackupHistory())
        {
            sprintf(MsgBuf_cn36, "Using '%s' as backup history.\n", PathOfBackupHistory);
            cn36_Verbose(MsgBuf_cn36);

            const char * DBSaveID=GetSaveIDFromDBMKnl(PathOfBackupHistory,
                                                      tTheDBResult.szLabel,
                                                      tTheDBResult.szDBStamp1Date,
                                                      tTheDBResult.szDBStamp1Time);

            if(DBSaveID==NULL)
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotGetSaveIDs.Set());    // rc allready 0
            else
            {
                sprintf(MsgBuf_cn36, "Got '%s' as save identifier.\n", DBSaveID);
                cn36_Verbose(MsgBuf_cn36);
            
                if(0!=GivePathOfExternalBackupHistory())
                {
                    sprintf(MsgBuf_cn36, "Using '%s' as external backup file.\n", PathOfExternalBackupHistory);
                    cn36_Verbose(MsgBuf_cn36);

                    if(!CollectToolSaveIDs() || NumberOfExternalSaveIDsFound!=Info.NumFiles)
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotGetSaveIDs.Set());    // rc allready 0
                    else
                    {
                        tni34_ATextFile TheLogFile(PathOfExternalBackupHistory, tni34_AFile::ToDiskAppend, 1, 1);

                        if(TheLogFile.IsGood())
                        {
                            char NewLogFileEntry[1024];
                            int i;

                            rc=1;

                            for(i=0; i<NumberOfExternalSaveIDsFound && rc; i++)
                            {
                                if(0==ExternalSaveID[i])
                                {
                                    rc=0; //end the loop
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                                }
                                else
                                {
                                    sprintf(NewLogFileEntry, "%s|%s|", DBSaveID, ExternalSaveID[i]);

                                    long BytesWritten;

                                    if(TheLogFile.WriteLine(NewLogFileEntry, BytesWritten))
                                    {
                                        sprintf(MsgBuf_cn36, "Wrote '%s' to the external backup file.\n", NewLogFileEntry);
                                        cn36_Verbose(MsgBuf_cn36);
                                    }
                                    else
                                    {
                                        rc=0; //end the loop
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotWriteToToolLogFile.Set(PathOfExternalBackupHistory));
                                    }
                                }
                            }

                            if(!TheLogFile.Close())
                                rc=0;
                        }   // else: rc already 0
                    }
                }   // else: rc already 0
            }
        }   // else: rc already 0

        if(rc)
            cn36_Verbose("Have updated external backup file successfully.\n\n", SubALevelFromPrefix);
        else
            cn36_Verbose("Could not update external backup file.\n\n", SubALevelFromPrefix);
    }
}


// -----------------------------------------------------------------------------
// member functions of class tcn35_BackupToolConnectorEnum
// -----------------------------------------------------------------------------

tcn35_BackupToolConnectorEnum::tcn35_BackupToolConnectorEnum(const AValue TheValue)
{
    Value=TheValue;
}

tcn35_BackupToolConnectorEnum::tcn35_BackupToolConnectorEnum(const tcn35_BackupToolConnectorEnum & Original)
{
    Value=Original.Value;
}

tcn35_BackupToolConnectorEnum::tcn35_BackupToolConnectorEnum(const tcn003_BackupHistoryMediumLine * pMediumLine)
{
    Value=DetermineToolFromMediumName(pMediumLine->szMEDIANAME);
}

tcn35_BackupToolConnectorEnum::tcn35_BackupToolConnectorEnum(const char * MediumName)
{
    Value=DetermineToolFromMediumName(MediumName);
}

tcn35_BackupToolConnectorEnum::AValue tcn35_BackupToolConnectorEnum::DetermineToolFromMediumName(const char *MediumName) const
{
    AValue rc=None;

    if(0!=MediumName)
    {
        if(0==cn36_StrNUprCmp(MediumName, "ADSM", strlen("ADSM"))) // medium starts with ADSM
            rc=ADSM;
        else
            if(0==cn36_StrNUprCmp(MediumName, "NSR", strlen("NSR"))) // medium starts with NSR
                rc=NSR;
            else
                if(0==cn36_StrNUprCmp(MediumName, "BACK", strlen("BACK")))    // medium starts with BACK
                    rc=BACK;
    }

    return rc;
}

tcn35_BackupToolConnectorEnum::operator tcn35_BackupToolConnectorEnum::AValue() const
{
    return Value;
}

int tcn35_BackupToolConnectorEnum::operator == (const tcn35_BackupToolConnectorEnum &Other) const
{
    return Value==Other.Value;
}

int tcn35_BackupToolConnectorEnum::NumberOfBackupToolConnectors()
{
    return int(None);
}


// -----------------------------------------------------------------------------
// functions from hcn35.h
// -----------------------------------------------------------------------------

int cn35_DetermineBackupTool(tcn35_BackupToolConnector::tcn35_ActionType ActionType,
                             const tcn35_SaveInfo &                      rInfo,
                             tcn35_BackupToolConnector * &               rpDeterminedTool)
{
    int rc=0;

    if(0==rInfo.pszMediumName)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::MediumHasNoName.Set());
    else
        rc=cn35_ConstructBackupToolConnector(ActionType,
                                             tcn35_BackupToolConnectorEnum(rInfo.pszMediumName),
                                             rInfo,
                                             rpDeterminedTool);

    return rc;
}

int cn35_ConstructBackupToolConnector(tcn35_BackupToolConnector::tcn35_ActionType    ActionType,
                                      const tcn35_BackupToolConnectorEnum          & BackupToolAsEnum,     
                                      const tcn35_SaveInfo                         & rInfo,
                                      tcn35_BackupToolConnector                   *& rpDeterminedTool)
{
    int rc=0;

    if(0!=rpDeterminedTool)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    else
    {
        rc=1;

        switch(BackupToolAsEnum)
        {
            case tcn35_BackupToolConnectorEnum::ADSM:
                rpDeterminedTool=new tcn351_ADSMConnector(ActionType, rInfo);
                break;

            case tcn35_BackupToolConnectorEnum::NSR:
                rpDeterminedTool=new tcn352_NetWorkerConnector(ActionType, rInfo);
                break;

            case tcn35_BackupToolConnectorEnum::BACK:
                rpDeterminedTool=new tcn353_Backint2Connector(ActionType, rInfo);
                break;

            case tcn35_BackupToolConnectorEnum::None:
                break;

            default:
            {
                rc=0;
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            }
            break;
        }

        if(rc && tcn35_BackupToolConnectorEnum::None!=BackupToolAsEnum)
        {
            if(0==rpDeterminedTool)
            {
                rc=0;
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());

                cn36_Verbose("No memory to allocate tool object.\n");
            }
            else
            {
                if(rpDeterminedTool->IsBad())
                    rc=0;
            }
        }

        if(rc && 0!=rpDeterminedTool)
        {
            sprintf(MsgBuf_cn36, "Using connection to %s.\n\n", rpDeterminedTool->GiveLongToolName());
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    return rc;
}

int cn35_IsExternalBackupTool(const tcn003_BackupHistoryMediumLine * pMediumLine)
{    
    tcn35_BackupToolConnectorEnum BackupToolType(pMediumLine->szMEDIANAME);

    return (tcn35_BackupToolConnectorEnum::None!=BackupToolType);
}

int cn35_IsExternalBackupTool(const tcn003_Medium * pMedium, const tcn003_MediaGroup * pMediaGroup)
{
    int rc=0;

    if(0!=pMedium)  //only one medium
        rc=(tcn35_BackupToolConnectorEnum::None!=tcn35_BackupToolConnectorEnum(pMedium->szName));
    else            //we have a group of media with at least one element
        rc=(tcn35_BackupToolConnectorEnum::None!=tcn35_BackupToolConnectorEnum(pMediaGroup->szName));

    return rc;
}
