/*******************************************************************************
  module:       vcn351.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing connection between dbmsrv and external archiving tool ADSM 
  last change:  2000-11-29 13:40
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include"hcn351.h"
#include"hcn36.h"
#include"hni34.h"

#include<time.h>
#include<stdio.h>

#if defined (WIN32)
static const char szOSLineDelimiter[3]={13, 10, '\0'};
#elif defined (UNIX)
const char   szOSLineDelimiter[2]={'\n','\0'};
#else
    #error vcn351.cpp was only coded for WIN32 and UNIX (define one of them).
#endif


// -----------------------------------------------------------------------------
// implementation of class tcn351_ADSMConnector
// -----------------------------------------------------------------------------

tcn351_ADSMConnector::tcn351_ADSMConnector(tcn35_ActionType TheActionType, const tcn35_SaveInfo & rInfo)
  :tcn35_BackupToolConnector(TheActionType, rInfo)
{
    Label=0;
    ADSMInput=0;
    const char * DefaultADSMDirectory=
    #if defined(WIN32)
                                "c:\\win32app\\ibm\\adsm\\adint2";
    #elif defined(UNIX)
                                "/usr/adint";
    #else
        #error tcn351_ADSMConnector::tcn351_ADSMConnector() only coded for UNIX and NT (define one of them).
    #endif

    cn36_StrAllocCpy(ADSMDirectory=0, DefaultADSMDirectory);

    ADSMProgram=0;
}

tcn351_ADSMConnector::~tcn351_ADSMConnector()
{
    cn36_StrDealloc(Label);
    cn36_StrDealloc(ADSMDirectory);
    cn36_StrDealloc(ADSMProgram);

    if(0!=ADSMInput)
        delete ADSMInput;
}

const char * tcn351_ADSMConnector::GiveToolName()
{
    return "ADSM";
}

const char * tcn351_ADSMConnector::GiveLongToolName()
{
    return "IBM's ADSM with client adint2";
}

int tcn351_ADSMConnector::CheckToolInstallation(int & IsInstalled)
{
    int rc=1;
    
    IsInstalled=1;
    
    cn36_Verbose("Checking existence and configuration of adint2.\n", AddALevelToPrefix);

    // The default value of the ADSM installation directory is set in the constructor
    // That default can be overridden by the 'ADINT' environment variable.
    // For security, it is exported if not found in the environment.

    rc=cn36_DetermineValueOfEnvironmentVariable("ADINT", ADSMDirectory, "adint2 directory", Info.pszDBName);

    if(rc) // all is ok, so far?
        if(!cn36_CheckDir(ADSMDirectory))
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::ADirectoryIsMissing.Set(ADSMDirectory));
            IsInstalled=0;
        }
        else
        {
            // The ADSM client 'adint2' must be an executable program in that directory.
            const char * pszADSMProgramDefault=
            #if defined (WIN32)
                                         "\\adint2.exe";
            #elif defined (UNIX)
                                         "/adint2";
            #else
            #error tcn351_ADSMConnector::ToolIsInstalled() was only coded for UNIX and NT (define one of them).
            #endif

            if(!cn36_StrAlloc(ADSMProgram, strlen(ADSMDirectory)+strlen(pszADSMProgramDefault)))
                rc=0;
            else
            {
                strcpy(ADSMProgram,ADSMDirectory);
                strcat(ADSMProgram,pszADSMProgramDefault);

                if(!cn36_CheckExe(ADSMProgram))
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsNotExecutable.Set(ADSMProgram));
                    IsInstalled=0;
                }
            }

            if(rc && IsInstalled)  // is all ok, so far?
            {
                sprintf(MsgBuf_cn36,"Using ADSM program '%s'.\n", ADSMProgram);
                cn36_Verbose(MsgBuf_cn36);

                // The default name of the options file is just a convention,
                // it can be overridden by the environment variable 'ADA_OPT'.
                // As the DBName is not known to the ADSM client, this MUST be exported
                // if it was set in the program (== not already found in the environment).

                char * ADSMOptionsFileName=0;
                
                if(!cn36_StrAlloc(ADSMOptionsFileName, strlen(ADSMDirectory)+1+4+strlen(Info.pszDBName)+4))
                    rc=0;
                else
                {
                    #if defined (WIN32)
                    sprintf(ADSMOptionsFileName, "%s\\init%s.opt", ADSMDirectory, Info.pszDBName);
                    #elif defined (UNIX)
                    sprintf(ADSMOptionsFileName, "%s/init%s.opt", ADSMDirectory, Info.pszDBName);
                    #else
                        #error tcn351_ADSMConnector::CheckToolInstallation() was only coded for UNIX and NT (define one of them).
                    #endif

                    rc=cn36_DetermineValueOfEnvironmentVariable("ADA_OPT", ADSMOptionsFileName, "ADSM options file", Info.pszDBName);

                    if(rc)  // is all ok, so far?
                        if(!cn36_CheckRead(ADSMOptionsFileName))
                        {
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsMissingOrUnreadable.Set(ADSMOptionsFileName));
                            IsInstalled=0;
                        }
                }
            }
        }

    cn36_Verbose((rc && IsInstalled)?"Check passed successful.\n\n":"Check failed.\n\n", SubALevelFromPrefix);

    return rc;
}

int tcn351_ADSMConnector::Prepare()
{
    int rc=1;

    switch(GiveActionType())
    {
        case ABackup:
        {
            if(BuildLabel() && BuildBackupToolCall())
            {
                ADSMInput=new tni34_APipe(tni34_APipe::Out, 1, 0); //create a temporary pipe, but do not open it now

                if(0==ADSMInput)
                {
                    rc=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                }
                else
                    if(ADSMInput->IsGood())
                    {
                        sprintf(MsgBuf_cn36, "Created temporary pipe '%s' as input for adint2.\n", ADSMInput->GiveName());
                        cn36_Verbose(MsgBuf_cn36);
                    }
                    else
                    {
                        rc=0;
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotCreateADSMInputPipe.Set());
                    }
            }
            else
                rc=0;
        }break;

        case ARestore:
        {
            rc=cn36_StrAllocCpy(Label, Info.ExternalSaveID);

            if(rc)
            {
                if(BuildBackupToolCall())
                    ADSMInput=0;
                else
                    rc=0;
            }

        }break;

        case GiveBackupInfo:
        {
            if(BuildLabel() && BuildBackupToolCall())
                ADSMInput=0;
            else
                rc=0;
        }break;

        default:
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
        }break;
    }

    if(rc)
    {
        if(0!=ToolProcessOutput)
            delete ToolProcessOutput;

        if(0!=ToolProcessError)
            delete ToolProcessError;

        ToolProcessOutput=new tni34_AFile(tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now
        ToolProcessError=new tni34_AFile(tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now

        if(0==ToolProcessOutput || 0==ToolProcessError)
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
        }
        else
        {
            if(ToolProcessOutput->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as output for adint2.\n", ToolProcessOutput->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;

            if(ToolProcessError->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as error output for adint2.\n", ToolProcessError->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;
        }
    }

    return rc;
}

int tcn351_ADSMConnector::StartTool()
{
    int i,rc=1;

    cn36_Verbose("Starting adint2.\n", AddALevelToPrefix);

    if(0==NumberOfRunningToolProcesses)
    {
        for(i=0;i<NumberOfCommandLines && rc;i++)
        {
            if(0!=ADSMInput)
            {
                sprintf(MsgBuf_cn36,
                        "Calling archiving tool process '%s <%s >>%s 2>>%s' ... ",
                        CommandLine[i],
                        ADSMInput->GiveName(),
                        ToolProcessOutput->GiveName(),
                        ToolProcessError->GiveName());
            }
            else
            {
                sprintf(MsgBuf_cn36,
                        "Calling archiving tool process '%s >>%s 2>>%s' ... ",
                        CommandLine[i],
                        ToolProcessOutput->GiveName(),
                        ToolProcessError->GiveName());
            }

            cn36_Verbose(MsgBuf_cn36);

            ToolProcess[i]=new tni33_ABackgroundProcess(CommandLine[i],
                                                        ADSMInput,          //0==ADSMInput is OK.
                                                        ToolProcessOutput,
                                                        1,
                                                        ToolProcessError,
                                                        1);

            if(0==ToolProcess[i])
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                rc=0;
            }
            else
                if(!ToolProcess[i]->WasStarted())
                    rc=0;
                else
                {
                    NumberOfRunningToolProcesses++;
                    ToolProcessWasRunning[i]=1;
                }

            if(rc)
                cn36_Verbose("OK.\n", NoPrefix);
            else
                cn36_Verbose("Failed.\n", NoPrefix);
        }

        if(rc && 1<=NumberOfRunningToolProcesses && 0!=ADSMInput)
            if(!ADSMInput->Open())
                rc=0;
    }
    else
        rc=0;

    cn36_Verbose(rc?"adint2 has been started successfully.\n\n":"Could not start adint2 correctly.\n\n",SubALevelFromPrefix);

    return rc;
}

int tcn351_ADSMConnector::SignalEndToTool()
{
    int rc=1;

    if(ToolProcess[0]->IsRunning() && ABackup==GiveActionType())
    {
        const char * ADSMMsg;
        long BytesWritten;

        if(RequestReplied==StatusOfRequest && 0==SQLCodeOfRequest)
            ADSMMsg="COMMIT \n";
        else
            ADSMMsg="ROLLBACK \n";

        int wrc=ADSMInput->Write(ADSMMsg, long(strlen(ADSMMsg)), BytesWritten);

        if(-1==wrc)
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotSignalEndToTool.Set());

        if(1!=rc)
            rc=0;
    }

    return rc;
}

int tcn351_ADSMConnector::CleanUpTool()
{
    //TODO?: use cn36_AppendFileToFile()
    int rc=1;

    if(0!=ADSMInput)
    {
        sprintf(MsgBuf_cn36,"Removing temporary adint2 input pipe %s ...", ADSMInput->GiveName());
        cn36_Verbose(MsgBuf_cn36);

        if(!ADSMInput->Remove())
        {
            cn36_Verbose(" Failed.\n", NoPrefix);
            rc=0;
        }
        else
            cn36_Verbose(" Done.\n", NoPrefix);

        delete ADSMInput;
        ADSMInput=0;
    }

    return rc;
}

int tcn351_ADSMConnector::CollectToolSaveIDs()
{
    int rc=1;

    if(cn36_StrAllocCpy(ExternalSaveID[0], Label))
        NumberOfExternalSaveIDsFound=1;
    else
    {
        NumberOfExternalSaveIDsFound=0;
        rc=0;
    }

    return rc;
}

int tcn351_ADSMConnector::GiveInformationOfLine(const char *   Line,
                                                const char * & EBIDInLine,
                                                size_t       & EBIDLength,
                                                const char * & TypeString,
                                                const char * & DateTimeString,
                                                int          & LineContainsEBID)
{
    int rc=1;

    LineContainsEBID=0;

    if(0!=Line)                                 // we have a Line
    {
        Line=cn36_FirstNonWhiteSpaceOf(Line);

        if(!strncmp(Line, Label, strlen(Label)-1)) // Label has form "HostName_DBName_*", -1 to ignore "*" of Label
        {
            const char * EndOfEBID=cn36_FirstWhiteSpaceOf(Line);

            EBIDInLine=Line;
            EBIDLength=EndOfEBID-Line;

            const char * StartOfDate=cn36_LoopedStrChr(Line, '_', 2);   // date starts behind second '_'
            const char * StartOfType=0;
            const char * EndOfType=0;

            if(0!=StartOfDate)
            {
                StartOfDate++;                                          // ignore the '_'
                StartOfType=cn36_LoopedStrChr(StartOfDate, '_', 2);     // the action(type) string starts behind the time
            }

            if(0!=StartOfType)
            {
                StartOfType++;                                          // ignore the '_'
                EndOfType=cn36_StrChr(StartOfType, '_');
            }

            if(0!=EndOfType)                                            // we have found all '_' we were looking for
            {   
                size_t DateTimeLength=StartOfType-1-StartOfDate;        // -1 for the '_' in front of StartOfType

                if(!cn36_StrAlloc(TempDateTimeString, 19))
                    rc=0;
                else
                {
                    int TempDateTimeConstructed=0;

                    if(17==DateTimeLength && cn36_StrNConsistsOfDigits(StartOfDate, 2))     // year is coded with only two digits, must be a migration backup from 6.2 or 7.1
                    {
                        int Year;

                        if(cn36_StrNToInt(Year, StartOfDate, 2))
                        {
                            if(Year<30)
                                strcpy(TempDateTimeString, "20");               // init missing century with 00
                            else
                                strcpy(TempDateTimeString, "19");

                            cn36_StrNCpy(TempDateTimeString+2, StartOfDate, DateTimeLength);    // +2 for missing century
                            TempDateTimeConstructed=1;
                        }
                    }
                    else
                        if(19==DateTimeLength)
                        {
                            cn36_StrNCpy(TempDateTimeString, StartOfDate, DateTimeLength);
                            TempDateTimeConstructed=1;
                        }

                    if(TempDateTimeConstructed)    // we have successfully constructed/copied the date to TempDateTimeString
                    {
                        char * FirstDot=TempDateTimeString+4;
                        char * SecondDot=TempDateTimeString+7;
                        char * StartOfTime=TempDateTimeString+10;
                        char * ThirdDot=TempDateTimeString+13;
                        char * FourthDot=TempDateTimeString+16;
                
                        *FirstDot=*SecondDot=*StartOfTime=*ThirdDot=*FourthDot='0'; //set temporary all possible non digits to '0'

                        if(cn36_StrConsistsOfDigits(TempDateTimeString))    // only digits should remain in the string
                        {
                            *FirstDot=*SecondDot='-';
                            *StartOfTime=' ';
                            *ThirdDot=*FourthDot=':';

                            DateTimeString=TempDateTimeString;

                            TypeString=ActionStringForActionStringN(StartOfType, EndOfType-StartOfType);

                            if(0!=TypeString)
                                LineContainsEBID=1;
                        }
                    }
                }
            }
        }
    }

    return rc;
}

int tcn351_ADSMConnector::BuildLabel()
{
    int rc=0;

    //TODO: Use switch
    if(ABackup==GiveActionType())
    {
        time_t ltime ;
        char   szDatTim[30];
        char   szPieces[20];

        // Get current date and time, convert according to timezone, make string.
        // Time separator is '.' instead of ':' because of NT name analysis!
        time(&ltime);
        strftime(szDatTim, sizeof(szDatTim), "%Y.%m.%d_%H.%M.%S", localtime(&ltime));

        if(Info.NumFiles>1)
            sprintf(szPieces, "_%dpcs", Info.NumFiles);
        else
            szPieces[0]='\0';

        if(cn36_StrAlloc(Label, strlen(Info.pszMachName)+1
                                +strlen(Info.pszDBName)+1
                                +strlen(szDatTim)+1
                                +strlen(Info.pszActionString)+1
                                +strlen(Info.pszMediumName)
                                +strlen(szPieces)))
        {
            sprintf(Label, "%s_%s_%s_%s_%s%s", Info.pszMachName , Info.pszDBName,
                    szDatTim, Info.pszActionString, Info.pszMediumName, szPieces);

            size_t iChars=strlen(Label);
            // date+time: 19 chars, 4 more components, 4 separators => min. 27 chars
            if(iChars<25)
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotBuildSaveLabel.SetI(int(iChars)));
            else
            {
                sprintf(MsgBuf_cn36,"Constructed save label '%s'.\n", Label);
                cn36_Verbose(MsgBuf_cn36);
                rc=1;
            }
        }
    }
    else
        if(GiveBackupInfo==GiveActionType())
        {
            size_t LengthOfLabel=strlen(Info.SourceDBHost)+1+strlen(Info.SourceDBName)+2;

            if(cn36_StrAlloc(Label, LengthOfLabel))
            {
                sprintf(Label, "%s_%s_*", Info.SourceDBHost, Info.SourceDBName);

                if(strlen(Label)!=LengthOfLabel)
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                else
                {
                    sprintf(MsgBuf_cn36,"Constructed save label '%s'.\n", Label);
                    cn36_Verbose(MsgBuf_cn36);
                    rc=1;
                }
            }
        }
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());

    return rc;
}

int tcn351_ADSMConnector::BuildBackupToolCall()
{
    char * Call=0;
    int rc=0;
    const char * Adint2ActionTypeFlag;

    if(NumberOfCommandLines!=0) //if old commandlines exists, free the memory
    {
        int i;

        for(i=0;i<NumberOfCommandLines;i++)
            if(0!=CommandLine[i])
            {
                delete [] CommandLine[i];
                CommandLine[i]=0;
            }
        
        NumberOfCommandLines=0;
    }

    switch(GiveActionType()) //determine adint2-action-flag
    {
        case ABackup:
            Adint2ActionTypeFlag="-b";
            break;
        case ARestore:
            Adint2ActionTypeFlag="-r";
            break;
        case GiveBackupInfo:
            Adint2ActionTypeFlag="-q";
            break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    }

    char szBlockSize[20];
    sprintf(szBlockSize,"%d",Info.BlockSize);

    int i;
    size_t Length=strlen(ADSMProgram)                       // "...\adint2"
                  +1+strlen(Adint2ActionTypeFlag)           // " -b", "-r" or "-q"
                  +4+strlen(Label);                         // " -n <label>"

    if(ABackup==GiveActionType() || ARestore==GiveActionType())
    {
        Length+=4+strlen(szBlockSize);                      // " -c <BlockSize>"
        Length+=Info.NumFiles*strlen(" -p ");               // " -p <pipe>" for every pipe 

        for(i=0;i<Info.NumFiles;i++)
            Length+=strlen(Info.ppszFiles[i]);
    }

    if(cn36_StrAlloc(Call, Length))                         // the new commandline has length Length
    {
        sprintf(Call, "%s %s -n %s", ADSMProgram, Adint2ActionTypeFlag, Label);

        if(ABackup==GiveActionType() || ARestore==GiveActionType())
        {
            strcat(Call, " -c ");
            strcat(Call, szBlockSize);

            for(i=0; i<Info.NumFiles; i++)
            {
                strcat(Call," -p ");
                strcat(Call,Info.ppszFiles[i]);
            }
        }

        if(Length!=strlen(Call))
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
        else
        {
            NumberOfCommandLines=1; // we start only one tool process, while using ADSM with adint2
            CommandLine[0]=Call;
            rc=1;

            sprintf(MsgBuf_cn36, "Constructed adint2 call '%s'.\n", Call);
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    return rc;
}
