/*******************************************************************************
  module:       vcn352.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing connection between dbmsrv and external backup tool
                NetWorker 
  last change:  2000-11-29 13:57
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include"hcn352.h"
#include"hni34.h"
#include<time.h>

#include"hcn36.h"


const char * tcn352_NetWorkerConnector::NSaveTimeEBIDPrefix="NST";
const char * tcn352_NetWorkerConnector::NSaveTimeEBIDPartDelimiter=" ";

// -----------------------------------------------------------------------------
// implementation of class tcn352_NetWorkerConnector
// -----------------------------------------------------------------------------

tcn352_NetWorkerConnector::tcn352_NetWorkerConnector(tcn35_ActionType TheActionType, const tcn35_SaveInfo & rInfo)
  :tcn35_BackupToolConnector(TheActionType, rInfo)
{
    const char * DefaultPathOfNSR_ENVFile=
    #if defined(WIN32)
                            "C:\\Program Files\\nsr\\sapdb\\env";
    #elif defined(UNIX)
                            "/nsr/sapdb/env";
    #else
        #error tcn352_NetWorkerConnector::tcn352_NetWorkerConnector() only coded for UNIX and NT (define one of them).
    #endif

    cn36_StrAllocCpy(PathOfNSR_ENVFile=0, DefaultPathOfNSR_ENVFile);

    PathOfNSR_HOMEDirectory=0;
    NameOfNSR_HOST=0;
    NameOfNSR_POOL=0;
    NSR_EXPIREValue=0;

    PathOfNSRSave=0;
    PathOfNSRRecover=0;
    PathOfNSRMminfo=0;

    StartSaveTime=long(time(0));
    EndSaveTime=0;
    
    DirectiveFileName=0;

    LevelFullForBackupType=DataMigration;
    VerboseFlagFor=SaveAndRecover;
    NSREBIDType=NSaveTime;
    TempEBID=0;

    UseNSROutput=1;

    if(1<Info.NumFiles)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseToManyPipes));
}

tcn352_NetWorkerConnector::~tcn352_NetWorkerConnector()
{
    cn36_StrDealloc(PathOfNSR_ENVFile);
    cn36_StrDealloc(PathOfNSR_HOMEDirectory);
    cn36_StrDealloc(NameOfNSR_HOST);
    cn36_StrDealloc(NameOfNSR_POOL);
    cn36_StrDealloc(NSR_EXPIREValue);
    cn36_StrDealloc(PathOfNSRSave);
    cn36_StrDealloc(PathOfNSRRecover);
    cn36_StrDealloc(PathOfNSRMminfo);
    cn36_StrDealloc(DirectiveFileName);

    cn36_StrDealloc(TempEBID);

    DeleteListOfErrorsToIgnore();
}

const char * tcn352_NetWorkerConnector::GiveToolName()
{
    return "NetWorker";
}

const char * tcn352_NetWorkerConnector::GiveLongToolName()
{
    return "Legato's NetWorker with save, recover and mminfo";
}

static int cn352AllocConstructAndCheckExe(char *& Path, const char * DirectoryPath, const char * Name, int & ExistsAndIsExecutable)
{
    int rc=1;

    ExistsAndIsExecutable=1;

    if(cn36_StrAlloc(Path, strlen(DirectoryPath)+strlen(Name)))
    {
        strcpy(Path, DirectoryPath);
        strcat(Path, Name);
        
        if(!cn36_CheckExe(Path))
        {
            ExistsAndIsExecutable=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsNotExecutable.Set(Path));
        }
    }
    else
        rc=0;

    return rc;
}

class tcn352_NSREnvFileEntry
{
  public:
    const char * KeyWord;
    char **      Destination; 
};

int tcn352_NetWorkerConnector::CheckToolInstallation(int & IsInstalled)
{
    int rc=1;

    IsInstalled=1;

    // The default value of the path of the NetWorker variables file is set in the constructor
    // That default can be overridden by the 'NSR_ENV' environment variable.
    // For security, it is exported if not found in the environment.    
    cn36_Verbose("Checking existence and configuration of NetWorker.\n", AddALevelToPrefix);

    DeleteListOfErrorsToIgnore();

    rc=cn36_DetermineValueOfEnvironmentVariable("NSR_ENV", PathOfNSR_ENVFile, "path of NetWorker's configuration file", Info.pszDBName);

    if(rc) // all is ok, so far?
        if(!cn36_CheckRead(PathOfNSR_ENVFile))
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsMissingOrUnreadable.Set(PathOfNSR_ENVFile));
            IsInstalled=0;
        }
        else
        {
            char * NSRFullLevelAsString=0;
            char * NSRNotVerboseAsString=0;
            char * NSREBIDTypeAsString=0;
            char * NSROutputScanAsString=0;
            char * NSRIgnoreError=0;

            tni34_ATextFile NSR_ENVFile(PathOfNSR_ENVFile, tni34_AFile::FromDisk, 0);

            if(NSR_ENVFile.IsGood())
            {
                const char *Line;

                tcn352_NSREnvFileEntry AcceptedNSREnvFileEntries[]={{"NSR_HOME",       &PathOfNSR_HOMEDirectory},
                                                                    {"NSR_HOST",       &NameOfNSR_HOST},
                                                                    {"NSR_POOL",       &NameOfNSR_POOL},
                                                                    {"NSR_EXPIRE",     &NSR_EXPIREValue},
                                                                    {"NSR_FULL",       &NSRFullLevelAsString},
                                                                    {"NSR_NOTVERBOSE", &NSRNotVerboseAsString},
                                                                    {"NSR_EBIDTYPE",   &NSREBIDTypeAsString},
                                                                    {"NSR_USEOUTPUT",  &NSROutputScanAsString},
                                                                    {"NSR_IGNOREERROR",&NSRIgnoreError},
                                                                    {0,                0}};

                while(rc && NSR_ENVFile.ReadLine(Line) && Line!=0)
                {
                    const char * NSRStart=cn36_FirstNonWhiteSpaceOf(Line); // ignore white spaces

                    if(!strncmp(NSRStart, "NSR_", strlen("NSR_"))) // all meaningful lines start with NSR_
                    {
                        char **Destination=0;
                        tcn352_NSREnvFileEntry * CurrentNSREnvFileEntry=AcceptedNSREnvFileEntries;

                        while(0==Destination && 0!=CurrentNSREnvFileEntry->KeyWord)
                        {
                            if(0==strncmp(NSRStart, CurrentNSREnvFileEntry->KeyWord, strlen(CurrentNSREnvFileEntry->KeyWord)))
                                Destination=CurrentNSREnvFileEntry->Destination;
                            else
                                CurrentNSREnvFileEntry++;
                        }

                        if(0==Destination)
                        {
                            cn36_Verbose("The following line of NetWorker's variables file has an unknown format and is ignored:\n", AddALevelToPrefix);
                            cn36_Verbose(Line);
                            cn36_Verbose("\n", NoPrefix);
                            cn36_Verbose(SubALevelFromPrefix);
                        }
                        else
                        {
                            const char *StartOfValue=cn36_FirstNonWhiteSpaceOf(cn36_FirstWhiteSpaceOf(NSRStart)); //ignore keyword and following whitespaces
                            const char *EndOfValue;
                            int         FoundValue=0;
                            
                            cn36_GetStringBetweenDoubleQuotesOrWhiteSpaces(StartOfValue, EndOfValue);

                            if('\0'==StartOfValue[0])   // empty strings are not useful for parameters
                            {
                                cn36_Verbose("The following line of NetWorker's variables file contains a format error and is ignored:\n", AddALevelToPrefix);
                                cn36_Verbose(Line);
                                cn36_Verbose("\n", NoPrefix);
                                cn36_Verbose(SubALevelFromPrefix);
                            }
                            else
                            {
                                rc=cn36_StrNAllocCpy(*Destination, StartOfValue, (EndOfValue-StartOfValue));
                            
                                if(rc)
                                {
                                    sprintf(MsgBuf_cn36,"Found NetWorker setting for '%s': '%s'.\n", CurrentNSREnvFileEntry->KeyWord, *Destination);
                                    cn36_Verbose(MsgBuf_cn36);
                                }

                                if(rc && Destination==&NSRIgnoreError)  // special handling for the list of errors to ignore
                                {
                                    if((0==ListOfErrorsToIgnore.Add(*Destination)))   // try to add the pointer to the string in the list
                                    {
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                                        rc=0;
                                    }
                                    else
                                        *Destination=0;  //forget about the string here to avoid freeing it (a pointer to the string is now in the list)
                                }   
                            }
                        }
                    }
                }
            }
            else
                rc=0;

            if(rc)  // IsInstalled==1 holds allways
            {
                if(0==PathOfNSR_HOMEDirectory || '\0'==PathOfNSR_HOMEDirectory[0])
                {
                    IsInstalled=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseCanNotDetermineNSR_HOME, PathOfNSR_ENVFile));
                }

                if(0==NameOfNSR_HOST || '\0'==NameOfNSR_HOST)
                {
                    IsInstalled=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseCanNotDetermineNSR_HOST, PathOfNSR_ENVFile));
                }
            }

            if(rc && IsInstalled)
                if(!cn36_CheckDir(PathOfNSR_HOMEDirectory))
                {
                    IsInstalled=0;

                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::ADirectoryIsMissing.Set(PathOfNSR_HOMEDirectory));

                    sprintf(MsgBuf_cn36, "The NetWorker directory '%s' is missing or no directory.\n", PathOfNSR_HOMEDirectory);
                    cn36_Verbose(MsgBuf_cn36);
                }

            if(rc && IsInstalled)
            {
                #if defined (WIN32)
                const char *NameOfSave="\\save.exe";
                const char *NameOfRecover="\\recover.exe";
                const char *NameOfMminfo="\\mminfo.exe";

                #elif defined (UNIX)
                const char *NameOfSave="/save";
                const char *NameOfRecover="/recover";
                const char *NameOfMminfo="/mminfo";

                #else
                    #error tcn352_NetWorkerConnector::ToolIsInstalled() was only coded for UNIX and NT (define one of them).
                #endif

                rc=cn352AllocConstructAndCheckExe(PathOfNSRSave, PathOfNSR_HOMEDirectory, NameOfSave, IsInstalled);

                if(rc && IsInstalled)
                    rc=cn352AllocConstructAndCheckExe(PathOfNSRRecover, PathOfNSR_HOMEDirectory, NameOfRecover, IsInstalled);

                if(rc && IsInstalled)
                    rc=cn352AllocConstructAndCheckExe(PathOfNSRMminfo, PathOfNSR_HOMEDirectory, NameOfMminfo, IsInstalled);

                if(rc && IsInstalled)
                {
                    cn36_Verbose("Using NetWorker programs:\n", AddALevelToPrefix);
                    sprintf(MsgBuf_cn36, "'%s'\n", PathOfNSRSave);
                    cn36_Verbose(MsgBuf_cn36);
                    sprintf(MsgBuf_cn36, "'%s'\n", PathOfNSRRecover);
                    cn36_Verbose(MsgBuf_cn36);
                    sprintf(MsgBuf_cn36, "'%s'\n", PathOfNSRMminfo);
                    cn36_Verbose(MsgBuf_cn36);
                    cn36_Verbose(SubALevelFromPrefix);
                }
            }

            if(rc && IsInstalled)
            {
                char * NameOfNSR_POOLFromEnvironment=0;

                rc=cn36_DetermineValueOfEnvironmentVariable("NSR_POOL", NameOfNSR_POOLFromEnvironment, "NetWorker pool", Info.pszDBName, 0);

                if(rc)
                {
                    if('\0'!=NameOfNSR_POOLFromEnvironment[0])                          // values from the environment are used instead of values from the configuration file
                        rc=cn36_StrAllocCpy(NameOfNSR_POOL, NameOfNSR_POOLFromEnvironment);
                    else
                    {
                        if(0==NameOfNSR_POOL)                       // if we have no value at all, it is the last chance to init NameOfNSR_POOL as required by the following code
                            rc=cn36_StrAllocCpy(NameOfNSR_POOL, "");
                    }
                }

                cn36_StrDealloc(NameOfNSR_POOLFromEnvironment);
            }

            if(rc && IsInstalled)
            {
                char * NSR_EXPIREValueFromEnvironment=0;

                rc=cn36_DetermineValueOfEnvironmentVariable("NSR_EXPIRE", NSR_EXPIREValueFromEnvironment, "NetWorker expiry period", Info.pszDBName, 0);

                if(rc)
                {
                    if('\0'!=NSR_EXPIREValueFromEnvironment[0])
                        rc=cn36_StrAllocCpy(NSR_EXPIREValue, NSR_EXPIREValueFromEnvironment);
                    else
                    {
                        if(0==NSR_EXPIREValue)
                            rc=cn36_StrAllocCpy(NSR_EXPIREValue, "");
                    }
                }

                cn36_StrDealloc(NSR_EXPIREValueFromEnvironment);
            }

            if(rc && IsInstalled)
            {
                if(0!=NSRFullLevelAsString && '\0'!=NSRFullLevelAsString[0])
                {
                    if(0==strcmp(NSRFullLevelAsString, "SAVEDATA") ||
                       0==strcmp(NSRFullLevelAsString, "DATA MIGRATION"))
                    {
                        LevelFullForBackupType=DataMigration;
                    }
                    else
                        if(0==strcmp(NSRFullLevelAsString, "SAVEDTNCHK") ||
                           0==strcmp(NSRFullLevelAsString, "DATA RECOVERY"))
                        {
                            LevelFullForBackupType=DataRecovery;
                        }
                        else
                            if(0==strcmp(NSRFullLevelAsString, "SAVEDATA+SAVEDTNCHK") ||
                               0==strcmp(NSRFullLevelAsString, "DATA MIGRATION+DATA RECOVERY"))
                            {
                                LevelFullForBackupType=DataMigrationAndDataRecovery;
                            }
                            else
                            {
                                sprintf(MsgBuf_cn36, "The unknown value '%s' for variable 'NSR_FULL' is ignored.\n", NSRFullLevelAsString);
                                cn36_Verbose(MsgBuf_cn36);
                            }
                }
                else
                    LevelFullForBackupType=DataMigration;

                if(0!=NSRNotVerboseAsString && '\0'!=NSRNotVerboseAsString[0])
                {
                    if(0==strcmp(NSRNotVerboseAsString, "NONE"))
                        VerboseFlagFor=SaveAndRecover;
                    else
                        if(0==strcmp(NSRNotVerboseAsString, "SAVE"))
                            VerboseFlagFor=Recover;
                        else
                            if(0==strcmp(NSRNotVerboseAsString, "RECOVER"))
                                VerboseFlagFor=Save;
                            else
                                if(0==strcmp(NSRNotVerboseAsString, "SAVE AND RECOVER"))
                                    VerboseFlagFor=None;
                                else
                                {
                                    sprintf(MsgBuf_cn36, "The unknown value '%s' for variable 'NSR_NOTVERBOSE' is ignored.\n", NSRNotVerboseAsString);
                                    cn36_Verbose(MsgBuf_cn36);
                                }
                }
                else
                    VerboseFlagFor=SaveAndRecover;

                if(0!=NSREBIDTypeAsString && '\0'!=NSREBIDTypeAsString[0])
                {
                    if(0==strcmp(NSREBIDTypeAsString, "SSID"))
                        NSREBIDType=SSID;
                    else
                        if(0==strcmp(NSREBIDTypeAsString, "NSAVETIME"))
                            NSREBIDType=NSaveTime;
                        else
                        {
                            sprintf(MsgBuf_cn36, "The unknown value '%s' for variable 'NSR_EBIDTYPE' is ignored.\n", NSREBIDTypeAsString);
                            cn36_Verbose(MsgBuf_cn36);
                        }
                }
                else
                    NSREBIDType=NSaveTime;

                if(0!=NSROutputScanAsString && '\0'!=NSROutputScanAsString[0])
                {
                    if(0==strcmp(NSROutputScanAsString, "SAVE"))
                        UseNSROutput=1;
                    else
                        if(0==strcmp(NSROutputScanAsString, "NONE"))
                            UseNSROutput=0;
                }
                else
                    UseNSROutput=1;
            }

            cn36_StrDealloc(NSRFullLevelAsString);
            cn36_StrDealloc(NSRNotVerboseAsString);
            cn36_StrDealloc(NSREBIDTypeAsString);
            cn36_StrDealloc(NSROutputScanAsString);
            cn36_StrDealloc(NSRIgnoreError);
        }

    cn36_Verbose((rc && IsInstalled)?"Check passed successful.\n\n":"Check failed.\n\n", SubALevelFromPrefix);

    return rc;
}

static class cn352LevelActionMapEntry
{
    public:
        const char * Level;
        const char * ActionString;
}
cn352LevelActionMap[]={
      {"full", "SAVEDATA"},
      {"incr", "SAVEPAGES"},
      {"1",    "SAVELOG"},
    //{"2",    "SAVELOGSEG"}, obsolete since 6.2.8, but do not reuse this level as long as possible
      {"3",    "SAVEDTNCHK"},
      {"4",    "SAVEPGNCHK"},
      {"5",    "SAVEHISTORY"},
      {0,      0}
};

const char * cn352LevelForAction(const char * ActionString)
{
    const cn352LevelActionMapEntry * Entry=cn352LevelActionMap;

    while(0!=Entry->Level && strcmp(Entry->ActionString, ActionString))
        Entry++;

    return Entry->Level;
}

const char * cn352ActionNForLevel(const char * Level, size_t LevelLength)
{
    const cn352LevelActionMapEntry * Entry=cn352LevelActionMap;
    
    while(0!=Entry->Level &&                                                                // not at end of list
          (LevelLength!=strlen(Entry->Level) || strncmp(Entry->Level, Level, LevelLength))  // strings differ in length or in letters
         )
    {
        Entry++;
    }

    return Entry->ActionString;
}

int tcn352_NetWorkerConnector::BuildBackupToolCall()
{
    int rc=1;

    if(NumberOfCommandLines!=0) //if old commandlines exists, free the memory
    {
        int i;
        for(i=0;i<NumberOfCommandLines;i++)
            if(0!=CommandLine[i])
            {
                delete [] CommandLine[i];
                CommandLine[i]=0;
            }
        
        NumberOfCommandLines=0;
    }

    if(ABackup==GiveActionType()) //write NetWorkers directives file for a backup
    {
        #if defined(WIN32)

        const char * DFileEntry=Info.ppszFiles[0];

        if(!cn36_StrAllocCpy(DirectiveFileName, "\\.nsr"))
            rc=0;

        #elif defined(UNIX)

        const char * EndOfDFileDirectory=cn36_BaseNameFrom(Info.ppszFiles[0]);
        const char * DFileEntry=EndOfDFileDirectory;

        if(!cn36_StrAlloc(DirectiveFileName, EndOfDFileDirectory-Info.ppszFiles[0]+strlen(".nsr")))
            rc=0;
        else
        {
            cn36_StrNCpy(DirectiveFileName, Info.ppszFiles[0], EndOfDFileDirectory-Info.ppszFiles[0]);
            strcat(DirectiveFileName, ".nsr");
        }

        #else
            #error tcn352_NetWorkerConnector::BuildBackupToolCall() only coded for WIN32 and UNIX, define one of them
        #endif

        if(rc)
        {
            tni34_ATextFile DFile(DirectiveFileName, tni34_AFile::ToDiskTruncate);

            if(DFile.IsGood())
            {
                long BytesWritten;
                char Line[1024];

                sprintf(Line, "rawasm: %s", DFileEntry);
                if(!DFile.WriteLine(Line, BytesWritten))
                    rc=0;

                if(!DFile.Close())
                    rc=0;
            }
            else
                rc=0;
        }

        if(rc)
        {
            sprintf(MsgBuf_cn36, "Constructed NetWorker's directives file '%s' successfully.\n", DirectiveFileName);
            cn36_Verbose(MsgBuf_cn36);
        }
        else
        {
            sprintf(MsgBuf_cn36, "Could not constructed NetWorker's directives file '%s'.\n", DirectiveFileName);
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    if(rc)  //count char of the new command line, allocate memory and construct new command line
    {
        size_t Length;
        int    UseVerboseFlag=0;

        NSREBIDTypeEnum UsedEBIDType;   //only used for restore
        const char * StartNSavetime;    //only used for restore with EBID=="NST..."
        const char * EndNSavetime;      //only used for restore with EBID=="NST..."
        const char * StartOfSaveClient; //only used for restore with EBID=="NST..."
        const char * EndOfSaveClient;   //only used for restore with EBID=="NST..."
        const char * StartOfEBID;       //only used for restore
        const char * EndOfEBID;         //only used for restore

        switch(GiveActionType())
        {
            case ABackup:
            {
                UseVerboseFlag=(Save==VerboseFlagFor || SaveAndRecover==VerboseFlagFor);    //must the "-v"-option be used

                Length=strlen(PathOfNSRSave);                                   // "...\save.exe" or ".../save"

                if(UseVerboseFlag)
                    Length+=3;                                                  // " -v"

                Length+=4+strlen(NameOfNSR_HOST);                               // " -s <NSR_HOST>"

                if('\0'!=NameOfNSR_POOL[0])
                    Length+=4+strlen(NameOfNSR_POOL);                           // " -b <NSR_POOL>"

                if('\0'!=NSR_EXPIREValue[0])
                    Length+=4+strlen(NSR_EXPIREValue);                          // " -e <NSR_EXPIRE>"

                Length+=4+strlen(Info.pszDBName);                               // " -N <dbname>"
                Length+=4+strlen(cn352LevelForAction(Info.pszActionString));    // " -l <Level>"
                Length+=4+strlen(DirectiveFileName);                            // " -f <dirfile>"
                Length+=1+strlen(Info.ppszFiles[0]);                            // " <pipename>"

                if(!cn36_StrAlloc(CommandLine[0], Length))
                    rc=0;
                else
                    sprintf(CommandLine[0],
                            "%s%s -s %s%s%s%s%s -N %s -l %s -f %s %s",
                            PathOfNSRSave,
                            UseVerboseFlag?" -v":"",
                            NameOfNSR_HOST,
                            '\0'==NameOfNSR_POOL[0]?"":" -b ",
                            '\0'==NameOfNSR_POOL[0]?"":NameOfNSR_POOL,
                            '\0'==NSR_EXPIREValue[0]?"":" -e ",
                            '\0'==NSR_EXPIREValue[0]?"":NSR_EXPIREValue,
                            Info.pszDBName,
                            cn352LevelForAction(Info.pszActionString),
                            DirectiveFileName,
                            Info.ppszFiles[0]);
            }break;

            case ARestore:
            {
                StartOfEBID=Info.ExternalSaveID;
                EndOfEBID=Info.ExternalSaveID+strlen(Info.ExternalSaveID);

                if(size_t(EndOfEBID-StartOfEBID)>strlen(NSaveTimeEBIDPrefix)+2+2*strlen(NSaveTimeEBIDPartDelimiter) &&                      //at least space for a NSaveTimeEBIDPrefix and two NSaveTimeEBIDPartDelimiter and 2 other chars (one for savetime and one for client)
                   0==cn36_StrNUprCmp(StartOfEBID, NSaveTimeEBIDPrefix, strlen(NSaveTimeEBIDPrefix)) &&                                     // the EBID starts with "NST"
                   0==strncmp(StartOfEBID+strlen(NSaveTimeEBIDPrefix), NSaveTimeEBIDPartDelimiter, strlen(NSaveTimeEBIDPartDelimiter)) &&   // which is followed with a NSaveTimeEBIDPartDelimiter
                   0!=strstr(StartOfEBID+strlen(NSaveTimeEBIDPrefix)+strlen(NSaveTimeEBIDPartDelimiter), NSaveTimeEBIDPartDelimiter))       // which is followed by something and another NSaveTimeEBIDPartDelimiter
                {
                    UsedEBIDType=NSaveTime;
                    StartNSavetime=StartOfEBID+strlen(NSaveTimeEBIDPrefix)+strlen(NSaveTimeEBIDPartDelimiter);
                    EndNSavetime=strstr(StartNSavetime, NSaveTimeEBIDPartDelimiter);
                    StartOfSaveClient=EndNSavetime+strlen(NSaveTimeEBIDPartDelimiter);
                    EndOfSaveClient=EndOfEBID;
                }
                else
                    if(0<(EndOfEBID-StartOfEBID) &&
                       cn36_StrNConsistsOfDigits(StartOfEBID, (EndOfEBID-StartOfEBID)))
                    {
                        UsedEBIDType=SSID;
                    }
                    else
                    {
                        tcn36_ADBMSrvError SyntaxError(tcn36_ADBMSrvError::CaseSyntaxErrorEBIDs);

                        rc=0;
                        TheDBMSrvErrorHandler_cn36.Report(SyntaxError.SetI(1)); //it's the 1st element, because we have only one element in the list
                    }

                if(rc)
                {
                    UseVerboseFlag=(Recover==VerboseFlagFor || SaveAndRecover==VerboseFlagFor);    //must the "-v"-option be used

                    Length=strlen(PathOfNSRRecover);                    // "...\recover.exe" or ".../recover"

                    if(UseVerboseFlag)
                        Length+=3;                                      // " -v"

                    Length+=4+strlen(NameOfNSR_HOST);                   // " -s <NSR_HOST>"

                    if(SSID==UsedEBIDType)                              // use of SSID
                        Length+=4+strlen(Info.ExternalSaveID);          // " -S <EBID>"
                    else                                                // use of nsavetime+client
                    {
                        Length+=7+(EndNSavetime-StartNSavetime);        // " -a -t <nsavetime>"
                        Length+=4+(EndOfSaveClient-StartOfSaveClient);  // " -c <save_client>
                    }

                    Length+=4+1+strlen(Info.ppszFiles[0]);              // " -iY <pipename>"
                       
                    if(!cn36_StrAlloc(CommandLine[0], Length))
                        rc=0;
                    else
                    {
                        if(SSID==UsedEBIDType)
                        {
                            sprintf(CommandLine[0],
                                    "%s%s -s %s -S %s -iY %s",             // -S -> use of SSID
                                    PathOfNSRRecover,
                                    UseVerboseFlag?" -v":"",
                                    NameOfNSR_HOST,
                                    Info.ExternalSaveID,
                                    Info.ppszFiles[0]);
                        }
                        else
                        {
                            size_t CurrPos;

                            sprintf(CommandLine[0],
                                    "%s%s -s %s -a -t ",             // -t -> use of nsavetime
                                    PathOfNSRRecover,
                                    UseVerboseFlag?" -v":"",
                                    NameOfNSR_HOST);
                            CurrPos=strlen(CommandLine[0]);

                            cn36_StrNCpy(CommandLine[0]+CurrPos, StartNSavetime, EndNSavetime-StartNSavetime);
                            CurrPos+=(EndNSavetime-StartNSavetime);

                            strcpy(CommandLine[0]+CurrPos, " -c ");
                            CurrPos+=strlen(CommandLine[0]+CurrPos);

                            cn36_StrNCpy(CommandLine[0]+CurrPos, StartOfSaveClient, EndOfSaveClient-StartOfSaveClient);
                            CurrPos+=(EndOfSaveClient-StartOfSaveClient);
                        
                            sprintf(CommandLine[0]+CurrPos,
                                    " -iY %s",
                                    Info.ppszFiles[0]);
                        }
                    }
                }
            }break;

            case GiveBackupInfo:
            {
                const char *ReportColumns;
                
                if(SSID==NSREBIDType)
                    ReportColumns="name(26),level(4),savetime(23),totalsize,ssid";  // 23 for "dd/mm/yyyy hh:mm:ss AM|PM"
                else
                    ReportColumns="name(26),level(4),savetime(23),totalsize,nsavetime";  // 23 for "dd/mm/yyyy hh:mm:ss AM|PM"

                Length=strlen(PathOfNSRMminfo)                      // "...\mminfo.exe" or ".../mminfo"
                       +3                                           // " -a"
                       +4+strlen(NameOfNSR_HOST)                    // " -s <NSR_HOST>"
                       +4+strlen(Info.SourceDBHost)                 // " -c <save_client>
                       +4+strlen(ReportColumns)                     // " -r "name(26),..."
                       +11                                          // " -q pssid=0"
                       +4+strlen(Info.SourceDBName)                 // " -N <save_dbname>"
                       +5;                                          // " -o t"
                
                if(!cn36_StrAlloc(CommandLine[0], Length))
                    rc=0;
                else
                    sprintf(CommandLine[0], "%s -a -s %s -c %s -r %s -q pssid=0 -N %s -o t", PathOfNSRMminfo,
                                                                                             NameOfNSR_HOST,
                                                                                             Info.SourceDBHost,
                                                                                             ReportColumns,
                                                                                             Info.SourceDBName);
            }break;

            default:
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                rc=0;
        }

        if(rc)
            if(strlen(CommandLine[0])!=Length)
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                cn36_StrDealloc(CommandLine[0]);
                rc=0;
            }
            else
            {
                NumberOfCommandLines=1;
                sprintf(MsgBuf_cn36, "Constructed NetWorker call '%s'.\n", CommandLine[0]);
                cn36_Verbose(MsgBuf_cn36);
            }
    }

    return rc;
}

int tcn352_NetWorkerConnector::Prepare()
{
    int rc=1;

    switch(GiveActionType())
    {
        case ABackup:
        case ARestore:
        case GiveBackupInfo:
            break;

        default:
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
        }break;
    }

    if(rc)
        rc=BuildBackupToolCall();

    if(rc)
    {
        ToolProcessOutput=new tni34_AFile(tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now
        ToolProcessError=new tni34_AFile(tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now

        if(0==ToolProcessOutput || 0==ToolProcessError)
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
        }
        else
        {
            if(ToolProcessOutput->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as output for NetWorker.\n", ToolProcessOutput->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;
        
            if(ToolProcessError->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as error output for NetWorker.\n", ToolProcessError->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;
        }

        NSROuputResult=NotChecked;  //just "created" the NetWorker output file, so it's not checked so far.
    }

    if(ABackup==GiveActionType())
    {
        if(rc)
        {
            // assure, that no two saves have the same time-stamp
            long CurrentTime=long(time(0));

            while(CurrentTime==StartSaveTime)
            {
                cn36_Sleep(1); //sleep a second
                CurrentTime=long(time(0));
            }

            StartSaveTime=CurrentTime;
        }
    }

    return rc;
}

int tcn352_NetWorkerConnector::CleanUpTool()
{
    int rc=1;

    if(0!=DirectiveFileName)
    {
        tni34_AFile DFile(DirectiveFileName, tni34_AFile::FromDisk, 0, 0); // do not open, do not create now

        if(DFile.Exists() && DFile.Remove())
        {
            sprintf(MsgBuf_cn36, "Removed NetWorker's directives file '%s'.\n", DirectiveFileName);
            cn36_Verbose(MsgBuf_cn36);
        }
        else
        {
            rc=0;

            sprintf(MsgBuf_cn36, "Could not remove NetWorker's directives file '%s'.\n", DirectiveFileName);
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    return rc;
}

int tcn352_NetWorkerConnector::CollectToolSaveIDs()
{
    int rc=0;

    NumberOfExternalSaveIDsFound=0;

    char *CmdLine=0;
    //const char * FixedParam=" -a -r name(26),nsavetime,level(4) -q pssid=0 -o t"; //<- use of nsavetime
    const char * FixedParam=" -a -r name(26),nsavetime,ssid,level(4) -q pssid=0 -o t"; //<- use of SSID

    cn36_Verbose("Requesting SaveID from NetWorker.\n", AddALevelToPrefix);
    
    size_t Length=strlen(PathOfNSRMminfo)           //"...\mminfo.exe" or ".../mminfo"
                  +4+strlen(NameOfNSR_HOST)         //" -s <NSR_HOST>
                  +4+strlen(Info.pszMachName)       //" -c <machine_name>
                  +strlen(FixedParam)               // see above
                  +4+strlen(Info.pszDBName);        //" -N <dbname>

    if(cn36_StrAlloc(CmdLine, Length))
    {
        sprintf(CmdLine,"%s -s %s -c %s%s -N %s",PathOfNSRMminfo
                                                ,NameOfNSR_HOST
                                                ,Info.pszMachName
                                                ,FixedParam
                                                ,Info.pszDBName);
        if(strlen(CmdLine)!=Length)
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            cn36_StrDealloc(CmdLine);
        }
        else
        {
            tni34_ATextFile MminfoOutput(tni34_AFile::FromDisk, 0, 0);

            if(MminfoOutput.IsGood() && MminfoOutput.Exists())
            {
                tni33_AProcess Mminfo(CmdLine, 0, &MminfoOutput);

                if(0==Mminfo.GiveReturnCode())
                {
                    EndSaveTime=long(time(0));

                    if(MminfoOutput.Open())
                    {
                        const char * Line=0;
                        rc=1;

                        while(rc && MminfoOutput.ReadLine(Line) && Line!=0)  // while no read error and not end of file
                        {
                            Line=cn36_FirstNonWhiteSpaceOf(Line);

                            if(!strncmp(Line, Info.pszDBName, strlen(Info.pszDBName)))
                            {
                                Line=cn36_FirstNonWhiteSpaceOf(cn36_FirstWhiteSpaceOf(Line)); //Line="nsavetime ..."
                                const char * EndOfThisItem=cn36_FirstWhiteSpaceOf(Line);

                                int SaveTime;
                                if(cn36_StrNToInt(SaveTime, Line, EndOfThisItem-Line))  // <- this is a year 2038 problem if int is signed 32 bit
                                {
                                    if(StartSaveTime<=SaveTime && SaveTime<=EndSaveTime)    // the savetime must be between begin and end of the backup
                                    {
                                        if(SSID==NSREBIDType)   // if we use SSID as identifier ...
                                        {
                                            Line=cn36_FirstNonWhiteSpaceOf(cn36_FirstWhiteSpaceOf(Line)); //... we must take the next item in the line (now Line="SSID ...")
                                            EndOfThisItem=cn36_FirstWhiteSpaceOf(Line);
                                            
                                            if(!cn36_StrNAllocCpy(ExternalSaveID[NumberOfExternalSaveIDsFound], Line, EndOfThisItem-Line))
                                                rc=0;
                                        }
                                        else
                                            rc=BuildEBIDFromNSavetime(ExternalSaveID[NumberOfExternalSaveIDsFound], Line, EndOfThisItem);

                                        if(rc)
                                        {
                                            sprintf(MsgBuf_cn36, "Found External Save ID '%s'.\n", ExternalSaveID[NumberOfExternalSaveIDsFound]);
                                            cn36_Verbose(MsgBuf_cn36);

                                            NumberOfExternalSaveIDsFound++; 
                                        }
                                        else
                                        {
                                            NumberOfExternalSaveIDsFound=0;
                                            rc=0;
                                        }
                                    } //else : we ignore older ID's from older saves
                                }
                                else
                                {
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfMminfoOutput.Set(Line));
                                    rc=0;
                                }
                            } //else : we ignore this line, because it is not starting with the dbname
                        }

                        if(rc && 0==NumberOfExternalSaveIDsFound)
                        {
                            rc=0;
                            cn36_Verbose("Output of mminfo contains no save information.\n");
                        }

                        if(!MminfoOutput.Remove()) //TODO: remove output in CleanUp()
                            rc=0;
                    } //else: rc already 0
                } //else: rc already 0
            } // else: rc already 0
        }
    } // else: rc already 0

    if(!rc)
    {
        sprintf(MsgBuf_cn36, "Could not get the External Save ID for the last save.\n");
        cn36_Verbose(MsgBuf_cn36);
    }
    
    cn36_Verbose(SubALevelFromPrefix);

    return rc;
}

int tcn352_NetWorkerConnector::CheckMedium()
{
    int rc=tcn35_BackupToolConnector::CheckMedium();

    if(rc)
    {
        if(Info.NumFiles>1)
            cn36_Verbose("Only media consiting of one pipe can be used with NetWorker at the moment.");

        //TODO? check for absolut name(s) of the pipe(s)
    }

    return rc;
}

int tcn352_NetWorkerConnector::GiveInformationOfLine(const char *   Line,
                                                     const char * & EBIDInLine,
                                                     size_t       & EBIDLength,
                                                     const char * & TypeString,
                                                     const char * & DateTimeString,
                                                     int          & LineContainsEBID)
{
    int rc=1;

    LineContainsEBID=0;

    if(0!=Line)
    {
        const char * Name=cn36_FirstNonWhiteSpaceOf(Line);
        const char * EndOfName=cn36_FirstWhiteSpaceOf(Name);

        size_t       LengthOfSourceDBName=strlen(Info.SourceDBName);

        if(!strncmp(Name, Info.SourceDBName, LengthOfSourceDBName) &&   // Line must start with the name of the saved database
           size_t(EndOfName-Name)==LengthOfSourceDBName)                // must start exactly with the name of the saved database
        {
            const char * Level=cn36_FirstNonWhiteSpaceOf(EndOfName);
            const char * EndOfLevel=cn36_FirstWhiteSpaceOf(Level);

            const char * SaveDate=cn36_FirstNonWhiteSpaceOf(EndOfLevel);
            const char * EndOfSaveDate=cn36_FirstWhiteSpaceOf(SaveDate);

            const char * SaveTime=cn36_FirstNonWhiteSpaceOf(EndOfSaveDate);
            const char * EndOfSaveTime=cn36_FirstWhiteSpaceOf(SaveTime);

            const char * AMPMString=cn36_FirstNonWhiteSpaceOf(EndOfSaveTime);   //is AM or PM following the time
            const char * EndOfAMPMString=cn36_FirstWhiteSpaceOf(AMPMString);

            const char * AM_cn36="AM";
            const char * PM_cn36="PM";

            int UsesAM=0, UsesPM=0;

            if(0==cn90StrNicmp(AMPMString, AM_cn36, EndOfAMPMString-AMPMString))
                UsesAM=1;
            else
                if(0==cn90StrNicmp(AMPMString, PM_cn36, strlen(PM_cn36)))
                    UsesPM=1;

            const char * TotalSize;    // TotalSize will not be tested or used
            const char * EndOfTotalSize;

            if(UsesAM || UsesPM)
                TotalSize=cn36_FirstNonWhiteSpaceOf(EndOfAMPMString);
            else
                TotalSize=cn36_FirstNonWhiteSpaceOf(EndOfSaveTime);

            EndOfTotalSize=cn36_FirstWhiteSpaceOf(TotalSize);

            const char * ID=cn36_FirstNonWhiteSpaceOf(EndOfTotalSize);  //can be SSID or NSaveTime
            const char * EndOfID=cn36_FirstWhiteSpaceOf(ID);
            
            if('\0'!=(*ID) &&                                   // ID must be not empty (Line must have enough parts)
               '\0'==(*cn36_FirstNonWhiteSpaceOf(EndOfID)) &&   // only white spaces are allowed behind ID
               cn36_StrNConsistsOfDigits(ID, EndOfID-ID))       // the ID must only consist of digits
            {
                if(SSID==NSREBIDType)
                {
                    EBIDInLine=ID;
                    EBIDLength=EndOfID-ID;
                }
                else
                {
                    rc=BuildEBIDFromNSavetime(TempEBID, ID, EndOfID);

                    if(rc)
                    {
                        EBIDInLine=TempEBID;    //EBIDInLine="NST <savetime> <machinename>"
                        EBIDLength=strlen(TempEBID);
                    }
                }

                if(rc)
                {
                    TypeString=cn352ActionNForLevel(Level, EndOfLevel-Level); // the level must be known and therefore correspond to an action

                    if(0!=TypeString)
                        TypeString=ActionStringForActionStringN(TypeString, strlen(TypeString));    //check transform into pointer to const from vcn35.cpp

                    if(0!=TypeString)
                    {
                        int YearWithFourDigits=0;

                        if(10==EndOfSaveDate-SaveDate)
                            YearWithFourDigits=1;

                        if((8==EndOfSaveDate-SaveDate || YearWithFourDigits) && // SaveDate must have form mm/dd/yy (-> 8 chars) or mm/dd/yyyy (-> 10 chars) 
                           8==EndOfSaveTime-SaveTime)                           // SaveTime must have form HH:MM:SS -> 8 chars
                        {
                            char SaveDateH[11];
                            char SaveTimeH[9];

                            cn36_StrNCpy(SaveDateH, SaveDate, EndOfSaveDate-SaveDate);   // make a copy because SaveDate is (const char *)
                            cn36_StrNCpy(SaveTimeH, SaveTime, 8);   // make a copy because SaveTime is (const char *)

                            SaveDateH[2]=SaveDateH[5]='0';
                            SaveTimeH[2]=SaveTimeH[5]='0';

                            if(cn36_StrConsistsOfDigits(SaveDateH) &&
                               cn36_StrConsistsOfDigits(SaveTimeH))
                            {
                                int SaveDateAsInt;
                                int SaveHourAsInt;

                                if(cn36_StrNToInt(SaveDateAsInt, SaveDateH, 8) &&
                                   cn36_StrNToInt(SaveHourAsInt, SaveTimeH, 2) &&
                                   cn36_StrAlloc(TempDateTimeString, 19))
                                {
                                    if(0<=SaveHourAsInt && SaveHourAsInt<24)
                                    {
                                        if(YearWithFourDigits)
                                            cn36_StrNCpy(TempDateTimeString, SaveDateH+6, 4);
                                        else
                                        {
                                            if(SaveDateAsInt%100<30)                // we deal with a Y2K issue
                                                strcpy(TempDateTimeString, "20");
                                            else
                                                strcpy(TempDateTimeString, "19");

                                            cn36_StrNCpy(TempDateTimeString+2, SaveDateH+6, 2);     // TempDateTimeString=="yyyy"
                                        }

                                        TempDateTimeString[4]='-';                                  // TempDateTimeString=="yyyy-"
                                        cn36_StrNCpy(TempDateTimeString+5, SaveDateH, 5);           // TempDateTimeString=="yyyy-mm0dd"
                                        TempDateTimeString[7]='-';                                  // TempDateTimeString=="yyyy-mm-dd"
                                        TempDateTimeString[10]=' ';                                 // TempDateTimeString=="yyyy-mm-dd "

                                        if(UsesPM && 0<=SaveHourAsInt && SaveHourAsInt<12)
                                        {
                                            SaveHourAsInt+=12;
                                            sprintf(TempDateTimeString+11, "%d", int(SaveHourAsInt));   // TempDateTimeString=="yyyy-mm-dd HH", can use this sprintf, because SaveHourAsInt>=12>10
                                            cn36_StrNCpy(TempDateTimeString+13, SaveTimeH+2, 6);        // TempDateTimeString=="yyyy-mm-dd HH0MM0SS"
                                        }
                                        else
                                            cn36_StrNCpy(TempDateTimeString+11, SaveTimeH, 8);      // TempDateTimeString=="yyyy-mm-dd HH0MM0SS"

                                        TempDateTimeString[13]=TempDateTimeString[16]=':';          // TempDateTimeString=="yyyy-mm-dd HH:MM:SS"
                                
                                        DateTimeString=TempDateTimeString;
                                        LineContainsEBID=1;
                                    }
                                }
                                else
                                    rc=0;
                            }
                        }
                    }
                }
            }
        }
    }

    return rc;
}

int tcn352_NetWorkerConnector::BuildEBIDFromNSavetime(char       * & EBID,
                                                      const char *   StartOfNSavetime,
                                                      const char *   EndOfNSavetime)
{
    int rc=1;

    size_t LengthOfNSavetime=EndOfNSavetime-StartOfNSavetime;

    if(!cn36_StrAlloc(EBID, strlen(NSaveTimeEBIDPrefix)     // try to get memory for a EBID of form "NST <nsavetime> <machinename>"
                            +strlen(NSaveTimeEBIDPartDelimiter)
                            +LengthOfNSavetime
                            +strlen(NSaveTimeEBIDPartDelimiter)
                            +strlen(Info.SourceDBHost))) 
    {
        rc=0;
    }
    else
    {                       //build the EBID "NST <nsavetime> <machinename>"
        size_t CurrPos;

        strcpy(EBID, NSaveTimeEBIDPrefix);                              // EBID="NST"
        CurrPos=strlen(NSaveTimeEBIDPrefix);

        strcpy(EBID+CurrPos, NSaveTimeEBIDPartDelimiter);               // EBID="NST "
        CurrPos+=strlen(NSaveTimeEBIDPartDelimiter);

        cn36_StrNCpy(EBID+CurrPos, StartOfNSavetime, LengthOfNSavetime);// EBID="NST <nsavetime>"
        CurrPos+=LengthOfNSavetime;

        strcpy(EBID+CurrPos, NSaveTimeEBIDPartDelimiter);               // EBID="NST <nsavetime> "
        CurrPos+=strlen(NSaveTimeEBIDPartDelimiter);
        
        strcpy(EBID+CurrPos, Info.SourceDBHost);                        // EBID="NST <nsavetime> <machinename>" (Info.SourceDBHost contains the current machine name on backup)
    }

    return rc;
}

int tcn352_NetWorkerConnector::ToolWasSuccessful()
{
    int rc=tcn35_BackupToolConnector::ToolWasSuccessful();  //check the NetWorker return code 

    if(rc && ABackup==GiveActionType() && UseNSROutput)     //if return code is 0 and we make a backup and we should look through NetWorkers output
    {
        if(FoundError==NSROuputResult)  // already searched the output of the NetWorker call and found an error
            rc=0;
        else
        {
            if(NotChecked==NSROuputResult)  //we have never before checked the current NetWorker output
            {
                sprintf(MsgBuf_cn36,"Analyzing output of %s.\n", GiveToolName());
                cn36_Verbose(MsgBuf_cn36, AddALevelToPrefix);

                if(!cn36_StrConsistsOfDigits(tTheDBResult.szBDPageCount) ||
                   !cn36_StrToUInt8(BDBackupSize, tTheDBResult.szBDPageCount))
                {
                    BDBackupSize=MAX_UINT8_CN36;    // this will lead at least to an error, but should not happen if db kernel sends correct replies
                }
                else
                    BDBackupSize*=Info.PageSize;
                
                int FoundCompletionMessage=0;

                {
                    tni34_ATextFile NSROutput(ToolProcessOutput->GiveName(), tni34_AFile::FromDisk, 0, 1);    //do not create, open now

                    if(NSROutput.IsGood())
                    {
                        const char *Line;

                        while(rc && NSROutput.ReadLine(Line) && Line!=0)    //for all lines of the file
                            rc=AnalyzeNSROutputLine(Line, FoundCompletionMessage);
                    }
                }

                if(rc)
                {
                    tni34_ATextFile NSRErrOutput(ToolProcessError->GiveName(), tni34_AFile::FromDisk, 0, 1);    //do not create, open now

                    if(NSRErrOutput.IsGood())
                    {
                        const char *Line;

                        while(rc && NSRErrOutput.ReadLine(Line) && Line!=0)    //for all lines of the file
                            rc=AnalyzeNSROutputLine(Line, FoundCompletionMessage);
                    }
                }

                if(0==FoundCompletionMessage)
                {
                    rc=0;   //we must find a completion message (if we can not open the file, we have not found one!)
                    sprintf(MsgBuf_cn36, "Could not find a completion message with the expected backup size in the %s output.\n", GiveToolName());
                }

                if(rc)
                {
                    sprintf(MsgBuf_cn36, "No error message found in %s output.\n\n", GiveToolName());
                    NSROuputResult=FoundNoError;
                }
                else
                {
                    sprintf(MsgBuf_cn36, "Found error message in %s output (see above).\n\n", GiveToolName());
                    NSROuputResult=FoundError;
                }

                cn36_Verbose(MsgBuf_cn36, SubALevelFromPrefix);
            }
            //else NetWorker output was already checked and no error was found -> return code is decisivly
        }
    }
    
    return rc;
}

int tcn352_NetWorkerConnector::AnalyzeNSROutputLine(const char * Line, int & FoundCompletionMessage)
{
    int rc=1;

    const char * SavePrefix="save:";

    if(0==strncmp(Line, SavePrefix, strlen(SavePrefix))) //Line starts with SavePrefix
    {
        const char * StartNextPart=cn36_FirstNonWhiteSpaceOf(Line+strlen(SavePrefix));
        const char * EndNextPart=cn36_FirstWhiteSpaceOf(StartNextPart);

        if(StartNextPart!=EndNextPart)  // there is something in the Line
        {
            if(FoundFileNameFollowedByWarning(Info.ppszFiles[0], StartNextPart))
                rc=0;
            else    //is it the completion message (save: <dbname> level=<level>, <size> B|KB|MB|GB|TB ..." )? 
            {
                if(FoundCompletionMessageWithCorrectSize(StartNextPart))
                {
                    FoundCompletionMessage=1;

                    sprintf(MsgBuf_cn36, "Have found an completion message in the %s output file line:\n%s\n", GiveToolName(), Line);
                    cn36_Verbose(MsgBuf_cn36);
                }
            }
        }
    }
    else
    {
        if(FoundFileNameFollowedByWarning(Info.ppszFiles[0], Line))     //in some NetWorker versions "warnings" are written without "save:"
            rc=0;
    }

    return rc;
}

int tcn352_NetWorkerConnector::FoundFileNameFollowedByWarning(const char * FileName,
                                                              const char * NSROutputLine)
{
    int rc=0;
    const char * BaseName=cn36_BaseNameFrom(FileName);
    const char * StartOfLine=cn36_FirstNonWhiteSpaceOf(NSROutputLine);
    const char * BehindTheFileName=0;

    if(0==cn36_StrStrNCmp(FileName, StartOfLine, strlen(FileName)))     //NSROutputLine starts with the pipe name
        BehindTheFileName=cn36_FirstNonWhiteSpaceOf(StartOfLine+strlen(FileName));
    else
        if(0==cn36_StrStrNCmp(BaseName, StartOfLine, strlen(BaseName))) //NSROutputLine starts with the pipe name without the path
            BehindTheFileName=cn36_FirstNonWhiteSpaceOf(StartOfLine+strlen(BaseName));

    if(0!=BehindTheFileName &&      //the line started with the FileName
       ':'==BehindTheFileName[0])   //and there is a ':' behind it. We do not check for further spaces and the error message itself, because the ':' should not be there on success.
    {
        ni31_ListElem<char *> *h=ListOfErrorsToIgnore.First();
        int                   ErrorIsValid=1;

        while(0!=h && ErrorIsValid)     //loop through the list of errors which must be ignored
        {
            if(0==strcmp(BehindTheFileName+1, h->GiveInfo()))
                ErrorIsValid=0;

            h=h->Next();
        }

        if(ErrorIsValid)    //if the current error was not in the list it is an error
        {
            sprintf(MsgBuf_cn36, "Interpreting '%s' form NetWorker output as error message for pipe '%s'.\n", NSROutputLine, Info.ppszFiles[0]);
            cn36_Verbose(MsgBuf_cn36);

            rc=1;
        }
    }

    return rc;
}

int tcn352_NetWorkerConnector::FoundCompletionMessageWithCorrectSize(const char * Line) //it's not the complete Line ("save: " is missing)
{    
    int rc=1;

    const char * StartNextPart=Line;
    const char * EndNextPart=cn36_FirstWhiteSpaceOf(StartNextPart);

    if(0!=cn36_StrStrNCmp(Info.pszDBName, StartNextPart, (EndNextPart-StartNextPart)))
        rc=0;

    if(rc)
    {
        const char * LevelKeyWord="level=";
        const char * ExpectedLevel=cn352LevelForAction(Info.pszActionString);

        StartNextPart=cn36_FirstNonWhiteSpaceOf(EndNextPart);
        EndNextPart=cn36_FirstWhiteSpaceOf(StartNextPart);

        if(size_t(EndNextPart-StartNextPart)<strlen(LevelKeyWord)+strlen(ExpectedLevel) ||          //next part is shorter than "level=" plus <expected level>
           size_t(EndNextPart-StartNextPart)>strlen(LevelKeyWord)+strlen(ExpectedLevel)+1 ||        //next part contains more than "level=<expected level>" or "level=<expected level>,"
            0!=strncmp(LevelKeyWord, StartNextPart, strlen(LevelKeyWord)) ||                        //next part does not start with "level="
            0!=strncmp(ExpectedLevel, StartNextPart+strlen(LevelKeyWord), strlen(ExpectedLevel)) || //"level=" is not followed by the expected NetWorker level string
            (EndNextPart==StartNextPart+strlen(LevelKeyWord)+strlen(ExpectedLevel)+1 &&             //next part contains more than "level=<expected level>"
             ','!= *(StartNextPart+strlen(LevelKeyWord)+strlen(ExpectedLevel))))                    //but does not end with ','
        {
            rc=0;
        }
    }

    if(rc)
    {
        StartNextPart=cn36_FirstNonWhiteSpaceOf(EndNextPart);
        EndNextPart=cn36_FirstWhiteSpaceOf(StartNextPart);


        teo00_Uint8 ReportedSize,
                    ReportedFraction=0,
                    FactorForUnit=1,
                    FactorForFractionUnit=0,
                    FactorBetweenUnits=1024;

        const char *Separator=StartNextPart;

        while(EndNextPart!=Separator &&
              ('.'!= (*Separator)) &&
              (','!= (*Separator)))
        {
            Separator++;
        }

        if(cn36_StrNConsistsOfDigits(StartNextPart, (Separator-StartNextPart)) &&                                   //next part starts with a number
           (EndNextPart==Separator || cn36_StrNConsistsOfDigits(Separator+1, (EndNextPart-(Separator+1)))) &&       //no separator or a number after the separator
           cn36_StrNToUInt8(ReportedSize, StartNextPart, (Separator-StartNextPart)) &&                              //could get the number into a teo00_Uint8
           (EndNextPart==Separator || cn36_StrNToUInt8(ReportedFraction, Separator+1, (EndNextPart-(Separator+1)))))//no separator or could get the fraction into a teo00_Uint8
        {
            StartNextPart=cn36_FirstNonWhiteSpaceOf(EndNextPart);
            EndNextPart=cn36_FirstWhiteSpaceOf(StartNextPart);

            if(0!=cn36_StrStrNUprCmp("B", StartNextPart, (EndNextPart-StartNextPart)) &&
               0!=cn36_StrStrNUprCmp("BYTE", StartNextPart, (EndNextPart-StartNextPart)))
            {
                FactorForUnit*=FactorBetweenUnits;
                FactorForFractionUnit=1;

                if(0!=cn36_StrStrNUprCmp("KB", StartNextPart, (EndNextPart-StartNextPart)))
                {
                    FactorForUnit*=FactorBetweenUnits;
                    FactorForFractionUnit*=FactorBetweenUnits;

                    if(0!=cn36_StrStrNUprCmp("MB", StartNextPart, (EndNextPart-StartNextPart)))
                    {
                        FactorForUnit*=FactorBetweenUnits;
                        FactorForFractionUnit*=FactorBetweenUnits;

                        if(0!=cn36_StrStrNUprCmp("GB", StartNextPart, (EndNextPart-StartNextPart)))
                        {
                            FactorForUnit*=FactorBetweenUnits;
                            FactorForFractionUnit*=FactorBetweenUnits;
                    
                            if(0!=cn36_StrStrNUprCmp("TB", StartNextPart, (EndNextPart-StartNextPart)))
                                rc=0;   // non of the known units fits -> we asume it is no completion message
                        }
                    }
                }
            }

            if(rc)
            {
                ReportedSize=ReportedSize*FactorForUnit+ReportedFraction*FactorForFractionUnit; //appply the units determined

                if(BDBackupSize>ReportedSize)
                {
                    const int SizeOfSizeAsString=30;    //21 would be enough for 2^64
                    char ReporedSizeAsString[SizeOfSizeAsString];
                    char DBSizeAsString[SizeOfSizeAsString];

                    rc=0;

                    if(!cn36_UInt8ToStrN(ReporedSizeAsString, SizeOfSizeAsString, ReportedSize) ||
                       !cn36_UInt8ToStrN(DBSizeAsString, SizeOfSizeAsString, BDBackupSize))
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    }
                    else
                    {
                        sprintf(MsgBuf_cn36, "According to line '%s', NetWorker has saved only %s bytes (the database kernel has saved %s bytes).\n", Line, ReporedSizeAsString, DBSizeAsString);
                        cn36_Verbose(MsgBuf_cn36);
                    }
                }
            }
        }
        else
            rc=0;
    }

    return rc;
}

void tcn352_NetWorkerConnector::DeleteListOfErrorsToIgnore()
{
    while(0!=ListOfErrorsToIgnore.First())                          //while something is in the list
    {
        delete ListOfErrorsToIgnore.First()->GiveInfo();            //delete the string the first element is pointing to
        ListOfErrorsToIgnore.Del();                                 //delete the firt list element itself
    }
}
