/*!*****************************************************************************
  module:       vcn353.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing connection between dbmsrv and Backint for SAP DB
                conforming backup tools
  last change:  2000-11-29 14:06
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hcn353.h"
#include "hni34.h"
#include "hcn42.h"

#include <limits.h>
#include <string.h>


// -----------------------------------------------------------------------------
// constants defined by the Backint interface
// -----------------------------------------------------------------------------

#define MAX_LENGTH_BID 16
#define MAX_LENGTH_BSIPATH 255

// -----------------------------------------------------------------------------
// implementation of class  tcn353_Backint2InputLine
// -----------------------------------------------------------------------------

class tcn353_Backint2InputLine
{
  public:
    enum BackintActionType{Backup, Inquire, Restore};

    tcn353_Backint2InputLine();
    ~tcn353_Backint2InputLine();

    int SetValues(const char * TheName,
                  int          IsAPipe,
                  const char * TheBID=0,
                  const char * TheDestination=0,
                  const char * EndOfDestination=0,
                  const char * TheSaveType=0,
                  const char * TheDateTime=0);

    int GenerateInputLine(char * & TheLine, BackintActionType Action);

    int TestAgianstOutputLine(const BackintActionType Action,
                              int &                   InputAndOutputLineAreEqual,
                              int &                   FoundForTheFirstTime,
                              const char *            OutFileName,
                              size_t                  OutFileNameLength,
                              const char *            OutFileBID,
                              size_t                  OutFileBIDLength);

    const char * GiveName() const;
    const char * GiveBID() const;
    const char * GiveSaveType() const;
    const char * GiveDateTime() const;
    int          WasFound() const;

  private:
    int    IsPipe;

    char * Name;
    char * Destination;
    char * BID;

    char * SaveType;
    char * DateTime;

    int AllreadyFound;
};

tcn353_Backint2InputLine::tcn353_Backint2InputLine()
{
    IsPipe=0;
    Name=0;
    Destination=0;
    BID=0;
    SaveType=0;
    DateTime=0;
    AllreadyFound=0;
}

tcn353_Backint2InputLine::~tcn353_Backint2InputLine()
{
    cn36_StrDealloc(Name);
    cn36_StrDealloc(Destination);
    cn36_StrDealloc(BID);
    cn36_StrDealloc(SaveType);
    cn36_StrDealloc(DateTime);
}

int tcn353_Backint2InputLine::SetValues(const char * TheName,
                                        int          IsAPipe,
                                        const char * TheBID,
                                        const char * TheDestination,
                                        const char * EndOfDestination,
                                        const char * TheSaveType,
                                        const char * TheDateTime)
{
    int rc=1;

    if(strlen(TheName)>MAX_LENGTH_BSIPATH) // Backint for SAP DB allows only 256 chars in a path
    {
        rc=0;
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::PipeOrFileNameToLong.Set(TheName));
    }

    //if(0!=TheBID && strlen(TheBID)>MAX_LENGTH_BID)        rc=0;

    if(0!=TheDestination && EndOfDestination-TheDestination>MAX_LENGTH_BSIPATH)
    {
        rc=0;

        char *HelpString;

        if(cn36_StrNAllocCpy(HelpString, TheDestination, EndOfDestination-TheDestination))
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::PipeOrFileNameToLong.Set(HelpString));
            cn36_StrDealloc(HelpString);
        }
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::PipeOrFileNameToLong.Set());
    }

    if(rc)
    {
        IsPipe=IsAPipe;
        
        rc=cn36_StrAllocCpy(Name, TheName);

        if(rc)
        {
            if(0!=TheBID)
                rc=cn36_StrAllocCpy(BID, TheBID);
            else
                cn36_StrDealloc(BID);
        }

        if(rc)
        {
            if(0!=TheDestination)
                rc=cn36_StrNAllocCpy(Destination, TheDestination, EndOfDestination-TheDestination);
            else
                cn36_StrDealloc(Destination);
        }

        if(rc)
        {
            if(0!=TheSaveType)
                rc=cn36_StrAllocCpy(SaveType, TheSaveType);
            else
                cn36_StrDealloc(SaveType);
        }

        if(rc)
        {
            if(0!=TheDateTime)
                rc=cn36_StrAllocCpy(DateTime, TheDateTime);
            else
                cn36_StrDealloc(DateTime);
        }
    }

    return rc;
}

int tcn353_Backint2InputLine::GenerateInputLine(char * & TheLine, BackintActionType Action)
{
    int rc=1;

    switch(Action)
    {
        case Backup:
            if(IsPipe)
            {
                const char * PipeKeyWord="#PIPE";

                if(!cn36_StrAlloc(TheLine, strlen(Name)+1+strlen(PipeKeyWord)))
                    rc=0;
                else
                    sprintf(TheLine, "%s %s", Name, PipeKeyWord);   // "<pipename> #PIPE"
            }
            else
            {
                if(!cn36_StrAllocCpy(TheLine, Name))    // "<filename>"
                    rc=0;
            }
            break;

        case Inquire:
        case Restore:
        {
            size_t Length;
            int    UseDestination=0;
            const char * EffectiveBID;

            if(0!=BID && BID[0]!='\0')
                EffectiveBID=BID;
            else
                EffectiveBID="#NULL";

            Length=strlen(EffectiveBID)+1+strlen(Name);

            if(Restore==Action && 0!=Destination && '\0'!=Destination[0])
            {
                Length+=1+strlen(Destination);
                UseDestination=1;
            }

            if(!cn36_StrAlloc(TheLine, Length))
                rc=0;
            else
            {
                if(UseDestination)
                    sprintf(TheLine, "%s %s %s", EffectiveBID, Name, Destination);  // "<BID>|#NULL <filename>|<pipename> <destination>"
                else
                    sprintf(TheLine, "%s %s", EffectiveBID, Name);                  // "<BID>|#NULL <filename>|<pipename>
            }
        }break;

        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
    }

    return rc;
}

int tcn353_Backint2InputLine::TestAgianstOutputLine(const BackintActionType Action,
                                                    int &                   InputAndOutputLineAreEqual,
                                                    int &                   FoundForTheFirstTime,
                                                    const char *            OutFileName,
                                                    size_t                  OutFileNameLength,
                                                    const char *            OutFileBID,
                                                    size_t                  OutFileBIDLength)
{
    int rc=1,
        BIDsAreEqual=0;

    FoundForTheFirstTime=InputAndOutputLineAreEqual=0;

    if(!cn36_StrStrNCmp(Name, OutFileName, OutFileNameLength)) // if it is the input and output have the same name
	{
        if(0==BID)
        {
            rc=cn36_StrNAllocCpy(BID, OutFileBID, OutFileBIDLength);
            BIDsAreEqual=1;
        }
        else
        {
            if(0==cn36_StrStrNCmp(BID, OutFileBID, OutFileBIDLength))
            {
                BIDsAreEqual=1;

                if(AllreadyFound)
                {
                    sprintf(MsgBuf_cn36, "BID '%s' was reported twice for '%s'! This is ignored.\n", BID, Name);
                    cn36_Verbose(MsgBuf_cn36);
                }
            }
        }

        switch(Action)
        {
            case Backup:
            case Restore:
                if(0==BIDsAreEqual)
                {
                    rc=0;

                    if(AllreadyFound)
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::TwoDifferentBIDs.Set(Name));
                    else
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::WrongBIDReturned.Set(Name));
                }
                else
                    InputAndOutputLineAreEqual=1;
            break;

            case Inquire:
                if(BIDsAreEqual)
                    InputAndOutputLineAreEqual=1;   // just ignore other pairs of BID/name while making an inquire!
            break;

            default:
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                rc=0;
        }

        if(rc && InputAndOutputLineAreEqual)
        {
            if(!AllreadyFound)
                FoundForTheFirstTime=1;

            AllreadyFound=1; // the Name (and BID) was found in the output file
        }
    }

    return rc;
}

const char * tcn353_Backint2InputLine::GiveName() const
{
    return Name;
}

const char * tcn353_Backint2InputLine::GiveBID() const
{
    return BID;
}

const char * tcn353_Backint2InputLine::GiveSaveType() const
{
    return SaveType;
}

const char * tcn353_Backint2InputLine::GiveDateTime() const
{
    return DateTime;
}

int tcn353_Backint2InputLine::WasFound() const
{
    return AllreadyFound;
}


// -----------------------------------------------------------------------------
// implementation of class  tcn353Backint2DBMKNLLine
// -----------------------------------------------------------------------------

class tcn353Backint2DBMKNLLine
{
  public:
    tcn353Backint2DBMKNLLine();
    ~tcn353Backint2DBMKNLLine();

    int TakeBackint2InfoFromDBMKNLLine(int & IsABackint2Line, const char * ADBMKNLLine, const char * NameOfDBMKNL);

    const char * GiveKeyLabel() const;
    const char * GiveType() const;
    const char * GiveDateTime() const;

  private:
    char * StringData;    // just for easier handling of dynamically allocated memory

    const char * KeyLabel;
    const char * DateTime;
    const char * TypeString;

    static const char SectionDelimiter;
};

const char tcn353Backint2DBMKNLLine::SectionDelimiter='|';

tcn353Backint2DBMKNLLine::tcn353Backint2DBMKNLLine()
    :StringData(0), KeyLabel(0), DateTime(0)
{
}

tcn353Backint2DBMKNLLine::~tcn353Backint2DBMKNLLine()
{
    cn36_StrDealloc(StringData);
}

int tcn353Backint2DBMKNLLine::TakeBackint2InfoFromDBMKNLLine(int & IsABackint2Line, const char * ADBMKNLLine, const char * NameOfDBMKNL)
{
    int rc=1;

    IsABackint2Line=0;

    if('\0'!=ADBMKNLLine[0])
    {
        const char * Medianame=cn36_LoopedStrChr(ADBMKNLLine, SectionDelimiter, 10);   //the media name starts after the 10th '|'

        if(0==Medianame || '\0'==Medianame[0])
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile.Set(NameOfDBMKNL));
        }
        else
        {                                                               // we have at least 10 '|' in the string, so the subsequent calls to cn36_LoopedStrChr() and cn36_LoopedStrChr can be made without danger
            tcn35_BackupToolConnectorEnum BackupToolType(++Medianame);

            if(tcn35_BackupToolConnectorEnum::BACK==BackupToolType) // was the backup of the ADBMKNLLine made with Backint for SAP DB?
            {
                const char * EndKeyLabel;
                
                KeyLabel   =ADBMKNLLine;
                EndKeyLabel=cn36_LoopedStrChr(KeyLabel, SectionDelimiter, 2);

                if(0==strncmp("SAVE ", EndKeyLabel+1, strlen("SAVE "))) // the action string must start with SAVE_ all other entries (RESTORE, HISTLOST, ...) must be ignored
                {
                    const char * EndDateTime;
                    const char * LogStatus;
                    const char * EndLogStatus;
                    int NeedLog=0;

                    IsABackint2Line=1;

                    DateTime   =cn36_LoopedStrChr(EndKeyLabel+1, SectionDelimiter, 3)+1;
                    EndDateTime=cn36_StrChr(DateTime+1, SectionDelimiter);

                    LogStatus   =cn36_LoopedStrChr(EndDateTime+1, SectionDelimiter, 3)+1;
                    EndLogStatus=cn36_StrChr(LogStatus+1, SectionDelimiter);

                    if(0==cn36_StrStrNCmp("YES", LogStatus, EndLogStatus-LogStatus))
                        NeedLog=1;
                    else
                        if(0!=cn36_StrStrNCmp("NO ", LogStatus, EndLogStatus-LogStatus) &&
                           0!=cn36_StrStrNCmp("   ", LogStatus, EndLogStatus-LogStatus))
                        {
                            rc=0;
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile.Set(NameOfDBMKNL));
                        }

                    if(rc)
                    {
                        const char * Label=cn36_StrChr(KeyLabel, SectionDelimiter)+1;

                        if(0==strncmp(Label, "DAT_", strlen("DAT_")))
                        {
                            if(NeedLog)
                                TypeString=tcn35_SaveInfo::Action_String_DataRecovery;
                            else
                                TypeString=tcn35_SaveInfo::Action_String_DataMigration;
                        }
                        else
                            if(0==strncmp(Label, "PAG_", strlen("PAG_")))
                            {
                                if(NeedLog)
                                    TypeString=tcn35_SaveInfo::Action_String_PagesRecovery;
                                else
                                    TypeString=tcn35_SaveInfo::Action_String_PagesMigration;
                            }
                            else
                                if(0==strncmp(Label, "LOG_",strlen("LOG_")))
                                    TypeString=tcn35_SaveInfo::Action_String_Log;
                                else
                                {
                                    rc=0;
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile.Set(NameOfDBMKNL));
                                }
                    }

                    if(rc)
                        rc=cn36_StrAlloc(StringData, (EndKeyLabel-KeyLabel)+1+(EndDateTime-DateTime)+1+strlen(TypeString));

                    if(rc)
                    {
                        char * CurrentPosition;

                        cn36_StrNCpy(StringData, KeyLabel, (EndKeyLabel-KeyLabel));
                        KeyLabel=StringData;

                        CurrentPosition=cn36_PositionOfTerminatingZero(StringData)+1;
                        cn36_StrNCpy(CurrentPosition, DateTime ,(EndDateTime-DateTime));
                        DateTime=CurrentPosition;

                        CurrentPosition=cn36_PositionOfTerminatingZero(CurrentPosition)+1;
                        strcpy(CurrentPosition, /*SaveType*/TypeString);
                        TypeString=CurrentPosition;
                    }
                }
                //else ignore RESTORE, HISTLOST and others
            }
        }
    }

    return rc;
}

const char * tcn353Backint2DBMKNLLine::GiveKeyLabel() const
{
    return KeyLabel;
}

const char * tcn353Backint2DBMKNLLine::GiveType() const
{
    return TypeString;
}

const char * tcn353Backint2DBMKNLLine::GiveDateTime() const
{
    return DateTime;
}

// -----------------------------------------------------------------------------
// implementation of class tcn353Backint2EBID
// -----------------------------------------------------------------------------

class tcn353Backint2EBID
{
  public:
    tcn353Backint2EBID();
    ~tcn353Backint2EBID();

    int TakeInfoFromString(int        & ContainsInfo,
                           const char * String,
                           const char   ListDelimiter);

    int TakeInfoFromStringList(int          & ContainsInfo,
                               const char * & CurrentPosition,
                               const char     ListDelimiter);

    const char * GiveUser() const;
    const char * GivePath() const;
    const char * GiveBID() const;

  private:
    char * StringData;    // just for easier handling of dynamically allocated memory

    const char * User;
    const char * BID;
    const char * Path;
};

tcn353Backint2EBID::tcn353Backint2EBID()
    :StringData(0), User(0), BID(0), Path(0)
{
}

tcn353Backint2EBID::~tcn353Backint2EBID()
{
    cn36_StrDealloc(StringData);
}

int tcn353Backint2EBID::TakeInfoFromString(int & ContainsInfo, const char * String, const char ListDelimiter/*size_t StringLength*/)
{
    const char * StringList=String;

    return TakeInfoFromStringList(ContainsInfo, StringList, ListDelimiter);
}

int tcn353Backint2EBID::TakeInfoFromStringList(int          & ContainsInfo,
                                               const char * & CurrentPosition,
                                               const char     ListDelimiter)
{
    int rc=1;

    ContainsInfo=0;

    if(0!=CurrentPosition)
    {
        User                  =cn36_FirstNonWhiteSpaceOf(CurrentPosition);  //go to the start of the User
        const char * EndOfUser=cn36_FirstWhiteSpaceOf(User);                //go to end of User

        BID                  =cn36_FirstNonWhiteSpaceOf(EndOfUser);         //go to start of BID
        const char * EndOfBID=cn36_FirstWhiteSpaceOf(BID);                  //go to end of BID

        Path=cn36_FirstNonWhiteSpaceOf(EndOfBID);                           //go to start of Path

        if('\0'!=(*Path) && ListDelimiter!=(*Path))
        {
            ContainsInfo=1;

            const char * EndOfPath=strchr(Path, ListDelimiter);

            if(0==EndOfPath)
                CurrentPosition=EndOfPath=cn36_PositionOfTerminatingZero(Path);
            else
                CurrentPosition=EndOfPath+1;

            rc=cn36_StrAlloc(StringData, (EndOfUser-User)+1+(EndOfBID-BID)+1+(EndOfPath-Path));

            if(rc)
            {
                char *h, *h2;

                cn36_StrNCpy(StringData, User, (EndOfUser-User));   // copy User to StringData
                h=StringData+(EndOfUser-User)+1;                    // h points to place for BID
                User=StringData;                                    // User points to new copy of user

                cn36_StrNCpy(h, BID, (EndOfBID-BID));               // copy BID to StringData
                h2=h+(EndOfBID-BID)+1;                              // h2 points to place for Path
                BID=h;                                              // BID points to new copy of backup ID

                cn36_StrNCpy(h2, Path, (EndOfPath-Path));           // copy Path to StrinData
                Path=h2;                                            // Path points to new copy of Path
            }
        }
    }

    return rc;
}

const char * tcn353Backint2EBID::GiveUser() const
{
    return User;
}

const char * tcn353Backint2EBID::GiveBID() const
{
    return BID;
}

const char * tcn353Backint2EBID::GivePath() const
{
    return Path;
}


// -----------------------------------------------------------------------------
// implementation of class tcn353Backint2DBMEBFLine
// -----------------------------------------------------------------------------

class tcn353Backint2DBMEBFLine
{
  public:
    tcn353Backint2DBMEBFLine();
    ~tcn353Backint2DBMEBFLine();

    int TakeBackint2InfoFromDBMEBFLine(int & ContainsInfo, const char * ADBMEBFLine, const char * NameOfDBMEBF);
    int EqualsInKeyAndLabel(const tcn353Backint2DBMKNLLine * DBMKNLInfo);

    const char * GiveUser() const;
    const char * GivePath() const;
    const char * GiveBID() const;

  private:
    tcn353Backint2EBID Backint2EBID;
    char * KeyLabel;

    static const char SectionDelimiter;
};

const char tcn353Backint2DBMEBFLine::SectionDelimiter='|';

tcn353Backint2DBMEBFLine::tcn353Backint2DBMEBFLine()
    :KeyLabel(0)
{
}

tcn353Backint2DBMEBFLine::~tcn353Backint2DBMEBFLine()
{
    cn36_StrDealloc(KeyLabel);
}

int tcn353Backint2DBMEBFLine::TakeBackint2InfoFromDBMEBFLine(int & ContainsInfo, const char * DBMEBFLine, const char * NameOfDBMEBF)
{
    int rc=1;

    ContainsInfo=0;

    if('\0'!=DBMEBFLine[0])
    {
        const char * ThirdDelimiter=cn36_LoopedStrChr(DBMEBFLine, SectionDelimiter, 3);

        if(0!=ThirdDelimiter && '\0'==ThirdDelimiter[1])    // the line must end exactly after the 3rd '|'
        {
            const char * EndOfKeyLabel=cn36_LoopedStrChr(DBMEBFLine, SectionDelimiter, 2);

            rc=Backint2EBID.TakeInfoFromString(ContainsInfo, EndOfKeyLabel+1, SectionDelimiter);

            if(rc && ContainsInfo)
                rc=cn36_StrNAllocCpy(KeyLabel, DBMEBFLine, EndOfKeyLabel-DBMEBFLine);
        }
        else
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseUnknownFormatOfDBMEbfFile, NameOfDBMEBF));
        }
    }
     
    return rc;
}

int tcn353Backint2DBMEBFLine::EqualsInKeyAndLabel(const tcn353Backint2DBMKNLLine * DBMKNLInfo)
{
    int rc=0;

    if(0==strcmp(KeyLabel, DBMKNLInfo->GiveKeyLabel()))
        rc=1;

    return rc;
}

const char * tcn353Backint2DBMEBFLine::GiveUser() const
{
    return Backint2EBID.GiveUser();
}

const char * tcn353Backint2DBMEBFLine::GiveBID() const
{
    return Backint2EBID.GiveBID();
}

const char * tcn353Backint2DBMEBFLine::GivePath() const
{
    return Backint2EBID.GivePath();
}


// -----------------------------------------------------------------------------
// implementation of class tcn353_Backint2Connector
// -----------------------------------------------------------------------------

tcn353_Backint2Connector::tcn353_Backint2Connector(tcn35_ActionType TheActionType, const tcn35_SaveInfo & rInfo)
  :tcn35_BackupToolConnector(TheActionType, rInfo),NumberOfBackupHistories(NumberOfBackupHistories_cn353)
{
    const char * BackintEnvFileDefaultName  ="bsi.env";
    const char * BackintInFileDefaultName   ="bsi.in";
    const char * BackintOutFileDefaultName  ="bsi.out";
    const char * BackintErrorFileDefaultName="bsi.err";

    cn36_ConstructCompletePath(PathToEnvFile=0, Info.DBRunDir, BackintEnvFileDefaultName);

    BackintPath=0;
    cn36_ConstructCompletePath(BackintInFile=0, Info.DBRunDir, BackintInFileDefaultName);
    cn36_ConstructCompletePath(BackintOutFile=0, Info.DBRunDir, BackintOutFileDefaultName);
    cn36_ConstructCompletePath(BackintErrorFile=0, Info.DBRunDir, BackintErrorFileDefaultName);
    BackintParamFile=0;

    BackintTimeOutSuccess=300;  // on success use a time out of 300 seconds (5 minutes), if nothing else is specified in the environment file
    BackintTimeOutFailure=300;  // same default value for error situations

    int i;

    NumberOfRequestedNames=0;

    BackintDataType=DataPipes;

    SuffixBackupHistory[0]=".knl";
    SuffixBackupHistory[1]=".ebf";
    SuffixOfCopyOfBackupHistory[0]=".knc";
    SuffixOfCopyOfBackupHistory[1]=".ebc";

    for(i=0; i<NumberOfBackupHistories; i++)
    {
        OriginalBackupHistoryMovedTo[i]=0;
        OriginalBackupHistoryName[i]=0;
        PathOfMigrationBackintHistory[i]=0;
        BackupHistoryState[i]=Unknown;
    }

    InFileItem=0;
    TempBID=0;
}

tcn353_Backint2Connector::~tcn353_Backint2Connector()
{
    cn36_StrDealloc(PathToEnvFile);

    cn36_StrDealloc(BackintPath);
    cn36_StrDealloc(BackintInFile);
    cn36_StrDealloc(BackintOutFile);
    cn36_StrDealloc(BackintErrorFile);
    cn36_StrDealloc(BackintParamFile);

    int i;

    DeleteBackintInputList();
    
    for(i=0; i<NumberOfBackupHistories; i++)
    {
        cn36_StrDealloc(OriginalBackupHistoryMovedTo[i]);
        cn36_StrDealloc(OriginalBackupHistoryName[i]);
        cn36_StrDealloc(PathOfMigrationBackintHistory[i]);
    }

    cn36_StrDealloc(TempBID);
}

const char * tcn353_Backint2Connector::GiveToolName()
{
    return "Backint for SAP DB";
}

const char * tcn353_Backint2Connector::GiveLongToolName()
{
    return "Backint for SAP DB Interface";
}

int tcn353_Backint2Connector::CheckToolInstallation(int & IsInstalled)
{
    int rc=1;

    IsInstalled=1;

    cn36_Verbose("Checking existence and configuration of Backint for SAP DB.\n", AddALevelToPrefix);

    rc=cn36_DetermineValueOfEnvironmentVariable("BSI_ENV",
                                                PathToEnvFile ,
                                                "path of the configuration file of Backint for SAP DB",
                                                Info.pszDBName,
                                                0);

    if(rc)
        if(!cn36_CheckRead(PathToEnvFile))
        {
            IsInstalled=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsMissingOrUnreadable.Set(PathToEnvFile));

            cn36_Verbose("Can not read the Backint for SAP DB configuration file.\n");
        }
        else
        {
            unsigned i=0;
            char * OriginalRundirectory=0;
            const unsigned NumberOfEnvKeyWords=8;
            const char * EnvKeyWord[NumberOfEnvKeyWords]={"BACKINT",
                                                          "INPUT",
                                                          "OUTPUT",
                                                          "ERROROUTPUT",
                                                          "PARAMETERFILE",
                                                          "TIMEOUT_SUCCESS",
                                                          "TIMEOUT_FAILURE",
                                                          "ORIGINAL_RUNDIRECTORY"};

            char ** StringValueForKeyWord[NumberOfEnvKeyWords]={&BackintPath,
                                                                &BackintInFile,
                                                                &BackintOutFile,
                                                                &BackintErrorFile,
                                                                &BackintParamFile,
                                                                0,
                                                                0,
                                                                &OriginalRundirectory};
            int * IntValueForKeyWord[NumberOfEnvKeyWords]={0,
                                                           0,
                                                           0,
                                                           0,
                                                           0,
                                                           &BackintTimeOutSuccess,
                                                           &BackintTimeOutFailure,
                                                           0};

            tni34_ATextFile TheEnvFile(PathToEnvFile,
                                       tni34_AFile::FromDisk,
                                       0);

            if(TheEnvFile.IsGood())
            {
                const char * Line=0;
                char * FirstWordInUpperLetters=0;

                sprintf(MsgBuf_cn36, "Reading the Backint for SAP DB configuration file '%s'.\n", PathToEnvFile);
                cn36_Verbose(MsgBuf_cn36, AddALevelToPrefix);

                while(rc && TheEnvFile.ReadLine(Line) && 0!=Line)
                {
                    unsigned j;
                    int      FoundAKeyWord=0;

                    const char * FirstWord=cn36_FirstNonWhiteSpaceOf(Line);
                    size_t LengthOfFirstWord=cn36_FirstWhiteSpaceOf(FirstWord)-FirstWord;

                    if(cn36_StrNAllocCpy(FirstWordInUpperLetters, FirstWord, LengthOfFirstWord))
                    {
                        cn36_StrUpr(FirstWordInUpperLetters);
                        
                        for(j=0; j<NumberOfEnvKeyWords; j++)
                            if(!strncmp(FirstWordInUpperLetters, EnvKeyWord[j], strlen(EnvKeyWord[j])))
                            {
                                const char * SecondWord=cn36_FirstNonWhiteSpaceOf(FirstWord+LengthOfFirstWord);
                                const char * EndOfSecondWord;
                                size_t       LengthOfSecondWord;
                                int          AFormatErrorOccured=0;

                                // if the value contains white spaces, it is enclosed in a pair of double quotes
                                cn36_GetStringBetweenDoubleQuotesOrWhiteSpaces(SecondWord, EndOfSecondWord);
                                LengthOfSecondWord=EndOfSecondWord-SecondWord;

                                FoundAKeyWord=1;

                                if(0!=StringValueForKeyWord[j])
                                {
                                    if('\0'==SecondWord[0])
                                        AFormatErrorOccured=1;
                                    else
                                        if(!cn36_StrNAllocCpy(*StringValueForKeyWord[j], SecondWord, LengthOfSecondWord))
                                            rc=0;
                                        else
                                        {
                                            sprintf(MsgBuf_cn36, "Found keyword '%s' with value '%s'.\n", EnvKeyWord[j], *StringValueForKeyWord[j]);
                                            cn36_Verbose(MsgBuf_cn36);
                                        }
                                }

                                if(0!=IntValueForKeyWord[j])
                                {
                                    if(!cn36_StrNToInt(*IntValueForKeyWord[j], SecondWord, LengthOfSecondWord))
                                        AFormatErrorOccured=1;
                                    else
                                    {
                                        sprintf(MsgBuf_cn36, "Found keyword '%s' with value '%d'.\n", EnvKeyWord[j], (int)(*IntValueForKeyWord[j]));
                                        cn36_Verbose(MsgBuf_cn36);
                                    }
                                }

                                if(AFormatErrorOccured)
                                {
                                    cn36_Verbose("The following line of the Backint for SAP DB configuration file contains a format error and is ignored:\n", AddALevelToPrefix);
                                    cn36_Verbose(Line);
                                    cn36_Verbose("\n", NoPrefix);
                                    cn36_Verbose(SubALevelFromPrefix);
                                }

                                j=NumberOfEnvKeyWords; // we have found a keyword no other will match
                            }

                        if(!FoundAKeyWord && Line[0]!='\0')
                        {
                            cn36_Verbose("The following line of the Backint for SAP DB configuration file does not start with a proper keyword and is ignored:\n", AddALevelToPrefix);
                            cn36_Verbose(Line);
                            cn36_Verbose("\n", NoPrefix);
                            cn36_Verbose(SubALevelFromPrefix);
                        }
                    }
                    else
                        rc=0;
                }

                cn36_StrDealloc(FirstWordInUpperLetters);

                cn36_Verbose("Finished reading of the Backint for SAP DB configuration file.\n", SubALevelFromPrefix);

                if(rc)
                {
                    const char * BackHist=GivePathOfBackupHistory();
                    const char * ExtBackHist=GivePathOfExternalBackupHistory();

                    if(0==BackHist || 0==ExtBackHist)
                        rc=0;
                    else
                    {
                        if(0!=OriginalRundirectory) //if found original rundirectory
                        {
                            if(!cn36_ConstructCompletePath(PathOfMigrationBackintHistory[0], OriginalRundirectory, cn36_BaseNameFrom(BackHist)) ||  //construct the original path
                               !cn36_ConstructCompletePath(PathOfMigrationBackintHistory[1], OriginalRundirectory, cn36_BaseNameFrom(ExtBackHist))) //construct the original path
                            {
                                rc=0;
                            }
                        }
                        else    //no original rundirectory was specified in the configuration file
                        {
                            if(!cn36_StrAllocCpy(PathOfMigrationBackintHistory[0], BackHist) ||
                               !cn36_StrAllocCpy(PathOfMigrationBackintHistory[1], ExtBackHist))
                            {
                                rc=0;
                            }
                        }
                    }
                }

                cn36_StrDealloc(OriginalRundirectory); //we do not need it anymore, if 0==OriginalRundirectory nothing is done

                if(rc)  //1==IsInstalled holds
                {
                    if(0==BackintPath)    
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnspecifiedBackintPath.Set(PathToEnvFile));
                        IsInstalled=0;
                    }
                    else
                    {
                        if(!cn36_CheckExe(BackintPath))
                        {
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsNotExecutable.Set(BackintPath));
                            IsInstalled=0;

                            sprintf(MsgBuf_cn36, "The Backint for SAP DB program '%s' is missing or not executable.\n", BackintPath);
                            cn36_Verbose(MsgBuf_cn36);
                        }
                    }

                    if(cn36_CheckExist(BackintInFile))
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileAlreadyExists.Set(BackintInFile));
                        IsInstalled=0;

                        sprintf(MsgBuf_cn36, "'%s' already exists and can not be used as input file of Backint for SAP DB.\n", BackintInFile);
                        cn36_Verbose(MsgBuf_cn36);
                    }

                    if(cn36_CheckExist(BackintOutFile))
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileAlreadyExists.Set(BackintOutFile));
                        IsInstalled=0;
                        
                        sprintf(MsgBuf_cn36, "'%s' already exists and can not be used as output file of Backint for SAP DB.\n", BackintOutFile);
                        cn36_Verbose(MsgBuf_cn36);
                    }

                    if(cn36_CheckExist(BackintErrorFile))
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileAlreadyExists.Set(BackintErrorFile));
                        IsInstalled=0;
                        
                        sprintf(MsgBuf_cn36, "'%s' already exists and can not be used as error output file of Backint for SAP DB.\n", BackintErrorFile);
                        cn36_Verbose(MsgBuf_cn36);
                    }

                    if(IsInstalled)
                    {
                        int MaximalTimeOut=int(INT_MAX)/(int(1000));

                        sprintf(MsgBuf_cn36, "Using '%s' as Backint for SAP DB program.\n", BackintPath);
                        cn36_Verbose(MsgBuf_cn36);

                        sprintf(MsgBuf_cn36, "Using '%s' as input file for Backint for .\n", BackintInFile);
                        cn36_Verbose(MsgBuf_cn36);

                        sprintf(MsgBuf_cn36, "Using '%s' as output file for Backint.\n", BackintOutFile);
                        cn36_Verbose(MsgBuf_cn36);

                        sprintf(MsgBuf_cn36, "Using '%s' as error output file for Backint.\n", BackintErrorFile);
                        cn36_Verbose(MsgBuf_cn36);

	                    if(0==BackintParamFile)
		                    cn36_Verbose("Using no parameter file for Backint for SAP DB.\n");
	                    else
                        {
		                    sprintf(MsgBuf_cn36, "Using '%s' as parameter file for Backint for SAP DB.\n", BackintParamFile);
                            cn36_Verbose(MsgBuf_cn36);
                        }

                        if(MaximalTimeOut<BackintTimeOutSuccess)
                        {
                            sprintf(MsgBuf_cn36, "Reducing value '%d' to its upper limit '%d'.\n", int(BackintTimeOutSuccess), int(MaximalTimeOut));
                            cn36_Verbose(MsgBuf_cn36);

                            BackintTimeOutSuccess=MaximalTimeOut;
                        }

                        sprintf(MsgBuf_cn36, "Using '%d' seconds as timeout for Backint in the case of success.\n", int(BackintTimeOutSuccess));
                        cn36_Verbose(MsgBuf_cn36);

                        if(MaximalTimeOut<BackintTimeOutFailure)
                        {
                            sprintf(MsgBuf_cn36, "Reducing value '%d' to its upper limit '%d'.\n", int(BackintTimeOutFailure), int(MaximalTimeOut));
                            cn36_Verbose(MsgBuf_cn36);

                            BackintTimeOutFailure=MaximalTimeOut;
                        }

                        sprintf(MsgBuf_cn36, "Using '%d' seconds as timeout for Backint in the case of failure.\n", int(BackintTimeOutFailure));
                        cn36_Verbose(MsgBuf_cn36);

                        if(0!=PathOfMigrationBackintHistory[0])
                        {
                            sprintf(MsgBuf_cn36, "Using '%s' as backup history of a database to migrate.\n", PathOfMigrationBackintHistory[0]);
                            cn36_Verbose(MsgBuf_cn36);
                        }

                        if(0!=PathOfMigrationBackintHistory[1])
                        {
                            sprintf(MsgBuf_cn36, "Using '%s' as external backup history of a database to migrate.\n", PathOfMigrationBackintHistory[1]);//PathOfMigrationExternalBackintHistory);
                            cn36_Verbose(MsgBuf_cn36);
                        }
                    }
                }
            }
            else
                rc=0;
        }

    cn36_Verbose((rc && IsInstalled)?"Check passed successful.\n\n":"Check failed.\n\n", SubALevelFromPrefix);

	return rc;
}

const char * BI_BACKUPStringForActionString(const char *ActionString)
{
    const char * rc=0;

    if(0==strcmp(ActionString, tcn35_SaveInfo::Action_String_DataMigration) ||
       0==strcmp(ActionString, tcn35_SaveInfo::Action_String_DataRecovery))
    {
        rc="FULL";
    }
    else
        if(0==strcmp(ActionString, tcn35_SaveInfo::Action_String_PagesMigration) ||
           0==strcmp(ActionString, tcn35_SaveInfo::Action_String_PagesRecovery))
        {
            rc="PARTIAL";
        }
        else
            if(0==strcmp(ActionString, tcn35_SaveInfo::Action_String_Log))
                rc="ARCHIVE";

    return rc;
}

int tcn353_Backint2Connector::Prepare()
{
    int rc=1,
        NumberOfInFileEntries;

    DeleteBackintInputList();

    if(!cn36_PutEnvironmentVariable("BI_CALLER", "DBMSRV") ||
       !cn36_PutEnvironmentVariable("BI_REQUEST", "NEW"))
    {
        rc=0;
    }

    if(rc)
        switch(GiveActionType())
        {
            case ABackup:   // on backup ...
            case ARestore:  // ... and restore
            {
                switch(BackintDataType)
                {
                    case DataPipes: //set the backup type (data, pages, log) in environment variable BI_BACKUP
                    {
                        const char * BI_BackupType=BI_BACKUPStringForActionString(Info.pszActionString);

                        if(0==BI_BackupType)
                        {
                            rc=0;
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                        }
                        else
                            if(!cn36_PutEnvironmentVariable("BI_BACKUP", BI_BackupType))
                                rc=0;
                    }break;

                    case HistoryFiles:
                        if(!cn36_PutEnvironmentVariable("BI_BACKUP", ""))
                            rc=0;
                    break;

                    default:
                    {
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                        rc=0;
                    }
                }
            }break;

            case GiveBackupInfo:
            {
                if(!cn36_PutEnvironmentVariable("BI_BACKUP", ""))
                    rc=0;
                else
                {
                    switch(BackintDataType)
                    {
                        case DataPipes:
                        {
                            int i, AllBackupHistoriesExist=1;

                            const char * PathDBMKNL=GivePathOfBackupHistory();
                            const char * PathDBMEBF=GivePathOfExternalBackupHistory();
        
                            if(0==PathDBMKNL || 0==PathDBMEBF)  // just get sure that the variables of the base class were filled successfully
                                rc=0;
                            else
                            {
                                rc=cn36_StrAllocCpy(OriginalBackupHistoryName[0], PathDBMKNL);
            
                                if(rc)
                                    rc=cn36_StrAllocCpy(OriginalBackupHistoryName[1], PathDBMEBF);

                                if(rc)
                                {
                                    for(i=0; i<NumberOfBackupHistories; i++)    // check the availability of all history files
                                    {
                                        if(cn36_CheckExist(OriginalBackupHistoryName[i]))
                                            BackupHistoryState[i]=Exist;
                                        else
                                        {
                                            BackupHistoryState[i]=NotExist;
                                            AllBackupHistoriesExist=0;
                                        }
                                    }
                        
                                    if(0!=strcmp(Info.SourceDBName, Info.pszDBName) ||  // source db is different from current db
                                       !AllBackupHistoriesExist)                        // the backup history or the external backup history or both do not exist
                                    {
                                        rc=BackupOrRestoreHistoryWithBackint();         // get the needed history files from Backint for SAP DB
                                    }
                                }
                            }
                        }break;

                        case HistoryFiles:
                        break;

                        default:
                        {
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                            rc=0;
                        }
                    }
                }
            }break;

            default:
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                rc=0;
            }
        }

    if(rc)
        rc=DetermineAndCountInputEntries(NumberOfInFileEntries);

    if(rc)
        rc=BuildBackupToolCall();

    if(rc)
    {
        if(0!=ToolProcessOutput)
            delete ToolProcessOutput;

        if(0!=ToolProcessError)
            delete ToolProcessError;

        ToolProcessOutput=new tni34_AFile(BackintOutFile, tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now
        ToolProcessError=new tni34_AFile(BackintErrorFile, tni34_AFile::ToDiskAppend, 0, 0); // do not create on open and do not open it now

        if(0==ToolProcessOutput || 0==ToolProcessError)
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
        }
        else
        {
            if(ToolProcessOutput->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as output for Backint for SAP DB.\n", ToolProcessOutput->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;

            if(ToolProcessError->IsGood())
            {
                sprintf(MsgBuf_cn36, "Created temporary file '%s' as error output for Backint for SAP DB.\n", ToolProcessError->GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
            else
                rc=0;
        }
    }

    if(rc)
    {
        tni34_ATextFile BackintIn(BackintInFile, tni34_AFile::ToDiskTruncate, 1, 0);

        if(BackintIn.IsGood() && BackintIn.Open())
        {
            char * Line=0;

            InFileItem=ListOfFilesProcessedByBackint.First();

            while(rc && 0!=InFileItem)
            {       
                switch(GiveActionType())
                {
                    case ABackup:
                        rc=InFileItem->GiveInfo()->GenerateInputLine(Line, tcn353_Backint2InputLine::Backup);
                        break;

                    case ARestore:
                        rc=InFileItem->GiveInfo()->GenerateInputLine(Line, tcn353_Backint2InputLine::Restore);
                        break;

                    case GiveBackupInfo:
                    {
                        switch(BackintDataType)
                        {
                            case DataPipes:
                                rc=InFileItem->GiveInfo()->GenerateInputLine(Line, tcn353_Backint2InputLine::Inquire);
                            break;

                            case HistoryFiles:
                                rc=InFileItem->GiveInfo()->GenerateInputLine(Line, tcn353_Backint2InputLine::Restore);
                            break;
                            
                            default:
                                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                                rc=0;
                        }
                    }
                    break;

                    default:
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                        rc=0;
                }

                if(rc)
                {
                    sprintf(MsgBuf_cn36, "Writing '%s' to the input file.\n", Line);
                    cn36_Verbose(MsgBuf_cn36);

                    if(!BackintIn.WriteLine(Line))
                    {
                        rc=0;
                        cn36_Verbose("Could not write the line to the input file '%s'.\n");
                    
                    }
                }

                InFileItem=InFileItem->Next();
            }
                    
            if(rc)
                NumberOfRequestedNames=NumberOfInFileEntries;
            else
                NumberOfRequestedNames=0;
            
            cn36_StrDealloc(Line);  // last chance to free the memory

            if(!BackintIn.Close())
            {
                rc=0;
                sprintf(MsgBuf_cn36, "Could not close the input file of Backint for SAP DB (%s).\n", BackintIn.GiveName());
                cn36_Verbose(MsgBuf_cn36);
            }
        }
    }

    return rc;
}

int tcn353_Backint2Connector::MaxWaitingTimeOfCurrentWaitingPhase()
{
    int rc;

    if(WaitForBackupTool==CurrentWaitingPhase || WaitForKilled==CurrentWaitingPhase)
    {
        if(DBActionWasSuccessful())
            rc=BackintTimeOutSuccess;
        else
            rc=BackintTimeOutFailure;
    }
    else
        rc=tcn35_BackupToolConnector::MaxWaitingTimeOfCurrentWaitingPhase();
    
    return rc;
}

int tcn353_Backint2Connector::DetermineAndCountInputEntries(int & NumberOfEntries)
{
    int rc=1,
        i;

    switch(GiveActionType())
    {
        case ABackup:
        {
            switch(BackintDataType)
            {
                case DataPipes:
                    NumberOfEntries=Info.NumFiles; // take all pipes of the medium

                    for(i=0; i<NumberOfEntries; i++)
                        rc=AppendToBackintInputList(Info.ppszFiles[i], 1);  // the i-th pipe path, 1==it is a pipe
                    break;
                case HistoryFiles:
                    if(2!=NumberOfBackupHistories)  // assure, that 2==NumberOfbackupHistories
                    {
                        rc=0;
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    }
                    else
                    {
                        const char * PathDBMKNL;
                        const char * PathDBMEBF;

                        NumberOfEntries=NumberOfBackupHistories;

                        PathDBMKNL=GivePathOfBackupHistory();
                        PathDBMEBF=GivePathOfExternalBackupHistory();

                        if(0==PathDBMKNL || 0==PathDBMEBF)  // just get sure that the variables of the base class were filled successfully
                            rc=0;
                        else
                        {
                            rc=AppendToBackintInputList(PathDBMKNL, 0);    // the backup history file, 0==it is no pipe

                            if(rc)
                                rc=AppendToBackintInputList(PathDBMEBF, 0); // the external backup history file, 0==it is no pipe
                        }
                    }
                    break;
                default:
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    rc=0;
            }
        }break;

        case ARestore:
        {
            tcn353Backint2EBID Backint2EBID;
            const char *CurrentPositionInEBIDList=Info.ExternalSaveID;
            int ContainsABackint2EBID;

            NumberOfEntries=0;

            cn36_Verbose("Got the following list of external backup ID's:\n", AddALevelToPrefix);
            cn36_Verbose("\"");
            cn36_Verbose(Info.ExternalSaveID, NoPrefix);  //don't sprintf it to MsgBuf_cn36 as usual, can be some 1000 characters (in worst case)
            cn36_Verbose("\"\n", NoPrefix);
            cn36_Verbose(SubALevelFromPrefix);

            do
            {
                rc=Backint2EBID.TakeInfoFromStringList(ContainsABackint2EBID,
                                                       CurrentPositionInEBIDList,
                                                       ',');

                if(rc)
                {
                    if(ContainsABackint2EBID)
                    {
                        if(0==NumberOfEntries)
                            rc=cn36_StrAllocCpy(Info.SourceDBName, Backint2EBID.GiveUser());
                        else
                            if(0!=strcmp(Info.SourceDBName, Backint2EBID.GiveUser()))
                            {
                                rc=0;
                                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseEBIDsOfDifferentDBs, Info.SourceDBName));    //all EBID's must contain the same user (=dbname)
                            }

                        if(rc)
                        {
                            if(NumberOfEntries>=Info.NumFiles)
                            {
                                rc=0;
                                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseTooManyEBIDs, Info.NumFiles)); //the list contains more EBID's than the medium pipes
                            }
                            else
                            {
                                if(0==strcmp(Info.ppszFiles[NumberOfEntries], Backint2EBID.GivePath()))
                                {
                                    rc=AppendToBackintInputList(Backint2EBID.GivePath(),
                                                                1,
                                                                Backint2EBID.GiveBID());
                                }
                                else
                                {
                                    rc=AppendToBackintInputList(Backint2EBID.GivePath(),
                                                                1,
                                                                Backint2EBID.GiveBID(),
                                                                Info.ppszFiles[NumberOfEntries],
                                                                cn36_PositionOfTerminatingZero(Info.ppszFiles[NumberOfEntries]));
                                }

                                NumberOfEntries++;
                            }
                        }
                    }
                    else
                    {
                        rc=0;
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseSyntaxErrorEBIDs, NumberOfEntries+1)); //the current EBID is no EBID
                    }
                }
            }
            while(rc && '\0'!=(*CurrentPositionInEBIDList));

            if(rc && NumberOfEntries<Info.NumFiles)
            {
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError(tcn36_ADBMSrvError::CaseTooFewEBIDs, Info.NumFiles)); //the list contains fewer EBID's than the medium pipes
                rc=0;
            }

        }break;

        case GiveBackupInfo:
        {
            if(rc)
            {
                switch(BackintDataType)
                {
                    case DataPipes:
                    {
                        NumberOfEntries=0;

                        //at first take the needed information from the backup history file (dbm.knl)
                        tni34_ATextFile DBMKNL(OriginalBackupHistoryName[0], tni34_AFile::FromDisk, 0, 0);    // do not create on open, do not open now

                        ni31_List<tcn353Backint2DBMKNLLine *> ListOfBackupsProcessedByBackint;
                        const char * Line;

                        if(!DBMKNL.IsGood() || !DBMKNL.Open())
                            rc=0;
                        else
                        {
                            tcn353Backint2DBMKNLLine * InfoOfDBMKNLLine=0;
                            
                            while(rc && DBMKNL.ReadLine(Line) && 0!=Line)
                            {
                                if(0==InfoOfDBMKNLLine)
                                    InfoOfDBMKNLLine=new tcn353Backint2DBMKNLLine;

                                if(0==InfoOfDBMKNLLine)
                                {
                                    rc=0;
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                                }
                                else
                                {
                                    int LineWasGeneratedWithBackint2;

                                    rc=InfoOfDBMKNLLine->TakeBackint2InfoFromDBMKNLLine(LineWasGeneratedWithBackint2, Line, DBMKNL.GiveName());

                                    if(rc && LineWasGeneratedWithBackint2)
                                    {
                                        if(rc && 0==ListOfBackupsProcessedByBackint.Add(InfoOfDBMKNLLine))    // try to add InfoOfDBMKNLLine to the list
                                        {
                                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                                            rc=0;
                                        }
                                        else
                                            InfoOfDBMKNLLine=0; // the pointer is now stored in the list
                                    }
                                }
                            }

                            if(0!=InfoOfDBMKNLLine)
                                delete InfoOfDBMKNLLine;

                            if(!DBMKNL.Close())
                            {
                                sprintf(MsgBuf_cn36, "Could not close backup history file %s.\n", DBMKNL.GiveName());
                                cn36_Verbose(MsgBuf_cn36);
                            }
                        }

                        if(rc)
                        {
                            //now check the content of the external backup history file against ListBackupsProcessedByBackint (constructed above)
                            tni34_ATextFile DBMEBF(OriginalBackupHistoryName[1], tni34_AFile::FromDisk, 0, 0);    // do not create on open, do not open now

                            if(!DBMEBF.IsGood() || !DBMEBF.Open())
                                rc=0;
                            else
                            {
                                tcn353Backint2DBMEBFLine LineInfo;
                                int                      LineContainesInfo;

                                while(rc && DBMEBF.ReadLine(Line) && 0!=Line)   // read next line until reached end of file
                                {
                                    rc=LineInfo.TakeBackint2InfoFromDBMEBFLine(LineContainesInfo, Line, DBMEBF.GiveName());

                                    if(rc && LineContainesInfo && 0==strcmp(Info.SourceDBName, LineInfo.GiveUser()))    // inquire only backups made by the db named by Info.SourceDBName
                                    {
                                        ni31_ListElem<tcn353Backint2DBMKNLLine *> * ListElement=ListOfBackupsProcessedByBackint.First();
                                        tcn353Backint2DBMKNLLine * KNLLine;

                                        while(rc && 0!=ListElement)
                                        {
                                            KNLLine=ListElement->GiveInfo();

                                            if(LineInfo.EqualsInKeyAndLabel(KNLLine))
                                            {
                                                rc=AppendToBackintInputList(LineInfo.GivePath(),
                                                                            1,
                                                                            LineInfo.GiveBID(),
                                                                            0,
                                                                            0,
                                                                            KNLLine->GiveType(),
                                                                            KNLLine->GiveDateTime());

                                                if(rc)
                                                    NumberOfEntries++;
                                            }

                                            ListElement=ListElement->Next();
                                        }
                                    }
                                }
                            }

                            if(!DBMEBF.Close())
                            {
                                sprintf(MsgBuf_cn36, "Could not close external backup history file %s.\n", DBMEBF.GiveName());
                                cn36_Verbose(MsgBuf_cn36);
                            }
                        }

                        // free the temporary list ListOfBackupsProcessedByBackint and the memory pointed to by this list
                        while(0!=ListOfBackupsProcessedByBackint.First())
                        {
                            delete ListOfBackupsProcessedByBackint.First()->GiveInfo();
                            ListOfBackupsProcessedByBackint.Del();
                        }

                        if(!BringHistoryFilesToFormerState())
                            rc=0;
                    }break;

                    case HistoryFiles:
                    {
                        if(0!=strcmp(Info.SourceDBName, Info.pszDBName))    // we are looking for backups of another database
                        {
                            NumberOfEntries=NumberOfBackupHistories;

                            for(i=0; rc && i<NumberOfBackupHistories; i++)  // all needed histories of the other database must be restored
                            {
                                if(cn36_CheckExist(OriginalBackupHistoryName[i]))   // if the history already exists...
                                {
                                    BackupHistoryState[i]=Exist;

                                    if(RenameABackupHistoryFile(OriginalBackupHistoryName[i], i))   // ...try to rename it
                                        BackupHistoryState[i]=Renamed;
                                    else
                                        rc=0;
                                }
                                else
                                    BackupHistoryState[i]=Recovered; //at the moment it is not existing (=NotExist; would be better), but here we set the state we desire (we would like to recover the file)!

                                if(rc)  // if existed or not, restore the corresponding history of the other database
                                    rc=AppendToBackintInputList(PathOfMigrationBackintHistory[i],   // the name of the i-th backup history file
                                                                0,                                  // it is no pipe
                                                                0,                                  // no specific BID (-> #NULL is used)
                                                                OriginalBackupHistoryName[i],       // we must use the destination feature
                                                                cn36_EndOfDirectory(OriginalBackupHistoryName[i]));
                            }  
                        }
                        else    //we are looking for backups of the same database
                        {
                            NumberOfEntries=0;

                            for(i=0; rc && i<NumberOfBackupHistories; i++)
                                if(NotExist==BackupHistoryState[i]) //==!cn36_CheckExist(OriginalBackupHistoryName[i])
                                {
                                    BackupHistoryState[i]=Recovered;
                                    rc=AppendToBackintInputList(OriginalBackupHistoryName[i], 0);  // the i-th backup history file, 0 == it is no pipe
                                    NumberOfEntries++;
                                }
                        }
                    }break;

                    default:
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                        rc=0;
                }
            }
        }break;

        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
        break;
    }

    return rc;
}

int tcn353_Backint2Connector::BringHistoryFilesToFormerState()
{
    int rc=1,
        i;

    if(GiveBackupInfo==GiveActionType())
    {
        // try to bring the backup history files in the same state as found at the beginning of the inquire
        for(i=0; i<NumberOfBackupHistories; i++)
        {
            switch(BackupHistoryState[i])
            {
                case Exist:     //it has existed and was not touched, or was touched and allready brought to former state
                case NotExist:  //it has not existed and was not touched, or was restored and allready deleted again
                break;

                case Recovered: //it has not existed before and was restored -> remove it
                    if(cn36_CheckExist(OriginalBackupHistoryName[i]))
                    {
                        sprintf(MsgBuf_cn36, "Removing recovered backup history file '%s'.\n", OriginalBackupHistoryName[i]);
                        cn36_Verbose(MsgBuf_cn36);

                        if(!cn36_Remove(OriginalBackupHistoryName[i]))
                            rc=0;
                        else
                            BackupHistoryState[i]=NotExist;
                    }
                break;

                case Renamed:   //it existed, was moved to another location and a history was restored ...
                    if(cn36_CheckExist(OriginalBackupHistoryName[i]))
                    {
                        sprintf(MsgBuf_cn36, "Removing recovered backup history file '%s'.\n", OriginalBackupHistoryName[i]);
                        cn36_Verbose(MsgBuf_cn36);

                        if(!cn36_Remove(OriginalBackupHistoryName[i]))  // remove the restored history ...
                            rc=0;
                    }
                
                    if(rc)
                    {
                        sprintf(MsgBuf_cn36, "Moving renamed backup history file from '%s' back to '%s'.\n", OriginalBackupHistoryMovedTo[i], OriginalBackupHistoryName[i]);
                        cn36_Verbose(MsgBuf_cn36);

                        if(!cn36_MoveFile(OriginalBackupHistoryMovedTo[i], OriginalBackupHistoryName[i]))   // ... and move the original history back 
                        {
                            BackupHistoryState[i]=Moved;
                            rc=0;
                        }
                        else
                            BackupHistoryState[i]=Exist;
                    }
                break;

                case Moved:
                    sprintf(MsgBuf_cn36, "Moving renamed backup history file from '%s' back to '%s'.\n", OriginalBackupHistoryMovedTo[i], OriginalBackupHistoryName[i]);
                    cn36_Verbose(MsgBuf_cn36);

                    if(!cn36_MoveFile(OriginalBackupHistoryMovedTo[i], OriginalBackupHistoryName[i]))   // move the original history back 
                        rc=0;
                    else
                        BackupHistoryState[i]=Exist;

                case Unknown:
                default:
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    rc=0;
            }
        }
    }

    return rc;
}

int tcn353_Backint2Connector::RenameABackupHistoryFile(const char * NameOfBackupHistoryFile,
                                                       int          NumberOfBackupHistoryFile)
{
    int rc=0;

    char * & DestName=OriginalBackupHistoryMovedTo[NumberOfBackupHistoryFile];    //just a shortcut
    
    if(cn36_StrAllocCpy(DestName, NameOfBackupHistoryFile)) //make a copy of the name
    {
        char * LastDot=strrchr(DestName, '.');  //search the last '.' in the name

        if(0!=LastDot && 0==strcmp(LastDot, SuffixBackupHistory[NumberOfBackupHistoryFile])) //the name must end with the expected suffix
        {
            strcpy(LastDot, SuffixOfCopyOfBackupHistory[NumberOfBackupHistoryFile]);    //change the suffix
                
            if(cn36_CheckExist(DestName))   //the destination file must not exist
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileAlreadyExists.Set(DestName));
            else
                if(cn36_MoveFile(NameOfBackupHistoryFile, DestName))
                    rc=1;
        }
        else
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set()); // we should always know the correct suffix
    }

    return rc;
}

int tcn353_Backint2Connector::AppendToBackintInputList(const char * TheName,
                                                       int          IsAPipe,
                                                       const char * BID,
                                                       const char * Destination,
                                                       const char * EndOfDestination,
                                                       const char * Type,
                                                       const char * DateTime)
{
    int rc=0;

    tcn353_Backint2InputLine * FileItem=new tcn353_Backint2InputLine;

    if(0==FileItem)
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
    else
    {
        rc=FileItem->SetValues(TheName, IsAPipe, BID, Destination, EndOfDestination, Type, DateTime);
                            
        if(rc && 0==ListOfFilesProcessedByBackint.Add(FileItem))    // try to add FileItem to the list
        {
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
            rc=0;
        }
        //else nothing to do, because rc!=0 holds already
    }

    return rc;
}

void tcn353_Backint2Connector::DeleteBackintInputList()
{
    ni31_ListElem<tcn353_Backint2InputLine *> * FileItem=ListOfFilesProcessedByBackint.First();

    while(0!=FileItem)  // while(we have a list element)
    {
        delete FileItem->GiveInfo();            //delete the tcn353_Backint2InputLine (were the list element's info part is pointing to)
        ListOfFilesProcessedByBackint.Del();    //delete the list element itself
        FileItem=ListOfFilesProcessedByBackint.First(); //get a pointer to the next list element
    }
}

int tcn353_Backint2Connector::BuildBackupToolCall()
{
    int rc=1;
    
    const char * BackintAction=0;
    const char * BackintUser=0;

    if(NumberOfCommandLines!=0) //if old commandlines exists, free the memory
    {
        int i;

        for(i=0;i<NumberOfCommandLines;i++)
            if(0!=CommandLine[i])
            {
                delete [] CommandLine[i];
                CommandLine[i]=0;
            }
        
        NumberOfCommandLines=0;
    }

    switch(GiveActionType()) //determine Backint-action
    {
        case ABackup:
            BackintAction="backup";     // backup data
        break;
        case ARestore:
            BackintAction="restore";    // restore a backup
        break;
        case GiveBackupInfo:
            switch(BackintDataType)
            {
                case DataPipes:
                    BackintAction="inquire";    // inquire info about former backups of data
                break;

                case HistoryFiles:
                    BackintAction="restore";    // restore the last version of history files (#NULL <history file name>)
                break;

                default:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    rc=0;
                }
            }
        break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
    }

    if(ABackup==GiveActionType())
        BackintUser=Info.pszDBName;
    else
        BackintUser=Info.SourceDBName;

    size_t Length=strlen(BackintPath)           // ...backint(.exe)
                  +4+strlen(BackintUser)        // " -u <DBName>"
                  +4+strlen(BackintAction)      // " -f <action>"
                  +8;                           // " -t file"

    if(0!=BackintParamFile && '\0'!=BackintParamFile[0])
        Length+=4+strlen(BackintParamFile);     // " -p <paramfile>

    Length+=4+strlen(BackintInFile);            // " -i <infile>"
    Length+=3;                                  // " -c"

    rc=cn36_StrAlloc(CommandLine[0], Length);

    if(rc)
    {
        sprintf(CommandLine[0], "%s -u %s -f %s -t file", BackintPath, BackintUser, BackintAction);

        if(0!=BackintParamFile && '\0'!=BackintParamFile[0])
        {
            strcat(CommandLine[0], " -p ");
            strcat(CommandLine[0], BackintParamFile);
        }

        strcat(CommandLine[0], " -i ");
        strcat(CommandLine[0], BackintInFile);
        
        strcat(CommandLine[0], " -c");

        if(Length!=strlen(CommandLine[0]))
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
        else
        {
            NumberOfCommandLines=1;

            sprintf(MsgBuf_cn36, "Constructed Backint for SAP DB call '%s'.\n", CommandLine[0]);
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    return rc;
}

int tcn353_Backint2Connector::PostOperationsAfterSuccess()
{
    int rc;

    switch(GiveActionType())
    {
        case ABackup:
            rc=BackupOrRestoreHistoryWithBackint();  //save the Backup History and the External Backup History File after every successful backup
        break;

        case GiveBackupInfo:
            rc=CollectToolSaveIDs();    // do not check NumberOfExternalSaveIDsFound against NumberOfRequestedNames is normal, because this is not useful on inquire
        break;

        default:
            rc=1;   // do nothing in all other cases
    }

    return rc;
}

int tcn353_Backint2Connector::CleanUpTool()
{
    int rc=1;
    
    tni34_AFile BackintIn(BackintInFile, tni34_AFile::FromDisk, 0, 0); // do not open, do not create now

    if(BackintIn.Exists())
    {
        if(BackintIn.Remove())
        {
            sprintf(MsgBuf_cn36, "Removed the %s input file '%s'.\n", GiveToolName(), BackintIn.GiveName());
            cn36_Verbose(MsgBuf_cn36);
        }
        else
        {
            rc=0;

            sprintf(MsgBuf_cn36, "Could not remove the %s input file '%s'.\n", GiveToolName(), BackintIn.GiveName());
            cn36_Verbose(MsgBuf_cn36);
        }
    }

    DeleteBackintInputList();   // free the memory used by the list

    if(!BringHistoryFilesToFormerState())
        rc=0;

    return rc;
}

int tcn353_Backint2Connector::CollectToolSaveIDs()
{
    int rc=1;

    NumberOfExternalSaveIDsFound=0; //we have not found any ID so far

    tni34_ATextFile BackintOut(BackintOutFile, tni34_AFile::FromDisk, 0, 0);    // do not open, do not create now

    if(BackintOut.IsGood())
    {
        if(BackintOut.Open())
        {
            const char * Line,
                       * CurrentBID;
            const char * Name;
            size_t       CurrentBIDLength,
                         NameLength;

            int LineContainsABID;

            while(rc && BackintOut.ReadLine(Line) && 0!=Line)
            {
                rc=GetBIDAndNameFromLine(Line, LineContainsABID, CurrentBID, CurrentBIDLength, Name, NameLength);

                if(rc && LineContainsABID)
                {
                    int FoundForTheFirstTime,
                        InputLineFound=0;

                    tcn353_Backint2InputLine::BackintActionType Action;
                    
                    switch(GiveActionType())
                    {
                        case ABackup:
                            Action=tcn353_Backint2InputLine::Backup;
                        break;

                        case ARestore:
                            Action=tcn353_Backint2InputLine::Restore;
                        break;

                        case GiveBackupInfo:
                            Action=tcn353_Backint2InputLine::Inquire;
                        break;

                        default:
                            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                            rc=0;
                    }

                    InFileItem=ListOfFilesProcessedByBackint.First();

                    while(rc && !InputLineFound && 0!=InFileItem)
                    {
                        if(rc)
                            rc=InFileItem->GiveInfo()->TestAgianstOutputLine(Action,
                                                                             InputLineFound,
                                                                             FoundForTheFirstTime,
                                                                             Name,
                                                                             NameLength,
                                                                             CurrentBID,
                                                                             CurrentBIDLength);

                        if(rc && FoundForTheFirstTime)
                        {
                            if(ABackup==GiveActionType())   // if we make a backup
                            {
                                rc=BuildAEBIDFrom(ExternalSaveID[NumberOfExternalSaveIDsFound], // copy the found BID. It will be written to the External Backup History
                                                  Info.pszDBName,
                                                  InFileItem->GiveInfo()->GiveBID(),
                                                  InFileItem->GiveInfo()->GiveName());

                                if(rc)
                                {

                                    sprintf(MsgBuf_cn36, "Found External Backup ID '%s'.\n", ExternalSaveID[NumberOfExternalSaveIDsFound]);
                                    cn36_Verbose(MsgBuf_cn36);
                                }
                            }
                    
                            NumberOfExternalSaveIDsFound++; // count the Number of returned BackupIDs regardless of the prformed action
                        }

                        InFileItem=InFileItem->Next();
                    }

                    if(rc && !InputLineFound)   // no appropriate line was found
					{
                        rc=0;
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::WrongFilesProcessed.Set(Name));
					}
                }
            }

            if(GiveBackupInfo!=GiveActionType())   // if we have made no inquire with backint
            {
                if(rc && NumberOfExternalSaveIDsFound<NumberOfRequestedNames)   // not all expected names were found in the output (NumberOfExternalSaveIDsFound<=NumberOfRequestedNames!)
                {                                                               // do not set rc=0, caller decides if this is an error
                    if(0<NumberOfExternalSaveIDsFound)
                        cn36_Verbose("Not all files were processed correctly by Backint for SAP DB.\n");
                    else
                        cn36_Verbose("No file was processed correctly by Backint for SAP DB.\n");    
                }
            }
            //else on inquire it is normal, that not all requested items are also mentioned in the output
        }
        else
            rc=0;
    }
    else
        rc=0;

    return rc;
}

int tcn353_Backint2Connector::BuildAEBIDFrom(char * &     EBID,
                                             const char * User,
                                             const char * BID,
                                             const char * Path)
{
    int rc;

    rc=cn36_StrAlloc(EBID, strlen(User)+1+strlen(BID)+1+strlen(Path));

    if(rc)
        sprintf(EBID, "%s %s %s", User, BID, Path);

    return rc;  
}

int tcn353_Backint2Connector::GetBIDAndNameFromLine(const char *   Line,
                                                    int          & ContainsABID,
                                                    const char * & BackupID,
                                                    size_t       & LengthOfBackupID,
                                                    const char * & Name,
                                                    size_t       & LengthOfName)
{
    int rc=1;
    const char * KeyWord;
    
    ContainsABID=0;

    switch(GiveActionType()) //determine Backint-action
    {
        case ABackup:
            KeyWord="#SAVED";
            break;
        case ARestore:
            KeyWord="#RESTORED";
            break;
        case GiveBackupInfo:
            switch(BackintDataType)
            {
                case DataPipes:
                    KeyWord="#BACKUP";  // inquired info about former backups of data
                break;

                case HistoryFiles:
                    KeyWord="#RESTORED";    // restored the last version of history files (#NULL <history file name>)
                break;

                default:
                {
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                    rc=0;
                }
            }
            break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
    }

    if(rc)
    {
        if(strlen(Line)>strlen(KeyWord) &&              // Is Line longer then the keyword?
           !strncmp(Line, KeyWord, strlen(KeyWord)) &&  // Starts Line with the keyword?
           isspace(Line[strlen(KeyWord)]))              // is a white space following the keyword?
        {
            const char * EndBackupID,
                       * EndName;

            BackupID   =cn36_FirstNonWhiteSpaceOf(Line+strlen(KeyWord));
            EndBackupID=cn36_FirstWhiteSpaceOf(BackupID);
            
            Name=cn36_FirstNonWhiteSpaceOf(EndBackupID);
            EndName=cn36_FirstWhiteSpaceOf(Name);

            rc=0;

            if('\0'!=Name[0]) // if this is true, '\0'!=BackupID[0] is also true
            {
                LengthOfBackupID=EndBackupID-BackupID;
                LengthOfName=EndName-Name;

                if(MAX_LENGTH_BID>LengthOfBackupID && 0<LengthOfBackupID && 
                   MAX_LENGTH_BSIPATH>LengthOfName && 0<LengthOfName)
                {
                    rc=1;
                    ContainsABID=1;
                }
            }
            
            if(!rc)
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::UnknownBackintOutputFormat.Set(Line));
        }
    }

    return rc;
}

int tcn353_Backint2Connector::PrepareToolOutputOfAvailableBackups()
{
    ListOfFilesProcessedByBackint.Reverse();            //Reverse the list to get the Backups in the right order (older backups first, newest backup at last)
    InFileItem=ListOfFilesProcessedByBackint.First();

    return 1;
}

int tcn353_Backint2Connector::GiveNextAvailableBackup(tcn003_BackupHistoryEBIDStatus & Status,
                                                      const char * & EBID,
                                                      size_t       & EBIDLength,
                                                      const char * & TypeString,
                                                      const char * & DateTimeString,
                                                      int          & ReachedEnd)
{
    int rc=1;

    if(0==InFileItem)
        ReachedEnd=1;
    else
    {
        tcn353_Backint2InputLine * TheLine=InFileItem->GiveInfo();
        
        ReachedEnd=0;

        if(TheLine->WasFound())  // did Backint for SAP DB return the BID and Name of this item in its inquire output
            Status=EBIDStatusAvailable;
        else
            Status=EBIDStatusUnavailable;


        rc=BuildAEBIDFrom(TempBID, Info.SourceDBName, TheLine->GiveBID(), TheLine->GiveName());

        if(rc)
        {
            EBID=TempBID;
            EBIDLength=strlen(EBID);

            TypeString=TheLine->GiveSaveType();
            DateTimeString=TheLine->GiveDateTime();
        }
        
        InFileItem=InFileItem->Next();
    }

    return rc;
}

int tcn353_Backint2Connector::CloseToolOutputOfAvailableBackups()
{
    cn36_StrDealloc(TempBID);
    return 1;
}

int tcn353_Backint2Connector::GiveInformationOfLine(const char *   Line,
                                                    const char * & EBIDInLine,
                                                    size_t       & EBIDLength,
                                                    const char * & TypeString,
                                                    const char * & DateTimeString,
                                                    int          & LineContainsEBID)
{
    int rc=0;

    return rc;
}

int tcn353_Backint2Connector::ToolWasSuccessful()
{
    int rc=1;

    if(ExistsToolProcesses())
        rc=0;
    else
    {
        int i,
            ToolReturncode;

        for(i=0; i<MAX_NUMBER_OF_TOOLPROCESSES_CN35 && rc!=0; i++)
        {
            if(0!=ToolProcess[i])
            {
                ToolReturncode=ToolProcess[i]->GiveReturnCode();

                if(0!=ToolReturncode && 1!=ToolReturncode)
                    rc=0;
            }
        }
    }
    
    return rc;
}

int tcn353_Backint2Connector::BackupOrRestoreHistoryWithBackint()
{
    int rc=1;

    switch(GiveActionType())
    {
        case ABackup:
            cn36_Verbose("\nMaking a backup of the Backup History files.\n", AddALevelToPrefix);
            break;
        case GiveBackupInfo:
            cn36_Verbose("\nRestoring the needed Backup History files.\n", AddALevelToPrefix);
            break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
    }

    if(rc)
    {
        if(ToolIsRunning())
        {
            rc=0;
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set()); // TODO : extra error for this case?
        }
        else
        {
            NumberOfRunningToolProcesses=0;

            cn36_Verbose("\n");

            if(ABackup==GiveActionType() && !MoveToolOutputToProtocolFile())    // move stdout and stderr of the backup operation to the protocol (dbm.ebp)
                rc=0;
                
            if(rc)
            {
                BackintDataTypeEnum OldBackintDataType=BackintDataType; // save the current data type
                BackintDataType=HistoryFiles;                           // set the type temporarily to tcn353_Backint2Connector::HistoryFiles

                cn36_Verbose("\n");

                switch(GiveActionType())
                {
                    case ABackup:
                        cn36_Verbose("Preparing backup.\n", AddALevelToPrefix);
                        rc=Prepare();
                        cn36_Verbose(rc?"Prepare passed successfully.\n\n":"Prepare failed.\n\n", SubALevelFromPrefix);
                        break;
                    case GiveBackupInfo:
                        cn36_Verbose("Preparing restore.\n", AddALevelToPrefix);
                        rc=Prepare();
                        cn36_Verbose(rc? "Prepare passed successfully.\n\n": "Prepare failed.\n\n", SubALevelFromPrefix);
                        break;
                    default:
                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                        rc=0;
                }

                if(rc)
                {
                    if(StartTool())
                    {
                        StatusOfTool=Started;
                    
                        if(CheckTool())
                        {
                            int CheckTime=5;

                            cn36_Verbose("Waiting for the end of Backint for SAP DB.\n", AddALevelToPrefix);

                            while(ToolIsRunning())                   // check backup tool periodically
                                cn36_Sleep(CheckTime++);

                            NumberOfRunningToolProcesses=0; //tool is not longer running, set NumberOfRunningToolProcesses=0 to allow further calls of the tool

                            cn36_Verbose("Ended the waiting.\n", SubALevelFromPrefix);

                            if(!ToolWasSuccessful())
                            {
                                rc=0;
                                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolFailed.SetI(SumOfToolExitCodes()));
                            }
                            else
                            {
                                cn36_Verbose("\n");
                                cn36_Verbose("Checking output of Backint for SAP DB.\n", AddALevelToPrefix);

                                rc=CollectToolSaveIDs();

                                if(rc && NumberOfExternalSaveIDsFound!=NumberOfRequestedNames)
                                {
                                    rc=0;
                                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::BackupToolFailed.Set());
                                }

                                cn36_Verbose(rc?"Have found all BID's as expected.\n\n":"Could not found all BID's as expected.\n\n", SubALevelFromPrefix);
                            }

                            if(!rc)
                                switch(GiveActionType())
                                {
                                    case ABackup:
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotSaveBackupHistoryFiles.Set());
                                        break;
                                    case GiveBackupInfo:
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CouldNotRestoreBackupHistoryFiles.Set());
                                        break;
                                    default:
                                        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
                                        rc=0;
                                }
                        }
                        else
                            rc=0;
                    }
                    else
                        rc=0;

                    if(GiveBackupInfo==GiveActionType())
                    {
                        DeleteBackintInputList();   //do not confuse the inquire with the restore of backup history files

                        if(!MoveToolOutputToProtocolFile()) // move stdout and stderr of the restore operation to the protocol (dbm.ebp)
                            rc=0;
                    }
                }

                BackintDataType=OldBackintDataType;                     // set the type back to the former value
            }
        }
    }

    switch(GiveActionType())
    {
        case ABackup:
            cn36_Verbose(rc?"Have saved the Backup History files successfully.\n\n":
                            "Could not save the Backup History files correctly.\n\n", SubALevelFromPrefix);
            break;
        case GiveBackupInfo:
            cn36_Verbose(rc?"Have restored the needed Backup History files successfully.\n\n":
                            "Could not restore the needed Backup History files correctly.\n\n", SubALevelFromPrefix);
            break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
            rc=0;
    }

    return rc;
}
