/*******************************************************************************
  module:       vcn36.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing basic functions used in vcn35.cpp, ..., vcn355.cpp
  last change:  2000-11-29 15:03
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#include "hni31.h"
#include "hcn36.h"
#include "hcn51.h"

#include<sys/types.h>
#include<sys/stat.h>
#if defined (WIN32)
    #include <io.h>
    #define         F_OK                    0x0000
    #define         W_OK                    0x0010
    #define         R_OK                    0x0100
#elif defined(UNIX)
    #include<unistd.h>    /* F_OK, R_OK, X_OK */
    #ifndef _S_IFDIR
        #define _S_IFDIR S_IFDIR       /* DEC Alpha */
    #endif
    #include<fcntl.h>
    #include<sys/stat.h>
    #include<sys/types.h>
    #include<sys/utsname.h>
    #include<ctype.h>
#endif
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<errno.h>
#include<time.h>
#include<limits.h>

int cn36_StrAlloc(char * & Dest, size_t Length)
{
    int rc=1;

    if(Dest!=0)
        delete [] Dest;

    Dest=new char [Length+1];

    if(0==Dest)
    {
        rc=0;
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
    }
    else
        Dest[0]='\0';

    return rc;
}

int cn36_StrAllocCpy(char * & Dest, const char *Src)
{
    int rc=cn36_StrAlloc(Dest, strlen(Src));

    if(rc)
        strcpy(Dest, Src);

    return rc;
}

int cn36_StrNAllocCpy(char * & Dest, const char *Src, size_t Length)
{
    int rc=cn36_StrAlloc(Dest, Length);

    if(rc)
        cn36_StrNCpy(Dest, Src, Length);

    return rc;
}

void cn36_StrDealloc(char * & String)
{
    if(0!=String)
    {
        delete [] String;
        String=0;
    }
}

static const char * cn36_StartOfTrailingWhiteSpaces(const char *s)
{
    const char *s1=s+strlen(s);

    while(s1!=s && isspace(*(s1-1)))
        s1--;

    return s1;
}

static const char * cn36_GetEnv(const char * cpszEnvironmentVariable)
{
    return getenv(cpszEnvironmentVariable);
}

static ni31_List<char *> ListOfUsedEnvironmentStrings_cn36;

static int cn36_PutEnv(const char * VariableName,
                       const char * VariableValue)
{
    int    rc=1;
    char * NewEnvString=0;

    if(!cn36_StrAlloc(NewEnvString, strlen(VariableName)+1+strlen(VariableValue)))  // allocate space for new part of environment
        rc=0;
    else
    {
        sprintf(NewEnvString, "%s=%s", VariableName, VariableValue);

        #if defined(WIN32)
        rc=!_putenv(NewEnvString);
        #elif defined(UNIX)
        rc=!putenv(NewEnvString);
        #else
            #error cn36_PutEnv() only coded for WIN32 and UNIX, define one of them
        #endif

        if(!rc) // could not make NewEnvString a part of the environment
        {
            cn36_StrDealloc(NewEnvString);  // free the memory 
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotExportEnvironment.Set(NewEnvString));
        }
        else    // the (*NewEnvString) is now part of the environment, (*NewEnvString) must not be freed as long as it remains in is the environment!!
        {
            ni31_ListElem<char *> *h=ListOfUsedEnvironmentStrings_cn36.First();
            int FoundInList=0;

            while(0!=h && !FoundInList) //look through the list of former parts of the environment (put by this function)
            {
                if(0==strncmp(h->GiveInfo(), VariableName, strlen(VariableName)) && //the environment allready contained a string for that variable
                   '='==h->GiveInfo()[strlen(VariableName)])
                {
                    char * OldEnvironmentString=h->GiveInfo();  // get the old string
                    cn36_StrDealloc(OldEnvironmentString);      //free the memory of the old string (as it is not longer used by the environment)

                    (*h)=NewEnvString; //place the new string in the list
                    
                    FoundInList=1;
                }

                h=h->Next();
            }

            if(!FoundInList)                                                // if the variable was puted for the first time to the environment
                if(0==ListOfUsedEnvironmentStrings_cn36.Add(NewEnvString))  //add the string to the list
                {
                    rc=0;
                    TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::OutOfMemory.Set());
                }
        }
    }

    return rc;
}

int cn36_PutEnvironmentVariable(const char * VariableName,
                                const char * VariableValue)
{
    int rc;

    sprintf(MsgBuf_cn36, "Setting environment variable '%s' to value '%s'.\n", VariableName, VariableValue);
    cn36_Verbose(MsgBuf_cn36);
    
    rc=cn36_PutEnv(VariableName, VariableValue);

    return rc;
}

int cn36_GetFileEnv(const char * FileEnvironmentVariable, const char * DBName, char * & Value)
{
    int rc=1;

    tsp00_DbNamec szDBName;
    tsp00_Namec VariableName;

    szDBName.rawAssign(DBName);
    VariableName.rawAssign(FileEnvironmentVariable);

    cn51DBMConfigValue CFGValue(szDBName, VariableName);

    if(!cn36_StrAllocCpy(Value, (const char *) tsp00_C512c(CFGValue)))
        rc=0;

    return rc;
}


int cn36_DetermineValueOfEnvironmentVariable(const char * const VariableName,
                                                   char * &     VariableValue,
                                             const char *       VariableComment,
                                             const char *       DBName,
                                                   int          ExportDefaultIfNotFound)
{
    int rc=1;

    char * FileEnvVar=0;

    if(0==VariableComment)
        VariableComment="unspecified variable";

    if(!cn36_GetFileEnv(VariableName, DBName, FileEnvVar))
        rc=0;
    else
    {
        const char * pszEnvVar=cn36_GetEnv(VariableName);

        if(0!=FileEnvVar && FileEnvVar[0]!='\0')    //if found a non empty value for the variable in dbm.cfg
        {
            rc=cn36_StrAllocCpy(VariableValue, FileEnvVar); //use the value

            if(rc)
            {
                sprintf(MsgBuf_cn36,"Using configuration variable '%s' = '%s'%s%s.\n", VariableName,
                                                                                       VariableValue,
                                                                                       0==VariableComment?"":" as ",
                                                                                       0==VariableComment?"":VariableComment);
                cn36_Verbose(MsgBuf_cn36);

                if(0==pszEnvVar || (0!=pszEnvVar && strcmp(pszEnvVar, FileEnvVar)))   // if environment variable has no or other value
                {                                                                     // export new value to the environment
                    sprintf(MsgBuf_cn36,
                            "Setting environment variable '%s'%s%s %s%s%sto configuration value '%s'.\n",
                            VariableName,
                            0==VariableComment?"":" for the ",
                            0==VariableComment?"":VariableComment,
                            0==pszEnvVar?"":"from '",
                            0==pszEnvVar?"":pszEnvVar,
                            0==pszEnvVar?"":"' ",
                            VariableValue);

                    cn36_Verbose(MsgBuf_cn36);

                    rc=cn36_PutEnv(VariableName, VariableValue);   
                }
            }
        }
        else
        {
            if(0==pszEnvVar)
            {
                if(0==VariableValue)
                {
                    VariableValue=new char[1]; // be sure, that it can be deleted by delete []
                    VariableValue[0]='\0';
                }

                if(ExportDefaultIfNotFound)
                {
                    sprintf(MsgBuf_cn36, "Setting environment variable '%s'%s%s to default '%s'.\n", VariableName,
                                                                                                     0==VariableComment?"":" for the ",
                                                                                                     0==VariableComment?"":VariableComment,
                                                                                                     VariableValue);
                    cn36_Verbose(MsgBuf_cn36);

                    rc=cn36_PutEnv(VariableName, VariableValue);
                }
            }
            else
            {
                rc=cn36_StrAllocCpy(VariableValue, pszEnvVar);

                if(rc)
                {
                    #if defined (WIN32)
                    sprintf(MsgBuf_cn36,"Using environment variable '%%%s%%' = '%s'%s%s.\n", VariableName,
                                                                                             VariableValue,
                                                                                             0==VariableComment?"":" as ",
                                                                                             0==VariableComment?"":VariableComment);
                    #elif defined (UNIX)
                    sprintf(MsgBuf_cn36,"Using environment variable '$%s' = '%s'%s%s.\n", VariableName,
                                                                                          VariableValue,
                                                                                          0==VariableComment?"":" as ",
                                                                                          0==VariableComment?"":VariableComment);
                    #else
                    #error cn351_DetermineValueOfEnvironmentVariable() was only coded for UNIX and NT (define one of them).
                    #endif

                    cn36_Verbose(MsgBuf_cn36);
                }
            }
        }
    }

    cn36_StrDealloc(FileEnvVar);

    return rc;
}

extern const char PathDelimiter_cn36;

int cn36_ConstructCompletePath(char * & CompletePath, const char * Directory, const char * FileName)
{
    int rc=0;

    size_t LengthOfDirectory=strlen(Directory);

    if(cn36_StrAlloc(CompletePath, LengthOfDirectory+1+strlen(FileName)))   // resevere space for the directory, \ or / (only inserted if needed) and the file name
    {
        strcpy(CompletePath, Directory);    // path starts with directory

        if(0==LengthOfDirectory ||                                  // if the directory is empty or
           (0<LengthOfDirectory &&                                  // (if at least one char in Directory
            PathDelimiter_cn36!=CompletePath[LengthOfDirectory-1])) // and last char is not \  or / respectivly)
        {
            CompletePath[LengthOfDirectory]=PathDelimiter_cn36; // append a \ or / respectivly
            ++LengthOfDirectory;
        }

        strcpy(CompletePath+LengthOfDirectory, FileName);

        rc=1;
    }

    return rc;
}


const int MaximalCharsForAnInt_cn36=20;

static const int MaxPrefixChars_cn36=256+1;
static const int CharsPerPrefixLevel_cn36=4;
static const int MaxPrefixLevels=(MaxPrefixChars_cn36-1)/CharsPerPrefixLevel_cn36;
static char LevelPrefix[MaxPrefixChars_cn36];
static int  CurrentPrefixLevel=0;

void cn36_Verbose(const char * pszString, tcn36_VerboseTyp PrefixModifier)
{
    if(SubALevelFromPrefix==PrefixModifier && CurrentPrefixLevel>0)
    {
        LevelPrefix[(CurrentPrefixLevel-1)*CharsPerPrefixLevel_cn36]='\0';
        CurrentPrefixLevel--;
    }

    if(PathOfProtFile_cn36[0]!='\0')
    {
        static int ClearProtFile=1;
        FILE * stream;
    
        if(ClearProtFile)
        {
            stream=fopen(PathOfProtFile_cn36, "w");
            ClearProtFile=0;

            memset(LevelPrefix, '\0', MaxPrefixChars_cn36);
        }
        else
            stream=fopen(PathOfProtFile_cn36, "a");

        if(0!=stream)
        {
            const char * NextNewLine;
            const char * CurrentPos=pszString;
            char FormatString[30];              // sould be enough for "%s%.<length>s", with length a number of type int
            const char * LineDelimiter="\n";
            const size_t LengthOfLineDelimiter=strlen(LineDelimiter);
            
            NextNewLine=strstr(CurrentPos, LineDelimiter);

            while(0!=NextNewLine)
            {
                sprintf(FormatString, "%%s%%.%ds", int(NextNewLine+LengthOfLineDelimiter-CurrentPos));                  // we do a sprintfn
                fprintf(stream, FormatString, NoPrefix==PrefixModifier?"":LevelPrefix, CurrentPos);

                CurrentPos=NextNewLine+LengthOfLineDelimiter;
                NextNewLine=strstr(CurrentPos, LineDelimiter);
            }

            if('\0'!=CurrentPos[0])
                fprintf(stream, "%s%s", NoPrefix==PrefixModifier?"":LevelPrefix, CurrentPos);

            fflush(stream);
            fclose(stream);
        }
    }

    if(AddALevelToPrefix==PrefixModifier && CurrentPrefixLevel<MaxPrefixLevels)
    {
        memset(LevelPrefix+CurrentPrefixLevel*CharsPerPrefixLevel_cn36, ' ', CharsPerPrefixLevel_cn36);
        CurrentPrefixLevel++;
    }
}

void cn36_Verbose(tcn36_VerboseTyp PrefixModifier)
{
    if(SubALevelFromPrefix==PrefixModifier && CurrentPrefixLevel>0)
    {
        LevelPrefix[(CurrentPrefixLevel-1)*CharsPerPrefixLevel_cn36]='\0';
        CurrentPrefixLevel--;
    }
    
    if(AddALevelToPrefix==PrefixModifier && CurrentPrefixLevel<MaxPrefixLevels)
    {
        memset(LevelPrefix+CurrentPrefixLevel*CharsPerPrefixLevel_cn36, ' ', CharsPerPrefixLevel_cn36);
        CurrentPrefixLevel++;
    }
}

char MsgBuf_cn36[1024];
char PathOfProtFile_cn36[MaxLengthOfProtFileName_cn36+1]={'\0',};

int cn36_AppendFileToFile(const char *Src, const char *Dest)
{
    int rc=1;

    tni34_AFile SrcFile(Src, tni34_AFile::FromDisk, 0, 1);

    if(SrcFile.IsGood())
    {   
        tni34_AFile DestFile(Dest, tni34_AFile::ToDiskAppend, 1, 1);

        if(DestFile.IsGood())
        {
            long BytesRead, BytesWritten;
            char Buf[1024];

            while(rc && SrcFile.Read((void*)Buf, 1024, BytesRead) && BytesRead>0)
            {
                if(!DestFile.Write((void*)Buf, BytesRead, BytesWritten) || BytesWritten!=BytesRead) //TODO: use "quiet"-versions of Write and Read
                    rc=0;// report this error after the LogFile was closed, because it will be reported to the LogFile !
            }
        }
        else
            rc=0;

        if(!DestFile.Close())
            rc=0;
    }
    else
        rc=0;

    return rc;
}

int cn36_MoveFile(const char * Src, const char *Dest)
{
    int rc=1;

    tni34_AFile File(Src, tni34_AFile::FromDisk, 0, 0);  // do not create and do not open

    if(File.IsGood())
        rc=File.Rename(Dest);
    else
        rc=0;

    return rc;
}

int cn36_Remove(const char *Name)
{
    int rc=0;

    tni34_AFile File(Name, tni34_AFile::ToDiskTruncate, 0, 0); // do not open, do not create now

    if(File.Exists())
    {
        if(File.Remove())
            rc=1;
    }
    else
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::AFileIsMissingOrUnreadable.Set(Name));

    if(0==rc)
    {
        sprintf(MsgBuf_cn36, "Could not remove the file '%s'.\n", Name);
        cn36_Verbose(MsgBuf_cn36);
    }

    return rc;
}

const char * cn36_GetHostName()
{
    const char * rc;

    #if defined (WIN32)
    DWORD MaxNameLngth=MAX_COMPUTERNAME_LENGTH+1; 
    static char ComputerName[MAX_COMPUTERNAME_LENGTH+1];

    if(GetComputerName(ComputerName, &MaxNameLngth))
        rc=ComputerName;
    else
    {
        rc=0;
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotGetComputerName.SetI((int)GetLastError()));
    }
    #elif defined (UNIX)
    static struct utsname UnameBuf;
    if(uname(&UnameBuf)>=0)
        rc=UnameBuf.nodename;
    else
    {
        rc=0;
        TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::CanNotGetComputerName.SetI(errno));
    }
    #else
        #error  cn36_GetHostName() only coded for WIN32 and UNIX, define one of them
    #endif

    return rc;
}

int cn36_CheckExist(const char * FileName)
{
    int Exists=1;

#if defined (WIN32)
    if(-1==_access(FileName, F_OK))
        Exists=0;
#elif defined (UNIX)
    if(-1==access(FileName, F_OK))
        Exists=0;
#else
#error vcn36.cpp only coded for WIN32 and UNIX, define one of them
#endif

    return Exists;
}

int cn36_CheckRead(const char * pszFileName)
{
    int IsReadable=1;

#if defined (WIN32)
    if( -1==_access(pszFileName, R_OK))
      IsReadable=0;
#elif defined (UNIX)
    if( -1==access(pszFileName, R_OK))
      IsReadable=0;
#else
#error vcn36.cpp only coded for WIN32 and UNIX, define one of them
#endif

    return IsReadable;
}

int cn36_CheckDir(const char * pszDirectoryName)
{
    int IsDirectory=1;

#if defined (WIN32)
    struct _stat s_stat;

    if( -1==_stat(pszDirectoryName, &s_stat) || 0==(s_stat.st_mode & _S_IFDIR))
        IsDirectory=0;
#elif defined (UNIX)
    struct stat s_stat;

    if( -1==stat(pszDirectoryName, &s_stat) || 0==(s_stat.st_mode & _S_IFDIR))
        IsDirectory=0;
#else
#error vcn36.cpp only coded for WIN32 and UNIX, define one of them
#endif

    return IsDirectory;
}

int cn36_CheckExe(const char * pszProgramName)
{
    int IsExecutable=1;

#if defined (WIN32)
    struct _stat s_stat;

    if( -1==_stat(pszProgramName, &s_stat) || 0==(s_stat.st_mode & _S_IEXEC)) 
        IsExecutable=0;
#elif defined (UNIX)
    if( -1==access( pszProgramName, X_OK))
        IsExecutable=0;
#else
#error vcn36.cpp only coded for WIN32 and UNIX, define one of them
#endif

    return IsExecutable;
}

void cn36_Sleep(int Seconds)
{
    sprintf(MsgBuf_cn36,"Waiting %d second%s ... ", (int)Seconds, Seconds>1?"s":"");
    cn36_Verbose(MsgBuf_cn36);

    cn361_Sleep(Seconds);

    cn36_Verbose("Done.\n", NoPrefix);
}

long cn36_GetCurrentTime()
{
    return (long)time(0);
}

long cn36_SecondsBetween(long time1, long time2)
{
    long rc=time2-time1;

    if(time1>time2)
        rc= -rc;

    return rc;
}

tcn36_DBMSrvErrorHandler TheDBMSrvErrorHandler_cn36;

tcn36_DBMSrvErrorHandler::tcn36_DBMSrvErrorHandler()
    :ReplyData(0), ReplyLen(0), LastError(OK_CN00), ReplyAlreadyConstructed(0), CurrentErrorPhase(ERR_CN00)
{
    SetNewErrorHandler(this);
}

void tcn36_DBMSrvErrorHandler::SetReplyArrayTo(char * pcReplyData, int * piReplyLen)
{
    ReplyData=pcReplyData;
    ReplyLen=piReplyLen;
    LastError=OK_CN00;
    ReplyAlreadyConstructed=0;
    CurrentErrorPhase=ERR_CN00;
}

static tcn00_Error cn36AnswerCN36Error( _TCHAR * szBuffer, int * pnDataLen, const tcn00_Error nError, _TCHAR * szInfo)
{
    int i=0;
    int j=0;

    struct
    {
        tcn00_Error   nValue;
        const _TCHAR* szText;
    }errorArray[] = ERROR_TO_TEXT_CN00;

    while((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue != nError))
        i++;

    _stprintf(szBuffer,
              _T("%s%s%d%s%s%s%s%s"),
              (OK_CN00==nError) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00, 
              LINE_SEPSTRING_CN00,
              nError, 
              TOKEN_SEPSTRING_CN00, 
              errorArray[i].szText, 
              LINE_SEPSTRING_CN00,
              0==szInfo? "" : szInfo,
              (0==szInfo || '\0'==szInfo[0])? "" : LINE_SEPSTRING_CN00);    // only print a second line separating sting if we realy have a second line

    *pnDataLen=int(_tcslen(szBuffer));

    cn36_Verbose("Constructed the following reply:\n", AddALevelToPrefix);

    sprintf(MsgBuf_cn36, "%s%s", ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00), LINE_SEPSTRING_CN00);
    cn36_Verbose(MsgBuf_cn36);
    sprintf(MsgBuf_cn36, "%d%s%s%s", nError, TOKEN_SEPSTRING_CN00, errorArray[i].szText, LINE_SEPSTRING_CN00);
    cn36_Verbose(MsgBuf_cn36);
    sprintf(MsgBuf_cn36, "%s\n", (szInfo == NULL ? "" : szInfo));
    cn36_Verbose(MsgBuf_cn36);
    cn36_Verbose(SubALevelFromPrefix);

    return nError;
}

void tcn36_DBMSrvErrorHandler::Report(const tni35_AError & Err)
{
    sprintf(MsgBuf_cn36, "Have encountered error %d:\n", int(CurrentErrorPhase));
    cn36_Verbose(MsgBuf_cn36, AddALevelToPrefix);

    Err.FillErrorMsgBuffer();

    cn36_Verbose(Err.GiveErrorMsgBuffer());
    cn36_Verbose("\n", NoPrefix);
    cn36_Verbose(SubALevelFromPrefix);
    
    if(0!=ReplyData && 0!=ReplyLen && !ReplyAlreadyConstructed)
    {
        cn36AnswerCN36Error(ReplyData, ReplyLen, CurrentErrorPhase, (char *) Err.GiveErrorMsgBuffer());
        ReplyAlreadyConstructed=1;
    }

    LastError=CurrentErrorPhase;
}

void tcn36_DBMSrvErrorHandler::SetPhaseTo(tcn00_Error Phase)
{
    CurrentErrorPhase=Phase;
}

tcn00_Error tcn36_DBMSrvErrorHandler::GiveLastError()
{
    return LastError;
}

void tcn36_ADBMSrvError::SetThisError(DBMSrvErrorCase CaseToSet)
{
    if(DBMSrvErrorToText[int(CaseToSet)].Case==CaseToSet)
        ThisError=DBMSrvErrorToText+int(CaseToSet);
    else
    {
#if defined (DEBUG_CN)
        cn36_Verbose("Warning: unexpected behaviour in tcn36_ADBMSrvError::SetThisError()");
#endif
        ThisError=DBMSrvErrorToText;
        while(ThisError->Case!=CaseToSet && ThisError->Case!=CaseInternalError)
            ThisError++;
    }
}

tcn36_ADBMSrvError::tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, DBMSrvErrorSetFlag SetFlag)
:tni35_AError(0) // do not set basis error, to avoid doing the same twice
{
    SetThisError(ErrCase);

    if(SetErrorNow==SetFlag)
        Set();                  // will set basis error to!    
}

tcn36_ADBMSrvError::tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, const char *ErrInfo)
:tni35_AError(ErrInfo, 0)
{
    SetThisError(ErrCase);

    Set(ErrInfo);
}

tcn36_ADBMSrvError::tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, int ErrInfo)
:tni35_AError(0)
{
    SetThisError(ErrCase);

    SQLCode=0;
    IntCode=ErrInfo;

    Set("");
}

tcn36_ADBMSrvError & tcn36_ADBMSrvError::SetI(int ErrInfo)
{
    SQLCode=0;
    IntCode=ErrInfo;

    Set("");

    return *this;
}

tcn36_ADBMSrvError & tcn36_ADBMSrvError::SetSQL(int TheSQLCode, const char * TheSQLErrText)
{
    SQLCode=TheSQLCode;
    IntCode=0;
    
    if(0!=TheSQLErrText)
    {
        const char *EndOfText=cn36_StartOfTrailingWhiteSpaces(TheSQLErrText);

        SetStrN(TheSQLErrText, EndOfText-TheSQLErrText);
    }
    else
        Set("");

    return *this;
}

tcn36_ADBMSrvError & tcn36_ADBMSrvError::SetSQLI(int TheSQLCode, int ErrInfo)
{
    SQLCode=TheSQLCode;
    IntCode=ErrInfo;

    Set("");

    return *this;
}

void tcn36_ADBMSrvError::FillErrorMsgBuffer() const
{        
    size_t LengthOfExtendedErrorMsg;

    switch(ThisError->InfoType)
    {
        case tcn36_ADBMSrvError::None:
            LengthOfExtendedErrorMsg=0;
        break;
        case AString:
            LengthOfExtendedErrorMsg=strlen(ThisError->ExtendedErrorText)-2+strlen(GiveErrorInfoBuffer());
        break;
        case AIntCode:
            LengthOfExtendedErrorMsg=strlen(ThisError->ExtendedErrorText)-2+MaximalCharsForAnInt_cn36;
        break;
        case AIntAndASQLCode:
        case ASQLCodeAndAInt:
            LengthOfExtendedErrorMsg=strlen(ThisError->ExtendedErrorText)-4+MaximalCharsForAnInt_cn36+MaximalCharsForAnInt_cn36;
        break;
        case ASQLCodeAndAString:
            LengthOfExtendedErrorMsg=strlen(ThisError->ExtendedErrorText)-4+MaximalCharsForAnInt_cn36+strlen(GiveErrorInfoBuffer());
        break;
        default:
            TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
    }

    if(ErrorMsgBufferLength_ni35-1 >= LengthOfExtendedErrorMsg)
    {
        switch(ThisError->InfoType)
        {
            case None:
                cn36_StrNCpy(ErrorMsgBuffer, ThisError->PlainErrorText, ErrorMsgBufferLength_ni35-1);
            break;
            case AString:
                sprintf(ErrorMsgBuffer, ThisError->ExtendedErrorText, GiveErrorInfoBuffer());
            break;
            case AIntCode:
                sprintf(ErrorMsgBuffer, ThisError->ExtendedErrorText, IntCode);
            break;
            case AIntAndASQLCode:
                sprintf(ErrorMsgBuffer, ThisError->ExtendedErrorText, IntCode, SQLCode);
            break;
            case ASQLCodeAndAInt:
                sprintf(ErrorMsgBuffer, ThisError->ExtendedErrorText, SQLCode, IntCode);
            break;
            case ASQLCodeAndAString:
                sprintf(ErrorMsgBuffer, ThisError->ExtendedErrorText, SQLCode, GiveErrorInfoBuffer());
            break;
            default:
                TheDBMSrvErrorHandler_cn36.Report(tcn36_ADBMSrvError::InternalError.Set());
        }
    }
    else
        cn36_StrNCpy(ErrorMsgBuffer, ThisError->PlainErrorText, ErrorMsgBufferLength_ni35-1);
}

tcn36_ADBMSrvError::DBMSrvErrorDescription tcn36_ADBMSrvError::DBMSrvErrorToText[]=
{
    {CaseOutOfMemory,
     None,
     "Out of memory.",
     0},


    {CaseCanNotExportEnvironment,
     AString,
     "Can not export an environment variable.",
     "Can not export the environment variable %s."},

    {CaseCanNotGetComputerName,
     AIntCode,
     "Can not determine the computer name.",
     "Can not determine the computer name (error %d)."},

    {CaseCanNotGetBlockingFactor,
     AString,
     "The blocking factor is not a number.",
     "The blocking factor %s is not a number."},

    {CaseCanNotGetPageSize,
     AString,
     "The page size is not a number.",
     "The page size %s is not a number."},


    {CaseADirectoryIsMissing,
     AString,
     "A directory is missing.",
     "The directory %s is missing."},

    {CaseAFileIsMissingOrUnreadable,
     AString,
     "A file is missing or unreadable.",
     "The file %s is missing or unreadable."},

    {CaseAFileIsNotExecutable,
     AString,
     "A file is not executable.",
     "The file %s is not executable."},

    {CaseAFileAlreadyExists,
     AString,
     "A file already exists.",
     "The file '%s' already exists."},


    {CaseToManyPipes,
     None,
     "The medium consists of to many pipes.",
     0},

    {CaseCouldNotCreateDataPipe,
     AString,
     "Could not create a data transfer pipe.",
     "Could not create the data transfer pipe '%s'."},

    {CaseCanNotCreateADSMInputPipe,
     AString,
     "Could not create the input pipe for ADSM.",
     "Could not create the input pipe '%s' for ADSM."},

    {CaseCanNotDetermineNSR_HOME,
     AString,
     "Could not determine value of 'NSR_HOME'.",
     "Could not determine value of 'NSR_HOME' from file '%s'."},

    {CaseCanNotDetermineNSR_HOST,
     AString,
     "Could not determine value of 'NSR_HOST'.",
     "Could not determine value of 'NSR_HOST' from file '%s'."},

    {CaseUnknowFormatOfMminfoOutput,
     AString,
     "Found unknown format of mminfo output.",
     "Found unknown format of mminfo output '%s'."},

    {CasePipeOrFileNameToLong,
     AString,
     "A pipe or file name is too long.",
     "The pipe or file name '%s' is too long."},

    {CaseUnspecifiedBackintPath,
     AString,
     "Path of the Backint for SAP DB program was not specified in the configuration file.",
     "Path of the Backint for SAP DB program was not specified in the configuration file '%s'."},

    {CaseTwoDifferentBIDs,
     AString,
     "Backint returned two different backup IDs for the same file or pipe.",
     "Backint returned two different backup IDs for '%s'."},

    {CaseWrongBIDReturned,
     AString,
     "Backint for SAP DB returned the wrong BID for a file or a pipe.",
     "Backint for SAP DB returned the wrong BID for '%s'."},
     
    {CaseWrongFilesProcessed,
     AString,
     "A wrong file was processed by Backint for SAP DB.",
     "A wrong file ('%s') was processed by Backint for SAP DB."},

    {CaseCouldNotSaveBackupHistoryFiles,
     None,
     "Could not save the Backup History or the External Backup History with Backint for SAP DB.",
     0},

    {CaseCouldNotRestoreBackupHistoryFiles,
     None,
     "Could not restore the Backup History or the External Backup History with Backint for SAP DB.",
     0},

    {CaseUnknownBackintOutputFormat,
     AString,
     "Found an unknown format in the Backint for SAP DB output file.",
     "Found an unknown format in the Backint for SAP DB output file in line '%s'."},

    {CaseSyntaxErrorEBIDs,
     AIntCode,
     "The list of external backup ID's contains not only external backup ID's (syntax error).",
     "The list of external backup ID's contains no external backup ID at list element %d."},

    {CaseEBIDsOfDifferentDBs,
     AString,
     "The list of external backup ID's contains ID's from different databases.",
     "The list of external backup ID's contains not only ID's from database %s."},

    {CaseTooManyEBIDs,
     AIntCode,
     "The list of external backup ID's contains too many ID's.",
     "The list of external backup ID's contains more than %d ID's."},

    {CaseTooFewEBIDs,
     AIntCode,
     "The list of external backup ID's contains too few ID's",
     "The list of external backup ID's contains less than %d ID's."},

    {CaseCanNotBuildSaveLabel,
     AString,
     "Could not construct save label.",
     "Could not construct save label (%s characters)."},

    {CaseCanNotSignalEndToTool,
     None,
     "Could not signal end to tool.",
     0},


    {CaseNoBackupIsRunning,
     None,
     "No backup is running.",
     0},

    {CaseBackupToolFailed,
     AIntCode,
     "The backup tool failed.",
     "The backup tool failed with %d as sum of exit codes."},

    {CaseBackupToolAndDBFailed,
     AIntAndASQLCode,
     "The backup tool and the database request failed.",
     "The backup tool failed with %d as sum of exit codes. The database request failed with error %d."},

    {CaseBackupToolKilled,
     AIntAndASQLCode,
     "The backup tool was killed.",
     "The backup tool was killed with %d as sum of exit codes. The database request ended with code %d."},

    {CaseBackupToolFailedAndDBKernelRequestCanceled,
     AIntAndASQLCode,
     "The backup tool failed and the database request was canceled.",
     "The backup tool failed with %d as sum of exit codes. The database request was canceled and ended with error %d."},

    {CaseBackupToolKilledAndDBKernelRequestCanceled,
     AIntAndASQLCode,
     "The backup tool was killed and the database request was canceled.",
     "The backup tool was killed with %d as sum of exit codes. The database request was canceled and ended with error %d."},

    {CaseDBKernelRequestCanceled,
     ASQLCodeAndAInt,
     "The database request was canceled.",
     "The database request was canceled and ended with error %d. The backup tool ended with %d as sum of exit codes."},

    {CaseCancelDBKernelRequestFailed,
     AIntCode,
     "A cancel of the database request failed.",
     "A cancel of the database request failed. The backup tool ended with %d as sum of exit codes."},

    {CaseBackupToolFailedAndCancelDBKernelRequestFailed,
     AIntCode,
     "The backup tool failed and a cancel of the database request failed.",
     "The backup tool failed with %d as sum of exit codes. A cancel of the database request failed."},

    {CaseBackupToolKilledAndCancelDBKernelRequestFailed,
     AIntCode,
     "The backup tool was killed and a cancel of the database request failed.",
     "The backup tool was killed with %d as sum of exit codes. A cancel of the database request failed."},
    
    {CaseSQLError,
     ASQLCodeAndAString,
     "The database was unable to fulfill a request.",
     "The database was unable to fulfill a request (%d, %s)."},

    {CaseCN00Error,
     AIntCode,
     "A baisc error occured.",
     "The error %d occured."},


    {CaseNoMediumFound,
     None,
     "Unable to determine backup medium.",
     0},

    {CaseMediumHasNoName,
     None,
     "Unable to determine name of medium.",
     0},


    {CaseUnknownFormatOfDBMKnlFile,
     AString,
     "Unknown format of backup history.",
     "Unknown format of backup history %s."},

    {CaseUnknownFormatOfDBMEbfFile,
     AString,
     "Unknown format of external backup history.",
     "Unknown format of external backup history %s."},

    {CaseCouldNotGetNameOfToolLogFile,
     AIntCode,
     "Could not determine the name of the external backup history file.",
     "Could not determine the name of the external backup history file (Error %d)."},

    {CaseCouldNotGetSaveIDs,
     None,
     "Could not get external backup ID's from the backup tool.\n",
     0},

    {CaseCouldNotWriteToToolLogFile,
     AString,
     "Could not write to external backup history file.",
     "Could not write to external backup history file %s."},

    {CaseNoLogPageInformation,
     None,
     "Have no correct log page information from database kernel.",
     0},
     

    {CaseCouldNotGetNameOfBackupHistory,
     AIntCode,
     "Could not determine the name of the backup history file.",
     "Could not determine the name of the backup history file (Error %d)."},

    {CaseCouldNotGetNameOfToolProtFile,
     None,
     "Could not determine name of external backup protocol file.",
     0},

    {CaseCouldNotMoveToolOutputToToolProtFile,
     AString,
     "Could not move output of tool to protocol file.",
     "Could not move output of tool from file %s to protocol file."},


    {CaseNeedExternalSaveIDForRestore,
     None,
     "An external backup ID is needed, but not supplied.",
     0},


    {CaseInternalError,
     None,
     "An internal error occured.",
     0}
};


tcn36_ADBMSrvError tcn36_ADBMSrvError::OutOfMemory               (tcn36_ADBMSrvError::CaseOutOfMemory, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotExportEnvironment   (tcn36_ADBMSrvError::CaseCanNotExportEnvironment, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotGetComputerName     (tcn36_ADBMSrvError::CaseCanNotGetComputerName, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotGetBlockingFactor   (tcn36_ADBMSrvError::CaseCanNotGetBlockingFactor, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotGetPageSize         (tcn36_ADBMSrvError::CaseCanNotGetPageSize, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::ADirectoryIsMissing       (tcn36_ADBMSrvError::CaseADirectoryIsMissing, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::AFileIsMissingOrUnreadable(tcn36_ADBMSrvError::CaseAFileIsMissingOrUnreadable, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::AFileIsNotExecutable      (tcn36_ADBMSrvError::CaseAFileIsNotExecutable, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::AFileAlreadyExists        (tcn36_ADBMSrvError::CaseAFileAlreadyExists, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotCreateDataPipe     (tcn36_ADBMSrvError::CaseCouldNotCreateDataPipe, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotCreateADSMInputPipe  (tcn36_ADBMSrvError::CaseCanNotCreateADSMInputPipe, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::UnknownFormatOfMminfoOutput(tcn36_ADBMSrvError::CaseUnknowFormatOfMminfoOutput, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::PipeOrFileNameToLong       (tcn36_ADBMSrvError::CasePipeOrFileNameToLong, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::UnspecifiedBackintPath     (tcn36_ADBMSrvError::CaseUnspecifiedBackintPath, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::TwoDifferentBIDs           (tcn36_ADBMSrvError::CaseTwoDifferentBIDs, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::WrongBIDReturned           (tcn36_ADBMSrvError::CaseWrongBIDReturned, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::WrongFilesProcessed        (tcn36_ADBMSrvError::CaseWrongFilesProcessed, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotSaveBackupHistoryFiles(tcn36_ADBMSrvError::CaseCouldNotSaveBackupHistoryFiles, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotRestoreBackupHistoryFiles(tcn36_ADBMSrvError::CaseCouldNotRestoreBackupHistoryFiles, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::UnknownBackintOutputFormat (tcn36_ADBMSrvError::CaseUnknownBackintOutputFormat, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotBuildSaveLabel       (tcn36_ADBMSrvError::CaseCanNotBuildSaveLabel, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CanNotSignalEndToTool      (tcn36_ADBMSrvError::CaseCanNotSignalEndToTool, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::NoBackupIsRunning          (tcn36_ADBMSrvError::CaseNoBackupIsRunning, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolFailed           (tcn36_ADBMSrvError::CaseBackupToolFailed, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolAndDBFailed      (tcn36_ADBMSrvError::CaseBackupToolAndDBFailed, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolKilled           (tcn36_ADBMSrvError::CaseBackupToolKilled, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolFailedAndDBKernelRequestCanceled(tcn36_ADBMSrvError::CaseBackupToolFailedAndDBKernelRequestCanceled, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolKilledAndDBKernelRequestCanceled(tcn36_ADBMSrvError::CaseBackupToolKilledAndDBKernelRequestCanceled, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::DBKernelRequestCanceled(tcn36_ADBMSrvError::CaseDBKernelRequestCanceled, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CancelDBKernelRequestFailed(tcn36_ADBMSrvError::CaseCancelDBKernelRequestFailed,tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolFailedAndCancelDBKernelRequestFailed(tcn36_ADBMSrvError::CaseBackupToolFailedAndCancelDBKernelRequestFailed,tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::BackupToolKilledAndCancelDBKernelRequestFailed(tcn36_ADBMSrvError::CaseBackupToolKilledAndCancelDBKernelRequestFailed,tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::SQLError                  (tcn36_ADBMSrvError::CaseSQLError, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CN00Error                 (tcn36_ADBMSrvError::CaseCN00Error, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::NoMediumFound             (tcn36_ADBMSrvError::CaseNoMediumFound, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::MediumHasNoName           (tcn36_ADBMSrvError::CaseMediumHasNoName, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::UnknownFormatOfDBMKnlFile (tcn36_ADBMSrvError::CaseUnknownFormatOfDBMKnlFile, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotGetNameOfToolLogFile(tcn36_ADBMSrvError::CaseCouldNotGetNameOfToolLogFile, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotGetSaveIDs        (tcn36_ADBMSrvError::CaseCouldNotGetSaveIDs, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotWriteToToolLogFile(tcn36_ADBMSrvError::CaseCouldNotWriteToToolLogFile, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotGetNameOfBackupHistory(tcn36_ADBMSrvError::CaseCouldNotGetNameOfBackupHistory, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotGetNameOfToolProtFile(CaseCouldNotGetNameOfToolProtFile, tcn36_ADBMSrvError::DoNotSetErrorNow);
tcn36_ADBMSrvError tcn36_ADBMSrvError::CouldNotMoveToolOutputToToolProtFile(CaseCouldNotMoveToolOutputToToolProtFile, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::NeedExternalSaveIDForRestore(tcn36_ADBMSrvError::CaseNeedExternalSaveIDForRestore, tcn36_ADBMSrvError::DoNotSetErrorNow);

tcn36_ADBMSrvError tcn36_ADBMSrvError::InternalError             (tcn36_ADBMSrvError::CaseInternalError, tcn36_ADBMSrvError::DoNotSetErrorNow);
