/*******************************************************************************
  module:       vcn361.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementing basic functions used in vcn35.cpp, ..., vcn355.cpp
  last change:  2000-11-29 15:22
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


#include "hni00b.h"
#include "hcn361.h"

#include <limits.h>
#include <ctype.h>

#if defined (WIN32)
#elif defined (UNIX)
    #include <unistd.h>
#else
    #error vcn361.cpp is only coded for WIN32 and UNIX (define one of them)
#endif


void cn36_StrNCpy(char * Dest, const char *Src, size_t Length)
{
    strncpy(Dest, Src, Length);
    Dest[Length]='\0';
}

void cn36_StrNCat(char * Dest, const char *Src, size_t Length)
{
    size_t Used=strlen(Dest);

    if(Length>Used)
        cn36_StrNCpy(Dest+Used, Src, Length-Used);
}

void cn36_StrUpr(char *String)
{
    #if defined (WIN32)
    _strupr(String);
    #elif defined (UNIX)
    while('\0'!= *String)
    {
        *String=toupper(*String);
        String++;
    }
    #else
        #error cn36_StrUpr() was only coded for UNIX and NT (define one of them).
    #endif
}


class tcn36_ADigitToIntMap
{
    public:
        enum MapType{NonDigitsZero, NonDigitsNegativ};

        tcn36_ADigitToIntMap(MapType Type)
        {
            int i;
            int NonDigitValue;

            if(Type==NonDigitsZero)
                NonDigitValue=0;
            else
                NonDigitValue=-1;

            for(i=0; i<=UCHAR_MAX; i++)
                Map[i]=NonDigitValue;

            Map[(unsigned char)'0']=0;
            Map[(unsigned char)'1']=1;
            Map[(unsigned char)'2']=2;
            Map[(unsigned char)'3']=3;
            Map[(unsigned char)'4']=4;
            Map[(unsigned char)'5']=5;
            Map[(unsigned char)'6']=6;
            Map[(unsigned char)'7']=7;
            Map[(unsigned char)'8']=8;
            Map[(unsigned char)'9']=9;
        }

        int operator[](unsigned char c) const
        {
            return Map[c];
        }

    private:
        int Map[int(UCHAR_MAX)+int(1)];
};

static tcn36_ADigitToIntMap DigitToIntMap_cn36(tcn36_ADigitToIntMap::NonDigitsZero);

static const char IntToCharMap_cn36[10]={'0','1','2','3','4','5','6','7','8','9'};

//converts Length chars from Buf to a I with MaxI as maximal possible value.
//returns 1 for success and 0 otherwise
template<class I> int cn36_StrTo(I & TheI, const char *Buf, size_t Length, I MaxI)
{
    int rc=1;

    TheI=0;
    const I MaxValueBeforeMultiply=MaxI/10;
    size_t Position=0;

    while(Position<Length)
    {
        if(isdigit(Buf[Position]))
            if(TheI<=MaxValueBeforeMultiply)
            {
                TheI*=10;
                I NextDigit=I(DigitToIntMap_cn36[Buf[Position]]);  // convert the character to an int (0..9)

                if(MaxI-NextDigit>=TheI) //test TheI+NextDigit<MaxI without risking an overflow
                    TheI+=NextDigit;
                else
                    rc=0;
            }
            else
                rc=0;

        Position++;
    }

    return rc;
}

int cn36_StrToInt(int & TheInt, const char *Buf)
{
    #if defined(WIN32)
    return cn36_StrTo<int>(TheInt, Buf, strlen(Buf), INT_MAX);
    #elif defined (UNIX)
    return cn36_StrTo(TheInt, Buf, strlen(Buf), INT_MAX); // the SNI-C++ compilers can not handle StrTo<int>
    #else
        #error cn36_StrToInt is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrNToInt(int & TheInt, const char *Buf, size_t Length)
{
    #if defined(WIN32)
    return cn36_StrTo<int>(TheInt, Buf, Length, INT_MAX);
    #elif defined (UNIX)
    return cn36_StrTo(TheInt, Buf, Length, INT_MAX);  // the SNI-C++ compilers can not handle StrTo<int>
    #else
        #error cn36_StrNToInt is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrNToLong(long & TheLong, const char *Buf, size_t Length)
{
    #if defined(WIN32)
    return cn36_StrTo<long>(TheLong, Buf, Length, long(LONG_MAX));  // long is needed by sun64
    #elif defined (UNIX)
    return cn36_StrTo(TheLong, Buf, Length, long(LONG_MAX));  // the SNI-C++ compilers can not handle StrTo<long>
    #else
        #error cn36_StrNToLong is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrToInt8(teo00_Int8 & TheInt8, const char *Buf)
{
    #if defined(WIN32)
    return cn36_StrTo<teo00_Int8>(TheInt8, Buf, strlen(Buf), MAX_INT8_CN36);
    #elif defined (UNIX)
    return cn36_StrTo(TheInt8, Buf, strlen(Buf), MAX_INT8_CN36);    // the SNI-C++ compilers can not handle cn36_StrTo<teo00_Int8>
    #else
        #error cn36_StrToInt8 is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrNToInt8(teo00_Int8 & TheInt8, const char *Buf, size_t Length)
{
    #if defined(WIN32)
    return cn36_StrTo<teo00_Int8>(TheInt8, Buf, Length, MAX_INT8_CN36);
    #elif defined (UNIX)
    return cn36_StrTo(TheInt8, Buf, Length, MAX_INT8_CN36); // the SNI-C++ compilers can not handle cn36_StrTo<teo00_Int8>
    #else
        #error cn36_StrNToInt8 is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrToUInt8(teo00_Uint8 & TheUInt8, const char * Buf)
{
    #if defined(WIN32)
    return cn36_StrTo<teo00_Uint8>(TheUInt8, Buf, strlen(Buf), MAX_UINT8_CN36);
    #elif defined (UNIX)
    return cn36_StrTo(TheUInt8, Buf, strlen(Buf), MAX_UINT8_CN36); // the SNI-C++ compilers can not handle cn36_StrTo<teo00_Uint8>
    #else
        #error cn36_StrToUInt8 is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_StrNToUInt8(teo00_Uint8 & TheUInt8, const char * Buf, size_t Length)
{
    #if defined(WIN32)
    return cn36_StrTo<teo00_Uint8>(TheUInt8, Buf, Length, MAX_UINT8_CN36);
    #elif defined (UNIX)
    return cn36_StrTo(TheUInt8, Buf, Length, MAX_UINT8_CN36);   // the SNI-C++ compilers can not handle cn36_StrTo<teo00_Uint8>
    #else
        #error cn36_StrNToUInt8 is only coded for WIN32 and UNIX (define one of them)
    #endif
}

int cn36_UInt8ToStrN(char *Buf, size_t Length, teo00_Uint8 i)
{
    int rc=1;

    char   String[100];
    size_t Pos=0;

    if(0==i)
        String[Pos++]='0';
    else
    {
        while(i>0)
        {
            String[Pos]=IntToCharMap_cn36[int(i%10)];
            Pos++;
            i/=10;
        }
    }

    if(Pos>Length-1)    // -1 for terminating '\0'
    {
        rc=0;         // Buf to small
        if(Length>0)
            Buf[0]='\0';
    }
    else
    {
        size_t Pos2=0;
        
        while(Pos>0)
        {
            Buf[Pos2]=String[Pos-1];
            Pos--;
            Pos2++;
        }

        Buf[Pos2]='\0';
    }    

    return rc;
}


int cn36_StrNConsistsOfDigits(const char *String, size_t Length)
{
    int rc=1;
    size_t i;
    
    if(0!=String)
    {
        for(i=0; rc && i<Length && '\0'!=String[i]; i++)
            if(!isdigit(String[i]))
                rc=0;
    }

    return rc;
}

int  cn36_StrConsistsOfDigits(const char *String)
{
    int rc=1;

    if(0!=String)
    {
        size_t i=0;

        while('\0'!=String[i] && rc)
        {
            if(!isdigit(String[i]))
                rc=0;

            i++;
        }
    }

    return rc;
}


char * cn36_PositionOfTerminatingZero(char * s)
{
    return s+strlen(s);
}

const char * cn36_PositionOfTerminatingZero(const char * s)
{
    return s+strlen(s);
}


char * cn36_FirstNonWhiteSpaceOf(char * s)
{
    while('\0'!=*s && isspace(*s))
        s++;

    return s;
}

const char * cn36_FirstNonWhiteSpaceOf(const char * s)
{
    while('\0'!=*s && isspace(*s))
        s++;

    return s;
}

char * cn36_FirstWhiteSpaceOf(char * s)
{
    while('\0'!=*s && !isspace(*s))
        s++;

    return s;
}

const char * cn36_FirstWhiteSpaceOf(const char * s)
{
    while('\0'!=*s && !isspace(*s))
        s++;

    return s;
}

void cn36_GetStringBetweenDoubleQuotesOrWhiteSpaces(const char * & Start, const char * & End)
{
    End=Start;

    if(0!=Start)
    {
        int FoundValue=0;
        const char DoubleQuote='\"';

        if(DoubleQuote==Start[0])   //if the value starts with an '\"', we take all chars up to a second '\"'
        {
            End=strchr(Start+1, DoubleQuote);  //+1 we look for the second '\"' not for the one in Start[0]

            if(0!=End)   //if we found a secound '\"', take the string between these to '\"'
            {
                Start++; //start now behind the first '\"', End is already correct and even empty strings are OK
                FoundValue=1;
            }
        }

        if(!FoundValue)
            End=cn36_FirstWhiteSpaceOf(Start);
    }
}

const char * cn36_StrChr(const char *String, char Character)
{
    if(0!=String)
        String=strchr(String, Character);

    return String;
}

char * cn36_StrChr(char *String, char Character)
{
    if(0!=String)
        String=strchr(String, Character);

    return String;
}

const char * cn36_LoopedStrChr(const char *String, char Character, int Number)
{
    if(0!=String && 0<Number)
    {
        String=strchr(String, Character);
        Number--;
    }

    while(0!=String && 0<Number)
    {
        String=strchr(String+1, Character); // String+1, because String points already to 'Character' from the previous strchr
        Number--;
    }

    return String;
}

char * cn36_LoopedStrChr(char *String, char Character, int Number)
{
    if(0!=String && 0<Number)
    {
        String=strchr(String, Character);
        Number--;
    }

    while(0!=String && 0<Number)
    {
        String=strchr(String+1, Character); // String+1, because String points already to 'Character' from the previous strchr
        Number--;
    }

    return String;
}

int cn36_StrStrNCmp(const char * String1, const char *String2, size_t Length2)
{
    int rc=strncmp(String1, String2, Length2);

    if(0==rc)
    {
        size_t Length1=strlen(String1);

        if(Length1>Length2)
            rc=1;
    }

    return rc;
}

int cn36_StrNUprCmp(const char *String1, const char *String2, size_t Length)
{
    int rc=0;
    size_t Pos=0;

    while(0==rc && Pos<Length && '\0'!=String2[Pos] && '\0'!=String1[Pos])
    {
        int S1=toupper(String1[Pos]),
            S2=toupper(String2[Pos]);
    
        if(S1<S2)
            rc=-1;
        else
            if(S1>S2)
                rc=1;
            
        Pos++;
    }

    if(0==rc && Pos<Length) //reached the end of one string. up to here both strings are equal
    {
        if('\0'==String1[Pos]) // at end of String1
        {
            if('\0'!=String2[Pos]) // String2 has not ended
                rc=-1;
        }
        else //at end of String2 (Pos==Length)||('\0'==String2[Pos])
            rc=1;   //String1 has not ended -> String2 must have ended (0==rc && Pos<Length && '\0'!=String1[Pos])
    }

    return rc;
}

int cn36_StrStrNUprCmp(const char * String1, const char *String2, size_t Length2)
{
    int rc=cn36_StrNUprCmp(String1, String2, Length2);

    if(0==rc)
    {
        size_t Length1=strlen(String1);

        if(Length1>Length2)
            rc=1;
    }

    return rc;
}

extern const char PathDelimiter_cn36;

#if defined (WIN32)
const char PathDelimiter_cn36='\\';
#elif defined (UNIX)
const char PathDelimiter_cn36='/';
#else
    #error vcn361.cpp is only coded for WIN32 and UNIX (define one of them)
#endif

const char * cn36_BaseNameFrom(const char *Path)
{
    const char * BaseName=strrchr(Path, PathDelimiter_cn36);

    if(0==BaseName)
        BaseName=Path;
    else
        BaseName++;

    return BaseName;
}

char * cn36_BaseNameFrom(char *Path)
{
    char * BaseName=strrchr(Path, PathDelimiter_cn36);

    if(0==BaseName)
        BaseName=Path;
    else
        BaseName++;

    return BaseName;
}

const char * cn36_EndOfDirectory(const char *Path)
{
    const char * End=strrchr(Path, PathDelimiter_cn36);

    if(0==End)
        End=Path;

    return End;
}

char * cn36_EndOfDirectory(char *Path)
{
    char * End=strrchr(Path, PathDelimiter_cn36);

    if(0==End)
        End=Path;

    return End;
}

void cn361_Sleep(int Seconds)
{
    #if defined (WIN32)
    Sleep(DWORD(Seconds)*1000);
    #elif defined (UNIX)
    sleep(Seconds);
    #else
        #error vcn361.cpp only coded for UNIX and WIN32 (define one of them)
    #endif
}
