#   SAP_DB.pm
#
#   Copyright (c) 2000, SAP AG
#
#   You may distribute under the terms of the
#   GNU Library General Public License
#
#
#    ========== licence begin LGPL
#    Copyright (C) 2000 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#


use SAP::DBTech::sapdb;
use strict;

{
    package DBD::SAP_DB;
    require DBI;

    my $VERSION = "7.2";

    my $drh = undef;       # holds driver handle once initialised
    my $err = 0;           # The $DBI::err value

    sub driver{
        return $drh if $drh;
        my($class, $attr) = @_;
        $class .= "::dr";
        ($drh) = DBI::_new_drh($class, {
            'Name' => 'SAP_DB',
            'Version' => $VERSION,
            'Attribution' => 'SAP DB driver by SAP AG',
            }, [ qw'example implementors private data']);
        $drh;
    }

    sub parseError {
        my ($handle, $fullString) = @_;
        my ($errcode, $errtext) = (-1, "unknown error");
        if ($fullString =~ /SQLError: \((-?[0-9]+)\) (.*) at /) {
            $errcode = $1;
            $errtext = $2;
        }
        else {
            $errtext = $fullString;
        }
        $handle->DBI::set_err ($errcode, $errtext);
    }

}


{   package DBD::SAP_DB::dr; # ====== DRIVER ======
    $DBD::SAP_DB::dr::imp_data_size = 0;
    # we use default (dummy) connect method

    sub connect {
        my($drh, $dbname, $user, $auth, $attr)= @_;

        # Some database specific verifications, default settings
        # and the like following here. This should only include
        # syntax checks or similar stuff where it's legal to
        # 'die' in case of errors.

        # create a 'blank' dbh (call superclass constructor)
        my $dbh = DBI::_new_dbh($drh, {
            'Name' => $dbname,
            'USER' => $user,
            'CURRENT_USER' => $user,
        });

        # Process attributes from the DSN; we assume ODBC syntax
        # here, that is, the DSN looks like var1=val1;...;varN=valN
        my @connectParts = split(/;/, $dbname);

        # Parse the DB part:
        my $host = ''; # Local connection by default.
        $dbname = shift @connectParts;
        if ($dbname =~ /^([^:]+):(.*)/) {
            $host = $1;
            $dbname = $2;
        }

        # Parse the options part:
        my @options;
        foreach my $var (@connectParts) {
            if ($var =~ /(.*?)=(,*)/) {
                $dbh->STORE($1, $2); # Not !!! $dbh->{$var} = $val;
                push (@options, $var);
            }
        }

        my $sapdb_session = undef;
        eval {
            my $optionString = @options?join('&', @options):'';

            $sapdb_session = sapdb::connect ($user, $auth,
                                             $dbname, $host, $optionString);
            $dbh->STORE ('sapdb_session', $sapdb_session);
        };
        if ($@) {
            DBD::SAP_DB::parseError ($dbh, $@);
            return undef;
        }
        return $dbh;
    }

    sub disconnect_all { }
    sub DESTROY { undef }
}


{   package DBD::SAP_DB::db; # ====== DATABASE ======
    $DBD::SAP_DB::db::imp_data_size = 0;

    sub disconnect {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->release ();
    }

    sub commit {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->commit ();
    }

    sub rollback {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->rollback ();
    }

    sub prepare {
        my($dbh, $statement)= @_;

        my $sapdb_prepared;
        eval {
            $sapdb_prepared = $dbh->FETCH ('sapdb_session')
                ->prepare ($statement);
        };
        if ($@) {
            DBD::SAP_DB::parseError ($dbh, $@);
            return undef;
        }
        my($outer, $sth) = DBI::_new_sth($dbh, {
            'Statement'     => $statement,
            }, [ qw'example implementors private data']);

        $sth->STORE ('sapdb_prepared', $sapdb_prepared);
        $sth->STORE ('sapdb_parameters', []);
        # $sth->STORE ('sapdb_io_indices', []);
        $outer;
    }

    sub FETCH {
        my ($dbh, $attrib) = @_;
        # In reality this would interrogate the database engine to
        # either return dynamic values that cannot be precomputed
        # or fetch and cache attribute values too expensive to prefetch.
        if ($attrib eq 'AutoCommit') {
            return $dbh->FETCH ('sapdb_session')->getAutocommit ();
        }
        # else pass up to DBI to handle
        return $dbh->DBD::_::db::FETCH($attrib);
        }

    sub STORE {
        my ($dbh, $attrib, $value) = @_;
        # would normally validate and only store known attributes
        # else pass up to DBI to handle
        if ($attrib eq 'AutoCommit') {
            $dbh->FETCH ('sapdb_session')->setAutocommit ($value);
            return 1;
        }
        if ($attrib =~ /^sapdb_/) {
            $dbh->{$attrib} = $value; # Yes, we are allowed to do this,
            return 1;                 # but only for our private attributes
        }
        return $dbh->DBD::_::db::STORE($attrib, $value);
    }

    sub DESTROY { undef }
}


{   package DBD::SAP_DB::st; # ====== STATEMENT ======
    $DBD::SAP_DB::st::imp_data_size = 0;

    sub execute {
        my ($sth, @args) = @_;
        my ($result, $parameters);

        eval {
            if (@args == 0) {
                $parameters = $sth->FETCH ('sapdb_parameters');
            }
            else {
                $parameters = \@args;
            }
            $result = $sth->FETCH ('sapdb_prepared')->execute ($parameters);
        };

        return 1 if $@ =~ /^SQLError: \(100\)/; # FF: This is not an error!

        if ($@) {
            DBD::SAP_DB::parseError ($sth, $@);
            return undef;
        }
        my $refresult = ref ($result);
        if (ref ($result) eq 'SapDB_ResultSet') {
            $sth->STORE ('sapdb_cursor', $result);
            # set cursor information
            my $colCount = $result->columnCount ();
            $sth->STORE ('NUM_OF_FIELDS', $colCount);
            my (@colnames, $i);
            my ($colname, $coltype, $collength, $colfrac);
            for ($i = 0; $i < $colCount; ++$i) {
               $colname = $result->columnInfo ($i)->[0];
               push (@colnames, $colname);
            }
            $sth->STORE ('sapdb_colnames', \@colnames);
        }
        else {
            return -1;  # no result count yet
        }
        1;
    }

    sub bind_param {
        my ($sth, $index, $value) = @_;
        my ($parameters, $parameterCount);

        $parameters = $sth->FETCH ('sapdb_parameters');
        while (@{$parameters} < $index) {
            push (@{$parameters}, undef);
        }
        $parameters->[$index - 1] = $value;
    }

    sub bind_col {
        my ($sth, $index, $reference) = @_;
        my ($colBindings) = getColBindings ($sth);

        $colBindings->[$index - 1] = $reference;
    }

    sub bind_columns {
        my ($sth, @references) = @_;
        my ($colBindings) = getColBindings ($sth);
        my ($i);

        if (! defined ($references[0])) {
            shift @references;
        }
        for ($i = 0; $i < @references; ++$i) {
            $colBindings->[$i] = $references[$i];
        }
    }

    sub getColBindings {
        my ($sth) = @_;
        my $colBindings = $sth->FETCH ('sapdb_colbindings');

        if (!defined ($colBindings)) {
            $colBindings = [];
            my $cursor = $sth->FETCH ('sapdb_cursor');
            my $colCount = $cursor->columnCount ();
            for (my $i = 0; $i < $colCount; ++$i) {
                push (@{$colBindings}, undef);
            }
            $sth->STORE ('sapdb_colbindings', $colBindings);
        }
        return $colBindings;
    }

#####################################################################
#     sub bin_param_inout {                                         #
#         my ($sth, $index, $reference, $max_len, $bind_type) = @_; #
#         my $io_indices = $sth->FETCH ('sapdb_io_indices');        #
#         $sth->bind_param                                          #
#         for ($i = 0; $i < @colBindings; ++$i) {                   #
#             if ($index == $io_indices->[$i]) {                    #
#                 last;                                             #
#             }                                                     #
#             if ($index > $io_indices->[$i]) {                     #
#                 splice (                                          #
#                 last;                                             #
#             }                                                     #
#         }                                                         #
#     }                                                             #
#####################################################################

    sub fetch {
        my($sth) = @_;

        my $cursor = $sth->FETCH ('sapdb_cursor');
        my $row = undef;
        eval {
            $row = $cursor->next() if defined $cursor; # FF: it might not be (empty retsult)
        };
        if ($@) {
            DBD::SAP_DB::parseError ($sth, $@);
            return undef;
        }
        if (!$row) {
            $sth->finish ();     # no more data so finish
        }
        if ($row) {
            my $colBindings = $sth->FETCH ('sapdb_colbindings');
            if (defined ($colBindings)) {
                my ($i, $reference);
                for ($i = 0; $i < @{$row}; ++$i) {
                    $reference = $colBindings->[$i];
                    if (defined ($reference)) {
                        $$reference = $row->[$i];
                    }
                }
            }
        }
        return $row;
    }

    sub fetchrow_hashref {
        my($sth) = @_;
        my $row = $sth->fetch();
        return undef unless $row;

        my $names = $sth->{'sapdb_colnames'};
        my @namearray = @{$names};
        my %result;
        for (my $i = 0; $i < @namearray; ++$i) {
            $result{$namearray[$i]} = $row->[$i];
        }
        return \%result;
    }

    sub finish {
        my($sth) = @_;
    }

    sub FETCH {
        my ($sth, $attrib) = @_;
        # would normally validate and only fetch known attributes
        # else pass up to DBI to handle
        if ($attrib =~ /^sapdb_/) {
            return $sth->{$attrib};
        } elsif ($attrib eq "NAME") {
            return $sth->{'sapdb_colnames'};
        }
        return $sth->DBD::_::st::FETCH($attrib);
    }

    sub STORE {
        my ($sth, $attrib, $value) = @_;
        # would normally validate and only store known attributes
        # else pass up to DBI to handle

        # FF: This is just a quick hack to disallow setting NUM_OF_FIELDS more than once.
        if ($attrib eq 'NUM_OF_FIELDS') {
            if (exists $sth->{sapdb_NUM_OF_FIELDS}) {
            return 1;
            } else {
            $sth->{sapdb_NUM_OF_FIELDS} = 1;
            }
        }

        if ($attrib =~ /^sapdb_/) {
            $sth->{$attrib} = $value; # Yes, we are allowed to do this,
            return 1;                 # but only for our private attributes
        }
        return $sth->DBD::_::st::STORE($attrib, $value);
    }

    sub DESTROY { undef }
}

1;
