#!/usr/bin/env x_python
# ireport.py
#
#    ========== licence begin LGPL
#    Copyright (C) 2000 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#


import sys
from string import find, index, join, ljust, split, strip, upper
import os
from cmd import Cmd

import sapdb
_version = 0.1

class InteractiveReport (Cmd):
    prompt = '==> '
    intro = """SAP DB Terminal Report %s
Type 'help' for help.
An empty line repeats the last command""" % _version

    cursor = None
    cards = None

    def __init__ (self, session, options):
        self.session = session
        for name, value in options.__dict__.items ():
            setattr (self, name, value)
        self.scroll = self.lines
        if self.cards:
            self.displayPage = self.displayCard
        else:
            self.displayPage = self.displayRows

    def do_help (self, cmdName):
        """show command help

        without argument: lists all commands
        with argument: shows help for the specified command
        """
        if cmdName:
            try:
                method = getattr (self, 'do_' + cmdName)
                print "Help for command '%s': %s" % (cmdName, method.__doc__)
            except AttributeError:
                print "No help for '" + cmdName + "'"
        else:
            items = self.__class__.__dict__.items ()
            items.sort ()
            for name, member in items:
                if (name [:3] == 'do_') and (name != 'do_EOF'):
                    self.methodHelp (name [3:], member)

    def methodHelp (self, name, member):
        doc = member.__doc__
        if doc == None:
            doc = ''
        try:
            summary = doc [:index (doc, '\n')]
            more = ' ...'
        except ValueError:
            summary = doc
            more = ''
        print '%-15s %s%s' % (name, summary, more)

    def do_exit (self, str):
        """exit program"""
        return 1

    do_EOF = do_exit

    def do_option (self, str):
        """change display options
        """
        blank = find (str, ' ')
        if blank == -1:
            blank = len (str)
        name = str [:blank]
        value = strip (str[blank:])
        if name in ['lines', 'scroll']:
            try:
                value = int (value)
            except ValueError:
                print 'option', name, 'must be numeric'
        elif name in []:
            pass
        else:
            print 'unknown option', name
        setattr (self, name, value)

    def do_sql (self, str):
        """execute a SQL command

        Displays the result if possible.
        """
        try:
            result = self.session.sql (str)
            if isinstance (result, sapdb.SapDB_ResultSet):
                self.setCursor (result)
                self.do_next ()
            else:
                if result:
                    print result
        except sapdb.SQLError, err:
            print err
            if err.errorPos > 0:
                print 'at', err.errorPos

    def default (self, str):
        if len (str) > 6:
            self.do_sql (str)
        else:
            Cmd.default (self, str)

    def setCursor (self, cursor):
        self.cursor = cursor
        self.cardFormat = CardFormat (cursor)
        self.rowFormat = RowFormat (cursor)

    def do_next (self, str = None):
        """display next page
        """
        if self.hasCursor ():
            self.displayPage (self.cursor.next ())

    def do_prev (self, str = None):
        """display previous page
        """
        if self.hasCursor ():
            row = self.cursor.relative ((- self.scroll * 2) + 1)
            self.displayPage (row)

    def do_top (self, str = None):
        """display first page
        """
        if self.hasCursor ():
            row = self.cursor.first ()
            self.displayPage (row)

    def do_bottom (self, str = None):
        """display last page
        """
        if self.hasCursor ():
            row = self.cursor.absolute (- self.scroll)
            self.displayPage (row)

    def do_card (self, str = None):
        """displays one row per screen
        """
        self.scroll = 1
        self.displayPage = self.displayCard
        self.onecmd ('next')

    def do_report (self, str = None):
        """displays one row per line
        """
        self.scroll = self.lines
        self.displayPage = self.displayRows

    def hasCursor (self):
        if self.cursor:
            result = 1
        else:
            result = None
            print "no result set available"
        return result

    def displayRows (self, row):
        write = self.rowFormat.write
        self.writeRow (row, write)
        next = self.cursor.next
        for i in xrange (self.scroll - 1):
            row = next ()
            if not row:
                return
            self.writeRow (row, write)

    def displayCard (self, row):
        self.writeRow (row, self.cardFormat.write)

    def writeRow (self, row, write):
        if not row:
            print '<<<End of Cursor>>>'
        else:
            write (row)

class Format:
    def getDescription (self, cursor):
        result = []
        for i in xrange (cursor.columnCount ()):
            result.append (list (cursor.columnInfo (i)))
        return result

    def write (self, row, stream = sys.stdout):
        list = []
        for (fmt, nullfmt), value in map (None, self.formats, row):
            if value == None:
                list.append (nullfmt)
            else:
                list.append (fmt % value)
        stream.write (join (list, ''))

class CardFormat (Format):
    def __init__ (self, cursor):
        formats = []
        description = self.getDescription (cursor)
        maxName = 0
        for name, kind, length, frac in description:
            maxName = max (maxName, len (name))
        for name, kind, length, frac in description:
            prefix = ljust (name, maxName)
            fmt = prefix + ': %s\n'
            nullfmt = prefix + ': ?\n'
            formats.append ((fmt, nullfmt))
        self.formats = formats

class RowFormat (Format):
    def __init__ (self, cursor):
        formats = []
        description = self.getDescription (cursor)
        for name, kind, length, frac in description:
            fmt = '|%s'
            nullfmt = '|?'
            formats.append ((fmt, nullfmt))
        fmt, bullfmt = formats [-1]
        formats [-1] = (fmt + '|\n', nullfmt + '|\n')
        self.formats = formats

class TutorialShell (InteractiveReport):
    prompt = 'Tutorial> '
    intro = None
    introText = """*Welcome to the tutorial.
*You can end the tutorial by typing Ctrl-Z on NT or Ctrl-D on Unix."""
    __tutorial = [
        ('sql', "enter a SQL command like 'sql SELECT * from messages'"),
        ('next', "advance to the next page by entering 'next'"),
        ('next', "you can repeat the last command by pressing <ENTER>"),
        ('bottom', "go to the end of the result set by entering 'bottom'"),
        ('prev', "return to the previous page by entering 'prev'"),
        ('top', "return to the beginning by entering 'top'"),
        ('card', "you can switch to a different display by entering 'card'"),
        ('next', "now, 'next' will display the next record"),
        ('report', "'report' switches back to a paged display"),
        ('next', "'next' will actually display the next page"),
    ]

    def __init__ (self, session, options):
        InteractiveReport.__init__ (self, session, options)
        self.ignoreUnmatch = None
        self.tutorialIndex = -1
        print self.introText
        self.advanceTutorial ()

    def getCmdPart (self, line):
        i, n = 0, len(line)
        while i < n and line[i] in self.identchars: i = i+1
        cmd, arg = line[:i], strip(line[i:])
        return cmd

    def onecmd (self, line):
        cmd = self.getCmdPart (line)
        result = InteractiveReport.onecmd (self, line)
        if cmd == self.allowedCommand:
            atEnd = self.advanceTutorial ()
            if atEnd:
                result = 1
        else:
            if self.ignoreUnmatch:
                pass   # no need to rub it in
            else:
                print ("*Nice try, but the tutorial continues with '"
                    + self.allowedCommand + "'")
        self.ignoreUnmatch = None
        return result

    def advanceTutorial (self, plus = 1):
        self.tutorialIndex = self.tutorialIndex + plus
        try:
            if not self.ignoreUnmatch:
                nextLesson = self.__tutorial [self.tutorialIndex]
                self.allowedCommand, text = nextLesson
                print '*Lesson', self.tutorialIndex + 1, ':',
                print text
            atEnd = None
        except IndexError:
            print ("*This concludes the tutorial. "
                + "You're on your own now.")
            atEnd = 1
        return atEnd

    def emptyline (self):
        self.ignoreUnmatch = 1
        InteractiveReport.emptyline (self)
        print '*That was easy, wasn\'t it? Now try a different command'
        self.advanceTutorial (0)
        self.ignoreUnmatch = 1

    def do_exit (self, line):
        result = InteractiveReport.do_exit (self, line)
        print '*Giving up already?'
        self.ignoreUnmatch = 1
        return result

    do_EOF = do_exit

def fatalError (msg):
    sys.stderr.write (msg + '\n')
    sys.exit (2)

def createSession (options):
    if not options.userinfo:
        fatalError ('user name must be specified (option -u)')
    if not options.dbname:
        fatalError ('dabase name must be specified (option -d)')
    name, pwd = split (options.userinfo, ',')
    try:
        if options.oracle:
            optstring = 'sqlmode=oracle'
        else:
            optstring = ''
        result = sapdb.connect (name, pwd, options.dbname, options.node, optstring)
    except sapdb.SQLError, err:
        if err.errorCode == -4008:
            name = upper (name)
            result = sapdb.connect (name, pwd, options.dbname, options.node)
    return result

def main (options, args):
    if options.tutorial:
        constructor = TutorialShell
    else:
        constructor = InteractiveReport
    report = constructor (createSession (options), options)
    if args:
        report.do_sql (join (args))
        if report.hasCursor ():
            report.onecmd ('next')
            report.cmdloop ('')
    else:
        report.cmdloop ()

def defaultLines ():
    try:
        lines = int (os.environ ['LINES'])
    except:
        lines = 24
    return lines

def defaultCols ():
    try:
        lines = int (os.environ ['COLUMNS'])
    except:
        lines = 80
    return lines

def _options ():
    return [
        # (optstring, varname, typechar, default, help)
        ('u', 'userinfo', ':', None, '<user>,<pwd>'),
        ('d', 'dbname', ':', None, 'database name'),
        ('n', 'node', ':', '', 'server name'),
        ('lines', None, ':', defaultLines (), 'lines per screen'),
        ('columns', None, ':', defaultCols (), 'columns per line'),
        ('version', None, '', _version, 'version information'),
        ('oracle', None, '', None, 'enable Oracle (tm) SQL syntax'),
        ('tutorial', None, '', None, 'start tutorial'),
        ]

if __name__ == '__main__':
    import optlib
    optlib.optMain2 (main, _options ())

