/*!*****************************************************************************
  module:       hni34.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: ni layer
  description:  classes handling files, directories and pipes
  last change:  1999-10-04 15:32
  version:      7.2
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined(_HNI34_H_)
#define _HNI34_H_


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hni00b.h"
#include "heo06.h"

#if defined (WIN32)
#include<windows.h>
#elif defined (UNIX)
#include<sys/types.h>
#include<dirent.h>
#else
    #error ni34.h only coded for WIN32 and UNIX (Define one of them.)
#endif


// -----------------------------------------------------------------------------
// class tni34_AFileSystemEntry
// -----------------------------------------------------------------------------

class tni34_AFileSystemEntry
{
  public:
    virtual int Exists();

    virtual int Check()=0;
    virtual int Create()=0;
    virtual int Remove()=0;
    
    const char * GiveName() const;
    int IsGood() const;

  protected:
    tni34_AFileSystemEntry(const char * name=0, int CreateOnCheck=0);
    virtual ~tni34_AFileSystemEntry();

    int SetNameTo(const char *name);
    int ResetNameTo(const char *NewName);

    int IsGoodFlag;
    int CreateOnCheckFlag;

    int IsOpenFlag;

  private:
    char *Name;
};


// -----------------------------------------------------------------------------
// types
// -----------------------------------------------------------------------------

#if defined (WIN32)
typedef WIN32_FIND_DATA tni34_ADirectoryEntry;
#elif defined (UNIX)
typedef dirent tni34_ADirectoryEntry;
#endif


// -----------------------------------------------------------------------------
// class tni34_ADirectory
// -----------------------------------------------------------------------------

class tni34_ADirectory:protected tni34_AFileSystemEntry
{
  public:
    tni34_ADirectory(const char *name, int CreateOnCheck=0, int CheckNow=1);
    ~tni34_ADirectory();

    int IsADirectory();

    int Check();
    int Create();
    int Open();
    int Read(const char * & NextEntry);
    int Remove();

  protected:
      tni34_ADirectoryEntry **DirectoryEntrys;          //TODO: should be replaced by a vector
      int                     NumberOfDirectoryEntrys;
      int                     PlaceForEntrys;
      int                     NumberOfCurrentReadEntry;

      void FreeDirectoryEntrys();
};


// -----------------------------------------------------------------------------
// class tni34_ADataEntry
// -----------------------------------------------------------------------------

class tni34_ADataEntry:public tni34_AFileSystemEntry
{
  public:
    virtual int Open()=0;
    virtual int Read(void *buf, long BytesToRead, long & BytesRead)=0;
    virtual int Write(const void *buf, long BytesToWrite, long & BytesWritten)=0;
    virtual int Close()=0;

    int IsOpen();

  protected:
    tni34_ADataEntry(const char * name=0, int CreateOnOpen=0, int CreateOnCheck=0);
    
    int CreateOnOpenFlag;
};


// -----------------------------------------------------------------------------
// constants and types
// -----------------------------------------------------------------------------

#if defined (WIN32)
typedef HANDLE tni34_AFileHandle;
const tni34_AFileHandle InvalidFileHandle_ni34=INVALID_HANDLE_VALUE;
#elif defined (UNIX)
typedef int tni34_AFileHandle;
const tni34_AFileHandle InvalidFileHandle_ni34=-1;
#endif


// -----------------------------------------------------------------------------
// class tni34_AFile
// -----------------------------------------------------------------------------

class tni34_AFile:public tni34_ADataEntry
{
  public:
    enum ADirection {FromDisk, ToDiskTruncate, ToDiskAppend};

    tni34_AFile(ADirection Dir=FromDisk,
                int        CreateOnOpen=1,
                int        OpenNow=1,
                int        CreateOnCheck=0,
                int        PrefBufSize=65536,
                int        MaxNumberOfRetries=1000);

    tni34_AFile(const char * name,
                ADirection   Dir=FromDisk,
                int          CreateOnOpen=1,
                int          OpenNow=1,
                int          CreateOnCheck=0,
                int          PrefBufSize=65536);

    ~tni34_AFile();
    
    int IsAFile();

    int Check();
    int Create();
    int Open();
    int Read(void *buf, long BytesToRead, long & BytesRead);
    int Write(const void *buf, long BytesToWrite, long & BytesWritten);
    int Close();
    int Remove();
    int Rename(const char * NameToMoveTo);
    
  private:
    ADirection DirectionOfFile;
    int PrefferedBufferSize;

    tni34_AFileHandle HandleForFile;

    int PureQuietCreate(const char * name);
    int PureCreate();
    int PureOpen();
    int PureQuietClose();
};


// -----------------------------------------------------------------------------
// class tni34_AVirtualFile
// -----------------------------------------------------------------------------

class tni34_AVirtualFile:public tni34_ADataEntry
{
  public:
    enum ADirection {FromDisk, ToDiskTruncate, ToDiskAppend};
    enum AOpenType {OpenNow, DoNotOpenNow};

    tni34_AVirtualFile(const char * name,
                       ADirection   Dir=FromDisk,
                       AOpenType    OpenFlag=OpenNow);

    ~tni34_AVirtualFile();

    int Exists();
    int IsAFile();

    int Check();
    int Create();

    int Open();
    int Read(void *buf, long BytesToRead, long & BytesRead);
    int Write(const void *buf, long BytesToWrite, long & BytesWritten);
    int Close();
    int Remove();

    int GetSize(teo00_Uint8 & FileSize);
    
  private:
    ADirection DirectionOfFile;

    tsp_int4 HandleForFile;
    int      HandleIsValid;
};


// -----------------------------------------------------------------------------
// class tni34_ATextFile
// -----------------------------------------------------------------------------

class tni34_ATextFile:public tni34_AFile
{
  public:
    tni34_ATextFile(ADirection   Dir=FromDisk,
                    int          CreateOnOpen=1,
                    int          OpenNow=1,
                    int          CreateOnCheck=0,
                    unsigned int EstimatedLineSize=1024,
                    int          PrefBufSize=65536,
                    int          MaxNumberOfRetries=1000);

    tni34_ATextFile(const char * name,
                    ADirection   Dir=FromDisk,
                    int          CreateOnOpen=1,
                    int          OpenNow=1,
                    int          CreateOnCheck=0,
                    unsigned int EstimatedLineSize=1024,
                    int          PrefBufSize=65536);

    ~tni34_ATextFile();

    int ReadLine(const char * & buf);
    int WriteLine(const char * buf, long & BytesWritten);
    int WriteLine(const char * buf);
    
  private:
    char * ReadBuffer;
    size_t SizeOfReadBuffer;
    size_t CurrentPositionInReadBuffer;
    int    EndOfFileReached;
    int    AllLinesRead;

    int  ReAllocReadBufferTo(size_t NewSize);
    void MoveToStartOfBuffer(char * Buffer, size_t StartOfSource, size_t EndOfSource);
};


// -----------------------------------------------------------------------------
// constants and types
// -----------------------------------------------------------------------------

#if defined (WIN32)
typedef HANDLE tni34_APipeHandle;
const tni34_APipeHandle InvalidPipeHandle_ni34=INVALID_HANDLE_VALUE;
#elif defined (UNIX)
typedef int tni34_APipeHandle;
const tni34_APipeHandle InvalidPipeHandle_ni34=-1;
#endif


// -----------------------------------------------------------------------------
// class tni34_APipe
// -----------------------------------------------------------------------------

class tni34_APipe:public tni34_ADataEntry
{
  public:
    enum ADirection {In, Out};

    tni34_APipe(ADirection Dir,
                int        CreateOnOpen=1,
                int        OpenNow=1,
                int        PrefBufSize=65536,
                int        MaxNumberOfRetries=1000);  //creates a temporary named pipe

    tni34_APipe(const char * name,
                ADirection   Dir,
                int          CreateOnOpen=1,
                int          OpenNow=1,
                int          PrefBufSize=65536);

    ~tni34_APipe();
    
    int IsAPipe();

    int Exists();
    int Check();
    int Create();
    int Open();
    int Read(void *buf, long BytesToRead, long & BytesRead);
    int Write(const void *buf, long BytesToWrite, long & BytesWritten);
    int IsDataAvailable();
    int Close();
    int Remove();
    int BreakAndRemovePipeIfOpened();
 
  private:
    ADirection DirectionOfPipe;
    int PrefferedBufferSize;
    int Created;

    tni34_APipeHandle HandleForPipe;

    int PureQuietCreate(const char * name);
    int PureCreate();
    int PureOpen();
    int PureQuietClose();
};


// -----------------------------------------------------------------------------
// class tni34_AVirtualPipe
// -----------------------------------------------------------------------------

class tni34_AVirtualPipe:public tni34_ADataEntry //TODO: usage of common base class with tni34_AVirtualFile
{
  public:
    enum ADirection {In, Out};
    enum AOpenType {OpenNow, DoNotOpenNow};

    tni34_AVirtualPipe(const char * name,
                       ADirection   Dir,
                       AOpenType    OpenFlag=OpenNow);

    ~tni34_AVirtualPipe();
    
    int IsAPipe();

    int Check();
    int Create();

    int Open();
    int Read(void *buf, long BytesToRead, long & BytesRead);
    int Write(const void *buf, long BytesToWrite, long & BytesWritten);
    int Close();
    int Remove();
    
  private:
    ADirection DirectionOfPipe;

    tsp_int4 HandleForPipe;
    int      HandleIsValid;
};


// -----------------------------------------------------------------------------
// inline functions
// -----------------------------------------------------------------------------

inline const char * tni34_AFileSystemEntry::GiveName() const
{
    return Name;
}

inline int tni34_AFileSystemEntry::IsGood() const
{
    return IsGoodFlag;
}

inline int tni34_ADataEntry::IsOpen()
{
    return IsOpenFlag;
}

#endif
