/*!*****************************************************************************
  module:       vni33.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: ni layer
  description:  implementation of classes for process handling
  last change:  1999-10-04 15:33
  version:      7.2
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include"hni33.h"
#include"hni35.h"

#include<string.h>

#include<ctype.h>
#include<stdio.h>
#if defined (WIN32)
#elif defined (UNIX)
#include<stdlib.h>
#include<unistd.h>
#include<sys/wait.h>
#include<fcntl.h>
#include<sys/types.h>
#include<signal.h>
#include<errno.h>
#else
    #error hni33.cpp only coded for WIN32 and UNIX (Define one of them.)
#endif

// -----------------------------------------------------------------------------
// private function: IsWhiteSpace
//
// description : decides that the argument is a white space character or not
// arguments   : c [in] character to check
// return value: 1 if c is a white space, 0 if it is not
// -----------------------------------------------------------------------------

int IsWhiteSpace(char c)
{
  return isspace((int)c);
}


// -----------------------------------------------------------------------------
// implementation of class tni33_ABackgroundProcess
// -----------------------------------------------------------------------------

tni33_ABackgroundProcess::tni33_ABackgroundProcess(const char        * Cmdline,
                                                   const tni34_AFile * StdIn,
                                                   const tni34_AFile * StdOut,
                                                   const int           AppendStdOut,
                                                   const tni34_AFile * StdErr,
                                                   const int           AppendStdErr)
{
    InitAndStartBackgroundProcess(Cmdline, StdIn, 1, StdOut, AppendStdOut, 1, StdErr, AppendStdErr, 1);
}

tni33_ABackgroundProcess::tni33_ABackgroundProcess(const char        * Cmdline,
                                                   const tni34_AFile * StdIn,
                                                   const tni34_APipe * StdOut,
                                                   const int           AppendStdOut,
                                                   const tni34_AFile * StdErr,
                                                   const int           AppendStdErr)
{
    InitAndStartBackgroundProcess(Cmdline, StdIn, 1, StdOut, AppendStdOut, 0, StdErr, AppendStdErr, 1);
}
    
tni33_ABackgroundProcess::tni33_ABackgroundProcess(const char        * Cmdline,
                                                   const tni34_APipe * StdIn,
                                                   const tni34_AFile * StdOut,
                                                   const int           AppendStdOut,
                                                   const tni34_AFile * StdErr,
                                                   const int           AppendStdErr)
{
    InitAndStartBackgroundProcess(Cmdline, StdIn, 0, StdOut, AppendStdOut, 1, StdErr, AppendStdErr, 1);
}
    
tni33_ABackgroundProcess::tni33_ABackgroundProcess(const char        * Cmdline,
                                                   const tni34_APipe * StdIn,
                                                   const tni34_APipe * StdOut,
                                                   const int           AppendStdOut,
                                                   const tni34_AFile * StdErr,
                                                   const int           AppendStdErr)
{
    InitAndStartBackgroundProcess(Cmdline, StdIn, 0, StdOut, AppendStdOut, 0, StdErr, AppendStdErr, 1);
}

void tni33_ABackgroundProcess::InitAndStartBackgroundProcess(const char             * Cmdline,
                                                             const tni34_ADataEntry * StdIn,
                                                             const int                StdInIsFile,
                                                             const tni34_ADataEntry * StdOut,
                                                             const int                AppendStdOut,
                                                             const int                StdOutIsFile,
                                                             const tni34_AFile      * StdErr,
                                                             const int                AppendStdErr,
                                                             const int                StdErrIsFile)
{
    Process=AInvalidProcessHandle_ni33;

    #if defined (WIN32)
    Thread=AInvalidProcessHandle_ni33;
    StandardInHandle=INVALID_HANDLE_VALUE;
    StandardOutHandle=INVALID_HANDLE_VALUE;
    StandardErrorHandle=INVALID_HANDLE_VALUE;
    
    OpenedStandardInHandle=0;
    OpenedStandardOutHandle=0;
    OpenedStandardErrorHandle=0;
    
    StandardInIsFile=StdInIsFile;
    StandardOutIsFile=StdOutIsFile;
    StandardErrorIsFile=StdErrIsFile;

    #elif defined (UNIX)
    #else
        #error  tni33_ABackgroundProcess::tni33_ABackgroundProcess() not coded for this operating system!
    #endif

    if(0==(CommandLine=new char[strlen(Cmdline)+1]))
        ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory);
    else
    {
        strcpy(CommandLine,Cmdline);
        
        #if defined (WIN32)
        SECURITY_ATTRIBUTES SecurityAttrib={sizeof(SECURITY_ATTRIBUTES), 0, TRUE};

        if(0!=StdIn)
        {
            StandardInHandle=CreateFile(StdIn->GiveName(),
                                        GENERIC_READ,
                                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                                        & SecurityAttrib,
                                        OPEN_EXISTING,
                                        FILE_ATTRIBUTE_NORMAL,
                                        NULL);

            OpenedStandardInHandle=1;
        }
        else
            StandardInHandle=GetStdHandle(STD_INPUT_HANDLE);

        if(INVALID_HANDLE_VALUE==StandardInHandle)
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
        else
        {
            if(0!=StdOut)
            {
                StandardOutHandle=CreateFile(StdOut->GiveName(),
                                             GENERIC_WRITE,
                                             FILE_SHARE_READ | FILE_SHARE_WRITE,
                                             & SecurityAttrib,
                                             AppendStdOut==0 ?  CREATE_ALWAYS : OPEN_ALWAYS,
                                             FILE_ATTRIBUTE_NORMAL,
                                             NULL);

                OpenedStandardOutHandle=1;

                if(AppendStdOut &&
                   INVALID_HANDLE_VALUE!=StandardOutHandle &&
                   0xFFFFFFFF==SetFilePointer(StandardOutHandle, 0, 0, FILE_END))
                {
                    CloseHandle(StandardOutHandle);
                    StandardOutHandle=INVALID_HANDLE_VALUE;
                }
            }
            else
                StandardOutHandle=GetStdHandle(STD_OUTPUT_HANDLE);

            if(INVALID_HANDLE_VALUE==StandardOutHandle)
                ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
            else
            {
                if(0!=StdErr)
                {
                    StandardErrorHandle=CreateFile(StdErr->GiveName(),
                                                   GENERIC_WRITE,
                                                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                   & SecurityAttrib,
                                                   AppendStdOut==0 ?  CREATE_ALWAYS : OPEN_ALWAYS,
                                                   FILE_ATTRIBUTE_NORMAL,
                                                   NULL);

                    OpenedStandardErrorHandle=1;

                    if(AppendStdErr &&
                       INVALID_HANDLE_VALUE!=StandardErrorHandle &&
                       0xFFFFFFFF==SetFilePointer(StandardErrorHandle, 0, 0, FILE_END))
                    {
                        CloseHandle(StandardErrorHandle);
                        StandardErrorHandle=INVALID_HANDLE_VALUE;
                    }
                }
                else
                    StandardErrorHandle=GetStdHandle(STD_ERROR_HANDLE);

                if(INVALID_HANDLE_VALUE==StandardErrorHandle)
                    ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
                else
                {
                    STARTUPINFO si={DWORD(sizeof(STARTUPINFO)),
                                    LPTSTR(0),
                                    LPTSTR(0),
                                    LPTSTR(0),
                                    DWORD(0),
                                    DWORD(0),
                                    DWORD(0),
                                    DWORD(0),
                                    DWORD(0),
                                    DWORD(0),
                                    DWORD(0),
                                    STARTF_USESTDHANDLES,
                                    WORD(0),
                                    WORD(0),
                                    LPBYTE(0),
                                    StandardInHandle,
                                    StandardOutHandle,
                                    StandardErrorHandle};
                    PROCESS_INFORMATION pi;

                    if(!CreateProcess(0, CommandLine, 0, 0, TRUE, 0, 0, 0, &si, &pi))
                    {
                        ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess,CommandLine));
                        Process=AInvalidProcessHandle_ni33;
                    }
                    else
                    {
                        Process=pi.hProcess;
                        Thread=pi.hThread;
                    }        
                }
            }
        }
        #elif defined (UNIX)

        int argc=0,p=0;
      
        while(CommandLine[p]!=0 && IsWhiteSpace(CommandLine[p]))
            p++;

        while(CommandLine[p]!=0)
        {
            while(CommandLine[p]!=0 && !IsWhiteSpace(CommandLine[p]))
                p++;
            argc++;
            while(CommandLine[p]!=0 && IsWhiteSpace(CommandLine[p]))
                p++;
        }

        if(argc<=0)
        {
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess,CommandLine));
            Process=AInvalidProcessHandle_ni33;
        }
        else
        {
            char **Arguments=new char * [argc+1];

            if(0==Arguments)
            {
                ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory);
                Process=AInvalidProcessHandle_ni33;
            }
            else
            {
                int i=0;
                for(i=0;i<=argc;i++)
                    Arguments[i]=0;

                p=0;
                i=0;

                while(CommandLine[p]!=0 && IsWhiteSpace(CommandLine[p]))
                    p++;

                int p1,ErrorOccured=0;
                
                while(CommandLine[p]!=0 && !ErrorOccured)
                {
                    p1=p;

                    while(CommandLine[p1]!=0 && !IsWhiteSpace(CommandLine[p1]))
                        p1++;

                    if(0==(Arguments[i]=new char [p1-p+1])) 
                    {
                        ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory);
                        Process=AInvalidProcessHandle_ni33;
                        ErrorOccured=1;
                    }
                    else
                    {
                        strncpy(Arguments[i],CommandLine+p,p1-p);
                        Arguments[i++][p1-p]=0;
                        p=p1;
                        while(CommandLine[p]!=0 && IsWhiteSpace(CommandLine[p]))
                            p++;
                    }
                }

                if(!ErrorOccured)
                {
                    if((Process=fork())<0)
                    {
                        ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess,CommandLine));
                        Process=AInvalidProcessHandle_ni33;
                    }
                    else
                        if(Process==0) // following code is executed only by the child process
                        {
                            int GotAllFileDescriptors=1;
                            int Desc;

                            if(0!=StdIn)
                            {
                                fclose(stdin);
                                if(0!=(Desc=open(StdIn->GiveName(), O_RDONLY)))
                                {
                                    ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
                                    Process=AInvalidProcessHandle_ni33;
                                    GotAllFileDescriptors=0;
                                }
                            }

                            if(GotAllFileDescriptors && 0!=StdOut)
                            {
                                fclose(stdout);

                                if(AppendStdOut)
                                    Desc=open(StdOut->GiveName(), O_WRONLY|O_APPEND);
                                else
                                    Desc=open(StdOut->GiveName(), O_WRONLY|O_TRUNC); 

                                if(1!=Desc)
                                {
                                    ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
                                    Process=AInvalidProcessHandle_ni33;
                                    GotAllFileDescriptors=0;
                                }
                            }
                            
                            if(GotAllFileDescriptors && 0!=StdErr)
                            {
                                fclose(stderr);

                                if(AppendStdErr)
                                    Desc=open(StdErr->GiveName(), O_WRONLY|O_APPEND);
                                else
                                    Desc=open(StdErr->GiveName(), O_WRONLY|O_TRUNC); 

                                if(2!=Desc)
                                {
                                    ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess, CommandLine));
                                    Process=AInvalidProcessHandle_ni33;
                                    GotAllFileDescriptors=0;
                                }
                            }
                        
                            if(GotAllFileDescriptors)
                            {
                                char *ProgramName=Arguments[0];
                                
                                execvp(ProgramName,Arguments);

                                //coming to this point means, that an error occurred. Report the reason and end the child-process.
                                ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotCreateChildProcess,CommandLine));
                                exit(1);
                            }
                            else
                                exit(1); // could not change StdIn, StdOut or StdErr for the child as requested, error was allready reported
                        }
                }        
                        
                for(i=0;i<argc;i++)
                    if(Arguments[i]!=0)
                        delete [] (Arguments[i]);

                delete [] Arguments;
            }
        }
        #else
            #error  tni33_ABackgroundProcess::tni33_ABackgroundProcess not coded for this operating system!
        #endif
    }

    if(AInvalidProcessHandle_ni33==Process)
        State=NotStarted;
    else
        State=Started;

    if(State!=Started)          // do not block the file handles if the process could not be started
        FreeFileHandles();
}

void tni33_ABackgroundProcess::FreeFileHandles()
{
    #if defined (WIN32)
    if(OpenedStandardInHandle && INVALID_HANDLE_VALUE!=StandardInHandle)    //close StandardInHandle also if it is a pipe (we get here only if the processes is dead already)
    {
        CloseHandle(StandardInHandle);
        StandardInHandle=INVALID_HANDLE_VALUE; // don't close it again
    }

    if(StandardOutIsFile && OpenedStandardOutHandle && INVALID_HANDLE_VALUE!=StandardOutHandle)
    {
        CloseHandle(StandardOutHandle);
        StandardOutHandle=INVALID_HANDLE_VALUE; // don't close it again
    }

    if(StandardErrorIsFile && OpenedStandardErrorHandle && INVALID_HANDLE_VALUE!=StandardErrorHandle)
    {
        CloseHandle(StandardErrorHandle);
        StandardErrorHandle=INVALID_HANDLE_VALUE; // don't close it again
    }
    #elif defined (UNIX)
    #else
        #error tni33_ABackgroundProcess::FreeFileHandles() not coded for this operating system!
    #endif
}

tni33_ABackgroundProcess::~tni33_ABackgroundProcess()
{
    if(CommandLine!=0)
        delete [] CommandLine;

    #if defined (WIN32)
    if(AInvalidProcessHandle_ni33!=Process)
        CloseHandle(Process);

    if(AInvalidProcessHandle_ni33!=Thread)
        CloseHandle(Thread);

    if(OpenedStandardInHandle && INVALID_HANDLE_VALUE!=StandardInHandle)
        CloseHandle(StandardInHandle);

    if(OpenedStandardOutHandle && INVALID_HANDLE_VALUE!=StandardOutHandle)
        CloseHandle(StandardOutHandle);

    if(OpenedStandardErrorHandle && INVALID_HANDLE_VALUE!=StandardErrorHandle)
        CloseHandle(StandardErrorHandle);
    
    #elif defined (UNIX)
    #else
        #error tni33_ABackgroundProcess::~tni33_ABackgroundProcess() not coded for this operating system!
    #endif
}

int tni33_ABackgroundProcess::Stop()
{
    // All kill operations are brute force without cleanup !
    int rc=0;

    if(State==Started)
    {
        #if defined (WIN32)
        if(TerminateProcess(Process, 0))
        {
            State=Stopped;
            if(GetExitCodeProcess(Process, &RetCode))
            {
                rc=1;
                CloseHandle(Process);               //TODO: check return code of CloseHandle()
                FreeFileHandles();
                Process=AInvalidProcessHandle_ni33;
            }
            else
            {
                ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotGetReturnCode, CommandLine));
                RetCode=InvalidReturnCode_ni33;
            }
        }
        else
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotStopProcess, CommandLine));

        #elif defined(UNIX)
	    errno=0;
        if(kill(Process, SIGKILL))
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotStopProcess, CommandLine));
        else
        {
            State=Stopped;
            
            int RCProcess;

            errno=0;

            //if(Process==waitpid(Process,&RCProcess,WNOHANG))
            if(Process==waitpid(Process, &RCProcess, 0))        // we hope, that the kill signal will end the process so that the waitpid-function will return, when the OS has done it's work
            {
                rc=1;
                Process=AInvalidProcessHandle_ni33;
            
                if(WIFEXITED(RCProcess))
                    RetCode=WEXITSTATUS(RCProcess);
                else
                    if(WIFSIGNALED(RCProcess))
                        RetCode=InvalidReturnCode_ni33;
                    else
                        RetCode=InvalidReturnCode_ni33;
            }
            else
            {
                ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotGetReturnCode, CommandLine));
                RetCode=InvalidReturnCode_ni33;
            }
        }
        #else
            #error tni33_ABackgroundProcess::Stop() was only coded for WIN32 and UNIX (define one of them).
        #endif
    }

    return rc;
}

int tni33_ABackgroundProcess::Wait()
{
    int rc=0;

    if(State==Started)
    {
        #if defined (WIN32)
        if(WAIT_FAILED==WaitForSingleObject(Process,INFINITE))
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseWaitForProcessFailed, CommandLine));
        if(GetExitCodeProcess(Process,&RetCode))
        {
            rc=1;
            CloseHandle(Process);
            FreeFileHandles();
            Process=AInvalidProcessHandle_ni33;
            State=Stopped;
        }
        else
        {
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseCouldNotGetReturnCode, CommandLine));
            RetCode=InvalidReturnCode_ni33;
        }

        #elif defined (UNIX)
        int RCProcess=0;

        if(Process==waitpid(Process,&RCProcess,0))
        {
            rc=1;
            Process=AInvalidProcessHandle_ni33;
            State=Stopped;
            
            if(WIFEXITED(RCProcess))
                RetCode=WEXITSTATUS(RCProcess);
            else
                if(WIFSIGNALED(RCProcess))
                    RetCode=InvalidReturnCode_ni33;
                else
                    RetCode=InvalidReturnCode_ni33;
        }
        else
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseWaitForProcessFailed,CommandLine));
        #else
            #error tni33_ABackgroundProcess::Wait() not coded for this operating system!
        #endif
    }
    else
        ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseProcessNotStarted, CommandLine));

    return rc;
}

int tni33_ABackgroundProcess::IsRunning()
{
    int rc=0;

    if(State==Started)
    {
        #if defined (WIN32)
        if(!GetExitCodeProcess(Process,&RetCode))
            ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseWaitForProcessFailed,CommandLine));
        else
        {
            if(STILL_ACTIVE==RetCode)
                rc=1;
            else
            {
                CloseHandle(Process);
                FreeFileHandles();
                Process=AInvalidProcessHandle_ni33;
                State=Stopped;
            }
        }

        #elif defined (UNIX)
        int RCProcess,
            RCWaitpid=waitpid(Process, &RCProcess, WNOHANG);

        if(0==RCWaitpid)
            rc=1;
        else
            if(Process!=RCWaitpid)
                ErrorHandler_ni35->Report(tni35_AStandardError(tni35_AStandardError::CaseWaitForProcessFailed,CommandLine));
            else
            {
                Process=AInvalidProcessHandle_ni33;
                State=Stopped;

                if(WIFEXITED(RCProcess))
                    RetCode=WEXITSTATUS(RCProcess);
                else
                    if(WIFSIGNALED(RCProcess))
                        RetCode=InvalidReturnCode_ni33;
                    else
                        RetCode=InvalidReturnCode_ni33;
            }
        #else
            #error tni33_ABackgroundProcess::Wait() not coded for this operating system!
        #endif
    }

    return rc;
}

void tni33_ABackgroundProcess::WaitForMultipleBackgroundProcesses(int NoProcs,tni33_ABackgroundProcess **Procs)
{
    int i;
    for(i=0;i<NoProcs;i++)
        Procs[i]->Wait();
}


// -----------------------------------------------------------------------------
// implementation of class tni33_AProcess
// -----------------------------------------------------------------------------

tni33_AProcess::tni33_AProcess(const char        * Cmdline,
                               const tni34_AFile * StdIn,
                               const tni34_AFile * StdOut,
                               const int           AppendStdOut,
                               const tni34_AFile * StdErr,
                               const int           AppendStdErr)
    :P(Cmdline, StdIn, StdOut, AppendStdOut, StdErr, AppendStdErr)
{
    if(P.WasStarted())
        P.Wait();
}
