/*!**************************************************************************

  module:    vpr01Trace.h

  -------------------------------------------------------------------------

  responsible: BurkhardD, MarcoP

  special area:   CPCDrv | Client | Trace SQL Commands 
  description: Management of trace descriptors
               output to tracefile
               methods for ON-/OFF switch of interface runtime

  last changed: 2000-10-19  14:17
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr01Trace.h"
#include "hsp77.h"
#include "gpr08.h"
#include "vpr08SharedMem.h"
#include "vpr100.h"
#include "vpr07c.h"
#include "gpr03.h"
#include "gpr01.h"
#include <stdio.h>
/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/
 static tpr08_sharedMem ashared_mem;
 static tpr08_sharedMem *myshared_mem = &ashared_mem;
/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

/*!
  Function:     pr01TraceIsTrace

  see also: 

  Description:  checks if trace is on 
  
  Arguments:
    sqlrap [in] sqlra area

  return value: BOOLEAN
    <UL>
        <LI>TRUE: trace is on
        <LI>FALSE: trace is off
    </UL>
 */
externC boolean pr01TraceIsTrace(sqlratype *sqlrap)
{
  sqltatype *tap = (sqlrap) ? sqlrap->rasqltap : NULL;
  if (!tap)
    return false;
  if (tap->tatrout != CPR_TRACE_LONG && tap->tatrout != CPR_TRACE_MODLONG)
    return false;
  return true;
}

/*!
  Function:     pr01TracePrintf

  see also: 

  Description:  Prints a given string to the tracefile 
  
  Arguments:
    sqlrap [in|out] sqlra area
    szFmt  [in] format string (sprintf like)
    ...
    
  return value: void
 */
externC void pr01TracePrintf(sqlratype * sqlrap, const char* szFmt, ...)
{
  va_list ap;
  sqltatype *tap = sqlrap->rasqltap;
  tsp_int2 *ll = &tap->tastr80l;
  va_start (ap, szFmt);
  *ll = sp77vsprintf(tap->tastr80, sizeof(tap->tastr80)-1, szFmt, ap);
  p08vfwritetrace (sqlrap);
}

/*!
  Function:     pr01CloseTraceFile

  see also: 

  Description:  Close the tracefile 
  
  Arguments:
    sqlca [in|out] sqlca area
    sqlxa [in|out] sqlxa area

  return value: void
 */
void pr01CloseTraceFile (sqlcatype *my_sqlca, sqlxatype *my_sqlxa)
{
   struct SQLERROR sqlemp;   
   /*close tracefile*/
   p03tvfclosetrace (my_sqlca->sqlrap, &sqlemp);
   if (sqlemp.eprerr == cpr_p_ok)
   {
      /*change trace to not initialized*/
      my_sqlca->sqlrap->rasqltap->tatracety = 0;
      if(myshared_mem->getPart(myshared_mem)!=0)
      {
        myshared_mem->part_set_irt_tracestatus(myshared_mem->part,0);
      }
   }
   else
   {
     /*error during close tracefile -> write message to trace*/
     if (my_sqlca->sqlrap->rasqltap->tatraceno != 0)
        pr01TracePrintf(my_sqlca->sqlrap, "ERROR CLOSING TRACEFILE: %s",sqlemp.etext);
   }    
}

/*!
  Function:     pr01OpenTraceFile

  see also: 

  Description:  Open the tracefile 
  
  Arguments:
    sqlca [in|out] sqlca area
    sqlxa [in|out] sqlxa area

  return value: BOOLEAN
    <UL>
        <LI>TRUE: Processing succeeds
        <LI>FALSE: Error in occurs
    </UL>
 */
BOOLEAN pr01OpenTraceFile (sqlcatype *my_sqlca, sqlxatype *my_sqlxa)
{
/* tsp_opcodes openType;
   struct SQLERROR sqlemp;
   
   if (myshared_mem->getopen_for_append(myshared_mem))
   {
     openType = SP5VF_APPEND;
   }  
   else 
   {
     openType = SP5VF_WRITE;
     myshared_mem->setopen_for_append(myshared_mem,TRUE);
   }*/ 
      
   /*open tracefile*/
   p01xtracefilecheck  (my_sqlca, my_sqlxa);
   return TRUE;
}


/*!
  Function:     pr01traceParameterRefresh

  see also: 

  Description: ON OFF and Switch between different trace types 
               of interface runtime durring running process

 
 irt trace status:  short int           momentane Traceart der IRT
                                        IRT beendet              = -1
                                        nicht initialisiert      =  0
                                        trace aus                =  1
                                        kurzer trace an          =  2
                                        langer trace an          =  3
                                        kurzer moduletrace an    =  4
                                        langer moduletrace an    =  5
                                        traceoptions aus profile =  6

  Arguments:
    sqlca [in|out] sqlca area
    sqlxa [in|out] sqlxa area

  return value: void
*/
externC void pr01traceParameterRefresh(sqlcatype *my_sqlca, sqlxatype *my_sqlxa)
{ 
/* if shared memory not initialize -> initialize */
   if ( ! myshared_mem   
     || !myshared_mem->header )
   {
 		if(! tpr08_sharedMem__init_sharedMem(myshared_mem, SHM_NOTERASE, NULL))
   		{
   		    /* error alloc shared mem               */
   		    /* only if tracefile open (PTS 1106543) */
   		    if (my_sqlca->sqlrap->rasqltap->tatraceno != 0)
   		    { 
                my_sqlca->sqlemp->elzu = sp1ce_notok;
                memset(my_sqlca->sqlemp->etext, ' ', sizeof(my_sqlca->sqlemp->elzu));
                if (strlen(myshared_mem->getErrortxt(myshared_mem)) > 0)
                    memcpy((char*)my_sqlca->sqlemp->etext, myshared_mem->getErrortxt(myshared_mem) , strlen(myshared_mem->getErrortxt(myshared_mem)));
 
        	    p08runtimeerror (my_sqlca,my_sqlxa,cpr_trace_error); 
        	}
            return;
    	} 
    	/*initialize lastTracetyp with current Tracetyp*/
    	myshared_mem->setlastTracetyp(myshared_mem, my_sqlca->sqlrap->rasqltap->tatracety);
   }
   
   /* if traceswitchcount changed */
   if (myshared_mem->traceswitchcount_changed(myshared_mem)) 
   {  /* if trace_all switched on */
      if (myshared_mem->header_get_trace_all(myshared_mem->header) == 'y')
      {
         BOOLEAN erg = TRUE;
         /* save current trace typ */
      	 myshared_mem->setlastTracetyp(myshared_mem, my_sqlca->sqlrap->rasqltap->tatracety);
     	 if ( my_sqlca->sqlrap->rasqltap->tatracety == 0 )
         /*open tracefile if nessesary*/
         { 
            /* long trace on */
      	    my_sqlca->sqlrap->rasqltap->tatracety = 3;
    	    erg = pr01OpenTraceFile (my_sqlca, my_sqlxa);
         }
         else
         {
           /* long trace on */
      	   my_sqlca->sqlrap->rasqltap->tatracety = 3;
    	 }
      }
      else 
      {  
         if (myshared_mem->getPart(myshared_mem)==0)
      	 /* if adress of part unknown -> find part */
         {
           if ( myshared_mem->findPart(myshared_mem,myshared_mem->getProcessid(myshared_mem)) )
           {  
           		myshared_mem->part_set_version(myshared_mem->part, atoi(pr07RelNr())); 
           } 
           else
           {       	   
             /*set to last tracetype because maybe it was switch from traceall == 'y' to 'n'*/
         	 if (    myshared_mem->getlastTracetyp(myshared_mem) == 0
         	      && my_sqlca->sqlrap->rasqltap->tatracety != 0)
         	 /*switch from on to off -> close the tracefile*/
         	 {
               pr01CloseTraceFile (my_sqlca, my_sqlxa);
         	 }    
             my_sqlca->sqlrap->rasqltap->tatracety = myshared_mem->getlastTracetyp(myshared_mem);        	
             return;
           }
         }
          
         if (   my_sqlca->sqlrap->rasqltap->tatrout 
             != myshared_mem->part_get_tracetyp(myshared_mem->part))
         /*someone has changed for this process*/
         {  
         	if (myshared_mem->part_get_tracetyp(myshared_mem->part)== -2)
         	/* only info about trace status */
         	{
         		myshared_mem->part_set_irt_tracestatus(myshared_mem->part, my_sqlca->sqlrap->rasqltap->tatrout);
         	}
         	else
         	/* change trace */
         	{	
         	    BOOLEAN erg = TRUE;
         	    if (   myshared_mem->part_get_irt_tracestatus(myshared_mem->part) == 0
         	        && myshared_mem->part_get_tracetyp(myshared_mem->part) > 1)
         	    /*switch from off to on -> open tracefile*/
         	    {   
         	        /* change trace */
                    my_sqlca->sqlrap->rasqltap->tatracety = myshared_mem->part_get_tracetyp(myshared_mem->part);
                    myshared_mem->part_set_irt_tracestatus(myshared_mem->part, myshared_mem->part_get_tracetyp(myshared_mem->part));
                    erg = pr01OpenTraceFile (my_sqlca, my_sqlxa);
           	    }    
         	    else if (   myshared_mem->part_get_irt_tracestatus(myshared_mem->part) > 1
         	             && myshared_mem->part_get_tracetyp(myshared_mem->part) == 0)
         	    /*switch from on to off -> close the tracefile*/
         	    {
                    pr01CloseTraceFile (my_sqlca, my_sqlxa);
         	    }    
         	    else
         	    {
             	 	/* switch from one trace type to another trace type*/
                	my_sqlca->sqlrap->rasqltap->tatracety = myshared_mem->part_get_tracetyp(myshared_mem->part);
     				myshared_mem->part_set_irt_tracestatus(myshared_mem->part, myshared_mem->part_get_tracetyp(myshared_mem->part));
                }
            	
            	/* comment to tracefile */ 
            	if (my_sqlca->sqlrap->rasqltap->tatraceno !=	0)
            	{  
            		pr01TracePrintf(my_sqlca->sqlrap, "");
            		pr01TracePrintf(my_sqlca->sqlrap, "TRACE TYPE HAS BEEN CHANGED TO: %s",myshared_mem->tracetyp_tochar(myshared_mem, myshared_mem->part_get_tracetyp(myshared_mem->part)));
         		}
         	}
         	/*update traceswitch*/
            myshared_mem->part_set_lasttraceswitch(myshared_mem->part, myshared_mem->header_get_traceswitchcount(myshared_mem->header));
         }
      }	  
   }
   return;
}   
/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/


