/*!================================================================

  module:       vpr03Part.c

 -------------------------------------------------------------------

  responsible:  BurkhardD

  special area: Functions to modify a Part

  description:  

  see also:

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr03Part.h"
#include "hsp26.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

/*!
  Function: pr03PartFind

  description: Liefert die Adresse des in part_kind angegebenen Parts.
               Optimierte Version, die mit einem Part Cache arbeitet.
	       Dieser PartCache muss nach abholen des Auftrages mit 
	       pr03PartInitCache initialisiert worden sein.
               
	       
  arguments:
      sqlra        [in]  Request Area.
      part_kind    [in]  Im Auftrag (PartCache) zu suchender Part.

  return:          Pointer auf den Part oder NULL fuer nicht gefunden.
 */

tsp1_part *pr03PartFind(sqlratype *sqlra, tsp1_part_kind_Param part_kind)
{
  tsp1_segment   *segp= (sqlra) ? (tsp1_segment*)sqlra->rasegptr : NULL;

  tsp1_part *part_ptr = NULL;

  if (!segp)
    return NULL;

  if (segp->variant.C_1.sp1s_segm_header_F.variant.C_1.sp1s_no_of_parts_F > 0) {
    if (part_kind < PartCache_mxpr00) { 
      part_ptr = sqlra->rasegmpartptr[part_kind];
    }
    else {
      /* part_kind is to big to fit in the cache */
      /* try to find directly */
      s26find_part(segp, part_kind, &part_ptr);
    }
  }
  return(part_ptr);
}

/*!
  Function: pr03PartFindDirect 

  description: Liefert die Adresse des in part_kind angegebenen Parts.
               Der Part wird im Gegensatz zu pr03PartFind direkt
	       im Auftrag gesucht.

  arguments:
      sqlra        [in]  Request Area.
      part_kind    [in]  Im Auftrag zu suchender Part.

  return:          Pointer auf den Part oder NULL fuer nicht gefunden.
 */

tsp1_part *pr03PartFindDirect(sqlratype *sqlra, tsp1_part_kind_Param part_kind)
{
  tsp1_segment   *segp= (sqlra) ? (tsp1_segment*)sqlra->rasegptr : NULL;
  tsp1_part* part_ptr = NULL;

  if (!segp)
    return NULL;

  if (segp->variant.C_1.sp1s_segm_header_F.variant.C_1.sp1s_no_of_parts_F > 0) {
    s26find_part(segp, part_kind, &part_ptr);
  }
  return(part_ptr);
}

/*!
  Function: pr03PartInitCache 

  description: Initialisiert den PartCache.
               Der PartCache ist ein Array von Pointern, mit Anfangsaddressen
	       der Parts im Auftrag.
	       sp1p_part_kind wird als Index in den PartCache benutzt.

  arguments:
      sqlra        [in]  Request Area. Nach dem Receive vom Kern.

  return:          ---
 */

void pr03PartInitCache(sqlratype *sqlra)
{
  tsp1_segment   *segp= (sqlra) ? (tsp1_segment*)sqlra->rasegptr : NULL;
  tsp1_part* part_ptr;
  tsp00_Int2 part_no;

  memset(sqlra->rasegmpartptr, 0, sizeof(sqlra->rasegmpartptr));  

  if (!segp)
    return;

  part_ptr = &segp->variant.C_1.sp1p_part_F;

  part_no=segp->variant.C_1.sp1s_segm_header_F.variant.C_1.sp1s_no_of_parts_F;
  while(part_ptr) {
    tsp00_Int2 part_kind = part_ptr->variant.C_2.sp1p_part_kind_F;
    if (part_kind < PartCache_mxpr00 ) {
      sqlra->rasegmpartptr[part_kind] = part_ptr;
    }    
    if (part_no > 1) {
      /* s26nextpart crashes if there is no nextpart available */
      s26nextpart(&part_ptr);
      part_no--;
    }
    else
      break;    
  }
}

void pr03PartPut(tsp1_part *partPtr, void *pBuf, tsp00_Int4 cbLen)
{
  memcpy(partPtr->variant.C_1.sp1p_buf_F, pBuf, cbLen);
  partPtr->sp1p_part_header.sp1p_buf_len = cbLen;
}
