/* @lastChanged: "1998-01-27 15:30"
 *
 * @filename:   vsp77.c
 * @purpose:    "sprintf functions"
 * @release:    7.1.0.0
 * @see:        "http://www.disney.com/"
 *
 * @copyright:  "Copyright by   SAP AG, 1997"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

#include <ctype.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>

#include "hsp77.h"

#define BETWEEN_MSP77(low, val, up) ((low <= val) && (val <= up))
#define MAX_MSP77(a,b)              ((a) > (b) ? (a) : (b))
#define MIN_MSP77(a,b)              ((a) < (b) ? (a) : (b))
#undef ROUTINE_DBG_MSP00
#define ROUTINE_DBG_MSP00(routine)

#ifdef TEST
#include <stdio.h>
FILE           *stdTrace;
#endif

typedef int cbool;
#define TRUE  1
#define FALSE 0
#define UNDEF (-1)

/*************************************
 *  U N I C O D E   ( D U M M Y )    *
 *************************************/

typedef void* t_memsetfunction (void*, const void*, int);
typedef short t_unichar;
typedef const t_unichar* t_cunistr;

/*
 * The following should be defined in a separate file (vjk11c in 6.2)
 *     the variable have different values depending on unicode
 *
 * The values here are for ASCII only
 */

static int local_unicode = 0;
static int local_charsize = 1;
static int local_chars [256];
static void* (*local_memcpy) (void*, const void*, size_t) = memcpy;
static void* (*local_memset) (void*, int, size_t) = memset;
static void* (*local_memchr) (const void*, int, size_t) = memchr;
static size_t (*local_strlen) (const char*) = strlen;
/* the following are called only in unicode mode */
static int (*local_asciicpy) (char*, int, t_cunistr, int) = NULL;
static int (*local_unicpy) (t_unichar*, const char*, int) = NULL;

/*
 * end unicode
 */

typedef struct t_printf_format {
    int             width;
    int             precision;
    int             argSize;
    int             argChar;
    struct {
        unsigned int    justifyLeft;
        unsigned int    sign;
        unsigned int    space;
        unsigned int    alternate;
        unsigned int    zeros;
        unsigned int    byteScale;
    } flags;
    /* unicode dependent */
    unsigned int    genUnicode;
    void           *(*callMemset) (void*, int, size_t);
    int             charSize;
} t_printf_format;

/*----------------------------------------*/

static int
sp77_putPadded (
    char               *buf,
    int                 bufLen,
    const char         *arg,
    int                 moveLen,
    t_printf_format    *printfFormat)
{
    /* ROUTINE_DBG_MSP00("sp77_putPadded"); */
    cbool           padLeft = FALSE;
    cbool           padRight = FALSE;
    int             padLen;
    int             padChar;
    int             blankChar = ' ';
    int             zeroChar  = '0';

    bufLen /= printfFormat->charSize;  /* convert to logical bufLen */
    padLen = MAX_MSP77 (0, printfFormat->width - moveLen);
    if (padLen > 0) {
        if (printfFormat->flags.justifyLeft)
            padRight = TRUE;
        else
            padLeft = TRUE;
        if (printfFormat->genUnicode) {
            blankChar = local_chars[' '];
            zeroChar = local_chars['0'];
        }
    }
    if (padLeft) {
        padLen = MIN_MSP77 (padLen, bufLen);
        if (printfFormat->flags.zeros)
            padChar = zeroChar;
        else
            padChar = blankChar;
        printfFormat->callMemset (buf, padChar, padLen);
        buf += padLen * printfFormat->charSize;
        bufLen -= padLen;
        if (bufLen <= 0)
            return padLen;
    }
    moveLen = MIN_MSP77 (moveLen, bufLen);
    if (printfFormat->genUnicode)
        local_unicpy ((t_unichar *) buf, arg, moveLen);
    else
        memcpy (buf, arg, moveLen);
    if (padRight && (bufLen > 0)) {
        buf += moveLen * printfFormat->charSize;
        bufLen -= moveLen;
        padLen = MIN_MSP77 (padLen, bufLen);
        printfFormat->callMemset (buf, blankChar, padLen);
    }
    return padLen + moveLen;
}

/*----------------------------------------*/

static int
sp77_putPaddedUnicode (
    char *buf,
    int bufLen,
    const char * arg,
    int moveLen,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_putPaddedUnicode"); */
    cbool           padLeft = FALSE;
    cbool           padRight = FALSE;
    int             padLen;
    int             padChar;
    int             blankChar = ' ';
    int             zeroChar  = '0';
    int             charsWritten;
    int             bytesWritten;

    bufLen /= printfFormat->charSize;  /* convert to logical bufLen */
    padLen = MAX_MSP77 (0, printfFormat->width - moveLen);
    if (padLen > 0) {
        if (printfFormat->flags.justifyLeft)
            padRight = TRUE;
        else
            padLeft = TRUE;
        if (printfFormat->genUnicode) {
            blankChar = local_chars[' '];
            zeroChar = local_chars['0'];
        }
    }
    if (padLeft) {
        padLen = MIN_MSP77 (padLen, bufLen);
        if (printfFormat->flags.zeros)
            padChar = zeroChar;
        else
            padChar = blankChar;
        printfFormat->callMemset (buf, padChar, padLen);
        buf += padLen * printfFormat->charSize;
        bufLen -= padLen;
        if (bufLen <= 0)
            return padLen;
    }
    moveLen = MIN_MSP77 (moveLen, bufLen);
    if (printfFormat->genUnicode) {
        local_memcpy (buf, arg, moveLen);
        charsWritten = moveLen;
        bytesWritten = moveLen * printfFormat->charSize;
    }
    else {
        charsWritten = local_asciicpy (buf, bufLen * printfFormat->charSize,
            (t_cunistr) arg, moveLen);
        bytesWritten = charsWritten;
    }
    if (padRight && (bufLen > 0)) {
        buf += bytesWritten;
        bufLen -= moveLen;
        padLen = MIN_MSP77 (padLen, bufLen);
        printfFormat->callMemset (buf, blankChar, padLen);
    }
    return padLen + charsWritten;
}

/*----------------------------------------*/

static const char lowDigits[17] =
   {'0', '1', '2', '3', '4',
   '5', '6', '7', '8', '9',
   'a', 'b', 'c', 'd', 'e', 'f', 'x'};
static const char upDigits[17] =
   {'0', '1', '2', '3', '4',
   '5', '6', '7', '8', '9',
   'A', 'B', 'C', 'D', 'E', 'F', 'X'};

/*----------------------------------------*/

static int
sp77_unsignedConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_unsignedConv"); */
    unsigned long   arg;
    char            tmpBuf[20];
    int             count = 0;
    unsigned        base;
    int             ownWritten = 0;
    t_unichar   *uniBuf;

    const char *       digits = lowDigits;

    switch (printfFormat->argSize) {
        case 'l':
        case 'L':
            arg = va_arg (*args, unsigned long);

            break;
        case 'h':
            arg = va_arg (*args, unsigned short);

            break;
        default:
            arg = va_arg (*args, unsigned int);

            break;
    }
    switch (printfFormat->argChar) {
        case 'o':
            base = 8;
            break;
        case 'u':
            base = 10;
            break;
        case 'x':
            base = 16;
            break;
        case 'X':
            base = 16;
            digits = upDigits;
            break;
        default:
            base = 10;
            break;
    }
    while (arg > 0) {
        tmpBuf[sizeof (tmpBuf) - count - 1] = digits[arg % base];
        ++count;
        arg /= base;
    }
    if (printfFormat->precision == UNDEF)
        printfFormat->precision = 1;
    while (count < printfFormat->precision) {
        tmpBuf[sizeof (tmpBuf) - count - 1] = '0';
        ++count;
    }
    if (printfFormat->flags.alternate) {
        switch (printfFormat->argChar) {
            case 'x':
            case 'X':
                if (printfFormat->flags.zeros
                    && !printfFormat->flags.justifyLeft
                    && (bufLen >= 2)) {
                    if (printfFormat->genUnicode) {
                        uniBuf = (t_unichar *) buf;
                        uniBuf[0] = local_chars['0'];
                        uniBuf[1] = local_chars[digits[16]];
                        buf += 2 * printfFormat->charSize;
                        bufLen -= 2 * printfFormat->charSize;
                    }
                    else {
                        buf[0] = '0';
                        buf[1] = digits[16];
                        buf += 2;
                        bufLen -= 2;
                    }
                    printfFormat->width -= 2;
                    ownWritten = 2;
                }
                else {
                    tmpBuf[sizeof (tmpBuf) - count - 1] = digits[16];
                    tmpBuf[sizeof (tmpBuf) - count - 2] = '0';
                    count += 2;
                }
                break;
            case 'o':
                if (tmpBuf[sizeof (tmpBuf) - count] != '0') {
                    tmpBuf[sizeof (tmpBuf) - count - 1] = '0';
                    ++count;
                }
                break;
            default:
                /* nothing required */
                break;
        }
    }
    return sp77_putPadded (buf, bufLen, tmpBuf + sizeof (tmpBuf) - count,
        count, printfFormat) + ownWritten;
}

/*----------------------------------------*/

static int
sp77_signedConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_signedConv"); */
    long            arg;
    char            signChar;
    char            tmpBuf[20];
    int             count = 0;
    int             ownWritten = 0;

    switch (printfFormat->argSize) {
        case 'l':
        case 'L':
            arg = va_arg (*args, long);

            break;
        case 'h':
            arg = va_arg (*args, short);

            break;
        default:
            arg = va_arg (*args, int);

            break;
    }
    if (arg < 0) {
        signChar = '-';
        arg = -arg;
    }
    else if (printfFormat->flags.sign)
        signChar = '+';
    else if (printfFormat->flags.space)
        signChar = ' ';
    else
        signChar = '\0';
    while (arg > 0) {
        tmpBuf[sizeof (tmpBuf) - count - 1] = (char) ((arg % 10) + '0');
        ++count;
        arg /= 10;
    }
    if (printfFormat->precision == UNDEF)
        printfFormat->precision = 1;
    while (count < printfFormat->precision) {
        tmpBuf[sizeof (tmpBuf) - count - 1] = '0';
        ++count;
    }
    if (signChar != '\0') {
        if (printfFormat->flags.zeros
            && !printfFormat->flags.justifyLeft
            && (bufLen >= 2)) {
            buf[0] = signChar;
            ++buf;
            --bufLen;
            --printfFormat->width;
            ownWritten = 1;
        }
        else {
            tmpBuf[sizeof (tmpBuf) - count - 1] = signChar;
            ++count;
        }
    }
    return sp77_putPadded (buf, bufLen, tmpBuf + sizeof (tmpBuf) - count,
        count, printfFormat) + ownWritten;
}

/*----------------------------------------*/

/*
 * 'F' and 'f' format do not differ at all...
 * Precision specifies the number of digits after the decimal point
 * Decimal point is not written if precision is 0
 */
static int sp77printDouble ( char  *tmpBuf,
        	  int    tmpBuflen,
		  int    precision,
		  double arg)
{
    int count = 0;
    register double digitValue;
    register double roundedArg;
    int    precDigits;

    roundedArg = arg;
    for ( precDigits = 0; precDigits < precision; precDigits++ )
    {
	roundedArg *= 10.;
    }
    roundedArg = floor(roundedArg+0.5);

    for ( precDigits = precision; precDigits > 0; precDigits-- )
    {
	digitValue = roundedArg;
	roundedArg = floor(roundedArg/10.);
	digitValue -= 10. * roundedArg;
	if ( digitValue < 0. )
	{
	    digitValue = 0.;
	}
        tmpBuf[tmpBuflen - count - 1] = (char) (((long)digitValue) + '0');
        ++count;
    }

    if ( precision > 0 )
    {
	tmpBuf[tmpBuflen - count - 1] = '.';
        ++count;
    }

    do
    {
	digitValue = roundedArg;
	roundedArg = floor(roundedArg/10.);
	digitValue -= 10. * roundedArg;
        tmpBuf[tmpBuflen - count - 1] = (char) (((long)digitValue) + '0');
        ++count;
    } while ( roundedArg > 0 );

    return count;
}

/*
 * 'E' and 'e' format differ only in 'ExponentChar'
 * Precision specifies the number of digits after the decimal point
 * Decimal point is not written if precision is 0
 */
int sp77printExponentDouble ( char  *tmpBuf,
        	  int    tmpBuflen,
		  int    precision,
		  double arg,
		  const char   ExponentChar )
{
    int count = 0;
    register double digitValue;
    register double roundedArg;
    register double shiftedArg;
    register double multiplier = 1.;
    int    precDigits;
    int    Exponent = 0;
    int    NegativExponent = 0;

    shiftedArg = arg;
    /*
     * Shift until shiftedArg is in interval ]1,10[
     */
    while ( shiftedArg >= 10. )
    {
	Exponent++;
	shiftedArg /= 10.;
    }
    while ( shiftedArg < 1. )
    {
	Exponent--;
	shiftedArg *= 10.;
    }

    /*
     * Now find the precision multiplier
     */
    for ( precDigits = 0; precDigits < precision; precDigits++ )
    {
	multiplier *= 10.;
    }

    /*
     * Now round the value to precision
     */
    roundedArg = floor((shiftedArg*multiplier)+0.5);
    /*
     * Due to rounding effect, now the value could be above 10
     */
    if ( ( roundedArg / multiplier ) >= 10. )
    {
	Exponent++;
        shiftedArg /= 10.;
	multiplier *= 10.;
        roundedArg = floor((shiftedArg*multiplier)+0.5);
    }

    /*
     * Show exponent
     */
    if ( Exponent < 0 )
    {
         Exponent = -Exponent;
	 NegativExponent = 1;
    }
    if ( Exponent < 10 )
    {
        tmpBuf[tmpBuflen - count - 1] = Exponent + '0';
        ++count;
        tmpBuf[tmpBuflen - count - 1] = '0';
        ++count;
    }
    else
    {
	do
        {
	    tmpBuf[tmpBuflen - count - 1] = (Exponent%10) + '0';
            ++count;
	    Exponent /= 10;
	} while( Exponent > 0 );
    }

    if ( NegativExponent )
    {
         tmpBuf[tmpBuflen - count - 1] = '-';
         ++count;
    }
    else
    {
         tmpBuf[tmpBuflen - count - 1] = '+';
         ++count;
    }

    tmpBuf[tmpBuflen - count - 1] = ExponentChar;
    ++count;

    /*
     * If precision is 0, no decimal point is shown
     */
    if ( precision <= 0 )
    {
        tmpBuf[tmpBuflen - count - 1] = (char) (((long)roundedArg) + '0');
        ++count;
    }
    else
    {

	for ( precDigits = precision; precDigits > 0; precDigits-- )
	{
	    digitValue = roundedArg;
	    roundedArg = floor(roundedArg/10.);
	    digitValue -= 10. * roundedArg;
	    tmpBuf[tmpBuflen - count - 1] = (char) (((long)digitValue) + '0');
	    ++count;
	}

	tmpBuf[tmpBuflen - count - 1] = '.';
	++count;

	do
	{
	    digitValue = roundedArg;
	    roundedArg = floor(roundedArg/10.);
	    digitValue -= 10. * roundedArg;
	    tmpBuf[tmpBuflen - count - 1] = (char) (((long)digitValue) + '0');
	    ++count;
	} while ( roundedArg > 0 );

    }

    return count;
}

/*
 * G format uses 'E' format if Exponet < -4 or >= precision
 *   else it uses 'F' format
 *
 * Precision specifies the number of significant digits!!!
 */
static int
sp77printSomeExponentDouble ( char  *tmpBuf,
        	  int    tmpBuflen,
		  int    precision,
		  double arg,
		  const char   ExponentChar )
{
    register double roundedArg;
    register double shiftedArg;
    register double multiplier = 1.;
    int    precDigits;
    int    Exponent = 0;

    shiftedArg = arg;
    /*
     * Shift until shiftedArg is in interval ]1,10[
     */
    while ( shiftedArg >= 10. )
    {
	Exponent++;
	shiftedArg /= 10.;
    }
    while ( shiftedArg < 1. )
    {
	Exponent--;
	shiftedArg *= 10.;
    }

    /*
     * Now find the precision multiplier
     */
    for ( precDigits = 0; precDigits < precision; precDigits++ )
    {
	multiplier *= 10.;
    }

    /*
     * Now round the value to precision
     */
    roundedArg = floor((shiftedArg*multiplier)+0.5);
    /*
     * Due to rounding effect, now the value could be above 10
     */
    if ( ( roundedArg / multiplier ) >= 10. )
    {
	Exponent++;
	/* calculation is redone in calls below... */
    }

    /*
     * Always at least one digit is significant
     */
    if ( precision == 0 )
    {
	precision = 1;
    }

    if ( Exponent < -4 || Exponent >= precision )
    {
	return sp77printExponentDouble ( tmpBuf, tmpBuflen,
					 precision - 1, arg,
				 	 (const char)(ExponentChar == 'G' ? 'E' : 'e') );
    }
    else
    {
	return sp77printDouble ( tmpBuf, tmpBuflen,
				precision - (Exponent + 1), arg );
    }
}

static int
sp77_doubleConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    double          arg;
    char            signChar;
    char            tmpBuf[250];
    int             count = 0;
    int             ownWritten = 0;

    arg = va_arg (*args, double);
    if (arg < 0) {
        signChar = '-';
        arg = -arg;
    }
    else if (printfFormat->flags.sign)
        signChar = '+';
    else if (printfFormat->flags.space)
        signChar = ' ';
    else
        signChar = '\0';
    if (signChar != '\0') {
        if (printfFormat->flags.zeros
            && !printfFormat->flags.justifyLeft
            && (bufLen >= 2)) {
            buf[0] = signChar;
            ++buf;
            --bufLen;
            --printfFormat->width;
            ownWritten = 1;
        }
        else {
            tmpBuf[0] = signChar;
            ++count;
        }
    }
    if ( printfFormat->argChar == 'f' || printfFormat->argChar == 'F' )
    {
    	count += sp77printDouble (tmpBuf + count,
        		sizeof (tmpBuf) - count,
			printfFormat->precision,
			arg);
    }
    else if ( printfFormat->argChar == 'e' || printfFormat->argChar == 'E' )
    {
    	count += sp77printExponentDouble ( tmpBuf + count,
        		sizeof (tmpBuf) - count,
			printfFormat->precision,
        		arg,
			(char)printfFormat->argChar );
    }
    else if ( printfFormat->argChar == 'g' || printfFormat->argChar == 'G' )
    {
    	count += sp77printSomeExponentDouble ( tmpBuf + count,
        		sizeof (tmpBuf) - count,
			printfFormat->precision,
        		arg,
			(char)printfFormat->argChar );
    }
    return sp77_putPadded (buf, bufLen, tmpBuf + sizeof(tmpBuf) - count,
        count, printfFormat) + ownWritten;
}

/*----------------------------------------*/

static int
sp77_cConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_cConv"); */
    char            arg;

    arg = (char) va_arg (*args, int);

    return sp77_putPadded (buf, bufLen, &arg, 1, printfFormat);
}

/*----------------------------------------*/

static int
sp77_pConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_pConv"); */
    printfFormat->argChar = 'X';
    return sp77_unsignedConv (buf, bufLen, args, printfFormat);
}

/*----------------------------------------*/

static const char * c_nullString = "(null)";
static const char c_nullStringUnicode[] =
{'\0', '(', '\0', 'n',
    '\0', 'u', '\0', 'l', '\0', 'l', '\0', ')', '\0', '\0'};

/*----------------------------------------*/

static int
sp77_sConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_sConv"); */
    const char *       argStr;
    int             moveLen;
    const char *       zeroFound;

    /* the argument is always a single tjk2_byte character string */
    argStr = va_arg (*args, char *);

    if (argStr == NULL)
        argStr = c_nullString;
    /* calculate real len */
    if (printfFormat->precision >= 0) {
        zeroFound = memchr (argStr, '\0', printfFormat->precision);
        if (zeroFound != NULL)
            moveLen = zeroFound - argStr;
        else
            moveLen = printfFormat->precision;
    }
    else
        moveLen = strlen (argStr);
    if (printfFormat->width <= 0)
        printfFormat->width = moveLen;
    return sp77_putPadded (buf, bufLen, argStr, moveLen, printfFormat);
}

/*----------------------------------------*/

static int
sp77_SConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_SConv"); */
    const char *       argStr;
    int             moveLen;
    const char *       zeroFound;

    if (!local_unicode)
        return sp77_sConv (buf, bufLen, args, printfFormat);
    /* in unicode mode, the argument is a unicode string    */
    /* otherwise, the arg is a single tjk2_byte character string */
    argStr = va_arg (*args, char *);
    if (argStr == NULL) {
        if (local_unicode)
            argStr = (const char *) c_nullStringUnicode;
        else
            argStr = c_nullString;
    }
    /* calculate real len */
    if (printfFormat->precision >= 0) {
        zeroFound = local_memchr (argStr, local_chars['0'], printfFormat->precision);
        if (zeroFound != NULL)
            moveLen = (zeroFound - argStr) / local_charsize;
        else
            moveLen = printfFormat->precision;
    }
    else
        moveLen = local_strlen (argStr);
    if (printfFormat->width <= 0)
        printfFormat->width = moveLen;
    return sp77_putPaddedUnicode (buf, bufLen, argStr, moveLen, printfFormat);
}

/*----------------------------------------*/

static int
sp77_putHexPadded (
    char               *buf,
    int                 bufLen,
    const char         *arg,
    int                 moveLen,
    t_printf_format    *printfFormat)
{
    /* ROUTINE_DBG_MSP00("sp77_putHexPadded"); */
    cbool           padLeft = FALSE;
    cbool           padRight = FALSE;
    int             padLen;
    int             padChar = ' ';
    int             i;
    const char *       digits = lowDigits;

    if (isupper (printfFormat->argChar)) {
        digits = upDigits;
    }
    padLen = MAX_MSP77 (0, printfFormat->width - (moveLen * 2));
    if (padLen > 0) {
        if (printfFormat->flags.justifyLeft)
            padRight = TRUE;
        else
            padLeft = TRUE;
    }
    if (padLeft) {
        padLen = MIN_MSP77 (padLen, bufLen);
        printfFormat->callMemset (buf, padChar, padLen);
        buf += padLen * printfFormat->charSize;
        bufLen -= padLen;
        if (bufLen <= 0)
            return padLen;
    }
    moveLen = MIN_MSP77 (moveLen, bufLen / 2);
    for (i = 0; i < moveLen; ++i) {
        buf [0] = digits [arg [i] / 16];
        buf [1] = digits [arg [i] % 16];
        buf += 2;
    }
    if (padRight && (bufLen > 0)) {
        bufLen -= moveLen * 2;
        padLen = MIN_MSP77 (padLen, bufLen);
        printfFormat->callMemset (buf, padChar, padLen);
    }
    return padLen + (moveLen * 2);
}

/*----------------------------------------*/

static int
sp77_hexConv (
    char *buf,
    int bufLen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_hexConv"); */
    const char *       argStr;
    int                moveLen;
    const char *       zeroFound;

    /* the argument is always a single tjk2_byte character string */
    argStr = va_arg (*args, char *);

    if (argStr == NULL)
        argStr = c_nullString;
    /* calculate real len */
    if (printfFormat->precision >= 0) {
        zeroFound = memchr (argStr, '\0', printfFormat->precision);
        if (zeroFound != NULL)
            moveLen = zeroFound - argStr;
        else
            moveLen = printfFormat->precision;
    }
    else
        moveLen = strlen (argStr);
    if (printfFormat->width <= 0)
        printfFormat->width = moveLen;
    return sp77_putHexPadded (buf, bufLen, argStr, moveLen, printfFormat);
}

/*----------------------------------------*/

int
sp77_parseWidth (
    const char **formatArg,
    va_list * args)
{
    /* ROUTINE_DBG_MSP00 ("sp77_parseWidth"); */
    const char     *format = *formatArg;
    int             result = 0;

    if (*format == '*') {
        ++(*formatArg);
        return va_arg (*args, int);
    }
    if (isdigit (*format)) {
        do {
            result = (result * 10) + (*format - '0');
            ++format;
            *formatArg = format;
        } while (isdigit (*format));
        return result;
    }
    else
        return 0;
}

/*----------------------------------------*/

const char     *
sp77_parseFormat (
    const char *format,
    t_printf_format * printfFormat,
    va_list * args)
{
    /* ROUTINE_DBG_MSP00 ("sp77_parseFormat"); */
    cbool           parsingFlags = TRUE;

    /*M2_ASSERT (*format == '%');*/
    printfFormat->width = UNDEF;
    printfFormat->precision = UNDEF;
    printfFormat->argSize = '\0';
    printfFormat->flags.justifyLeft = FALSE;
    printfFormat->flags.sign = FALSE;
    printfFormat->flags.space = FALSE;
    printfFormat->flags.alternate = FALSE;
    printfFormat->flags.zeros = FALSE;
    printfFormat->flags.byteScale = FALSE;
    while (parsingFlags) {
        ++format;
        switch (*format) {
            case '-':
                printfFormat->flags.justifyLeft = TRUE;
                break;
            case '+':
                printfFormat->flags.sign = TRUE;
                break;
            case ' ':
                printfFormat->flags.space = TRUE;
                break;
            case '#':
                printfFormat->flags.alternate = TRUE;
                break;
            case '0':
                printfFormat->flags.zeros = TRUE;
                break;
            case '\'':
                printfFormat->flags.byteScale = TRUE;
                break;
            default:
                parsingFlags = FALSE;
                break;
        }
    }
    printfFormat->width = sp77_parseWidth (&format, args);
    if (*format == '.') {
        ++format;
        printfFormat->precision = sp77_parseWidth (&format, args);
    }
    if ((*format == 'h') || (*format == 'l') || (*format == 'L')) {
        printfFormat->argSize = *format;
        ++format;
    }
    printfFormat->argChar = *format;
    if ((*format == 'e') || (*format == 'E')
     || (*format == 'f') || (*format == 'F')
     || (*format == 'g') || (*format == 'G'))
    {
        if ( printfFormat->precision == UNDEF )
	{
	    printfFormat->precision = 6;
	}
    }
    else
    {
	if (printfFormat->flags.byteScale) {
            if ( printfFormat->width != UNDEF )
	    	printfFormat->width /= printfFormat->charSize;
            if ( printfFormat->precision != UNDEF )
		printfFormat->precision /= printfFormat->charSize;
	}
    }
    ++format;
#ifdef TEST
    fprintf (stdTrace, "Format width: %d; prec: %d; ( ",
        printfFormat->width, printfFormat->precision);
    if (printfFormat->flags.justifyLeft)
        fprintf (stdTrace, "ljust ");
    if (printfFormat->flags.sign)
        fprintf (stdTrace, "sign ");
    if (printfFormat->flags.space)
        fprintf (stdTrace, "space ");
    if (printfFormat->flags.alternate)
        fprintf (stdTrace, "alt ");
    if (printfFormat->flags.zeros)
        fprintf (stdTrace, "zeros ");
    if (printfFormat->flags.byteScale)
        fprintf (stdTrace, "byteScale ");
    fprintf (stdTrace, ") -> %c\n", printfFormat->argChar);
#endif
    return format;
}

/*----------------------------------------*/

typedef int     (*sp77_convFunction) (char *buf, int buflen, va_list * args,
    t_printf_format * printfFormat);
typedef struct tjk_convFunctionStruct {
    sp77_convFunction f;
} tjk_convFunctionStruct;

/*----------------------------------------*/

static int
sp77_invalidChar (
    char *buf,
    int buflen,
    va_list * args,
    t_printf_format * printfFormat)
{
    /* ROUTINE_DBG_MSP00 ("sp77_invalidChar"); */
    return UNDEF;
}

/*----------------------------------------*/

/*lint -esym(528,dummy) */
static tjk_convFunctionStruct dummy =
{sp77_invalidChar};
static tjk_convFunctionStruct sp77_convLower[26] =
{
    {sp77_invalidChar},          /* 'a' */
    {sp77_invalidChar},          /* 'b' */
    {sp77_cConv},                /* 'c' */
    {sp77_signedConv},           /* 'd' */
    {sp77_doubleConv},           /* 'e' */
    {sp77_doubleConv},           /* 'f' */
    {sp77_doubleConv},           /* 'g' */
    {sp77_invalidChar},          /* 'h' */  /* used as modifier */
    {sp77_signedConv},           /* 'i' */
    {sp77_invalidChar},          /* 'j' */
    {sp77_invalidChar},          /* 'k' */
    {sp77_invalidChar},          /* 'l' */  /* used as modifier */
    {sp77_invalidChar},          /* 'm' */
    {sp77_invalidChar},          /* 'n' */
    {sp77_unsignedConv},         /* 'o' */
    {sp77_pConv},                /* 'p' */
    {sp77_invalidChar},          /* 'q' */
    {sp77_hexConv},              /* 'r' */
    {sp77_sConv},                /* 's' */
    {sp77_invalidChar},          /* 't' */
    {sp77_unsignedConv},         /* 'u' */
    {sp77_invalidChar},          /* 'v' */
    {sp77_invalidChar},          /* 'w' */
    {sp77_unsignedConv},         /* 'x' */
    {sp77_invalidChar},          /* 'y' */
    {sp77_invalidChar}           /* 'z' */
};

static tjk_convFunctionStruct sp77_convUpper[26] =
{
    {sp77_invalidChar},          /* 'A' */
    {sp77_invalidChar},          /* 'B' */
    {sp77_invalidChar},          /* 'C' */
    {sp77_invalidChar},          /* 'D' */
    {sp77_doubleConv},           /* 'E' */
    {sp77_doubleConv},           /* 'F' */
    {sp77_doubleConv},           /* 'G' */
    {sp77_invalidChar},          /* 'H' */
    {sp77_invalidChar},          /* 'I' */
    {sp77_invalidChar},          /* 'J' */
    {sp77_invalidChar},          /* 'K' */
    {sp77_invalidChar},          /* 'L' */  /* used as modifier */
    {sp77_invalidChar},          /* 'M' */
    {sp77_invalidChar},          /* 'N' */
    {sp77_invalidChar},          /* 'O' */
    {sp77_invalidChar},          /* 'P' */
    {sp77_invalidChar},          /* 'Q' */
    {sp77_hexConv},              /* 'R' */
    {sp77_SConv},                /* 'S' */
    {sp77_invalidChar},          /* 'T' */
    {sp77_invalidChar},          /* 'U' */
    {sp77_invalidChar},          /* 'V' */
    {sp77_invalidChar},          /* 'W' */
    {sp77_unsignedConv},         /* 'X' */
    {sp77_invalidChar},          /* 'Y' */
    {sp77_invalidChar}           /* 'Z' */
};

/*----------------------------------------*/

int
sp77_vsprintf (
    cbool genUnicode,
    char *buf,
    int buflen,
    const char *format,
    va_list args)
{
    /* ROUTINE_DBG_MSP00 ("sp77_vsprintf"); */
    char           *ptr = buf;
    const char     *textStart;
    int             converted = 0;
    t_printf_format printfFormat;
    int             charsWritten;
    cbool           error = FALSE;
    int             literalLen;
    cbool           quickExit = FALSE;

    printfFormat.genUnicode = genUnicode;
    if (genUnicode) {
        printfFormat.callMemset = local_memset;
        printfFormat.charSize = local_charsize;
    }
    else {
        printfFormat.callMemset = memset;
        printfFormat.charSize = 1;
    }
    while ((*format != '\0') && (buflen > 0) && !error) {
        /*
         * look for format
         */
        textStart = format;
        while ((*format != '\0') && (*format != '%')) {
            ++format;
        }
        /*
         * copy literal text
         */
        if (format > textStart) {
            literalLen = format - textStart;
            /* check length */
            /* move bytes */
            if (!printfFormat.genUnicode) {
                if (literalLen > buflen - 1) {
                    literalLen = buflen - 1;
                    quickExit = TRUE;
                }
                memcpy (ptr, textStart, literalLen);
                ptr += literalLen;
                buflen -= literalLen;
                if (quickExit) {
                    break;
                }
            }
            else {
                if (literalLen * local_charsize > buflen - local_charsize) {
                    literalLen = (buflen / local_charsize) - 1;
                    quickExit = TRUE;
                }
                local_unicpy ((t_unichar *) ptr, (const char *) textStart, literalLen);
                ptr += literalLen * local_charsize;
                buflen -= literalLen * local_charsize;
                if (quickExit) {
                    break;
                }
            }
        }
        if (*format == '%') {
            /*
             * parse format
             */
            format = sp77_parseFormat (format, &printfFormat, &args);
            /*
             * call conversion
             */
            switch (printfFormat.argChar) {
                case '\0':
                    error = TRUE;
                    break;
                case '%':
                    *ptr = '%';
                    ++ptr;
                    --buflen;
                    break;
                case 'n':
                    switch (printfFormat.argSize) {
                        case 'h':
                            *va_arg (args, short *) = (short) (ptr - buf);

                            break;
                        case 'l':
                            *va_arg (args, long *) = ptr - buf;

                            break;
                        case '\0':
                            *va_arg (args, int *) = ptr - buf;

                            break;
                        default:
                            error = TRUE;
                            break;
                    }
                    break;
                default:
                    if (BETWEEN_MSP77 ('a', printfFormat.argChar, 'z')) {
                        charsWritten = sp77_convLower[printfFormat.argChar - 'a']
                            .f (ptr, buflen, &args, &printfFormat);
                    }
                    else if (BETWEEN_MSP77 ('A', printfFormat.argChar, 'Z')) {
                        charsWritten = sp77_convUpper[printfFormat.argChar - 'A']
                            .f (ptr, buflen, &args, &printfFormat);
                    }
                    else {
                        charsWritten = UNDEF;
                    }
                    if (charsWritten >= 0) {
                        ptr += charsWritten * printfFormat.charSize;
                        buflen -= charsWritten * printfFormat.charSize;
                        --converted;
                    }
                    else
                        error = TRUE;
                    break;
            }
#ifdef TRACE
            fprintf (stdTrace, "restformat: '%s'\n", format);
#endif
        }
    }
    *ptr = '\0';
    if (genUnicode)
        ptr[1] = '\0';
    if (error)
        return converted;
    else
        return (ptr - buf) / printfFormat.charSize;
}

/*----------------------------------------*/

int
sp77vsprintf (
    char *buf,
    int buflen,
    const char *format,
    va_list args)
{
    /* ROUTINE_DBG_MSP00 ("sp77vsprintf"); */
    return sp77_vsprintf (FALSE, buf, buflen, format, args);
}

/*----------------------------------------*/

int
sp77sprintf (
    char *buf,
    int buflen,
    const char *format,
    ...)
{
    /* ROUTINE_DBG_MSP00 ("sp77sprintf"); */
    int             result;
    va_list         ap;

    va_start (ap, format);
    result = sp77_vsprintf (FALSE, buf, buflen, format, ap);
    va_end (ap);
    return result;
}

/*----------------------------------------*/

int
sp77vsunicodef (
    char *buf,
    int buflen,
    const char *format,
    va_list args)
{
    /* ROUTINE_DBG_MSP00 ("sp77vsunicodef"); */
    return sp77_vsprintf (local_unicode, buf, buflen, format, args);
}

/*----------------------------------------*/

int
sp77sunicodef (
    char *buf,
    int buflen,
    const char *format,
    ...)
{
    /* ROUTINE_DBG_MSP00 ("sp77sunicodef"); */
    int             result;
    va_list         ap;

    va_start (ap, format);
    result = sp77_vsprintf (local_unicode, buf, buflen, format, ap);
    va_end (ap);
    return result;
}


/*----------------------------------------*/

#ifdef TEST

static int
sprintfLib (
    char *buf,
    int buflen,
    const char *format,
    ...)
{
    /* ROUTINE_DBG_MSP00 ("sprintfLib"); */
    int             result;
    va_list         ap;

    va_start (ap, format);
    result = vsprintf (buf, format, ap);
    va_end (ap);
    return result;
}

typedef int     sprintfFunction (char *, int, const char *,...);
typedef struct t_sprintfFunctionRec {
    sprintfFunction *f;
} t_sprintfFunctionRec;

static t_sprintfFunctionRec sprintfF[2] =
{
    {sp77sprintf},
    {sprintfLib}
};

#define OWN     0
#define LIB     1

static const char * origins[] =
{"own", "lib"};

/*----------------------------------------*/

static void
sp77_testStrings (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testStrings"); */
    char            buf[2000];
    int             i;
    const char     *someString = "someString";
    const char     *otherString = "otherString";
    const char     *simpleFormat = "%s";
    const char     *complexFormat = "%20.5s";
    const char     *testFormat = "%s %% %s";

    for (i = OWN; i <= LIB; ++i) {
        sprintfF[i].f (buf, sizeof (buf), testFormat, someString, otherString);
        fprintf (stdout, "%s: '%s'\n", origins[i], buf);
    }
}

/*----------------------------------------*/

static void
sp77_testChars (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testChars"); */
    char            buf[2000];
    int             i;
    int             j;
    char            ch;
    const char *       testFormat = "Char <%3c>";

    for (j = 32; j < 125; ++j) {
        ch = (char) j;
        for (i = OWN; i <= LIB; ++i) {
            sprintfF[i].f (buf, sizeof (buf), testFormat, ch);
            fprintf (stdout, "%s: '%s'\n", origins[i], buf);
        }
    }
}

#define c_unsignedFormats 3

/*----------------------------------------*/

static void
sp77_testUnsigned (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testUnsigned"); */
    char            buf[2000];
    int             i;
    int             findex;
    unsigned int    arg = 300;
    const char     *testFormat[c_unsignedFormats];

    testFormat[0] = "%0#10u";
    testFormat[1] = "%0#10X";
    testFormat[2] = "%0#10o";
    for (findex = 0; findex < c_unsignedFormats; ++findex) {
        for (i = OWN; i <= LIB; ++i) {
            sprintfF[i].f (buf, sizeof (buf), testFormat[findex], arg);
            fprintf (stdout, "%s: '%s'\n", origins[i], buf);
        }
    }
}

/*----------------------------------------*/

#define c_signedCases 5

static void
sp77_testSigned (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testSigned"); */
    char            buf[2000];
    int             i;
    int             argIndex;
    int             arg[c_signedCases];
    const char     *testFormat = "%5.*d";

    arg[0] = 0;
    arg[1] = 1;
    arg[2] = -1;
    arg[3] = 100;
    arg[4] = -100;
    for (argIndex = 0; argIndex < c_signedCases; ++argIndex) {
        for (i = OWN; i <= LIB; ++i) {
            sprintfF[i].f (buf, sizeof (buf), testFormat, 3, arg[argIndex]);
            fprintf (stdout, "%s: '%s'\n", origins[i], buf);
        }
    }
}

/*----------------------------------------*/

#define c_doubleArgs 4
#define c_doubleFormats 9

static void
sp77_testDouble (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testDouble"); */
    char            buf[2000];
    int             i;
    int             aindex;
    int             findex;
    double          args[c_doubleArgs];
    const char     *testFormat[c_doubleFormats];

    args[0] = 12345760.01234567;
    args[1] = 0.000001234567;
    args[2] = 12.3456789012345678901234567890;
    args[3] = 0.01234567;

    testFormat[0] = "%20.14f";
    testFormat[7] = "%f";
    testFormat[1] = "%e";
    testFormat[2] = "%g";
    testFormat[3] = "%.0f";
    testFormat[4] = "%.0e";
    testFormat[5] = "%.0g";
    testFormat[6] = "%.7g";
    testFormat[8] = "%20.14E";

    for (aindex = 0; aindex < c_doubleArgs; ++aindex) {
        for (findex = 0; findex < c_doubleFormats; ++findex) {
            for (i = OWN; i <= LIB; ++i) {
                sprintfF[i].f (buf, sizeof (buf), testFormat[findex], args[aindex]);
                fprintf (stdout, "%s: '%s'\n", origins[i], buf);
            }
        }
    }
}

/*----------------------------------------*/

const fillCharC = '-';

static void
sp77_checkBuffer (
    const char * buf,
    int size)
{
    int i;
    for (i = 0; i < size; ++i) {
        if (buf [i] != fillCharC) {
            fprintf (stdout, "mismatch at: %d\n", i);
        }
    }
}

/*----------------------------------------*/

static void
sp77_testOverflow (void)
{
    char    underflowArea [500];
    char    buf [40];
    char    overflowArea [500];

    memset (underflowArea, fillCharC, sizeof (underflowArea));
    memset (overflowArea, fillCharC, sizeof (overflowArea));
    sp77sprintf (buf, sizeof (buf),
        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx%s",
        "parameterString");
    fprintf (stdout, "overflow: [%d] %s\n", strlen (buf), buf);
    sp77_checkBuffer (underflowArea, sizeof (underflowArea));
    sp77_checkBuffer (overflowArea, sizeof (overflowArea));
}

/*----------------------------------------*/

#if defined (UNICODE_SUPPORT)
static void
sp77_testUnicode (void)
{
    /* ROUTINE_DBG_MSP00 ("sp77_testUnicode"); */
    t_unichar    uniBuf[100];
    const char     *someString = "some String";
    void           *argBuf;
    int             len;
    char            buf[2000];
    const char     *format[6];
    int             i;
    int             outlen;
    cbool           sourceUnicode = TRUE;
    cbool           genUnicode = TRUE;

    if (sourceUnicode) {
        local_gounicode (NULL, 0, 0, 0);
        argBuf = uniBuf;
        len = strlen (someString);
        local_unicpy (uniBuf, someString, len);
        uniBuf[len] = 0;
    }
    else {
        argBuf = someString;
    }
    format[0] = "'%S'";
    format[1] = "'%20S'";
    format[2] = "'%.5S'";
    format[3] = "'%20.5S'";
    format[4] = "'%-'20S'";
    format[5] = "'%-20S'";
    for (i = 0; i < 6; ++i) {
        if (genUnicode) {
            outlen = sp77sunicodef (buf, sizeof (buf), format[i], argBuf);
            fwrite (buf, outlen * local_charsize, 1, stdout);
        }
        else {
            outlen = sp77sprintf (buf, sizeof (buf), format[i], argBuf);
            fwrite (buf, outlen, 1, stdout);
            fputc ('\n', stdout);
        }
    }
}
#endif

/*----------------------------------------*/

int
main ()
{
    /* ROUTINE_DBG_MSP00 ("main"); */
    stdTrace = fopen ("testf.prt", "w");
    sp77_testStrings ();
    sp77_testChars ();
    sp77_testUnsigned ();
    sp77_testSigned ();
    sp77_testDouble();
    sp77_testOverflow ();
#if defined (UNICODE_SUPPORT)
    sp77_testUnicode ();
#endif
    fclose (stdTrace);
    return 0;
}
#endif

